/******************************************************************************
 * $Header: /CommonBe/agmsmith/Programming/AGMSRAMFileSystem/RCS/AGMSRAMFileSystem.c,v 1.156 2004/04/04 02:00:16 agmsmith Exp agmsmith $
 *
 * This is a RAM based file system for BeOS 5.0.  It stores files and
 * directories directly in memory, rather than going through a device driver to
 * access storage space.  It implements most of the BFS features except for
 * indices and queries, plus it adds AGMS links (an improvement over UNIX®
 * style hard links and Macintosh® style aliases).
 *
 * Copyright © 2001 by Alexander G. M. Smith.  Macintosh is a registered trade
 * mark of Apple Computer, Inc.  UNIX is a registered trademark of The Open
 * Group.  POSIX is a registered trademark of the IEEE.  To see the copyright
 * and trademark symbols, make sure you are using UTF-8 encoding to view this
 * file.
 *
 * This file system is free software; you can redistribute it and/or modify it
 * under the terms of the GNU Lesser General Public License as published by the
 * Free Software Foundation; either version 2.1 of the License, or (at your
 * option) any later version.
 *
 * This file system is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
 * or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public
 * License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with this file system; if not, write to the Free Software Foundation,
 * Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA.
 *
 * Send e-mail with suggestions for improvement to agmsmith@rogers.com.  If
 * that address doesn't work, search the web for "Alexander G. M. Smith" to
 * find my home page and contact info.
 *
 *
 * Features:
 *
 * Because AGMSRAMFileSystem uses actual file sizes rather than rounding up to
 * the next larger block size, the amount of memory used corresponds to the
 * amount of data stored in the system (the main advantage this file system
 * over the previous AGMSRAMDiskDevice project).
 *
 * It supports the usual BFS (Be File System) features.  These include Unix
 * style operations on files, directories and symbolic links.  Plus, the extra
 * Be specific live change notification mechanism and file attributes are
 * implemented.  There's also a fair bit of Unix style security implemented;
 * but as BeOS uses the root user account for everything, it won't make much
 * difference in practice.  It's also designed for multiple simultaneous thread
 * access, so multiple programs working with different files won't interfere
 * with each other.  However, queries and indices aren't implemented yet.
 *
 * The main nifty keen cool new feature is the implementation of hard links,
 * which are missing from BFS.  Rather, a variation of hard links officially
 * called AGMS links (Alias Generated Morphing Symbolic links), like Macintosh
 * aliases but better, is implemented.  The main idea is to let you put your
 * files inside multiple directories; it's useful for filing your data so that
 * you can find your aunt's letter about the cottage in both a directory of
 * aunt related things and a directory of cottage related things.  Unlike
 * Macintosh aliases or symbolic links, they are valid even if you delete the
 * original file - one of the AGMS links will morph and become the file itself.
 * Unlike Unix hard links, it is possible to have an AGMS hard link refering to
 * a parent directory - cycles in the file name hierarchy are supported
 * correctly.  Finally, because they appear to the OS as symbolic links (which
 * happen to morph once in a while), they are compatible with existing
 * software.
 *
 *
 * Usage:
 *
 * Use the separately provided Be application to start up (mount is the
 * technical term) the file system (it provides an extra argument needed during
 * startup).  The normal unmount command (or GUI equivalent) can be used to
 * unmount it.  The application will also ensure that the file system code is
 * in the /boot/home/config/add-ons/kernel/file_systems directory before
 * mounting.
 *
 * AGMS links are created by making a hard link (since there aren't any hard
 * links in BeOS I can ursurp that function safely).  The command line command
 * is "ln existing_file_path_name new_link_name" and for directories you need
 * to use "ln -d existing_directory_path_name new_link_name" (that's an
 * artifact of the Unix way of preventing accidental hard linking of
 * directories).  Each file system item shows up as a regular file (or
 * directory etc) in one directory and as fake symbolic links in the others.
 * Ideally I'd like to have the AGMS links show up as a special kind of file
 * type (specified by the read status rstat() function), but most software
 * won't understand that, thus the fake symbolic links (until someone rewrites
 * the OS and utilities).  The use of fake symbolic links for secondary
 * references to a file/directory (rather than hard links) lets existing
 * software work without getting stuck in directory cycles.  You'll also see
 * backwards links such as ".." to the main parent directory of an object (a
 * hard link as usual), and "..." as the second one (a fake symbolic link to
 * the directory containing the first AGMS link, should really be a hard link
 * but "..." isn't treated like ".." by the OS and utilities so I can't do
 * that), and "...." as the parent directory containing the second AGMS link,
 * and so on.  Since you can make a cycle of directories (a directory with
 * itself as a subdirectory, even indirectly through other subdirectories,
 * which would cause infinite loop problems with Unix), there is extra code to
 * allow for that sort of situation.  It will prevent anything bad from
 * happening, such as deleting the last link between the rest of the file
 * system and a cycle.  It also morphs AGMS links into hard links and
 * vice-versa if it needs to change the main parent directory ".." as part of
 * the process of keeping the ".." chain to the root as pure hard links.
 *
 *
 * Programming guidelines that don't fit elsewhere:
 *
 * Implementation is based indirectly on the POSIX® functionality, as it tries
 * to behave the same as the BFS (the AGMSRAMFSMounter test program is used for
 * comparing functionality) and also uses the BSDI (Berkeley Software Design
 * Incorporated, now acquired by Wind River Systems Incorporated) manual pages
 * for functionality documentation (see http://www.bsdi.com/bsdi-man for them).
 *
 * Locking of multiple file system objects is preferably done starting from the
 * higher level directories and working down.  If you need to lock both a file
 * and the directory it is in, you lock the parent directory before locking the
 * object in the directory.  Hopefully this will avoid some deadlocks, though
 * with cycles in the graph of directories, that won't work.  Instead, with
 * cycles possible, you need to rely on the timeout feature of WriteLock().
 *
 * Locking is preferably done in the callback functions that the VFS calls (the
 * ones named ARFS_*), rather than deeper inside subroutines and obscure
 * places.
 *
 *
 * Things To Do and Ideas:
 *
 * Limit memory used to a total maximum.  Either file space used or also
 * include overhead of data structures.
 *
 * Load and dump in XML format, with compression and AGMS link support.
 *
 *
 * Considerations on Hard Links:
 *
 * I considered several ways of implementing hard links, half implemented some,
 * and decided that other ways were better.
 *
 * The traditional Unix way of doing hard links just sets the relevant
 * directory entries (containing a name string and an inode number) to all
 * reference the same inode number (unlike BeOS, Unix inodes don't store the
 * true name of the file).  The entries don't even have to use the same name.
 * They only allow the root user to make hard links to directories, and require
 * an extra flag to the "ln" command for safety.  Presumably the root user
 * knows enough to avoid making cycles.  Pure hard links also have the problem
 * of sharing attributes of the target, such as the position of the icon, which
 * causes problems if there are links from different directories (the icon will
 * have the same position in all directory windows, which would look bad in
 * most of them).
 *
 * Unix and BeOS symbolic links store a path string identifying the target
 * (either relative or absolute path) along with the name (doesn't have to be
 * the same as the target) in the directory.  The OS and utilities realise that
 * it's not the real file and shouldn't be followed in recursive traversals.
 * It works reasonably well, but if you move or rename the file, it stops
 * working because the path string doesn't change.
 *
 * I much later found out about Macintosh OS "aliases".  They are a lot like
 * Unix hard links (renaming or moving the file doesn't cause problems), except
 * that they show up as aliases (names displayed in italic) and the OS and
 * utilities know they aren't the real file or item, thus avoiding recursive
 * loop problems.  But when you delete the target of an alias, the alias that
 * remains is broken (using it brings up an error message saying it can't find
 * the target).
 *
 * One idea (now abandonned) I came up with was something I call "unique hard
 * links".  Along the way, I found out that the BeOS desktop expects that the
 * true name of an object to match the name in the directory (otherwise draging
 * icons to move them doesn't work correctly).  While having multiple names for
 * a file sounds cool to programmers, it's confusing to users, causes problems
 * with the BeOS desktop software, and similar functionality can be implemented
 * with ordinary symbolic links, so I'm open to dropping that requirement.
 * "Unique hard links" are hard links which require that the name in the
 * directory is the same as the true name of the file or directory item.  Thus
 * the item can only appear once in a directory, a minor loss in flexability.
 * If you rename a file, all the directories it is in also change their idea of
 * the name to match (thus renaming requires a check for duplicate names in all
 * parent directories).  This is nice for users - they see the same name
 * everywhere.  However, recursive directory operations (such as using "zip" to
 * compress a directory tree into an archive file) will still get stuck in an
 * infinite loop (until the combined path length exceeds the maximum).
 * Deleting items is a lot more complex - it needs to check that removing a
 * link to a directory won't isolate a section of the file system graph from
 * the root.
 *
 * Another idea I came up with and abandonned was something I call "persistent
 * symbolic links", which turn out to be a lot like Mac aliases.  These show up
 * as symbolic links to the file/directory item being referenced, but
 * internally are hard links.  I require that all items have one ordinary path
 * from the root directory all the way down to the item (thus imposing a tree
 * structure on the messy directed graph of items).  That ordinary path is what
 * shows up in the path string of the fake symbolic links.  Internally, each
 * item has an array of parent directories.  The first directory (array[0]) is
 * treated as the main parent directory (works like the normal parent
 * directory).  The remaining parent directories are auxiliary ones in which
 * the item shows up as a fake symbolic link - the directory listing / Walk()
 * code lists it as a fake symbolic link if the true name is different from the
 * directory's idea of the name or if the directory isn't the main directory of
 * the item.  That way recursive directory listings wouldn't get stuck in
 * infinite loops and other operations would work normally.  The trouble is
 * that you have to store the full ordinary path to the item in all the hard
 * links, and if the main path changed, all the pseudo-symbolic hard links have
 * to be updated.  Alternatively you could build the full path string
 * dymamically by doing a get_node/lock/read name/unlock/put_node for each main
 * parent directory up to the root.  For simple operations such as Walk(), that
 * seems a bit excessive.  It also brings more chances for deadlock if some
 * other thread is modifying any of those directories.  Either way, the string
 * that's returned could be obsolete before it gets used if some other thread
 * renames any of the directories (symbolic links have the same problem).  Even
 * worse, when one of the directories in the main path to the root gets its
 * main link deleted, all items attached to it need to have their main parent
 * directory changed so that it uses a directory which has an ordinary path
 * from the root (or if it is the only parent directory of some items, the
 * deletion has to return the directory-not-empty error).  This is similar to
 * the check for isolation of a section of the directory graph.
 *
 * An aside on VNodeIDs and Fake Symlinks.  This is a trick to have a directory
 * sometimes show up as a symbolic link and sometimes show up as itself, but
 * without creating a real symbolic link.  It's a feature that I originally
 * needed for "persistent symbolic links" but now need for for backward
 * references to AGMS link home directories (the "...", "...." and so on extra
 * parent directories that correspond to AGMS links targetting a directory).
 * Actually you could make an AGMS link named "..." but that's awkward and
 * isn't similar to how ".." is handled (and I don't want to store separate
 * attributes for the "..." directory icon position etc).  I essentially just
 * want readdir() to list "..." as a symbolic link to the appropriate parent
 * directory.  I found out that readdir() gives the caller the vnodeid of the
 * item and its parent, but doesn't specify what kind of file it is so I can't
 * simply tell the caller that it is a symbolic link.  D'oh!  Instead, "ls"
 * calls rstat() to find out if an item is a symbolic link, file or directory.
 * The trouble is that rstat() takes a pointer to a shadow vnode, and if we
 * used the actual item's pointer, it wouldn't be a symbolic link, it would be
 * whatever it is.  So, after much head scratching, I've decided to hand out
 * fake vnodeids when a symbolic link is required in walk() and readdir().
 * Sure, a few vnodes for fake symbolic links may hang around in memory after
 * deleting the real file, but that shouldn't cause any problems.  The fake
 * vnodeid contains a mangled vnodeid of the AGMS link FSO.  One way to mangle
 * it is to set the high 32 bits to non-zero and leave the lower 32 bits set to
 * the normal vnodeid.  Another way is to add one to the target vnodeid, which
 * is normally an even multiple of 8.  I'll use the add-one technique.  When
 * one of those mangled vnodeids is converted to a shadow vnode pointer,
 * read_vnode() will use the pointer to the file system object plus 1.  Since
 * memory allocation is done on 8 byte boundaries, we can separate real FSO
 * pointers from fake symbolic link shadow vnode pointers by seeing if the
 * lower 3 bits are zero.  rstat() will then be able to return info on the file
 * if the provided address is 0 modulo 8, or return info on a simulated
 * symbolic link if the pointer is 1 modulo 8.
 *
 * Now, after half way implementing some other ideas, I'm leaning towards MacOS
 * style aliases with a few extra features.  I've also made up a name for them:
 * AGMS links (Alias Generated Morphing Symbolic links).  Alternatively it
 * could stand for Absolutely Grand Morphing Symbolic links, or Amazingly Great
 * Morphing Symbolic links :-).  Not coincidentally, they are also my initials.
 * AGMS links will appear as symbolic links to the OS and users, but internally
 * they are file system objects themselves which just store a pointer to the
 * target file system object (or a vnodeid - same thing) rather than a path
 * string.  When the OS asks for the path, it will dynamically return a string
 * with the current true path.  If the target is deleted from its main (true)
 * parent directory, it will change the target's main parent directory to be
 * some other parent directory, replacing the AGMS link object in that
 * directory with the actual target file (the target's true name will also
 * change to match the AGMS link it is replacing).  Cycle detection will
 * prevent the user from deleting more than one file/directory at a time yet
 * allow them to delete redundant links to non-empty directories.
 *
 *
 * $Log: AGMSRAMFileSystem.c,v $
 * Revision 1.156  2004/04/04 02:00:16  agmsmith
 * Hack for kernel memory allocation bug - leave a hole in memory
 * before the space allocated for the RAM data.  That way whatever
 * the kernel seems to need has space to grow into.  Well, about
 * 60MB of space in this version.
 *
 * Revision 1.155  2004/03/31 01:54:21  agmsmith
 * No, go back to 64MB allocation size.
 *
 * Revision 1.154  2004/03/27 15:17:21  agmsmith
 * Try a smaller 32MB chunk size when allocating, so it doesn't
 * waste as much memory for empty RAM disks.  Hope that's enough
 * to avoid the kernel memory fragmentation bugs.
 *
 * Revision 1.153  2004/03/27 15:08:27  agmsmith
 * Change memory allocation chunk size so that kernel memory space doesn't
 * get as fragmented (which makes the kernel and some devices malfunction,
 * like not being able to mount new DOS disks).
 *
 * Revision 1.152  2004/03/21 18:20:38  agmsmith
 * Tracker expects queries for the date modified to be values in
 * seconds, not microseconds.  So humour it.
 *
 * Revision 1.151  2004/02/28 18:40:30  agmsmith
 * When writing an attribute, if the new type is zero
 * then keep the old type.
 *
 * Revision 1.150  2004/02/28 18:07:01  agmsmith
 * More forgiving and fancier type checking when reading
 * attributes (allows CSTR to read MIMS attributes etc).
 *
 * Revision 1.149  2004/02/28 17:21:51  agmsmith
 * Limit the length of debug printouts of subtrees when parsing, to avoid
 * crashing the kernel.  Also fix attribute reading so that if the type
 * of the attribute is zero, it doesn't need to match the requested type.
 *
 * Revision 1.148  2003/05/30 20:47:13  agmsmith
 * Dumping and loading under construction.
 *
 * Revision 1.147  2003/05/27 21:46:31  agmsmith
 * Added function for converting binary to escape codes.
 *
 * Revision 1.146  2003/05/27 00:23:20  agmsmith
 * Added parsing for dump file mount options.
 *
 * Revision 1.145  2003/05/24 21:04:16  agmsmith
 * New command line arguments to mount() - more generic keyword based
 * rather than raw fixed size binary data.
 *
 * Revision 1.144  2003/05/21 23:07:03  agmsmith
 * Added the RemoveBackslashEscapeCodes function, as part of the
 * new mount options parsing.
 *
 * Revision 1.143  2003/05/14 21:33:56  agmsmith
 * Tested multiple keyword indices, just needed a few changes to
 * error messages!
 *
 * Revision 1.142  2003/05/12 20:10:43  agmsmith
 * Don't print common error messages.
 *
 * Revision 1.141  2003/05/11 21:01:42  agmsmith
 * Now does query optimization for wildcard queries.  Needed to make
 * the range start and end keys dynamically allocated so that they
 * could hold truncated and otherwise modified strings (can't have
 * the wildcard characters in the iteration limits).
 *
 * Revision 1.140  2003/05/08 01:23:50  agmsmith
 * Range optimization now present, but needs testing.
 *
 * Revision 1.139  2003/05/05 21:14:32  agmsmith
 * Pretend to be a persistent file system so that Tracker will do queries
 * on the RAM file system volume.  Also do more conservative file memory
 * allocation - only allocating extra if the increase in size is less than
 * 2K.  Also use old or new name for query notifications during file rename,
 * not just the old name all the time.
 *
 * Revision 1.138  2003/05/04 01:16:25  agmsmith
 * Now with live queries!
 *
 * Revision 1.137  2003/05/04 00:40:30  agmsmith
 * Rearranged tokenizing code to remove forward references.
 *
 * Revision 1.136  2003/05/04 00:36:33  agmsmith
 * Live queries under construction.
 *
 * Revision 1.135  2003/05/01 21:55:37  agmsmith
 * Added locking functions for the live query list.
 *
 * Revision 1.134  2003/04/27 00:03:21  agmsmith
 * Add a unique code after the fake name used in indexes when seen as
 * directories, so that several entries in the index that have the same
 * key (such as files the same size in the size index) appear to be
 * different to old programs.  Also need to check if a particular key
 * and value are in the AVL tree.
 *
 * Revision 1.133  2003/04/25 17:43:00  agmsmith
 * Update the name index when renaming a file.
 *
 * Revision 1.132  2003/04/25 16:30:43  agmsmith
 * Print error messages when things go wrong for standard index updates.
 *
 * Revision 1.131  2003/04/25 01:32:49  agmsmith
 * Avoid a deadlock with queries holding a read lock on the index for
 * too long by having them back off and try again.
 *
 * Revision 1.130  2003/04/23 20:25:43  agmsmith
 * Save vnode_ids of the standard indices so that they can be used without
 * having to lock the .indices directory and root directory, which was
 * causing deadlock problems.
 *
 * Revision 1.129  2003/04/21 00:07:49  agmsmith
 * Now date last modified, accessed and inode changed are updated.
 *
 * Revision 1.128  2003/04/19 16:05:09  agmsmith
 * Moved code around to fix dependencies.
 *
 * Revision 1.127  2003/04/19 15:52:39  agmsmith
 * Adding date modified, accessed, etc. updates.
 *
 * Revision 1.126  2003/04/12 20:35:33  agmsmith
 * Changes to file size now change the entry in the "size" index.
 *
 * Revision 1.125  2003/04/11 01:32:07  agmsmith
 * Now removes name, size, date modified virtual attributes when
 * deleting a file.
 *
 * Revision 1.124  2003/04/10 19:57:31  agmsmith
 * Added a retry with a delay for deadlocked rename or delete operations.
 *
 * Revision 1.123  2003/04/10 17:33:13  agmsmith
 * Removed debug code for checking FSO validity after Locking.
 *
 * Revision 1.122  2003/04/10 17:01:36  agmsmith
 * Added a semaphore to file cookies.
 *
 * Revision 1.121  2003/02/18 03:38:05  agmsmith
 * Added a retry after a random delay for WriteLock and
 * ConvertReadLockToWriteLock.  Seems to work, avoiding a deadlock
 * which happens when you view a directory in Tracker while
 * simultaneously doing copyattr -d a bunch of files to it (it
 * renames them from a temporary name and Tracker has fun trying
 * to keep up).
 *
 * Revision 1.120  2003/02/18 00:31:32  agmsmith
 * Improve debugging of deadlocks - display time it took to wait and file name.
 *
 * Revision 1.119  2003/02/17 22:05:49  agmsmith
 * Use an on-stack semi-global variable for working memory when doing a
 * recursive parse tree evaluation, since it can get 50 levels deep when
 * doing some common queries from Tracker.
 *
 * Revision 1.118  2003/01/13 03:18:30  agmsmith
 * Query code is starting to work, can do simple comparison queries.  Ran
 * into debug bugs, can't dprintf more than about 1K before it crashes,
 * and Tracker generates queries longer than that!
 *
 * Revision 1.117  2003/01/12 03:49:06  agmsmith
 * Value comparison for queries under construction.  Also finished
 * up parsing of the <> and >< unary operators.
 *
 * Revision 1.116  2003/01/10 22:53:00  agmsmith
 * Adding parse tree evaluation and the new attribute exists <> and
 * attribute missing >< unary relational operators.
 *
 * Revision 1.115  2003/01/07 00:04:04  agmsmith
 * Query infrastructure in place, but doesn't actually test files against the
 * parse tree yet, or do optimization of the query.
 *
 * Revision 1.114  2003/01/04 22:45:08  agmsmith
 * Upper and lower bounds, not start and end, make the iteration simpler.
 *
 * Revision 1.113  2003/01/04 04:31:59  agmsmith
 * Query evaluation under construction, adding iteration ranges
 * to the generic open directory functions.
 *
 * Revision 1.112  2002/12/31 22:59:32  agmsmith
 * Parser now runs and makes parse trees!
 *
 * Revision 1.111  2002/12/31 19:09:20  agmsmith
 * Parsing almost finished.
 *
 * Revision 1.110  2002/12/30 21:50:16  agmsmith
 * Parser under construction.
 *
 * Revision 1.109  2002/12/16 21:38:45  agmsmith
 * Under construction from months ago.
 *
 * Revision 1.108  2002/08/16 19:03:27  agmsmith
 * Parse tree under construction.
 *
 * Revision 1.107  2002/08/12 00:10:06  agmsmith
 * Better check it in before RCS gets more confused.
 *
 * Revision 1.106  2002/08/09 01:08:47  agmsmith
 * Added tokenizing of query strings.
 *
 * Revision 1.105  2002/07/04 19:37:28  agmsmith
 * Changed error handling for update errors.
 *
 * Revision 1.104  2002/07/03 23:34:29  agmsmith
 * Added index directory routines, and stat and remove index.
 *
 * Revision 1.103  2002/07/02 21:11:01  agmsmith
 * More index stuff.
 *
 * Revision 1.102  2002/07/02 17:17:21  agmsmith
 * Indices are now visible as directories, with their contents as
 * fake symbolic links.
 *
 * Revision 1.101  2002/07/01 20:09:50  agmsmith
 * Now adds attributes to an index, and removes them too.  Found out that
 * floating point attributes don't print correctly (sprintf doesn't do
 * floats or doubles).
 *
 * Revision 1.100  2002/06/29 21:41:18  agmsmith
 * Indices under construction.
 *
 * Revision 1.99  2002/06/24 21:04:48  agmsmith
 * Now uses generic directory routines for directory and attribute listing.
 *
 * Revision 1.98  2002/06/23 17:54:55  agmsmith
 * Generic directory listing under construction.
 *
 * Revision 1.97  2002/06/12 19:40:31  agmsmith
 * Added plus and minus infinity value test option for iteration.
 *
 * Revision 1.96  2002/06/08 20:34:08  agmsmith
 * Allow opening in write mode of non-file things, even though you
 * can't actually write (or read) them.
 *
 * Revision 1.95  2002/06/06 22:33:48  agmsmith
 * Changed readdir to use a separate iterator function for adding dirents,
 * more efficient for large batches and enables code reuse for attribute
 * and index content listing.
 *
 * Revision 1.94  2002/05/24 16:42:43  agmsmith
 * Added stat_attr.
 *
 * Revision 1.93  2002/05/23 22:05:09  agmsmith
 * Added rename_attr.
 *
 * Revision 1.92  2002/05/23 19:03:26  agmsmith
 * Added remove_attr.
 *
 * Revision 1.91  2002/05/13 02:53:20  agmsmith
 * Added read_attr.
 *
 * Revision 1.90  2002/05/12 18:45:24  agmsmith
 * Added write_attr.
 *
 * Revision 1.89  2002/04/30 00:12:35  agmsmith
 * In the middle of adding attributes.
 *
 * Revision 1.88  2002/04/29 22:41:00  agmsmith
 * Don't store volume name as part of path to root, since the volume
 * name can be changed.  Now partly works, except if you rename the
 * root directory without changing the volume name to match (BeOS
 * usually will do both).
 *
 * Revision 1.87  2002/04/28 21:58:22  agmsmith
 * Added wfsstat, starting attributes.
 *
 * Revision 1.86  2002/04/26 21:48:29  agmsmith
 * Better change notification for rename and delete, plus
 * debug messages for change notifications.
 *
 * Revision 1.85  2002/04/26 20:45:30  agmsmith
 * Working on notification for rename and delete, currently wrong.
 *
 * Revision 1.84  2002/04/21 21:10:24  agmsmith
 * Notification messages being added, mostly B_STAT_CHANGED.
 *
 * Revision 1.83  2002/03/31 20:09:44  agmsmith
 * Added system for limiting the total file size, to avoid using
 * up too much memory and causing a kernel crash.
 *
 * Revision 1.82  2002/03/17 20:10:30  agmsmith
 * AGMS link stuff working now, including llink via ioctl.
 *
 * Revision 1.81  2002/03/16 19:38:56  agmsmith
 * Need to remove deleted AGMS link from target's parent list.
 *
 * Revision 1.80  2002/03/16 16:06:18  agmsmith
 * Bug fix for "...." walk not working correctly.
 *
 * Revision 1.79  2002/03/13 19:34:18  agmsmith
 * Minor cleanup.
 *
 * Revision 1.78  2002/03/11 18:02:49  agmsmith
 * Debugging rename and delete.
 *
 * Revision 1.77  2002/03/10 16:50:30  agmsmith
 * Now compiles, and symbolic links (readlink) fixed.
 *
 * Revision 1.76  2002/03/06 23:57:04  agmsmith
 * Finished a batch of typing for rename and delete, most of the
 * work pushed into the change main parent subroutine.
 *
 * Revision 1.75  2002/02/23 20:58:28  agmsmith
 * Under construction, now up to getting write-locks.
 *
 * Revision 1.74  2002/02/14 22:58:49  agmsmith
 * Fixed a few bugs.
 *
 * Revision 1.73  2002/02/08 22:47:43  agmsmith
 * Rename and delete still under construction, but it compiles.
 *
 * Revision 1.72  2002/02/04 21:07:49  agmsmith
 * Under construction.
 *
 * Revision 1.71  2002/01/31 21:55:56  agmsmith
 * AVL tree routines now use an unsigned int as the value, and
 * the callbacks have an extra level of indirection for the value.
 * Also have to use TempUINT to work around a compiler bug with
 * casting the returned unsigned int into a pointer.
 *
 * Revision 1.70  2002/01/31 18:10:41  agmsmith
 * Bug fix.
 *
 * Revision 1.69  2002/01/29 22:44:33  agmsmith
 * Minor tweaks.
 *
 * Revision 1.68  2002/01/29 19:18:57  agmsmith
 * Allow opening of non-files (including fake symbolic links) since
 * the Tracker seems to need to open things even if it doesn't use
 * them for reading or writing.
 *
 * Revision 1.67  2002/01/28 00:56:02  agmsmith
 * AGMS links starting to work, but Tracker's Get Info crashes
 * when it tries to analyse one of them.  Also "..." doesn't
 * show up in referenced directory in Tracker but does appear
 * when you use "ls -la" to look at the directory.
 *
 * Revision 1.66  2002/01/27 02:24:37  agmsmith
 * It compiles, but still working on rename and delete, and adding
 * AGMS links to the infrastructure (need to do fake symbolic links).
 *
 * Revision 1.65  2002/01/22 19:49:24  agmsmith
 * Resuming work after the ReiserFS mailing list debates.
 *
 * Revision 1.64  2001/12/31 22:59:39  agmsmith
 * Rename and delete guts under reconstruction.
 *
 * Revision 1.63  2001/12/31 19:22:03  agmsmith
 * Under construction.
 *
 * Revision 1.62  2001/12/30 20:10:19  agmsmith
 * Updated documentation for new AGMS Link terminology.
 *
 * Revision 1.61  2001/12/27 22:49:52  agmsmith
 * Still contemplating aliases and cycles as part of rename and delete.
 *
 * Revision 1.60  2001/12/22 22:51:08  agmsmith
 * Reworking rename and delete docs to use aliases.
 *
 * Revision 1.59  2001/12/22 00:52:43  agmsmith
 * Reconstructing things to use aliases rather than parent directories.
 *
 * Revision 1.58  2001/12/19 22:49:15  agmsmith
 * More thinking about Macintosh aliases and a decision to use
 * something like them but better.  Updated documents with thoughts.
 *
 * Revision 1.57  2001/12/17 22:51:07  agmsmith
 * Time to add fake symbolic links and vnodes.
 *
 * Revision 1.56  2001/12/16 03:57:33  agmsmith
 * Compiles, but symbolic links for things with multiple parent
 * directories isn't working right.
 *
 * Revision 1.55  2001/12/16 00:16:18  agmsmith
 * Hard links and symbolic links are mostly in, and it compiles.
 *
 * Revision 1.54  2001/12/15 17:04:19  agmsmith
 * Symbolic links under construction.
 *
 * Revision 1.53  2001/12/09 16:02:59  agmsmith
 * Still intermittently under construction.
 *
 * Revision 1.52  2001/12/01 19:08:13  agmsmith
 * *** empty log message ***
 *
 * Revision 1.51  2001/11/21 23:09:19  agmsmith
 * Under construction, but compiles.
 *
 * Revision 1.50  2001/11/20 19:18:10  agmsmith
 * Under construction.
 *
 * Revision 1.49  2001/11/16 21:39:24  agmsmith
 * Rename and delete under construction.
 *
 * Revision 1.48  2001/11/15 18:44:00  agmsmith
 * More pseudocode changes.
 *
 * Revision 1.47  2001/11/05 22:14:17  agmsmith
 * Comments and some data structures for the rename and delete fancy
 * code have been added.
 *
 * Revision 1.46  2001/10/30 20:01:08  agmsmith
 * More minor cleanup.
 *
 * Revision 1.45  2001/10/30 18:28:39  agmsmith
 * Changed return type of PermissionAllowed to an error code.
 *
 * Revision 1.44  2001/10/16 22:08:12  agmsmith
 * Changed return code from AVLDupAdd to specify separate codes for
 * adding a new node and trying to add an existing key/value.
 *
 * Revision 1.43  2001/10/16 21:42:44  agmsmith
 * Changed locking functions to return an error code.
 *
 * Revision 1.42  2001/10/09 20:53:09  agmsmith
 * *** empty log message ***
 *
 * Revision 1.41  2001/10/09 20:50:13  agmsmith
 * More renaming code, and added a description of unique persistent symbolic
 * links.  Need to rewrite a lot to fit that in now...
 *
 * Revision 1.40  2001/10/06 21:15:08  agmsmith
 * Initial version of renaming.
 *
 * Revision 1.39  2001/10/06 15:56:12  agmsmith
 * Tuning for handling big files.
 *
 * Revision 1.38  2001/09/16 01:26:10  agmsmith
 * Fixed handling of big file sizes and handle running out of
 * memory better when making big files.
 *
 * Revision 1.37  2001/09/15 22:18:51  agmsmith
 * wstat largely finished.
 *
 * Revision 1.36  2001/09/15 17:59:29  agmsmith
 * Fixed up append mode writes.
 *
 * Revision 1.35  2001/09/14 20:59:43  agmsmith
 * *** empty log message ***
 *
 * Revision 1.34  2001/09/13 21:55:50  agmsmith
 * wstat is under construction.
 *
 * Revision 1.33  2001/09/12 22:46:25  agmsmith
 * Removed unnecessary umask code.
 *
 * Revision 1.32  2001/09/10 21:25:59  agmsmith
 * After finding out that Create gets called on existing files,
 * combined create and open.
 *
 * Revision 1.31  2001/09/10 18:43:26  agmsmith
 * Added unlink (actually rmdir and unlink are now one function).
 *
 * Revision 1.30  2001/09/09 16:49:08  agmsmith
 * Experiments with trying to find umask value, none work.
 *
 * Revision 1.29  2001/09/07 00:07:50  agmsmith
 * Minor changes, mostly permission stuff.
 *
 * Revision 1.28  2001/09/01 23:43:51  agmsmith
 * Got reading and writing working!
 *
 * Revision 1.27  2001/09/01 20:52:47  agmsmith
 * Read and write implemented, but not tested.
 *
 * Revision 1.26  2001/08/23 23:59:44  agmsmith
 * Working on creating and reading files.  Writing next...
 *
 * Revision 1.25  2001/08/23 00:00:13  agmsmith
 * Create under construction again.
 *
 * Revision 1.24  2001/08/22 14:00:42  agmsmith
 * *** empty log message ***
 *
 * Revision 1.23  2001/08/22 03:05:41  agmsmith
 * Changed debug messages that list a file to call a standard
 * function for printing file names and info (including permissions
 * which were needed for permission testing).
 *
 * Revision 1.22  2001/08/21 01:56:08  agmsmith
 * Was working on creating files, sidetracked to add permission tests.
 *
 * Revision 1.21  2001/07/31 21:59:07  agmsmith
 * Under construction.
 *
 * Revision 1.20  2001/07/29 23:46:15  agmsmith
 * Added utility routines for managing storage of file contents data.
 *
 * Revision 1.19  2001/07/29 17:57:20  agmsmith
 * Under construction.
 *
 * Revision 1.18  2001/07/28 19:07:48  agmsmith
 * Added rmdir command and tests.
 *
 * Revision 1.17  2001/07/28 00:42:27  agmsmith
 * Don't allow creation of directories with names of all periods.  Also
 * improved readdir so that it returns multiple dirents if requested.
 *
 * Revision 1.16  2001/07/27 20:55:55  agmsmith
 * Busy adding data files.
 *
 * Revision 1.15  2001/07/26 21:27:37  agmsmith
 * Minor bugs fixed, updated comments to reflect get_vnode knowledge.
 *
 * Revision 1.14  2001/07/26 19:24:51  agmsmith
 * Before fiddling around with get_vnode.
 *
 * Revision 1.13  2001/07/26 02:14:54  agmsmith
 * First working version, whoo-hoo!  Makedir works, and you can
 * get a listing of a directory.  Mount and unmount work too.
 *
 * Revision 1.12  2001/07/23 01:08:24  agmsmith
 * opendir series of functions updated.
 *
 * Revision 1.11  2001/07/22 20:15:42  agmsmith
 * Adding and removing directories added.  Grind continues
 * while listening to Cross Country Checkup on summer job stories.
 *
 * Revision 1.10  2001/07/22 04:55:12  agmsmith
 * Working on mkdir and all that that implies.
 *
 * Revision 1.9  2001/07/21 23:28:31  agmsmith
 * File system objects starting to take shape...
 *
 * Revision 1.8  2001/07/20 15:50:49  agmsmith
 * Under construction.
 *
 * Revision 1.7  2001/07/15 21:21:43  agmsmith
 * Add a more sophisticated disk volume record validity check.
 *
 * Revision 1.6  2001/07/12 17:45:20  agmsmith
 * Now partly works, you can mount and unmount it and see
 * the empty root directory.
 *
 * Revision 1.5  2001/07/08 18:54:36  agmsmith
 * Under construction.
 *
 * Revision 1.4  2001/07/06 00:23:57  agmsmith
 * Mount and unmount getting there, adding vnodes.
 *
 * Revision 1.3  2001/07/02 21:58:46  agmsmith
 * Busy setting up memory pool and mount instance data.
 *
 * Revision 1.2  2001/05/31 20:45:47  agmsmith
 * Have a bare minimum working version - always says it failed to
 * start up.  But at least it is recognised by the system.
 *
 * Revision 1.1  2001/05/14 14:17:03  agmsmith
 * Initial revision
 */

#include <KernelExport.h>  /* Threads, semaphores, etc. */
#include <area_malloc.h>   /* Memory pool library. */
#include <fs_index.h>      /* For index_info. */
#include <Errors.h>        /* Various return codes. */
#include <TypeConstants.h> /* For type enum items like B_STRING_TYPE. */
#define B_MIME_STRING_TYPE 'MIMS' /* Since we can't use C++ish Mime.h */
#define B_QUERY_UPDATE 'QUPD' /* From the AppDefs.h definition. */
#include <ctype.h>         /* For isspace. */
#include <errno.h>         /* For errno. */
#include <stdio.h>         /* For sprintf. */
#include <string.h>        /* For strcat. */
#include <stdlib.h>        /* For atol. */
#include <zlib.h>          /* For data compression. */

#include "fsproto.h"
#include "AGMSRAMFSExports.h"
#include "AVLDupTree.h"
#include "MatchUTF8.h"


#define DPRINT_NON_ERROR_MESSAGES 0
#define DPRINT_READ_AND_WRITE_OPERATIONS 0
#define DPRINT_NOTIFICATIONS 0
#define DPRINT_ALL_LOCKS 0
  /* 0 or 1.  1 turns on the printing (via dprintf - to the system log) of
  debug messages for normal operations, which can make things slow (since it
  may be printing to the serial port at 19200 baud).  0 makes the code only
  print error conditions and special messages. */

#define MAX_SIMULTANEOUS_READERS 1000
  /* Number of threads allowed to simultaneously read from a file, directory,
  disk volume or whatnot.  The access semaphore is initialised to this value,
  readers just need 1 unit available to get through, writers need to grab all
  MAX_SIMULTANEOUS_READERS units from the semaphore, thus guaranteeing only one
  writer at a time. */

#if DPRINT_NON_ERROR_MESSAGES
  #define DEADLOCK_TIMEOUT 30000000
#else
  #define DEADLOCK_TIMEOUT 3000000
#endif
  /* Timeout in microseconds to wait while acquiring a semaphore for a
  write-lock (read-locks don't have timeouts, but then you can do multiple
  read-locks).  Most file system operations should be quick, particularly since
  it's a RAM file system, so 3 seconds is way more than enough (unless
  debugging).  If the timeout happens, it usually means that there is a
  deadlock.  Rarely it is due to two threads trying to get access to each
  other's locked items.  More often it's a bug in the code where one thread
  mistakenly tries to get multiple locks to the same file. */

#define FILE_SIZE_CHANGE_NOTIFICATION_TIME_PERIOD 2000000
  /* Number of microseconds to stay quiet between sending notification messages
  that the file size has changed, and also updating the "size" index.  This
  avoids having too many notification messages if the user is continuously
  writing to the file, yet avoids the cheap trick of only sending the
  notification when the file is closed.  Only used for files; directories and
  other things don't notify about size changes and aren't in the "size" index.
  Should be a Human perceivable time delay - the GUI display of the size will
  update this often. */

/* Numbers for identifying each kind of file system object. */

typedef enum FileSystemObjectTypesEnum
{
  FSO_TYPE_FILE = 0,
  FSO_TYPE_DIRECTORY,
  FSO_TYPE_SYMBOLIC_LINK,
  FSO_TYPE_AGMS_LINK,
  FSO_TYPE_INDEX,
  FSO_TYPE_MAX
} FileSystemObjectTypes;

static const char * g_ObjectTypeNames [] =
{
  "file",
  "directory",
  "symbolic link",
  "AGMS link",
  "index"
};


static const char g_AreaMallocModuleName [] = B_AREA_MALLOC_MODULE_NAME;
  /* To avoid having copies of this long string... */

static const char g_MaxDotName [NAME_MAX] =
  "................................................................" /* 64 */
  "................................................................" /* 64 */
  "................................................................" /* 64 */
  "..............................................................."; /* 63 */
  /* A string with the longest possible name of a parent directory, consisting
  of 255 periods followed by a NUL for a total of 256 == NAME_MAX bytes.
  When generating strings for parent directories, we just return a pointer
  to the appropriate period n bytes from the end of this string.  It is
  at address (g_MaxDotName + NAME_MAX - 1 - DotCount). */

static bool g_UseVirtualMemory = true;
  /* TRUE to use virtual memory when allocating, which allows more information
  to be stored, but can make things really slow before crashing when there is
  no more memory.  FALSE to use real locked down physical memory, which can
  crash the system quickly if too much is used, but which otherwise avoids the
  swapping delays of virtual memory. */


static const char g_IndicesDirectoryName [] = ".Indices";
  /* The name of the directory, under the root directory, where the indices are
  stored.  This directory will be automatically created at start-up.  The user
  can delete it if they wish, and if they recreate it with this name, the new
  directory will get used. */

typedef enum StandardIndexNumbersEnum
{
  STDINUM_NAME = 0,
  STDINUM_SIZE,
  STDINUM_LASTMOD,
  STDINUM_MAX
} StandardIndexNumbers;

static const char g_NameOfNameIndex [] = "name";
static const char g_NameOfSizeIndex [] = "size";
static const char g_NameOfLastModifiedIndex [] = "last_modified";
static const char *g_NamesOfStandardIndices [STDINUM_MAX] =
{
  g_NameOfNameIndex, g_NameOfSizeIndex, g_NameOfLastModifiedIndex
};
  /* The names of various standard indices, for convenient access and saving
  memory space. */

static const type_code g_TypesOfStandardIndices [STDINUM_MAX] =
{
  B_STRING_TYPE, /* "name" index. */
  B_UINT64_TYPE, /* "size" index. */
  B_INT64_TYPE /* "last_modified" is not UNIX 32 bit B_TIME_TYPE. */
};

static const char g_NoParentsDirectoryName [] =
  ".no parent directory because file has been deleted but is still open";
  /* When getting the true path for a file which has no parent directory, they
  appear to be in this fake root subdirectory (it doesn't exist) in the fake
  symlinks pointing to them. */


typedef enum DateKindsEnum
{   /* In most frequently changed to least frequently changed order. */
    DK_ACCESS = 0,
    DK_MODIFIED,
    DK_INODE_CHANGED,
    DK_CREATED,
    DK_MAX
} DateKinds;

typedef enum DateKindsFlagsEnum
{
    DK_ACCESS_FLAG = 1,
    DK_MODIFIED_FLAG = 2,
    DK_INODE_CHANGED_FLAG = 4,
    DK_CREATED_FLAG = 8
} DateKindsFlags;


/* All things (files, directories, symbolic links, AGMS links, indices) in the
file system have this common generic file system object data.  The vnodeID is
the address of the start of this common data record (high 32 bits are zero,
unless running on a 64 bit processor?), and it is coincidentally a multiple of
8 (the low 3 bits are used for flags to mark fake objects, zero for a real
object).  This structure can be inside bigger more complex records, thus
behaving a lot like a base class for you C++ programmers (there are also
functions which implement the various constructors (one for each subclass) and
a virtual destructor). */

#define FILE_SYSTEM_OBJECT_MAGIC 'FSOm'

#define VALID_FILE_SYSTEM_OBJECT(fsop) ((fsop) != NULL && \
  ((fsop)->magicCode ^ (uint32) (fsop)) == FILE_SYSTEM_OBJECT_MAGIC)

#define INVALID_FILE_SYSTEM_OBJECT(fsop) ((fsop) == NULL || \
  ((fsop)->magicCode ^ (uint32) (fsop)) != FILE_SYSTEM_OBJECT_MAGIC)

typedef struct FileSystemObjectStruct
  FileSystemObjectRecord, *FileSystemObjectPointer;

typedef struct DirectoryObjectStruct
  DirectoryObjectRecord, *DirectoryObjectPointer;

typedef struct AGMSLinkObjectStruct
  AGMSLinkObjectRecord, *AGMSLinkObjectPointer;

typedef struct QueryStateStruct
  QueryStateRecord, *QueryStatePointer;


struct FileSystemObjectStruct
{
  FileSystemObjectTypes objectType;
    /* Identifies whether this is a file, directory, link, etc. */

  sem_id accessSemID;
    /* Multiple reader or single writer access control semaphore, or -1 if
    this object's vnode record isn't loaded (so we don't have semaphores for
    all files, just the active ones). */

  char *trueNamePntr;
    /* The real name of the object.  This points to a buffer just big enough to
    hold the UTF-8 encoded string plus the usual trailing NUL, which is
    allocated when the file system object is created.  If the name changes, the
    string has to get reallocated to match the new size. */

  uint32 magicCode;
    /* To see if this file system object is valid, exclusive-or magicCode with
    the address of the start of the FileSystemObjectRecord (treat the pointer
    as a number) and check that the result is equal to
    FILE_SYSTEM_OBJECT_MAGIC. */

  DirectoryObjectPointer mainParentDirectory;
    /* This is the "main" parent directory for the object, usually the
    directory it was created in.  It is the one you get when you use "cd .." in
    a directory.  This object is also listed using an ordinary entry in that
    parent directory (not an indirect link).  It is updated to always be a
    parent directory which has a direct path (using multiple ".." links) to the
    top level root of the file system (it can't be a directory which is purely
    a subdirectory, to avoid cycles).  The root directory of the disk volume
    has NULL as its main parent (even if it has AGMS links pointing to it). */

  AGMSLinkObjectPointer *parentLinkArray;
    /* This points to an array of pointers, one for each AGMS link that is
    targetting this object.  There is a NULL entry after the last one in the
    array.  The array is always reallocated after the number of links changes
    as I presume that most items only have a few references so that won't waste
    too much time (a slow linear search is used to find them too).  If there
    are no links (the normal case), then we don't bother allocating an array
    with just a NULL in it, instead this pointer to the array is NULL.  You can
    think of the AGMS links pointing to this object as extra parent
    directories, so in a directory listing (if this object is a directory) the
    directories that the extra links come from show up as fake symbolic links
    (pointing to the home directories of the AGMS links) named "...", "....",
    "....." and so on. */

  AVLDupTreePointer attributesTree;
    /* This is NULL if there are no attributes, otherwise it contains the names
    and values of all attributes.  The name string is the key to the tree, the
    value is a pointer to a dymamically allocated AttributeDataRecord, which
    contains the type and size of the attribute and its contents. */

  mode_t modeBits;
    /* Access mode bits.  It contains rwx (read/write/execute) flags for user,
    group, other permissions.  It also has bits to say if the thing is a file
    or directory or symbolic link or a few others.  See the comments in
    <stat.h> for the meanings of the bits.  Currently we don't do anything with
    the sticky, setuid and setgid flags, other than storing them. */

  uid_t uid; /* User id of the owner of this thing. */
  gid_t gid; /* Group id of the owner of this thing. */
  bigtime_t atime; /* Last access time (time of last read), in microseconds. */
  bigtime_t mtime; /* Last data modification time (time of last write). */
  bigtime_t ctime; /* Last change in file status (inode changed) time. */
  bigtime_t crtime; /* Creation time, in microseconds like all the others. */
};



/* This file system object variation stores a directory containing various
other file system objects.  It has a sorted tree which associates names with
pointers to file system objects (subdirectories or files etc). */

struct DirectoryObjectStruct
{
  FileSystemObjectRecord fso;
    /* Generic file system object info, must be at the start of the struct. */

  AVLDupTreePointer directory;
    /* A list of all the files and other items in the directory.  Keys are name
    strings (Note: they are copies of the object's true name, which wastes
    space), values are a FileSystemObjectPointer for the relevant file, subdir
    or link etc.  Parent directories ("..", "..." and so on) aren't listed in
    this directory; they are generated on the fly for directory listings. */
};



/* This file system object variation stores an ordinary file.  Blocks of memory
are strung together to contain the file's data, with simplified special case
storage for zero or one block of memory. */

typedef struct FileObjectStruct
{
  FileSystemObjectRecord fso;
    /* Generic file system object info, must be at the start of the struct. */

  uint32 fileSize;
    /* Current size of the file, as the user sees it.  Note that we are using a
    32 bit size since you can't allocate more than that with 32 bit pointers.
    For that matter, the way addresses are set up, the maximum kernel space
    memory allocation is under 2 gigabytes.  The largest user space allocation
    you can actually make is less than 900MB. */

  uint32 fileSizeInIndex;
    /* The file size that this file is registered as being, in the index.
    Updated usually only when the file is closed (in free cookie), or an
    attempt is made to rstat the file, or when enough time has passed between
    write operations. */

  uint32 totalAllocated;
    /* Total number of bytes allocated in all the memory blocks.  Will be
    larger than or equal to the file size. */

  uint8 *singleBlockPntr;
    /* A pointer to a single block of memory holding the entire file.  Only
    used when there is 1 block, otherwise this is NULL (the 0 or >= 2 block
    situations).  When a file is closed (actually when nobody is using it - see
    write_vnode), if it has multiple blocks then an attempt will be made to
    copy them all into one big block (just big enough to hold the file) to
    avoid having the overhead of the blocks collection sorted data structure
    and reclaim over-allocated space. */

  AVLDupTreePointer blocksCollection;
    /* A list of all the memory blocks storing data for the file, in order of
    increasing position in the file.  The key is the offset in the file of the
    start of the block and value is a pointer to the beginning of the memory
    block.  NULL if there are less than 2 blocks in use.  Yes, a fancy data
    structure so that seeks to an arbitrary position will be fast no matter how
    fragmented the file is. */

  bigtime_t timeOfLastSizeChangeNotification;
    /* The time (obtained from system_time()) when the file size was last
    changed and reported via the notification system (and the "size" index is
    updated).  For every write operation, if the current time is more than
    FILE_SIZE_CHANGE_NOTIFICATION_TIME_PERIOD microseconds larger than this
    field, a notification will be sent. */

  bool sizeChangedButNotificationNotSent;
    /* Used for ensuring that the final size change is sent.  If a size change
    notification has been skipped because it wasn't time for it, this field
    gets set to TRUE.  Then when the file is closed, a final size change
    notification (and "size" index update) will be done if this is TRUE.
    Whenever a notification is actually sent, this field is set to FALSE. */

  bool fileActuallyRead;
    /* Used for updating the date last accessed.  If some data is actually read
    from the file, then this gets set, and the date gets updated when the file
    is closed. */

} FileObjectRecord, *FileObjectPointer;



/* This file system object variation stores a symbolic link.  It contains a
string which identifies to the location where the actual item will be found.
BeOS will take care of evaluating that string, which could be an absolute path
or relative path name. */

typedef struct SymbolicLinkObjectStruct
{
  FileSystemObjectRecord fso;
    /* Generic file system object info, must be at the start of the struct. */

  char *symbolicPathPntr;
    /* A buffer containing UTF-8 characters plus a trailing NUL that identifies
    the place that this symbolic link points to.  Allocated by the file system
    when the object is created, and freed when it is destroyed. */

} SymbolicLinkObjectRecord, *SymbolicLinkObjectPointer;



/* This file system object variation stores an AGMS link.  It contains a
pointer to the FSO of the actual target item.  The target FSO also stores a
pointer back to this AGMS link so that it knows about its references and can
list them as extra parent directories. */

struct AGMSLinkObjectStruct
{
  FileSystemObjectRecord fso;
    /* Generic file system object info, must be at the start of the struct. */

  FileSystemObjectPointer targetFSOPntr;
    /* The thing that the AGMS link is refering to.  The target in other words.
    When users examine the link, it will appear to be a symbolic link with the
    true path of the target, wherever it currently is and whatever it is
    called. */
};


/* This file system object variation stores an index.  It has a key type,
usually B_STRING_TYPE, but it can also be numbers, dates and other things which
the AVLDupTree library supports.  To save space, only the first 256 bytes
(includes the NUL at the end of the string) of a string are stored in the
index.  It has a sorted tree which associates the keys with pointers to file
system objects.  If the name of the index itself matches the name of an
attribute being added or removed, then that attribute value will be used as the
key, and the pointer to the FSO the attribute is attached to will be used as
the value.  That key/value pair will be added or removed when the attribute is
added or removed.  Duplicate keys are allowed; the value is used to distinguish
between them.  For bonus points, if this is a keyword index, the attribute
contents are treated as a list of keywords to be added to the index (so a file
could appear multiple times in a keyword enabled index).  Besides supporting
the BeOS index interface, the index also appears to the user as a directory,
containing the keys as names (numeric keys converted to a readable string) and
the values as a fake symbolic link to the corresponding actual file. */

typedef struct IndexObjectStruct
{
  FileSystemObjectRecord fso;
    /* Generic file system object info, must be at the start of the struct. */

  type_code keyType;
    /* What kind of thing does this index index?  Strings?  Numbers?  Dates? */

  int normalAttributeSize;
    /* How big is the attribute data that corresponds to keyType?  This field
    is the number of bytes used for that type of data, 0 if it is a variable
    length string. */

  AVLDupTreePointer treePntr;
    /* The indexed list of things with this attribute.  The key is the
    attribute contents (up to 256 bytes) and the value is the file that the
    attribute is on. */

  bool interpretAsMultipleKeywords;
    /* FALSE for regular use, TRUE to interpret the attribute contents as a
    list of keywords to be separately indexed.  For strings, the keywords are
    separated by control characters or spaces.  For fixed size types, the
    attribute value is treated as an array; the size divided by the
    normalAttributeSize is the number of keywords. */

} IndexObjectRecord, *IndexObjectPointer;



/* This structure stores the semi-global state information about a particular
disk volume.  Each disk volume corresponds to a successful use of the mount
operation, mounting a disk into the OS's file system name space.  Semi-global
stuff about a particular disk volume, including mutual exclusion semaphores
for multitasking safety, are stored here.  Allocated by mount, freed by
unmount, used by just about everything else. */

#define DISK_VOLUME_MAGIC 'DiVo'

#define VALID_DISK_VOLUME(dvp) ((dvp) != NULL && \
  ((dvp)->nameSpaceID ^ (dvp)->magicCode) == DISK_VOLUME_MAGIC)

#define INVALID_DISK_VOLUME(dvp) ((dvp) == NULL || \
  ((dvp)->nameSpaceID ^ (dvp)->magicCode) != DISK_VOLUME_MAGIC)

#define mallocDVP(dvp,amount) \
  ((dvp)->mallocPntr ((dvp)->volumePoolID, (amount)))

#define freeDVP(dvp,pntr) \
  ((dvp)->freePntr ((dvp)->volumePoolID, (pntr)))

typedef struct DiskVolumeStruct
{
  nspace_id nameSpaceID;
    /* Identifies this particular disk volume to BeOS.  Seems to be a small
    integer which increments every time a disk volume is mounted. */

  uint32 magicCode;
    /* For identifying valid records when given an arbitrary pointer.  It's
    valid if nameSpaceID^magicCode == DISK_VOLUME_MAGIC, just use the
    VALID_DISK_VOLUME or INVALID_DISK_VOLUME macros to check. */

  bool readOnly;
    /* TRUE if this disk is completely read-only, FALSE for read-write.  The
    user specifies this as one of the mount flags and possibly can change it
    via wfsstat. */

  const void *volumePoolID;
    /* The memory pool for all memory allocations.  There is a separate pool
    for each mount-instance so that unmounting a disk volume can easily free
    all related memory.  I was thinking of having separate pools for each file
    and directory, but that wouldn't improve performance that much and would
    exhaust the BeOS limit on the number of system memory areas. */

  status_t (*deletePoolPntr) (const void *pool_id);
  void * (*mallocPntr) (const void *pool_id, size_t size);
  void (*freePntr) (const void *pool_id, void *ptr);
    /* Pointers to the memory pool module functions.  I suppose having them as
    globals might work if they are the same for different instances of
    get_module, but this way is safer. */

  sem_id liveQueryListSemaphoreID;
    /* Controls access to the live query list.  It is set up to allow multiple
    readers or just one writer, using the ReadLockLiveQueryList and
    WriteLockLiveQueryList functions and the corresponding unlock functions.
    Set to -1 if the semaphore isn't initialised (an abnormal situation). */

  QueryStatePointer liveQueryListHead;
    /* Points to the first live query in the list.  The list isn't sorted,
    since all live queries have to be processed for any attribute change
    anyways.  NULL if the list is empty.  I had considered making a .queries
    directory and using that to hold the list, but the queries are too
    ephemeral to be of much use to the user and it would cause deadlock
    problems.  Later there may be query objects that behave like directories,
    which will get their own entries added to this list as part of their
    creation. */

  int32 maxTotalFileSize;
    /* This is the maximum total for all the file sizes.  It limits the RAM
    file system so that it doesn't use up all memory when someone accidentally
    writes a big file to it.  Specified as a mount parameter.  Doesn't include
    memory allocated for directories, the file data structures or anything
    else, just the actual data contents.  Not even the extra buffer size used
    in open files is counted.  To do it more properly, the total amount of
    memory allocated should be used as the limit, but the area malloc module
    doesn't provide it or limit it. */

  vint32 totalFileSize;
    /* The current total of all the file sizes.  Use atomic_add to update it,
    to avoid multithreading problems. */

  DirectoryObjectPointer rootFSOPntr;
    /* The root directory of this disk.  The name of the disk volume is taken
    from the root directory's true name.  This directory always has NULL as its
    main parent directory, even if it has AGMS links pointing to it, to avoid
    confusion when looking for cycles.  Similarly, if you want to lock the
    whole disk, the closest approximation to that is to lock the root
    directory, though that won't stop operations which don't access the root
    directory (most of them), so it really isn't that useful.  Wa'd ya expect,
    it's supposed to be a multithreaded file system! */

  vnode_id rootVNodeID;
    /* VNodeID of the root directory for convenient access. */

  vnode_id standardIndicesVnodes [STDINUM_MAX];
    /* VNodeIDs of the standard indices.  NULL if the index doesn't exist
    (happens if the user deletes it, gets reset if the user creates a new index
    with the right name and type).  This saves having to lock the root and then
    the .Indices directory when you just want to write lock the date last
    modified index.  This results in some annoying deadlock situations.  So now
    you can lock the standard indices directly by using these vnodes (also do a
    get_vnode first). */

  char *pathToRootParent;
    /* The absolute path to the parent directory containing the root directory.
    Since there aren't any BeOS functions for finding it out, it's passed as a
    mount argument from the custom mount program (and we hope nobody moves the
    directory while the file system is active).  Includes leading slash but not
    a trailing slash.  NUL terminated.  Can be the empty string if mounted at
    the top level (the normal case).  Prepended to internal paths (for
    converting AGMS links into absolute path symbolic links) to get an OS wide
    absolute path starting with a slash. */

  int stringLengthOfPathToRootParent;
    /* Length of the pathToRootParent string, to avoid doing strlen too many
    times. */

  const char *dumpFilePath;
    /* When unmounting, dump the RAM file system to this file.  If it is NULL
    then don't dump.  Default is NULL. */

  const char *dumpCompressionMethod;
    /* Which compression method to use.  Either "none" or "zlibdeflate".
    Default is "none". */

  int dumpZlibCompressionLevel;
    /* Which compression level to use for zlib compression of the dump file.
    Varies from 0 (none) to 9 (maximum).  Default is 6. */

  int stringStorageSize;
    /* The number of bytes for extra string storage space allocated immediately
    after this record. */

  char stringStorageSpace [1];
    /* The string storage array, which extends past the end of this record by
    stringStorageUsed bytes (array size is 1 rather than zero to avoid some
    compiler problems, wasting one byte).  Holds copies of command line string
    arguments. */

} DiskVolumeRecord, *DiskVolumePointer;



/* The data for each attribute is stored in this variable sized record.  The
total size of the attribute structure is AttributeDataRecord.size + sizeof
(AttributeDataRecord). */

typedef struct AttributeDataStruct
{
  uint32 size;    /* Number of bytes of data, can be zero. */
  type_code type; /* What primitive type of attribute it is. */
  uint8 value[0]; /* The value immediately follows this structure in memory. */
} AttributeDataRecord, *AttributeDataPointer;


/* A version of an AttributeDataRecord for static allocation (usually on the
stack, the fixed size wastes space) of an attribute with space for the
associated data. */

typedef struct PremadeAttributeStruct
{
  AttributeDataRecord attributeData;
  union StorageUnion
  {
    char              nameString [NAME_MAX];
    uint32            uint32data;
    uint64            uint64data;
  } storage;
} PremadeAttributeRecord, *PremadeAttributePointer;


/* The different kinds of directories and directory-like things we can list. */

typedef enum GenericDirVariationsEnum
{
  GDV_PLAIN_DIRECTORY = 0,
  GDV_ATTRIBUTE_LISTING,
  GDV_INDEX_CONTENTS,
  GDV_QUERY,
  GDV_MAX
} GenericDirVariations;


/* A record for controlling iteration state when traversing an AVL tree
(commonly used for listing directories).  Basically a convenient way of storing
the arguments for AVLDupIterate: starting, ending, traversal direction and a
few other odds and ends. */

typedef struct AVLIterationRangeStruct
{
  bool ascendingOrder;
    /* TRUE to traverse in ascending order, FALSE for descending. */

  AVLDupThingRecord lowerKey;
    /* The iteration goes from the lower to the upper key if in ascending mode,
    or goes from the upper to the lower when descending.  If a string type data
    is used for lowerKey then the strings are presumed to be allocated and
    deallocated elsewhere.  Same for other data types; the iteration functions
    don't allocate or deallocate these upper and lower Things. */

  bool includeThingEqualToLower;
    /* TRUE if the lower key is an allowed value to find in the iteration, or
    FALSE if the iteration starts/stops at the next larger key.  This is useful
    for implementing queries where attribute < value or attribute <= value. */

  bool infiniteLower;
    /* TRUE if minus infinity is used for the lower bound of the iteration.  In
    other words, ignore lowerKey and start at the beginning when ascending or
    stop at the end of the tree when descending.  FALSE to use lowerKey to set
    the lower bound of the iteration. */

  AVLDupThingRecord upperKey;
  bool includeThingEqualToUpper;
  bool infiniteUpper;

} AVLIterationRangeRecord, *AVLIterationRangePointer;



/* Directory cookies keep track of the position being examined while a user is
reading a directory / list of attributes / index / query.  Allocated in
opendir, freed in free_dircookie. */

typedef struct GenericDirCookieStruct
{
  FileSystemObjectPointer fsoPntr;
    /* This has to match the container object (usually a directory, the
    iteration index for queries) in subsequent calls.  It's used as a validity
    check to make sure the user isn't giving us stale cookies.  They also don't
    taste very good. */

  sem_id mutexSemID;
    /* This semaphore protects the cookie.  Its count is initialised to 1.  It
    is used for ensuring that only one thread uses the cookie at one time
    (mutual exclusion), so that a thread reading the directory (which updates
    the last read position in the cookie) doesn't clash with a thread rewinding
    the directory (resetting the last read position to the beginning).
    Normally threads don't share cookies so this really isn't necessary.  But
    just in case someone uses the same directory handle and cookie in multiple
    threads...  A negative value means that the semaphore hasn't been
    allocated. */

  GenericDirVariations listingVariety;
    /* Identifies which variation of a directory we are listing.  This implies
    the way to find the AVL tree containing the items we are listing (can't
    cache the tree pointer since the tree can disappear between calls).  It
    also controls whether parent directories are listed (normal directory
    listings) or not (attribute listings and other specialised listings). */

  type_code keyType;
    /* What kind of things are used for the keys in the AVLDupTree being
    listed?  Normally will be B_STRING_TYPE. */

  QueryStatePointer queryPntr;
    /* For queries this contains the parse tree and other things needed for
    deciding whether a particular file should be included in the listing.  NULL
    for non-queries. */

  bool directoryOpen;
    /* This flag is TRUE while the directory is open for reading.  FALSE when
    it is closed, so that other threads trying to read from it (which exist if
    the user is passing the cookie around between his threads) get an error
    message. */

  AVLIterationRangeRecord iterationRange;
    /* Specifies the start and end limits of the iteration, and direction.
    Query versions need to have the range key Things deallocated when the query
    is closed, since they may be allocated strings which have had the wildcards
    removed from them. */

  int nextParentDirIndex;
    /* Keeps track of how many parent directories have been listed so far and
    also controls the tree iteration mode.  Usually it starts out at zero, so
    readdir lists "." (the FSO itself) as the 0th, ".." (mainParentDirectory)
    when nextParentDirIndex == 1, "..." as the 2nd (the AGMS link from
    parentLinkArray[0]), and so on for later links.  Then when there are no
    more parent directories, nextParentDirIndex gets changed to -1 and
    iteration starts from the beginning of the iteration range.  After the
    first batch of names is returned, this gets set to -2 and the next
    iteration starts from just after the previous one (see lastKey).  It stays
    at -2 for the rest of the directory listing operations. */

  AVLDupThingRecord lastKey;
    /* When reading the directory from the AVL tree, this field stores the last
    key (usually corresponding to a file name) returned to the user.  Then the
    next time readdir is called, it resumes the iteration just after lastKey /
    lastValue.  String keys (B_STRING_TYPE and B_MIME_STRING_TYPE) avoid a
    dynamic allocation and use lastName to store the actual string, with the
    key containing a pointer to it.  Note that we need to copy the key since
    the user could delete the corresponding file and tree node between calls to
    readdir.  For non-string keys, the keys are assumed to not need allocation
    by the generic thing routines. */

  char lastName [NAME_MAX];
    /* The lastName is the last name returned to the user when listing the
    actual contents of the directory (not used when listing the parent dot
    directories).  For string type keys, the string is stored separately in
    lastName (rather than indirectly in lastKey) to avoid the overhead and
    fragmentation of allocating and deallocating memory for each readdir.  On
    the next readdir call, when in tree traversal mode, it will resume with the
    next name after this in alphabetical order.  That does mean that the tree
    can only use NAME_MAX bytes (256, includes NUL at end) per string, thus
    string attributes need to be cut down to that size when added to the index
    (otherwise iteration will get stuck in an infinite loop at a name that's
    longer than NAME_MAX).  For non-string things, the readable string
    equivalent is stored in this field for convenience, though the actual value
    is in lastKey. */

  unsigned int lastValue;
    /* The value (usually a FileSystemObjectPointer) associated with the key
    last returned to the user.  We need both a key and a value so that listings
    with duplicate keys can resume at the right point. */

} GenericDirCookieRecord, *GenericDirCookiePointer;



/* Open file cookies keep track of some of the state of an open file or
directory (yes, you can open directories but not actually read/write them).
But not all state; the file position is tracked separately by the VFS which
passes us a file position offset to use for reads and writes.  So, basically
just track read and write access (for read only files or write only files).
Allocated in open, freed in free_cookie. */

typedef struct FileCookieStruct
{
  FileSystemObjectPointer fsoPntr;
    /* This has to match the file or directory or whatever the thing is in
    subsequent calls, as a kind of validity check.  It is the unmodified value;
    the lower 3 bits can be non-zero to indicate fake symbolic links. */

  sem_id mutexSemID;
    /* This semaphore protects the cookie.  Its count is initialised to 1.  It
    is used for ensuring that only one thread uses the cookie at one time
    (mutual exclusion), so that a thread using the cookie (which really only
    means changes to fileIsOpen and read / write / append mode) doesn't clash
    with other threads.  Normally threads don't share cookies so this really
    isn't necessary.  But just in case someone uses the same file handle and
    cookie in multiple threads...  A negative value means that the semaphore
    hasn't been allocated.  Previously there was no semaphore and we would
    write lock the file (or directory), but this causes an unnecessary
    bottleneck when used on a directory that Tracker is displaying while files
    are being added. */

  bool fileIsOpen;
    /* This flag is TRUE while the file or thing is open.  FALSE when it is
    closed, so that other threads trying to use it (which exist if the user is
    passing the cookie around between his threads) get an error message.  This
    implies you should lock the cookie mutex before checking the flag. */

  bool readingAllowed;
  bool writingAllowed;
    /* These control the availability of reading and writing operations.
    Usually set according to the omode flags passed in the open/create call,
    but can be changed by setflags/fcntl.  Will be set to false for non-files
    to prevent impossible operations. */

  bool appendMode;
    /* TRUE if writes are done in append mode (always add to the end of the
    file and ignore the position provided by the OS).  FALSE for normal writes,
    at the OS provided position. */

} FileCookieRecord, *FileCookiePointer;



/* Forward references. */

int ARFS_open (void *ns, void *node, int omode, void **cookie);



/******************************************************************************
 * Return a string describing a file or other file system object.  The file
 * system object's name, permissions, owner, etc, are described.  The string is
 * actually one of four global buffers (static allocation) so you can only use
 * it four times per dprintf statement.  Four buffers are needed for the rename
 * case (moved file, source and destination directories, replaced file).  Not
 * multitasking safe - the name can be overwritten if several programs are
 * printing names simultaneously.  Adding a semaphore here would only reduce
 * the occurance, not fix it (consider several threads printing at once, if 4
 * prints happen from other threads, then the fourth print will wrap around and
 * modify the first thread's buffer).
 */

static const char *FSOAsString (FileSystemObjectPointer OriginalPntr)
{
  FileSystemObjectPointer FSOPntr;
  bool                    IsFakeSymbolicLink = false;
  volatile static char   *LastString;
  mode_t                  ModeFlags;
  AGMSLinkObjectPointer   LinkPntr;
  AGMSLinkObjectPointer  *LinkPntrPntr;
  char                   *String;
  /* Buffer size includes full name 256, symbolic path 1024, etc. */
  static char             StringA [2048];
  static char             StringB [2048];
  static char             StringC [2048];
  static char             StringD [2048];
  char                   *WritePntr;

  /* Switch the output between one of the four global buffers.  Do this in a
  short bit of code (read and quickly write back to the global variable
  LastString) to reduce some multitasking overlap problems. */

  String = (char *) LastString; /* Cast to throw away volatileness. */
  if (String == StringA)
    String = StringB;
  else if (String == StringB)
    String = StringC;
  else if (String == StringC)
    String = StringD;
  else
    String = StringA;
  LastString = String;

  if (OriginalPntr == NULL)
  {
    strcpy (String, "NULL-FSO");
  }
  else
  {
    /* Check for fake symbolic links and extract the real object. */

    if ((7 & (unsigned int) OriginalPntr) == 1)
    {
      IsFakeSymbolicLink = true;
      FSOPntr = (void *) (((unsigned int) OriginalPntr) - 1);
    }
    else
      FSOPntr = OriginalPntr;

    /* Print info about the real object. */

    if (INVALID_FILE_SYSTEM_OBJECT (FSOPntr))
    {
      sprintf (String, "INVALID-FSO $%X", (unsigned int) OriginalPntr);
    }
    else
    {
      WritePntr = String;

      if (IsFakeSymbolicLink)
      {
        sprintf (WritePntr, "fake symbolic link $%X to ",
          (unsigned int) OriginalPntr);
        WritePntr += strlen (WritePntr);
      }

      ModeFlags = FSOPntr->modeBits;

      sprintf (WritePntr, "%s $%X (%s) %c%c%c%c%c%c%c%c%c%c%c%c U%d G%d",
        (FSOPntr->objectType < 0 || FSOPntr->objectType >= FSO_TYPE_MAX)
        ? "Invalid-Thing-Type" : g_ObjectTypeNames [FSOPntr->objectType],
        (unsigned int) FSOPntr,
        (FSOPntr->trueNamePntr == NULL) ? "NULL-Name" : FSOPntr->trueNamePntr,
        (ModeFlags & S_ISUID) ? 'u' : '-', /* Set user */
        (ModeFlags & S_ISGID) ? 'g' : '-', /* Set group */
        (ModeFlags & S_ISVTX) ? 's' : '-', /* Sticky / sticky swapfile text */
        (ModeFlags & S_IRUSR) ? 'r' : '-',
        (ModeFlags & S_IWUSR) ? 'w' : '-',
        (ModeFlags & S_IXUSR) ? 'x' : '-',
        (ModeFlags & S_IRGRP) ? 'r' : '-',
        (ModeFlags & S_IWGRP) ? 'w' : '-',
        (ModeFlags & S_IXGRP) ? 'x' : '-',
        (ModeFlags & S_IROTH) ? 'r' : '-',
        (ModeFlags & S_IWOTH) ? 'w' : '-',
        (ModeFlags & S_IXOTH) ? 'x' : '-',
        FSOPntr->uid,
        FSOPntr->gid);
      WritePntr += strlen (WritePntr);

      /* Append the symbolic link path if this is a symbolic link. */

      if (FSOPntr->objectType == FSO_TYPE_SYMBOLIC_LINK)
      {
        sprintf (WritePntr, " -> \"%s\"",
          ((SymbolicLinkObjectPointer) FSOPntr)->symbolicPathPntr);
        WritePntr += strlen (WritePntr);
      }

      /* Append a bit of info about the target FSO if this is an AGMS link. */

      if (FSOPntr->objectType == FSO_TYPE_AGMS_LINK)
      {
        sprintf (WritePntr, " -> $%X \"%s\"",
        (int) ((AGMSLinkObjectPointer) FSOPntr)->targetFSOPntr,
        INVALID_FILE_SYSTEM_OBJECT (((AGMSLinkObjectPointer) FSOPntr)->
        targetFSOPntr) ? "Invalid!" :
        ((AGMSLinkObjectPointer) FSOPntr)->targetFSOPntr->trueNamePntr);
        WritePntr += strlen (WritePntr);
      }

      /* Append a list of the links that refer to this link. */

      if ((LinkPntrPntr = FSOPntr->parentLinkArray) != NULL)
      {
        strcpy (WritePntr, " with extra parents (");
        WritePntr += strlen (WritePntr);
        while ((LinkPntr = *LinkPntrPntr) != NULL)
        {
          if (WritePntr - String > sizeof (StringA) - 300)
          {
            strcpy (WritePntr, "...");
            WritePntr += strlen (WritePntr);
            break;
          }
          sprintf (WritePntr, "$%X \"%s\"%s",
            (unsigned int) LinkPntr, LinkPntr->fso.trueNamePntr,
            (LinkPntrPntr[1] == NULL) ? "" : ", ");
          WritePntr += strlen (WritePntr);
          LinkPntrPntr++;
        }
        strcpy (WritePntr, ")");
        WritePntr += strlen (WritePntr);
      }
    }
  }

  return String;
}



/******************************************************************************
 * Return a string describing a four character type code as both a number and
 * four characters.  Replace unprintable characters with a "?".  Use global
 * string storage, but with slight multithreading allowed.
 */

static const char *TypeCodeAsString (type_code TypeCode)
{
  int                     i;
  volatile static char   *LastString;
  int                     Letter;
  char                   *String;
  static char             StringA [40];
  static char             StringB [40];
  static char             StringC [40];
  static char             StringD [40];
  uint32                  Temp32;

  /* Switch the output between one of the four global buffers.  Do this in a
  short bit of code (read and quickly write back to the global variable
  LastString) to reduce some multitasking overlap problems. */

  String = (char *) LastString; /* Cast to throw away volatileness. */
  if (String == StringA)
    String = StringB;
  else if (String == StringB)
    String = StringC;
  else if (String == StringC)
    String = StringD;
  else
    String = StringA;
  LastString = String;

  Temp32 = TypeCode;
  for (i = 0; i < 4; i++)
  {
    Letter = (char) Temp32;
    Temp32 >>= 8;
    if (Letter < 32 || Letter > 127)
      Letter = '?';
    String [3 - i] = Letter;
  }

  sprintf (String + 4, " ($%X)", (unsigned int) TypeCode);

  return String;
}



/******************************************************************************
 * Find the true path name of a file system object.  This is the absolute path
 * that makes use of the path through ".." directories from the object to the
 * root to identify the object.  Writes the result to the user provided buffer,
 * returns TRUE if successful, FALSE if the buffer was too small (leaves buffer
 * untouched).  Because it doesn't lock the files, the result could be
 * incorrect if files are deleted / renamed shortly after the string is
 * generated.  But that's always been true with symbolic links.
 */

static status_t GetTruePath (
  DiskVolumePointer DiskVolumePntr,
  FileSystemObjectPointer FSOPntr,
  char *Buffer,
  int BufferSize)
{
  status_t                ErrorCode;
  bool                    GetVNodeDone = false;
  FileSystemObjectPointer PathFSOPntr;
  char                   *PathNamePntr;
  int                     StringLength;
  status_t                ReturnCode = -1;
  char                    TempPath [PATH_MAX];
  vnode_id                VNodeID = 0;
  void                   *VoidPntr;

  /* Prepare our temporary buffer, where the path name will be built up in
  reverse order.  We could be fancy and use the user's buffer, but there should
  be enough stack space for a 1K string. */

  memset (TempPath, '?', sizeof (TempPath));
  PathNamePntr = TempPath + sizeof (TempPath);
  *(--PathNamePntr) = 0; /* NUL at end of string. */
  PathFSOPntr = FSOPntr;

  /* Add on parent directory names, stop when root directory of the file system
  is reached (include it - it contains the volume name, which could change), or
  stop if we run out of space (possibly because of an infinite loop in the file
  system data, though that shouldn't happen).  Also stop early for files which
  don't have a parent directory. */

  while (PathNamePntr > TempPath /* There is space left in string buffer */)
  {
    /* Do a get_vnode just to make sure that the file system object isn't
    deleted while we are looking at it (which would lead to bad pointer
    dereferencing and a crash). */

    VNodeID = (unsigned int) PathFSOPntr;
    ErrorCode = get_vnode (DiskVolumePntr->nameSpaceID, VNodeID, &VoidPntr);
    if (ErrorCode != 0)
    {
      dprintf ("ARFS:GetTruePath get_vnode failed for %s, code %d!\n",
        FSOAsString (PathFSOPntr), (int) ErrorCode);
      ReturnCode = ErrorCode;
      goto ErrorExit;
    }
    GetVNodeDone = true;
    if (VoidPntr != PathFSOPntr)
    {
      dprintf ("ARFS:GetTruePath get_vnode returned some other node, looking "
        "for %s, got %s!\n",
        FSOAsString (PathFSOPntr), FSOAsString (VoidPntr));
      ReturnCode = ErrorCode;
      goto ErrorExit;
    }

    /* Even though get_vnode was successful, still check the pointer.  It could
    accidentally be a fake symbolic link or some other future non-FSO, which we
    can't handle. */

    if (INVALID_FILE_SYSTEM_OBJECT (PathFSOPntr))
    {
      dprintf ("ARFS:GetTruePath invalid FSO $%X while finding absolute path "
        "name to %s, got to \"%s\".\n",
        (unsigned int) PathFSOPntr, FSOAsString (FSOPntr), PathNamePntr);
      goto ErrorExit;
    }

    /* The file name string could change here if the file is being renamed, but
    we don't want to lock the file since it may have been write-locked by the
    caller (that would cause a deadlock).  So, if the string is wrong, it won't
    crash but the results will be wrong.  But hey, the string will be wrong if
    someone renames the file between the time this function returns and the
    time the caller uses the string. */

    StringLength = strlen (PathFSOPntr->trueNamePntr);
    PathNamePntr -= StringLength;
    if (PathNamePntr >= TempPath) /* If buffer not full, after subtract. */
      memcpy (PathNamePntr, PathFSOPntr->trueNamePntr, StringLength);
    if (PathNamePntr > TempPath) /* If buffer not full, before subtract. */
      *(--PathNamePntr) = '/'; /* Separate parts with '/'. */

    put_vnode (DiskVolumePntr->nameSpaceID, VNodeID);
    GetVNodeDone = false;

    /* Advance to the parent directory. */

    if (PathFSOPntr == &DiskVolumePntr->rootFSOPntr->fso)
      break; /* Reached the root - stop. */
    PathFSOPntr = &PathFSOPntr->mainParentDirectory->fso;

    /* Check if this thing has no parents, happens for open deleted files (and
    the root directory, but we've already checked for it).  There isn't a valid
    path for such an object, so make up a fake one which is a dummy directory
    sibling of the root directory. */

    if (PathFSOPntr == NULL)
    {
      StringLength = strlen (g_NoParentsDirectoryName);
      PathNamePntr -= StringLength;
      if (PathNamePntr >= TempPath) /* If buffer not full, after subtract. */
        memcpy (PathNamePntr, g_NoParentsDirectoryName, StringLength);
      if (PathNamePntr > TempPath) /* If buffer not full, before subtract. */
        *(--PathNamePntr) = '/'; /* Separate parts with '/'. */
      break;
    }
  }

  /* Prepend mount point's parent directory path. */

  PathNamePntr -= DiskVolumePntr->stringLengthOfPathToRootParent;
  if (PathNamePntr >= TempPath)
  {
    if (DiskVolumePntr->stringLengthOfPathToRootParent > 0)
      memcpy (PathNamePntr, DiskVolumePntr->pathToRootParent,
      DiskVolumePntr->stringLengthOfPathToRootParent);
  }
  else /* Path string too long to fit in temp buffer. */
  {
    dprintf ("ARFS:GetTruePath Absolute path to %s is too long, max of %d "
      "bytes, got to \"%s\".\n", FSOAsString (FSOPntr),
      (int) sizeof (TempPath), TempPath);
    goto ErrorExit;
  }

  if (BufferSize < sizeof (TempPath) - (PathNamePntr - TempPath))
  {
    /* Buffer is too small to hold the path. */

    dprintf ("ARFS:GetTruePath Absolute path to %s (which is \"%s\") doesn't "
      "fit in provided buffer of size %d.\n",
      FSOAsString (FSOPntr), PathNamePntr, BufferSize);
    ReturnCode = ENOMEM;
    goto ErrorExit;
  }

  strcpy (Buffer, PathNamePntr);
  ReturnCode = 0; /* Success! */

ErrorExit:
  if (GetVNodeDone)
    put_vnode (DiskVolumePntr->nameSpaceID, VNodeID);
  return ReturnCode;
}



/******************************************************************************
 * Find the size of the object as far as the user sees it.  For use by rstat
 * and the virtual size attribute.  Non-file things also have their own
 * slightly different sizes so that dumb POSIX programs can tell them apart,
 * even if they all appear to be symbolic links.  Files report the size of the
 * file as it is in the index, so that what you see in directory listings
 * matches what's in the index.  Both will be updated once in a while as the
 * file grows and when the file is closed.
 */

static uint64 GetUserFileSize (FileSystemObjectPointer FSOPntr)
{
  uint64 FileSize = 0;

  if (1 == (7 & (unsigned int) FSOPntr))
    FileSize = 2; /* A fake symbolic link. */
  else if (FSOPntr->objectType == FSO_TYPE_AGMS_LINK)
    FileSize = 1;
  else if (FSOPntr->objectType == FSO_TYPE_SYMBOLIC_LINK)
    FileSize = 0;
  else if (FSOPntr->objectType == FSO_TYPE_DIRECTORY)
    FileSize =
    AVLDupGetTreeCount (((DirectoryObjectPointer) FSOPntr)->directory);
  else if (FSOPntr->objectType == FSO_TYPE_INDEX)
    FileSize = AVLDupGetTreeCount (((IndexObjectPointer) FSOPntr)->treePntr);
  else if (FSOPntr->objectType == FSO_TYPE_FILE)
    FileSize = ((FileObjectPointer) FSOPntr)->fileSizeInIndex;
  else
    dprintf ("ARFS:GetUserFileSize: Bug!  Don't know how to "
    "find the size of %s.  Using zero.\n", FSOAsString (FSOPntr));

  return FileSize;
}



/******************************************************************************
 * Waits a random amount of time and then returns.  The delay is between 4
 * seconds and 10 seconds.  The SavedRandomState should be thread local for
 * multitasking safety, and initialised to zero for the first use.
 */

static void RandomDelay (unsigned int *SavedRandomState)
{
  unsigned int RandomNumber = *SavedRandomState;
  bigtime_t    SleepTime;

  if (RandomNumber == 0)
  {
    /* Set up a semi-random seed based on the current time and the address in
    the stack pointer (which should be somewhat different for all threads). */
    RandomNumber = system_time ();
    RandomNumber += (unsigned int) &RandomNumber;
  }

  /* Generate the next random number using a linear congruential method. */

  RandomNumber = RandomNumber * 69069 /* Good for 32 bit values */ + 1;
  *SavedRandomState = RandomNumber;

  SleepTime = 4000000 + RandomNumber /
    (((unsigned int) -1 /* max unsigned int */) / 6000000U);

#if DPRINT_NON_ERROR_MESSAGES
  dprintf ("ARFS:RandomDelay: Sleeping for %Ld microseconds.\n", SleepTime);
#endif

  snooze (SleepTime);

#if DPRINT_NON_ERROR_MESSAGES
  dprintf ("ARFS:RandomDelay: Woke up after sleeping for %Ld microseconds.\n",
    SleepTime);
#endif
}



/******************************************************************************
 * Lock a file system object for writing.  This means grabbing the semaphore
 * for a count of MAX_SIMULTANEOUS_READERS so that we know no readers are using
 * it.  You should also do a get_vnode or otherwise ensure that the OS won't
 * flush the vnode from memory, which would deallocate the semaphore.
 *
 * A timeout and retry count scheme is used so that deadlocks don't cause
 * everything to stop (though you will eventually get the EDEADLK error if that
 * strategy doesn't work).  Besides the deadlock of two programs traversing a
 * cycle in the file tree in different directions and getting in each other's
 * way, which this scheme can't fix, a more common one is due to the conversion
 * of a read lock to a write lock.  For example, thread A gets a read lock on a
 * directory (the semaphore count goes from MAX_SIMULTANEOUS_READERS to
 * MAX_SIMULTANEOUS_READERS - 1), thread B tries for a write lock and blocks
 * (BeOS subtracts MAX_SIMULTANEOUS_READERS from the semaphore value, resulting
 * in a count of -1, even though thread B is not running!), thread A then tries
 * to convert the read lock to a write lock (by subtracting the remaining count
 * of MAX_SIMULTANEOUS_READERS - 1), but has to wait because the semaphore is
 * less than that.  Another way of looking at it is that BeOS semaphores don't
 * allow out of order execution.  To get around that order of execution
 * problem, the semaphore acquire operation backs off to let the other thread
 * get its work done.  This is implemented by undoing the attempted lock (a
 * side effect of the timeout) and waiting a random amount of time before
 * trying again.  For more complex deadlock situations, it will give up after a
 * certain number of iterations.  A total average of 30 seconds delay is used,
 * so that normal operations have time to complete (should be OK for a RAM file
 * system, maybe a bit short for slow devices like magnetic tape or some
 * removeable media).
 *
 * Returns zero if successful, an error code if the
 * semaphore is missing or the lock timed out or was aborted by the user
 * hitting control-C.
 */

static status_t WriteLock (FileSystemObjectPointer FSOPntr)
{
  status_t           ErrorCode;
  int                RetryCount = 0;
  unsigned int       SavedRandomState = 0;
#if DPRINT_NON_ERROR_MESSAGES || DPRINT_ALL_LOCKS
  int32              SemCount = 0;
  bigtime_t          WaitingTime = 0;
#endif

  if (FSOPntr->accessSemID < 0)
  {
    dprintf ("ARFS:WriteLock can't work with %s, semaphoreID is %d.\n",
      FSOAsString (FSOPntr), (int) FSOPntr->accessSemID);
    return B_BAD_SEM_ID;
  }

  while (true)
  {
#if DPRINT_NON_ERROR_MESSAGES || DPRINT_ALL_LOCKS
    get_sem_count (FSOPntr->accessSemID, &SemCount);
    if (SemCount < MAX_SIMULTANEOUS_READERS || RetryCount > 0 ||
    DPRINT_ALL_LOCKS)
    {
      dprintf ("ARFS:WriteLock Semaphore count is %d for %s, attempt #%d.  "
        "May block.\n", (int) SemCount, FSOAsString (FSOPntr), RetryCount);
      WaitingTime = system_time ();
    }
#endif

    ErrorCode = acquire_sem_etc (FSOPntr->accessSemID, MAX_SIMULTANEOUS_READERS,
      B_CAN_INTERRUPT | B_RELATIVE_TIMEOUT /* flags */, DEADLOCK_TIMEOUT);

#if DPRINT_NON_ERROR_MESSAGES || DPRINT_ALL_LOCKS
    if (SemCount < MAX_SIMULTANEOUS_READERS || RetryCount > 0)
    {
      /* If message printed earlier, continue describing the situation. */
      WaitingTime = system_time () - WaitingTime;
      get_sem_count (FSOPntr->accessSemID, &SemCount);
      dprintf ("ARFS:WriteLock Continuing execution after a delay of %Ld "
        "microseconds, count is now %d, for %s, attempt #%d.\n",
        WaitingTime, (int) SemCount, FSOAsString (FSOPntr), RetryCount);
    }
#endif

    if (ErrorCode == 0)
      return 0;

    if (ErrorCode != B_TIMED_OUT)
    {
      /* Semaphore deleted, control-C interrupt, etc. */
      dprintf ("ARFS:WriteLock failed to aquire semaphore for %s, "
        "semaphoreID is %d, error %d.\n",
        FSOAsString (FSOPntr), (int) FSOPntr->accessSemID, (int) ErrorCode);
      return ErrorCode;
    }

    if (++RetryCount > 3)
    {
      dprintf ("ARFS:WriteLock timed out too many times, maybe there is a "
        "real deadlock?  Failed to aquire semaphore for %s, semaphoreID "
        "is %d, returning error %d (EDEADLK).\n",
        FSOAsString (FSOPntr), (int) FSOPntr->accessSemID, (int) EDEADLK);
      return EDEADLK; /* Convert to a more meaningful file system error. */
    }

    RandomDelay (&SavedRandomState);
  }

  return -1;
}



/******************************************************************************
 * Unlock a file system object previously locked for writing.  Returns 0 if
 * successful, a negative error code if it fails.
 */

static status_t WriteUnlock (FileSystemObjectPointer FSOPntr)
{
  status_t ErrorCode;

#if DPRINT_ALL_LOCKS
  dprintf ("ARFS:WriteUnlock entered for %s.\n", FSOAsString (FSOPntr));
#endif

  if (FSOPntr->accessSemID < 0)
  {
    dprintf ("ARFS:WriteUnlock can't work with %s, semaphoreID is %d.\n",
      FSOAsString (FSOPntr), (int) FSOPntr->accessSemID);
    return B_BAD_SEM_ID;
  }

  ErrorCode = release_sem_etc (FSOPntr->accessSemID, MAX_SIMULTANEOUS_READERS,
    0 /* flags */);
  if (ErrorCode != 0)
  {
    dprintf ("ARFS:WriteUnlock failed to release semaphore for %s, "
      "semaphoreID is %d, error %d.\n",
      FSOAsString (FSOPntr), (int) FSOPntr->accessSemID, (int) ErrorCode);
    return ErrorCode;
  }

  return 0;
}



/******************************************************************************
 * Lock a file system object for reading.  This means grabbing the semaphore
 * for a count of 1 so that multiple readers can access the object (semaphore
 * was initialised to a large count).  Double locking from the same thread
 * shouldn't really be done, but it will work most of the time.  Don't need to
 * worry about deadlocks, since writers are the ones that cause that problem
 * (unless you get zillions simultaneous readers deadlocking somehow).  If
 * Tentatively is TRUE then it will only try to get a read lock for a short
 * time, then give up and return EDEADLK.  FALSE means it keeps on trying
 * forever.  The general policy is that writelocks will wait for readlocks (and
 * back off the writelock and retry if it takes too long), since readlocks are
 * usually used for operations which will finish quickly.  But sometimes a
 * readlock may get in the way of writelocks (such as a query evaluation
 * readlocking the index for a long time), so the tentative option exists.
 * Returns zero if successful, a negative error code if the semaphore is
 * missing or the user hits control-C while it is waiting or it times out.
 */

static status_t ReadLock (FileSystemObjectPointer FSOPntr, bool Tentatively)
{
  status_t  ErrorCode;
#if DPRINT_NON_ERROR_MESSAGES || DPRINT_ALL_LOCKS
  int32     SemCount = 0;
  bigtime_t WaitingTime = 0;
#endif

  if (FSOPntr->accessSemID < 0)
  {
    dprintf ("ARFS:ReadLock can't work with %s, semaphoreID is %d.\n",
      FSOAsString (FSOPntr), (int) FSOPntr->accessSemID);
    return B_BAD_SEM_ID;
  }

#if DPRINT_NON_ERROR_MESSAGES || DPRINT_ALL_LOCKS
  get_sem_count (FSOPntr->accessSemID, &SemCount);
  if (SemCount < MAX_SIMULTANEOUS_READERS || DPRINT_ALL_LOCKS)
  {
    dprintf ("ARFS:ReadLock Semaphore count is %d for %s.  "
      "%s\n", (int) SemCount,
      FSOAsString (FSOPntr),
      (SemCount <= 0) ? "May block." : "Simultaneous access.");
    if (SemCount <= 0)
      WaitingTime = system_time ();
  }
#endif

  ErrorCode = acquire_sem_etc (FSOPntr->accessSemID,
    1 /* just a count of one for readers */,
    Tentatively
      ? (B_CAN_INTERRUPT | B_RELATIVE_TIMEOUT)
      : B_CAN_INTERRUPT /* flags */,
    DEADLOCK_TIMEOUT / 6 /* short timeout */);

#if DPRINT_NON_ERROR_MESSAGES || DPRINT_ALL_LOCKS
  if (SemCount <= 0) /* If blocking message printed earlier. */
  {
    WaitingTime = system_time () - WaitingTime;
    get_sem_count (FSOPntr->accessSemID, &SemCount);
    dprintf ("ARFS:ReadLock Continuing execution after a delay of %Ld "
      "microseconds, count is now %d, for %s.\n",
      WaitingTime, (int) SemCount, FSOAsString (FSOPntr));
  }
#endif

  if (ErrorCode != 0)
  {
    if (ErrorCode == B_TIMED_OUT)
      ErrorCode = EDEADLK;
    dprintf ("ARFS:ReadLock failed to aquire semaphore for %s, "
      "semaphoreID is %d, error %d.\n",
      FSOAsString (FSOPntr), (int) FSOPntr->accessSemID, (int) ErrorCode);
    return ErrorCode;
  }

  return 0;
}



/******************************************************************************
 * Unlock a file system object previously locked for reading.  Returns zero if
 * successful, or an error code if it fails.
 */

static status_t ReadUnlock (FileSystemObjectPointer FSOPntr)
{
  status_t ErrorCode;

  if (FSOPntr->accessSemID < 0)
  {
    dprintf ("ARFS:ReadUnlock can't work with %s, semaphoreID is %d.\n",
      FSOAsString (FSOPntr), (int) FSOPntr->accessSemID);
    return B_BAD_SEM_ID;
  }

#if DPRINT_ALL_LOCKS
  dprintf ("ARFS:ReadUnlock entered for %s.\n", FSOAsString (FSOPntr));
#endif

  ErrorCode = release_sem (FSOPntr->accessSemID);
  if (ErrorCode != 0)
  {
    dprintf ("ARFS:ReadUnlock failed to release semaphore for %s, "
      "semaphoreID is %d, error %d.\n",
      FSOAsString (FSOPntr), (int) FSOPntr->accessSemID, (int) ErrorCode);
    return ErrorCode;
  }

  return 0;
}



/******************************************************************************
 * Convert a read lock into a write lock.  The normal technique is to get read
 * locks on all things you're thinking of changing (so that they won't get
 * deleted or modified while you are examining them), then once everything is
 * figured out, convert them to write locks and do the planned operations.
 * That way all the modifications get done; you don't have to worry about doing
 * some operations and then having a lock fail and then have to undo those
 * operations.  Returns zero if succesful, an error code if it fails.  After a
 * deadlock failure or when aborted by the user with a control-C interrupt, the
 * read lock remains active.
 */

static status_t ConvertReadLockToWriteLock (FileSystemObjectPointer FSOPntr)
{
  status_t           ErrorCode;
  int                RetryCount = 0;
  unsigned int       SavedRandomState = 0;
#if DPRINT_NON_ERROR_MESSAGES || DPRINT_ALL_LOCKS
  int32              SemCount = 0;
  bigtime_t          WaitingTime = 0;
#endif

  if (FSOPntr->accessSemID < 0)
  {
    dprintf ("ARFS:ConvertReadLockToWriteLock can't work with %s, "
      "semaphoreID is %d.\n",
      FSOAsString (FSOPntr), (int) FSOPntr->accessSemID);
    return B_BAD_SEM_ID;
  }

  while (true)
  {
#if DPRINT_NON_ERROR_MESSAGES || DPRINT_ALL_LOCKS
    get_sem_count (FSOPntr->accessSemID, &SemCount);
    if (SemCount < MAX_SIMULTANEOUS_READERS - 1 || RetryCount > 0 ||
    DPRINT_ALL_LOCKS)
    {
      dprintf ("ARFS:ConvertReadLockToWriteLock Semaphore count is %d "
        "for %s, attempt #%d.  May block.\n",
        (int) SemCount, FSOAsString (FSOPntr), RetryCount);
      WaitingTime = system_time ();
    }
#endif

    ErrorCode = acquire_sem_etc (FSOPntr->accessSemID,
      MAX_SIMULTANEOUS_READERS - 1 /* Note the minus one - that's the core. */,
      B_CAN_INTERRUPT | B_RELATIVE_TIMEOUT /* flags */, DEADLOCK_TIMEOUT);

#if DPRINT_NON_ERROR_MESSAGES || DPRINT_ALL_LOCKS
    if (SemCount < MAX_SIMULTANEOUS_READERS - 1 || RetryCount > 0)
    {
      /* If message printed earlier, continue describing the situation. */
      WaitingTime = system_time () - WaitingTime;
      get_sem_count (FSOPntr->accessSemID, &SemCount);
      dprintf ("ARFS:ConvertReadLockToWriteLock Continuing execution after a "
        "delay of %Ld microseconds, count is now %d, for %s, attempt #%d.\n",
        WaitingTime, (int) SemCount, FSOAsString (FSOPntr), RetryCount);
    }
#endif

    if (ErrorCode == 0)
      return 0;

    if (ErrorCode != B_TIMED_OUT)
    {
      dprintf ("ARFS:ConvertReadLockToWriteLock failed to convert lock for "
        "%s, semaphoreID is %d, error %d.\n",
        FSOAsString (FSOPntr), (int) FSOPntr->accessSemID, (int) ErrorCode);
      return ErrorCode;
    }

    if (++RetryCount > 3)
    {
      dprintf ("ARFS:ConvertReadLockToWriteLock timed out too many times, "
        "maybe there is a real deadlock?  Failed to convert semaphore "
        "for %s, semaphoreID is %d, returning error %d (EDEADLK).\n",
        FSOAsString (FSOPntr), (int) FSOPntr->accessSemID, (int) EDEADLK);
      return EDEADLK; /* Convert to a more meaningful file system error. */
    }

    RandomDelay (&SavedRandomState);
  }

  return -1;
}



/******************************************************************************
 * Get a read lock on the live query list.  Returns zero if successful, a
 * negative error code if it fails (usually because the user hit control-C).
 */

static status_t ReadLockLiveQueryList (
  DiskVolumePointer DiskVolumePntr)
{
  status_t  ErrorCode;
#if DPRINT_NON_ERROR_MESSAGES || DPRINT_ALL_LOCKS
  int32     SemCount = 0;
  bigtime_t WaitingTime = 0;
#endif

  if (DiskVolumePntr->liveQueryListSemaphoreID < 0)
  {
    dprintf ("ARFS:ReadLockLiveQueryList: liveQueryListSemaphoreID is %d.\n",
      (int) DiskVolumePntr->liveQueryListSemaphoreID);
    return B_BAD_SEM_ID;
  }

#if DPRINT_NON_ERROR_MESSAGES || DPRINT_ALL_LOCKS
  get_sem_count (DiskVolumePntr->liveQueryListSemaphoreID, &SemCount);
  if (SemCount < MAX_SIMULTANEOUS_READERS || DPRINT_ALL_LOCKS)
  {
    dprintf ("ARFS:ReadLockLiveQueryList: Semaphore count is %d.  %s\n",
      (int) SemCount, (SemCount <= 0) ? "May block." : "Simultaneous access.");
    if (SemCount <= 0)
      WaitingTime = system_time ();
  }
#endif

  ErrorCode = acquire_sem_etc (DiskVolumePntr->liveQueryListSemaphoreID,
    1 /* just a count of one for readers */, B_CAN_INTERRUPT /* flags */, 0);

#if DPRINT_NON_ERROR_MESSAGES || DPRINT_ALL_LOCKS
  if (SemCount <= 0) /* If blocking message printed earlier. */
  {
    WaitingTime = system_time () - WaitingTime;
    get_sem_count (DiskVolumePntr->liveQueryListSemaphoreID, &SemCount);
    dprintf ("ARFS:ReadLockLiveQueryList: Continuing execution after a "
      "delay of %Ld microseconds, count is now %d.\n",
      WaitingTime, (int) SemCount);
  }
#endif

  if (ErrorCode != 0)
  {
    if (ErrorCode == B_TIMED_OUT)
      ErrorCode = EDEADLK;
    dprintf ("ARFS:ReadLockLiveQueryList: Failed to aquire semaphore, "
      "error %d.\n", (int) ErrorCode);
    return ErrorCode;
  }

  return 0;
}


/******************************************************************************
 * Undo the read lock for the live query list.  Returns zero if successful, or
 * an error code if it fails.
 */

static status_t ReadUnlockLiveQueryList (
  DiskVolumePointer DiskVolumePntr)
{
  status_t ErrorCode;

  if (DiskVolumePntr->liveQueryListSemaphoreID < 0)
  {
    dprintf ("ARFS:ReadUnlockLiveQueryList: liveQueryListSemaphoreID is %d.\n",
      (int) DiskVolumePntr->liveQueryListSemaphoreID);
    return B_BAD_SEM_ID;
  }

#if DPRINT_ALL_LOCKS
  dprintf ("ARFS:ReadUnlockLiveQueryList: entered.\n");
#endif

  ErrorCode = release_sem (DiskVolumePntr->liveQueryListSemaphoreID);
  if (ErrorCode != 0)
  {
    dprintf ("ARFS:ReadUnlockLiveQueryList: Failed to release semaphore, "
      "semaphoreID is %d, error %d.\n",
      (int) DiskVolumePntr->liveQueryListSemaphoreID, (int) ErrorCode);
    return ErrorCode;
  }

  return 0;
}



/******************************************************************************
 * Lock the live query list for modifications by a single thread.  Returns zero
 * if successful, an error code if the semaphore is missing or was aborted by
 * the user hitting control-C (no timeout delay feature).
 */

static status_t WriteLockLiveQueryList (
  DiskVolumePointer DiskVolumePntr)
{
  status_t           ErrorCode;
#if DPRINT_NON_ERROR_MESSAGES || DPRINT_ALL_LOCKS
  int32              SemCount = 0;
  bigtime_t          WaitingTime = 0;
#endif

  if (DiskVolumePntr->liveQueryListSemaphoreID < 0)
  {
    dprintf ("ARFS:WriteLockLiveQueryList: SemaphoreID is %d.\n",
      (int) DiskVolumePntr->liveQueryListSemaphoreID);
    return B_BAD_SEM_ID;
  }

#if DPRINT_NON_ERROR_MESSAGES || DPRINT_ALL_LOCKS
  get_sem_count (DiskVolumePntr->liveQueryListSemaphoreID, &SemCount);
  if (SemCount < MAX_SIMULTANEOUS_READERS || DPRINT_ALL_LOCKS)
  {
    dprintf ("ARFS:WriteLockLiveQueryList: Semaphore count is %d.  "
      "May block.\n", (int) SemCount);
    WaitingTime = system_time ();
  }
#endif

  ErrorCode = acquire_sem_etc (DiskVolumePntr->liveQueryListSemaphoreID,
    MAX_SIMULTANEOUS_READERS, B_CAN_INTERRUPT /* flags */, 0);

#if DPRINT_NON_ERROR_MESSAGES || DPRINT_ALL_LOCKS
  if (SemCount < MAX_SIMULTANEOUS_READERS)
  {
    /* If message printed earlier, continue describing the situation. */
    WaitingTime = system_time () - WaitingTime;
    get_sem_count (DiskVolumePntr->liveQueryListSemaphoreID, &SemCount);
    dprintf ("ARFS:WriteLockLiveQueryList: Continuing execution after "
      "a delay of %Ld microseconds, count is now %d.\n",
      WaitingTime, (int) SemCount);
  }
#endif

  if (ErrorCode != 0)
    dprintf ("ARFS:WriteLockLiveQueryList: failed to aquire semaphore, "
      "liveQueryListSemaphoreID is %d, error %d.\n",
      (int) DiskVolumePntr->liveQueryListSemaphoreID, (int) ErrorCode);

  return ErrorCode;
}



/******************************************************************************
 * Undo a write lock on the live query list.  Returns 0 if successful, a
 * negative error code if it fails.
 */

static status_t WriteUnlockLiveQueryList (
  DiskVolumePointer DiskVolumePntr)
{
  status_t ErrorCode;

#if DPRINT_ALL_LOCKS
  dprintf ("ARFS:WriteUnlockLiveQueryList: entered.\n");
#endif

  if (DiskVolumePntr->liveQueryListSemaphoreID < 0)
  {
    dprintf ("ARFS:WriteUnlockLiveQueryList: SemaphoreID is %d.\n",
      (int) DiskVolumePntr->liveQueryListSemaphoreID);
    return B_BAD_SEM_ID;
  }

  ErrorCode = release_sem_etc (DiskVolumePntr->liveQueryListSemaphoreID,
    MAX_SIMULTANEOUS_READERS, 0 /* flags */);
  if (ErrorCode != 0)
    dprintf ("ARFS:WriteUnlockLiveQueryList: failed to release semaphore, "
      "liveQueryListSemaphoreID is %d, error %d.\n",
      (int) DiskVolumePntr->liveQueryListSemaphoreID, (int) ErrorCode);

  return ErrorCode;
}



/******************************************************************************
 * Determine if the current thread and associated user is allowed to have the
 * given type of access (R_OK, W_OK, X_OK) to the given file/directory.  The
 * F_OK flag is zero and is just for a check that the file exists.  If you
 * specify multiple flags ORed together, then permission is granted if any of
 * the flags is matched.  Returns zero if they are allowed, a negative error
 * code if they don't have permission.
 */

static status_t PermissionAllowed (
  DiskVolumePointer DiskVolumePntr,
  FileSystemObjectPointer FSOPntr,
  int AccessType)
{
  int   CombinedFlags;
  bool  IsRootUser;
  int   RWXMask;
  uid_t UserID;

  AccessType = (AccessType & 7); /* Just care about low 3 bits of category. */

  if (AccessType == 0)
    return 0; /* The F_OK case, return OK if it exists, which it does. */

  /* Set up the mask for allowing file bits to be checked.  Basically just
  allow through the low 3 bits (rwx).  If the whole file system is read-only,
  only allow through the rx bits. */

  if (DiskVolumePntr->readOnly)
    RWXMask = (S_IROTH | S_IXOTH);
  else
    RWXMask = S_IRWXO;

  /* Have to pass at least one access request, so if AccessType is equal to
  R_OK | W_OK then the file has to be readable or writeable or both.  Note that
  the permission bits are 3 bits, the lowest 3 are for "other", the next 3 are
  for "group" and the next larger 3 are for "user".  Beyond that the bits are
  used for other purposes.  First check the Other bits. */

  CombinedFlags = (FSOPntr->modeBits & RWXMask);
  if (CombinedFlags & AccessType)
    return 0;

  UserID = geteuid ();
  IsRootUser = (UserID == 0);

  /* Root user always gets read and write permission, but not execute (one of
  the other categories has to have execute, which is why those IsRootUser is
  there in the group and user tests) since we don't want to execute
  non-executable files. */

  if (IsRootUser)
  {
    CombinedFlags |= ((S_IROTH | S_IWOTH) & RWXMask);
    if (CombinedFlags & AccessType)
      return 0;
  }

  if (IsRootUser || FSOPntr->gid == getegid ())
  {
    CombinedFlags |= ((FSOPntr->modeBits >> 3) & RWXMask);
    if (CombinedFlags & AccessType)
      return 0;
  }

  if (IsRootUser || FSOPntr->uid == UserID)
  {
    CombinedFlags |= ((FSOPntr->modeBits >> 6) & RWXMask);
    if (CombinedFlags & AccessType)
      return 0;
  }

  dprintf ("ARFS:PermissionAllowed %s "
    "doesn't include requested permission $%X for user %d group %d.\n",
    FSOAsString (FSOPntr), AccessType, UserID, getegid ());

  if (DiskVolumePntr->readOnly && (AccessType & W_OK))
    return EROFS; /* Attempt to write to a read only file system. */

  return EACCES; /* Generic permission failure error. */
}



#if 0 /* Not needed, BeOS handles the umask for us. */
/******************************************************************************
 * Apply the umask to the given permission bits.  The low 9 bits of the
 * permission are rwxrwxrwx flags, with the rwx for "other" being in the lowest
 * valued bits and user in the highest.  Other bits in the permission word
 * aren't affected by this operation.  The umask is a per-process setting,
 * which modifies the default permissions used when creating new files and
 * directories to restrict them further.  A umask of zero does nothing.  When a
 * bit in the umask is set, the corresponding bit in the permissions is forced
 * to zero.  A typical umask value is 022 (in octal numbers) which effectively
 * leaves the "user" bits alone, and clears the w bit in the "group" and
 * "other" permissions, thus by default preventing other people from writing to
 * your files.
 */

static mode_t ApplyUMask (mode_t InputPermissions)
{
  mode_t PreviousMask;

#if 0
  /* Get the value of the umask setting, awkwardly have to set it to get it. */

  PreviousMask = umask (0777 /* A safe temporary no-permissions mask */);
  umask (PreviousMask); /* Undo our temporary mask change. */
#else /* BeOS doesn't have access to the umask value in kernel mode. */
  PreviousMask = 0;
#endif

  /* Apply the mask, ignoring extra bits past the 9 permission bits. */

  PreviousMask &= (S_IRWXU | S_IRWXG | S_IRWXO); /* Remove extra bits. */
  PreviousMask = ~PreviousMask; /* Negate the mask so 1's becomes 0's. */
  return (InputPermissions & PreviousMask);
}
#endif



/******************************************************************************
 * Internal function to deallocate one attribute as a tree iteration callback.
 */

static bool DeallocateOneAttribute (
  AVLDupThingConstPointer KeyPntr,
  const unsigned int *ValuePntr,
  void *ExtraData)
{
  DiskVolumePointer DiskVolumePntr = ExtraData;
  AttributeDataPointer AttrPntr = (AttributeDataPointer) *ValuePntr;

#if DPRINT_NON_ERROR_MESSAGES
  dprintf ("ARFS:DeallocateOneAttribute Deallocating \"%s\", "
    "type %s, size %ld.\n",
    AVLDupGetStringPntrFromThing (*KeyPntr), TypeCodeAsString (AttrPntr->type),
    AttrPntr->size);
#endif

  freeDVP (DiskVolumePntr, AttrPntr);

  return true; /* Continue iterating. */
}



/******************************************************************************
 * Deallocate all the attributes and frees the tree that stores them.
 */

static void DeallocateAllAttributes (
  DiskVolumePointer DiskVolumePntr,
  FileSystemObjectPointer FSOPntr)
{
  if (FSOPntr->attributesTree == NULL)
    return;

  AVLDupIterate (FSOPntr->attributesTree, true,
    NULL, 0, 0, false,
    NULL, 0, 0, false,
    DeallocateOneAttribute, DiskVolumePntr);

  AVLDupFreeTree (FSOPntr->attributesTree);
  FSOPntr->attributesTree = NULL;
}



/******************************************************************************
 * Internal function to deallocate one storage block as a tree iteration
 * callback.
 */

static bool DeallocateOneBlock (
  AVLDupThingConstPointer KeyPntr,
  const unsigned int *ValuePntr,
  void *ExtraData)
{
  DiskVolumePointer DiskVolumePntr = ExtraData;

  freeDVP (DiskVolumePntr, (void *) *ValuePntr);

  return true; /* Continue iterating. */
}



/******************************************************************************
 * Deallocate all the storage blocks in a file's block collection.  Doesn't
 * update size used or anything else, just deallocates the blocks and frees the
 * collection data structures.
 */

static void DeallocateBlockCollection (
  DiskVolumePointer DiskVolumePntr,
  FileObjectPointer FileFSOPntr)
{
  if (FileFSOPntr->blocksCollection == NULL)
    return;

  AVLDupIterate (FileFSOPntr->blocksCollection, true,
    NULL, 0, 0, false,
    NULL, 0, 0, false,
    DeallocateOneBlock, DiskVolumePntr);

  AVLDupFreeTree (FileFSOPntr->blocksCollection);
  FileFSOPntr->blocksCollection = NULL;
}



/******************************************************************************
 * Deallocate all the storage space used by the file's contents and set the
 * size to zero.  Can be used as part of deleting a file, or for the truncate
 * operation.  Not many checks done, so lock the file elsewhere etc.
 */

static void DeallocateFileContents (
  DiskVolumePointer DiskVolumePntr,
  FileObjectPointer FileFSOPntr)
{
  DeallocateBlockCollection (DiskVolumePntr, FileFSOPntr);

  if (FileFSOPntr->singleBlockPntr != NULL)
  {
    freeDVP (DiskVolumePntr, FileFSOPntr->singleBlockPntr);
    FileFSOPntr->singleBlockPntr = NULL;
  }

  atomic_add (&DiskVolumePntr->totalFileSize, -FileFSOPntr->fileSize);

  FileFSOPntr->fileSize = 0;
  FileFSOPntr->totalAllocated = 0;
}



/******************************************************************************
 * Find the buffer address for the given offset within a file's contents.
 * Returns a pointer to the place in the buffer corresponding to the offset and
 * the number of bytes you can write there before hitting the end of that
 * particular portion of the buffer.  Returns NULL if the offset is past the
 * end of the allocated space or something has been corrupted.
 */

static uint8 * GetFileBufferAtOffset (
  FileObjectPointer FileFSOPntr,
  uint32 FileOffset,
  uint32 *BytesAvailablePntr)
{
  AVLDupThingRecord       Key;
  AVLDupThingConstPointer NextKeyPntr;
  uint8                  *NextValue;
  uint32                  OffsetOfNextBlock;
  uint32                  OffsetOfPreviousBlock;
  AVLDupThingConstPointer PreviousKeyPntr;
  uint8                  *PreviousValue;
  unsigned int            TempUINT;

  if (FileOffset >= FileFSOPntr->totalAllocated)
    return NULL; /* Also includes zero allocated space case. */

  if (FileFSOPntr->singleBlockPntr != NULL)
  {
    *BytesAvailablePntr = FileFSOPntr->totalAllocated - FileOffset;
    return FileFSOPntr->singleBlockPntr + FileOffset;
  }

  Key.int32Thing = FileOffset + 1;
  PreviousKeyPntr = &Key;
  TempUINT =
    AVLDupFindPrevious (FileFSOPntr->blocksCollection, &PreviousKeyPntr);
  PreviousValue = (uint8 *) TempUINT;

  if (PreviousValue == NULL)
  {
    dprintf ("ARFS:GetFileBufferAtOffset Bug!  "
      "Unable to find block containing file offset %u in %s!\n",
      (unsigned int) FileOffset, FSOAsString (&FileFSOPntr->fso));
    return NULL;
  }
  OffsetOfPreviousBlock = PreviousKeyPntr->int32Thing;

  NextKeyPntr = PreviousKeyPntr;
  TempUINT = AVLDupFindNext (FileFSOPntr->blocksCollection, &NextKeyPntr);
  NextValue = (uint8 *) TempUINT;

  if (NextValue == NULL) /* If the block was the last block, no next one. */
    OffsetOfNextBlock = FileFSOPntr->totalAllocated;
  else
    OffsetOfNextBlock = NextKeyPntr->int32Thing;

  if (OffsetOfPreviousBlock >= OffsetOfNextBlock)
  {
    dprintf ("ARFS:GetFileBufferAtOffset Bug!  "
      "Block at offset %u doesn't come before block at offset %u!  "
      "While looking for file offset %u in %s.\n",
      (unsigned int) OffsetOfPreviousBlock, (unsigned int) OffsetOfNextBlock,
      (unsigned int) FileOffset, FSOAsString (&FileFSOPntr->fso));
    return NULL;
  }

  if (FileOffset < OffsetOfPreviousBlock ||
  FileOffset >= OffsetOfNextBlock)
  {
    dprintf ("ARFS:GetFileBufferAtOffset Bug!  "
      "Block at offset %u doesn't contain file offset %u in %s!\n",
      (unsigned int) OffsetOfPreviousBlock, (unsigned int) FileOffset,
      FSOAsString (&FileFSOPntr->fso));
    return NULL;
  }

  *BytesAvailablePntr = OffsetOfNextBlock - FileOffset;
  return PreviousValue + (FileOffset - OffsetOfPreviousBlock);
}



/******************************************************************************
 * Allocate some extra space for the file's data storage.  You can allocate
 * more than the actual size of the file, if you feel it will be using it soon
 * (you can shrink it to the actual size needed later).  This function will
 * allocate a new memory block with the given size, and append it to the
 * blockCollection list of blocks (created if not present, even if there will
 * only be one block in it - which happens when you start with an empty file).
 * Returns TRUE if successful, FALSE if out of memory or corrupted data
 * structures.
 */

static bool AddMemoryToFile (
  DiskVolumePointer DiskVolumePntr,
  FileObjectPointer FileFSOPntr,
  uint32 AmountToAllocate)
{
  AVLDupThingRecord Key;
  uint8            *NewBlockPntr = NULL;

  if (AmountToAllocate <= 0)
  {
    dprintf ("ARFS:AddMemoryToFile Bad input parameters!\n");
    goto ErrorExit;
  }

  NewBlockPntr = mallocDVP (DiskVolumePntr, AmountToAllocate);
  if (NewBlockPntr == NULL)
  {
    dprintf ("ARFS:AddMemoryToFile Out of memory while allocating %u bytes "
      "to be added to %s.\n",
      (unsigned int) AmountToAllocate, FSOAsString (&FileFSOPntr->fso));
    goto ErrorExit;
  }

  /* Fill the new block with zeroes so that the user doesn't see remenants of
  some other program's data. */

  memset (NewBlockPntr, 0, AmountToAllocate);

  if (FileFSOPntr->blocksCollection == NULL)
  {
    /* Switching over from a single block or no blocks to the AVLTree of
    blocks, indexed by the starting offset of each block.  The value for each
    entry in the tree is a pointer to the start of the block.  First create the
    AVL tree data structures. */

    FileFSOPntr->blocksCollection = AVLDupAllocTree (
      B_INT32_TYPE, DiskVolumePntr->mallocPntr, DiskVolumePntr->freePntr,
      DiskVolumePntr->volumePoolID);

    if (FileFSOPntr->blocksCollection == NULL)
    {
      dprintf ("ARFS:AddMemoryToFile Unable to create blocks collection AVL "
        "tree for %s, out of memory?\n", FSOAsString (&FileFSOPntr->fso));
      goto ErrorExit;
    }

    /* Add the existing block, if any, to the tree. */

    if (FileFSOPntr->singleBlockPntr != NULL)
    {
      Key.int32Thing = 0; /* Offset of the original whole-file block. */
      if (AVLDupAdd (FileFSOPntr->blocksCollection, &Key,
      (unsigned int) FileFSOPntr->singleBlockPntr) < 0)
      {
        dprintf ("ARFS:AddMemoryToFile Unable to add original block to AVL "
          "tree for %s, out of memory?\n", FSOAsString (&FileFSOPntr->fso));
        AVLDupFreeTree (FileFSOPntr->blocksCollection);
        FileFSOPntr->blocksCollection = NULL;
        goto ErrorExit;
      }
      FileFSOPntr->singleBlockPntr = NULL; /* Block is now in the tree. */
    }
  }

  /* Add the new block to the tree. */

  Key.int32Thing = FileFSOPntr->totalAllocated; /* Offset of new block. */
  if (AVLDupAdd (FileFSOPntr->blocksCollection, &Key,
  (unsigned int) NewBlockPntr) < 0)
  {
    dprintf ("ARFS:AddMemoryToFile Unable to add new block to AVL "
      "tree for %s, out of memory?\n", FSOAsString (&FileFSOPntr->fso));
    goto ErrorExit;
  }

  FileFSOPntr->totalAllocated += AmountToAllocate;
  return true;

ErrorExit:
  if (NewBlockPntr != NULL)
    freeDVP (DiskVolumePntr, NewBlockPntr);
  return false;
}



/******************************************************************************
 * Allocate more space in the file if there isn't enough to meet the desired
 * total.  Add enough to get to the given total allocation, though if memory is
 * short, it may use several allocations of smaller chunks (in case memory is
 * fragmented).  Returns TRUE if successful, FALSE if it wasn't able to
 * allocate the full amount (may have done a partial allocation).
 */

static bool EnlargeFileTotalAllocation (
  DiskVolumePointer DiskVolumePntr,
  FileObjectPointer FileFSOPntr,
  uint32 DesiredTotalFileAllocatedSpace)
{
  uint32 AmountToAllocate;
  bool   FoundSomeMemory;
  uint32 SmallestAllocationSize;

  AmountToAllocate =
    DesiredTotalFileAllocatedSpace - FileFSOPntr->totalAllocated;

  SmallestAllocationSize = AmountToAllocate / 16;
  if (SmallestAllocationSize < 64)
    SmallestAllocationSize = 64;

  /* Try allocating memory in smaller and smaller chunks until it gets some (in
  case of fragmentation).  But no point in allocating lots of small pieces, it
  usually just means we are out of memory (like when you try to allocate 1000
  gigabytes worth but the OS can only provide 0.5GB and smaller chunks).  So,
  if the smallest available size is too small to be useful, stop. */

  while (DesiredTotalFileAllocatedSpace > FileFSOPntr->totalAllocated)
  {
    FoundSomeMemory = false;

    do /* At least one iteration, for small <64 byte requests. */
    {
      if (AddMemoryToFile (DiskVolumePntr, FileFSOPntr, AmountToAllocate))
      {
        FoundSomeMemory = true;
        break; /* Successfully added the memory. */
      }

      dprintf ("ARFS:EnlargeFileTotalAllocation while expanding %s to "
        "%d bytes, allocation of %d bytes failed.\n",
        FSOAsString (&FileFSOPntr->fso),
        (int) DesiredTotalFileAllocatedSpace,
        (int) AmountToAllocate);

      AmountToAllocate /= 2; /* Try half as much. */
    } while (AmountToAllocate >= SmallestAllocationSize);

    if (!FoundSomeMemory)
      break; /* The well is dry, no more memory at all. */
  }

  if (DesiredTotalFileAllocatedSpace > FileFSOPntr->totalAllocated)
  {
    dprintf ("ARFS:EnlargeFileTotalAllocation wasn't able to allocate "
      "memory for requested file total of %d (got %d now) for %s.\n",
      (int) DesiredTotalFileAllocatedSpace,
      (int) FileFSOPntr->totalAllocated,
      FSOAsString (&FileFSOPntr->fso));
    return false;
  }

  return true;
}



/******************************************************************************
 * Packs a file's data by fitting all the data in the block collection into one
 * memory block (or none), and not allocating any more than the amount needed
 * for the file's size.
 */

static void PackFileMemory (
  DiskVolumePointer DiskVolumePntr,
  FileObjectPointer FileFSOPntr)
{
  uint32 Amount;
  int    BlocksPacked;
  uint8 *DestPntr;
  uint8 *NewBlockPntr;
  uint32 Offset;
  uint8 *SrcPntr;
  int    TotalBlockCount;

  if (FileFSOPntr->fileSize > FileFSOPntr->totalAllocated)
  {
    dprintf ("ARFS:PackFileMemory Bug!  File size is larger than amount of "
      "memory allocated for %s.\n", FSOAsString (&FileFSOPntr->fso));
    return;
  }

  if (FileFSOPntr->blocksCollection == NULL)
    return; /* Already packed into zero or one blocks. */

  TotalBlockCount = AVLDupGetTreeCount (FileFSOPntr->blocksCollection);

  if (FileFSOPntr->singleBlockPntr != NULL)
  {
    dprintf ("ARFS:PackFileMemory Bug!  The %s already has a single "
      "block $%X.  Throwing it away...\n", FSOAsString (&FileFSOPntr->fso),
      (unsigned int) FileFSOPntr->singleBlockPntr);
    FileFSOPntr->singleBlockPntr = NULL;
  }

  if (FileFSOPntr->fileSize > 0)
  {
    NewBlockPntr = mallocDVP (DiskVolumePntr, FileFSOPntr->fileSize);
    if (NewBlockPntr == NULL)
    {
      dprintf ("ARFS:PackFileMemory Out of memory while allocating %u bytes, "
        "leaving %s unpacked.\n",
        (unsigned int) FileFSOPntr->fileSize, FSOAsString (&FileFSOPntr->fso));
      return;
    }

    BlocksPacked = 0;
    DestPntr = NewBlockPntr;
    Offset = 0;
    while (Offset < FileFSOPntr->fileSize)
    {
      SrcPntr = GetFileBufferAtOffset (FileFSOPntr, Offset, &Amount);
      if (SrcPntr == NULL || Amount <= 0)
        break;
      if (Amount > FileFSOPntr->fileSize - Offset)
        Amount = FileFSOPntr->fileSize - Offset;
      memcpy (DestPntr, SrcPntr, Amount);
      BlocksPacked++;
      DestPntr += Amount;
      Offset += Amount;
    }
    FileFSOPntr->singleBlockPntr = NewBlockPntr;
#if DPRINT_NON_ERROR_MESSAGES
    dprintf ("ARFS:PackFileMemory Packed %d blocks of %d into one, "
      "reducing memory from %d to %d bytes.\n",
      BlocksPacked, TotalBlockCount,
      (int) FileFSOPntr->totalAllocated, (int) FileFSOPntr->fileSize);
#endif
  }

  DeallocateBlockCollection (DiskVolumePntr, FileFSOPntr);
  FileFSOPntr->totalAllocated = FileFSOPntr->fileSize;
}



/******************************************************************************
 * Initialise a file system object's contents.  Presumably it has been
 * allocated elsewhere, perhaps as part of a larger structure.  Note that it is
 * initialised into an unloaded state (the semaphore isn't created).  Also it
 * doesn't set the parent directory, which will presumably be done when adding
 * it to the parent directory.  Same for the access mode flags - set to zero.
 * Returns TRUE if successful, FALSE if it ran out of memory or you specified
 * bad arguments.
 */

static bool InitialiseFSO (
  DiskVolumePointer DiskVolumePntr,
  FileSystemObjectPointer FSOPntr,
  FileSystemObjectTypes ObjectType,
  const char *TrueNamePntr)
{
  int    StringLength;

  if (FSOPntr == NULL || TrueNamePntr == NULL)
  {
    dprintf ("ARFS:InitialiseFSO Unexpected NULL arguments!\n");
    return false;
  }

  /* Set all fields to zero, many default to zero so this is convenient. */

  memset (FSOPntr, 0, sizeof (FileSystemObjectRecord));

  /* Set the file type and related flags.  The permission bits remain at zero
  so that the caller can OR in the bits they want. */

  FSOPntr->objectType = ObjectType;
  if (ObjectType == FSO_TYPE_FILE)
    FSOPntr->modeBits = S_IFREG;
  else if (ObjectType == FSO_TYPE_DIRECTORY)
    FSOPntr->modeBits = S_IFDIR;
  else if (ObjectType == FSO_TYPE_SYMBOLIC_LINK)
    FSOPntr->modeBits = S_IFLNK;
  else if (ObjectType == FSO_TYPE_AGMS_LINK)
    FSOPntr->modeBits = /* S_IFAGMS in the future */ S_IFLNK;
  else if (ObjectType == FSO_TYPE_INDEX)
    FSOPntr->modeBits = S_IFDIR; /* Appears as a directory to the user. */
  else
  {
    dprintf ("ARFS:InitialiseFSO unknown object type of %d!\n",
      (int) ObjectType);
    return false;
  }

  /* Copy the true name string. */

  StringLength = strlen (TrueNamePntr);
  FSOPntr->trueNamePntr = mallocDVP (DiskVolumePntr, StringLength + 1);
  if (FSOPntr->trueNamePntr == NULL)
  {
    dprintf ("ARFS:InitialiseFSO unable to allocate TrueName!\n");
    return false;
  }
  strcpy (FSOPntr->trueNamePntr, TrueNamePntr);

  /* Set ownership user and group from effective user and group.  The group is
  often taken from the parent directory's group for new files and directories,
  but we'll let the caller replace the group with that if they desire. */

  FSOPntr->uid = geteuid ();
  FSOPntr->gid = getegid ();

  FSOPntr->atime = real_time_clock_usecs ();
  FSOPntr->mtime = FSOPntr->ctime = FSOPntr->crtime = FSOPntr->atime;

  /* Set various other smaller non-zero settings. */

  FSOPntr->accessSemID = -1;
  FSOPntr->magicCode = (FILE_SYSTEM_OBJECT_MAGIC ^ (uint32) FSOPntr);
  return true;
}



/******************************************************************************
 * Deallocate the things inside a generic file system object.  Presumably it
 * has already been unlinked from parent directories and indices.  Doesn't
 * actually delete the FSO record, which could be part of some bigger
 * structure.  C++ programmers should think of this as being the base class
 * destructor.  Returns 0 if successful, otherwise returns a negative error
 * code.
 */

static status_t UninitialiseFSO (
  DiskVolumePointer DiskVolumePntr,
  FileSystemObjectPointer FSOPntr)
{
  status_t ErrorCode;
  status_t ReturnCode = 0;

  if (FSOPntr == NULL)
  {
    dprintf ("ARFS:UninitialiseFSO Unexpected NULL arguments!\n");
    return -1;
  }

  if (FSOPntr->accessSemID != -1)
  {
    dprintf ("ARFS:UninitialiseFSO for %s, potential bug, access semaphore "
      "#%d exists!\n", FSOAsString (FSOPntr), (int) FSOPntr->accessSemID);
    ReturnCode = -3;

    ErrorCode = delete_sem (FSOPntr->accessSemID);
    if (ErrorCode < 0)
    {
      dprintf ("ARFS:UninitialiseFSO for %s, failed to delete access "
        "semaphore #%d, code %d!\n",
        FSOAsString (FSOPntr), (int) FSOPntr->accessSemID, (int) ErrorCode);
      ReturnCode = ErrorCode;
    }
    FSOPntr->accessSemID = -1;
  }

  DeallocateAllAttributes (DiskVolumePntr, FSOPntr);

  if (FSOPntr->trueNamePntr == NULL)
  {
    dprintf ("ARFS:UninitialiseFSO for %s, potential bug, "
      "True Name is NULL!\n", FSOAsString (FSOPntr));
    ReturnCode = -2;
  }
  else
  {
    freeDVP (DiskVolumePntr, FSOPntr->trueNamePntr);
    FSOPntr->trueNamePntr = NULL;
  }

  FSOPntr->magicCode++; /* Make the magic code invalid. */

  return ReturnCode;
}



/******************************************************************************
 * Allocate and initialise a directory file system object.  Returns the new FSO
 * if successful, NULL if it ran out of memory or you specified bad arguments.
 */

static DirectoryObjectPointer AllocateDirectoryFSO (
  DiskVolumePointer DiskVolumePntr,
  const char *TrueNamePntr)
{
  bool                   FSOInitialised = false;
  DirectoryObjectPointer NewDirFSOPntr = NULL;

  NewDirFSOPntr = mallocDVP (DiskVolumePntr, sizeof (DirectoryObjectRecord));
  if (NewDirFSOPntr == NULL)
  {
    dprintf ("ARFS:AllocateDirectoryFSO while doing \"%s\", "
      "ran out of memory!\n", TrueNamePntr);
    goto ErrorExit;
  }

  if (!InitialiseFSO (DiskVolumePntr, &NewDirFSOPntr->fso, FSO_TYPE_DIRECTORY,
  TrueNamePntr))
  {
    dprintf ("ARFS:AllocateDirectoryFSO while doing \"%s\", unable to "
      "initialise the generic file system object!\n", TrueNamePntr);
    goto ErrorExit;
  }
  FSOInitialised = true;

  NewDirFSOPntr->directory = AVLDupAllocTree (B_STRING_TYPE,
    DiskVolumePntr->mallocPntr, DiskVolumePntr->freePntr,
    DiskVolumePntr->volumePoolID);
  if (NewDirFSOPntr->directory == NULL)
  {
    dprintf ("ARFS:AllocateDirectoryFSO while doing \"%s\", unable to "
      "create a new AVL tree!\n", TrueNamePntr);
    goto ErrorExit;
  }

  /* Successful! */
  goto NormalExit;

ErrorExit:
  if (FSOInitialised)
    UninitialiseFSO (DiskVolumePntr, &NewDirFSOPntr->fso);
  if (NewDirFSOPntr != NULL)
    freeDVP (DiskVolumePntr, NewDirFSOPntr);
  NewDirFSOPntr = NULL;
NormalExit:
  return NewDirFSOPntr;
}



/******************************************************************************
 * Allocate and initialise an ordinary user file object.  Returns the new FSO
 * if successful, NULL if it ran out of memory or you specified bad arguments.
 */

static FileObjectPointer AllocateFileFSO (
  DiskVolumePointer DiskVolumePntr,
  const char *TrueNamePntr)
{
  bool               FSOInitialised = false;
  FileObjectPointer  NewFileFSOPntr = NULL;

  NewFileFSOPntr = mallocDVP (DiskVolumePntr, sizeof (FileObjectRecord));
  if (NewFileFSOPntr == NULL)
  {
    dprintf ("ARFS:AllocateFileFSO while doing \"%s\", "
      "ran out of memory!\n", TrueNamePntr);
    goto ErrorExit;
  }

  if (!InitialiseFSO (DiskVolumePntr, &NewFileFSOPntr->fso, FSO_TYPE_FILE,
  TrueNamePntr))
  {
    dprintf ("ARFS:AllocateFileFSO while doing \"%s\", unable to "
      "initialise the generic file system object!\n", TrueNamePntr);
    goto ErrorExit;
  }
  FSOInitialised = true;

  NewFileFSOPntr->fileSize = 0;
  NewFileFSOPntr->fileSizeInIndex = 0;
  NewFileFSOPntr->totalAllocated = 0;
  NewFileFSOPntr->singleBlockPntr = NULL;
  NewFileFSOPntr->blocksCollection = NULL;
  NewFileFSOPntr->timeOfLastSizeChangeNotification = system_time ();
  NewFileFSOPntr->sizeChangedButNotificationNotSent = false;
  NewFileFSOPntr->fileActuallyRead = false;

  /* Successful! */
  goto NormalExit;

ErrorExit:
  if (FSOInitialised)
    UninitialiseFSO (DiskVolumePntr, &NewFileFSOPntr->fso);
  if (NewFileFSOPntr != NULL)
    freeDVP (DiskVolumePntr, NewFileFSOPntr);
  NewFileFSOPntr = NULL;
NormalExit:
  return NewFileFSOPntr;
}



/******************************************************************************
 * Allocate and initialise a symbolic link file system object.  Returns the new
 * FSO if successful, NULL if it ran out of memory or you specified bad
 * arguments.
 */

static SymbolicLinkObjectPointer AllocateSymbolicLinkFSO (
  DiskVolumePointer DiskVolumePntr,
  const char *TrueNamePntr,
  const char *SymbolicPathPntr)
{
  bool                      FSOInitialised = false;
  SymbolicLinkObjectPointer NewSymlinkFSOPntr = NULL;
  int                       StringLength;

  if (SymbolicPathPntr == NULL)
  {
    dprintf ("ARFS:AllocateSymbolicLinkFSO NULL string for symbolic name!\n");
    goto ErrorExit;
  }

  NewSymlinkFSOPntr =
    mallocDVP (DiskVolumePntr, sizeof (SymbolicLinkObjectRecord));
  if (NewSymlinkFSOPntr == NULL)
  {
    dprintf ("ARFS:AllocateSymbolicLinkFSO while doing \"%s\", "
      "ran out of memory!\n", TrueNamePntr);
    goto ErrorExit;
  }

  if (!InitialiseFSO (DiskVolumePntr, &NewSymlinkFSOPntr->fso,
  FSO_TYPE_SYMBOLIC_LINK, TrueNamePntr))
  {
    dprintf ("ARFS:AllocateSymbolicLinkFSO while doing \"%s\", unable to "
      "initialise the generic file system object!\n", TrueNamePntr);
    goto ErrorExit;
  }
  FSOInitialised = true;

  StringLength = strlen (SymbolicPathPntr);
  NewSymlinkFSOPntr->symbolicPathPntr =
    mallocDVP (DiskVolumePntr, StringLength + 1);
  if (NewSymlinkFSOPntr->symbolicPathPntr == NULL)
  {
    dprintf ("ARFS:AllocateSymbolicLinkFSO while doing \"%s\", unable to "
      "allocate space for symbolic path \"%s\"!\n",
      TrueNamePntr, SymbolicPathPntr);
    goto ErrorExit;
  }
  strcpy (NewSymlinkFSOPntr->symbolicPathPntr, SymbolicPathPntr);

  /* Successful! */
  goto NormalExit;

ErrorExit:
  if (FSOInitialised)
    UninitialiseFSO (DiskVolumePntr, &NewSymlinkFSOPntr->fso);
  if (NewSymlinkFSOPntr != NULL)
    freeDVP (DiskVolumePntr, NewSymlinkFSOPntr);
  NewSymlinkFSOPntr = NULL;
NormalExit:
  return NewSymlinkFSOPntr;
}



/******************************************************************************
 * Allocate and initialise an AGMS link file system object, pointing to the
 * given target file system object (can be NULL if you want to set it later, in
 * either case the caller has to add it to the target's parentLinkArray).
 * Returns the new FSO if successful, NULL if it ran out of memory or you
 * specified bad arguments.
 */

static AGMSLinkObjectPointer AllocateAGMSLinkFSO (
  DiskVolumePointer DiskVolumePntr,
  const char *TrueNamePntr,
  FileSystemObjectPointer TargetFSOPntr)
{
  bool                  FSOInitialised = false;
  AGMSLinkObjectPointer NewAGMSlinkFSOPntr = NULL;

  if (INVALID_FILE_SYSTEM_OBJECT (TargetFSOPntr))
  {
    dprintf ("ARFS:AllocateAGMSLinkFSO while making \"%s\", "
      "found out that target %s is invalid.\n",
      TrueNamePntr, FSOAsString (TargetFSOPntr));
    goto ErrorExit;
  }

  NewAGMSlinkFSOPntr =
    mallocDVP (DiskVolumePntr, sizeof (AGMSLinkObjectRecord));
  if (NewAGMSlinkFSOPntr == NULL)
  {
    dprintf ("ARFS:AllocateAGMSLinkFSO while doing \"%s\", "
      "ran out of memory!\n", TrueNamePntr);
    goto ErrorExit;
  }

  if (!InitialiseFSO (DiskVolumePntr, &NewAGMSlinkFSOPntr->fso,
  FSO_TYPE_AGMS_LINK, TrueNamePntr))
  {
    dprintf ("ARFS:AllocateAGMSLinkFSO while doing \"%s\", unable to "
      "initialise the generic file system object!\n", TrueNamePntr);
    goto ErrorExit;
  }
  FSOInitialised = true;

  NewAGMSlinkFSOPntr->targetFSOPntr = TargetFSOPntr;

  /* Successful! */
  goto NormalExit;

ErrorExit:
  if (FSOInitialised)
    UninitialiseFSO (DiskVolumePntr, &NewAGMSlinkFSOPntr->fso);
  if (NewAGMSlinkFSOPntr != NULL)
    freeDVP (DiskVolumePntr, NewAGMSlinkFSOPntr);
  NewAGMSlinkFSOPntr = NULL;
NormalExit:
  return NewAGMSlinkFSOPntr;
}



/******************************************************************************
 * Allocate and initialise an index file system object.  Returns the new FSO
 * if successful, NULL if it ran out of memory or you specified bad arguments.
 */

static IndexObjectPointer AllocateIndexFSO (
  DiskVolumePointer DiskVolumePntr,
  const char *TrueNamePntr,
  type_code KeyType,
  bool InterpretAsMultipleKeywords)
{
  bool               FSOInitialised = false;
  IndexObjectPointer NewIndexFSOPntr = NULL;
  int                TypeSize;

  TypeSize = AVLDupGetThingTypeSize (KeyType);
  if (TypeSize < 0)
  {
    dprintf ("ARFS:AllocateIndexFSO: Type code of %s isn't a recognized "
      "data type.  Can't make index \"%s\".\n",
      TypeCodeAsString (KeyType), TrueNamePntr);
    goto ErrorExit;
  }

  NewIndexFSOPntr = mallocDVP (DiskVolumePntr, sizeof (IndexObjectRecord));
  if (NewIndexFSOPntr == NULL)
  {
    dprintf ("ARFS:AllocateIndexFSO: While doing \"%s\", "
      "ran out of memory!\n", TrueNamePntr);
    goto ErrorExit;
  }

  if (!InitialiseFSO (DiskVolumePntr, &NewIndexFSOPntr->fso, FSO_TYPE_INDEX,
  TrueNamePntr))
  {
    dprintf ("ARFS:AllocateIndexFSO: While doing \"%s\", unable to "
      "initialise the generic file system object!\n", TrueNamePntr);
    goto ErrorExit;
  }
  FSOInitialised = true;

  NewIndexFSOPntr->keyType = KeyType;
  NewIndexFSOPntr->normalAttributeSize = TypeSize;
  NewIndexFSOPntr->interpretAsMultipleKeywords = InterpretAsMultipleKeywords;
  NewIndexFSOPntr->treePntr = AVLDupAllocTree (KeyType,
    DiskVolumePntr->mallocPntr, DiskVolumePntr->freePntr,
    DiskVolumePntr->volumePoolID);
  if (NewIndexFSOPntr->treePntr == NULL)
  {
    dprintf ("ARFS:AllocateIndexFSO: While doing \"%s\", unable to "
      "create a new AVL tree!\n", TrueNamePntr);
    goto ErrorExit;
  }

  /* Successful! */
  goto NormalExit;

ErrorExit:
  if (FSOInitialised)
    UninitialiseFSO (DiskVolumePntr, &NewIndexFSOPntr->fso);
  if (NewIndexFSOPntr != NULL)
    freeDVP (DiskVolumePntr, NewIndexFSOPntr);
  NewIndexFSOPntr = NULL;
NormalExit:
  return NewIndexFSOPntr;
}



/******************************************************************************
 * Deallocate this file system object, uninitialising the contents and freeing
 * all the storage used.  Operates in the reverse order of construction of the
 * object (does high level things first, then cleans up lower levels, much like
 * a C++ virtual destructor).  Works for all types of file system object.
 * Returns 0 if successful, an error code if something went wrong.
 */

static status_t DeallocateAnyFSO (
  DiskVolumePointer DiskVolumePntr,
  FileSystemObjectPointer FSOPntr)
{
  status_t ErrorCode;
  status_t ReturnCode = 0;

#if DPRINT_NON_ERROR_MESSAGES
  dprintf ("ARFS:DeallocateAnyFSO Deallocating %s.\n",
    FSOAsString (FSOPntr));
#endif

  if (FSOPntr->objectType == FSO_TYPE_FILE)
  {
    FileObjectPointer FileFSOPntr = (FileObjectPointer) FSOPntr;
    DeallocateFileContents (DiskVolumePntr, FileFSOPntr);
  }
  else if (FSOPntr->objectType == FSO_TYPE_DIRECTORY)
  {
    DirectoryObjectPointer DirectoryFSOPntr = (DirectoryObjectPointer) FSOPntr;

    AVLDupFreeTree (DirectoryFSOPntr->directory);
    DirectoryFSOPntr->directory = NULL;
  }
  else if (FSOPntr->objectType == FSO_TYPE_SYMBOLIC_LINK)
  {
    SymbolicLinkObjectPointer SymlinkFSOPntr;
    SymlinkFSOPntr = (SymbolicLinkObjectPointer) FSOPntr;
    if (SymlinkFSOPntr->symbolicPathPntr != NULL)
    {
      freeDVP (DiskVolumePntr, SymlinkFSOPntr->symbolicPathPntr);
      SymlinkFSOPntr->symbolicPathPntr = NULL;
    }
  }
  else if (FSOPntr->objectType == FSO_TYPE_AGMS_LINK)
  {
    AGMSLinkObjectPointer AGMSLinkFSOPntr;
    AGMSLinkFSOPntr = (AGMSLinkObjectPointer) FSOPntr;
    AGMSLinkFSOPntr->targetFSOPntr = NULL;
  }
  else if (FSOPntr->objectType == FSO_TYPE_INDEX)
  {
    IndexObjectPointer IndexFSOPntr = (IndexObjectPointer) FSOPntr;

    AVLDupFreeTree (IndexFSOPntr->treePntr);
    IndexFSOPntr->treePntr = NULL;
  }
  else
  {
    dprintf ("ARFS:DeallocateAnyFSO %s has an unknown "
      "object type of %d!\n", FSOAsString (FSOPntr), FSOPntr->objectType);
    ReturnCode = -1;
  }

  /* Clean up the basic file system object portion, and mark it as invalid. */

  ErrorCode = UninitialiseFSO (DiskVolumePntr, FSOPntr);
  if (ErrorCode < 0) ReturnCode = ErrorCode;

  /* And finally deallocate the whole record. */

  freeDVP (DiskVolumePntr, FSOPntr);

  return ReturnCode;
}



/******************************************************************************
 * Activate this file system object, changing it from a dormant state to one
 * ready for use - just allocate the semaphore if it isn't already there (it
 * should not be).  Returns 0 if successful, an error code if something went
 * wrong (file system object left in a deactivated state).
 */

static status_t ActivateFSO (FileSystemObjectPointer FSOPntr)
{
  status_t ErrorCode;

  if (FSOPntr->accessSemID != -1)
  {
    dprintf ("ARFS:ActivateFSO Potential Bug!  File system object %s is "
      "already active, the semaphore #%d already exists!\n",
      FSOAsString (FSOPntr), (int) FSOPntr->accessSemID);
    return 0;
  }

  ErrorCode = create_sem (MAX_SIMULTANEOUS_READERS, FSOPntr->trueNamePntr);
  if (ErrorCode < 0)
  {
    dprintf ("ARFS:ActivateFSO Unable to create access semaphore for "
      "%s, error code %d!\n",
      FSOAsString (FSOPntr), (int) ErrorCode);
    return ErrorCode;
  }

  FSOPntr->accessSemID = ErrorCode;

  /* Set the ownership to the kernel, so that the semaphore doesn't disappear
  when the caller dies. */

  ErrorCode = set_sem_owner (FSOPntr->accessSemID, B_SYSTEM_TEAM);
  if (ErrorCode < 0)
  {
    dprintf ("ARFS:ActivateFSO Unable to kernelize ownership of semaphore for "
      "%s, error code %d!\n",
      FSOAsString (FSOPntr), (int) ErrorCode);
    delete_sem (FSOPntr->accessSemID);
    FSOPntr->accessSemID = -1;
    return ErrorCode;
  }

  return 0;
}



/******************************************************************************
 * Deactivate this file system object - just free the semaphore.  Returns 0 if
 * successful, an error code if something went wrong.
 */

static status_t DeactivateFSO (FileSystemObjectPointer FSOPntr)
{
  status_t ErrorCode;

  if (FSOPntr->accessSemID < 0)
  {
    dprintf ("ARFS:DeactivateFSO Potential Bug!  File system object %s is "
      "already inactive, semaphore #%d (should be -1).\n",
      FSOAsString (FSOPntr), (int) FSOPntr->accessSemID);
    return 0;
  }

  ErrorCode = delete_sem (FSOPntr->accessSemID);
  FSOPntr->accessSemID = -1;

  if (ErrorCode < 0)
  {
    dprintf ("ARFS:DeactivateFSO Unable to delete access semaphore for "
      "file system object %s, error code %d!\n",
      FSOAsString (FSOPntr), (int) ErrorCode);
    return ErrorCode;
  }

  return 0;
}



/******************************************************************************
 * Add a new AGMS link reference to the file system object's array of parent
 * links.  That's all it does, no checking validity, locking, or updating of
 * indices etc.  Internally the links are stored in a dynamically allocated
 * array (plus a NULL pointer at the end of the array to mark the end).  It's
 * done this simple way because there usually aren't any links, and I don't
 * expect to see more than a dozen links to a particular file.  If there are
 * more, then we'd need a full featured balanced sorted AVL tree to hold them.
 * Returns TRUE if successful, FALSE if it runs out of memory or the AGMS link
 * is already listed.
 */

static bool AddAGMSLinkReferenceToFSO (
  DiskVolumePointer DiskVolumePntr,
  FileSystemObjectPointer FSOPntr,
  AGMSLinkObjectPointer NewLinkPntr)
{
  AGMSLinkObjectPointer *LinkPntrPntr;
  int                    NumberOfLinks;

  /* Have to add to the parentLinkArray, which may or may not exist.  Find out
  how many links we will be needing. */

  NumberOfLinks = 1; /* For the new one we are adding. */
  if (FSOPntr->parentLinkArray != NULL)
  {
    /* Add the count of the existing links, and check that it isn't trying to
    add the same one twice. */

    for (LinkPntrPntr = FSOPntr->parentLinkArray;
    *LinkPntrPntr != NULL;
    LinkPntrPntr++)
    {
      if (*LinkPntrPntr == NewLinkPntr)
      {
        dprintf ("ARFS:AddAGMSLinkReferenceToFSO Bug!  %s is "
          "already listed as a parent link of %s.\n",
          FSOAsString (&NewLinkPntr->fso),
          FSOAsString (FSOPntr));
        return false;
      }
      NumberOfLinks++;
    }
  }

  /* Allocate the new array of refering AGMS links, plus space for a NULL
  pointer at the end. */

  LinkPntrPntr = mallocDVP (DiskVolumePntr,
    (NumberOfLinks + 1) * sizeof (AGMSLinkObjectPointer));
  if (LinkPntrPntr == NULL)
  {
    dprintf ("ARFS:AddAGMSLinkReferenceToFSO unable to add "
      "%s to AGMS link array of %s, out of memory for link "
      "array with %d entries.\n",
      FSOAsString (&NewLinkPntr->fso),
      FSOAsString (FSOPntr),
      NumberOfLinks);
    return false;
  }

  /* Copy the previous links, if any. */

  if (NumberOfLinks >= 2)
    memcpy (LinkPntrPntr, FSOPntr->parentLinkArray,
    (NumberOfLinks - 1) * sizeof (AGMSLinkObjectPointer));

  /* Add our new link, and the NULL marker at the end. */

  LinkPntrPntr[NumberOfLinks - 1] = NewLinkPntr;
  LinkPntrPntr[NumberOfLinks] = NULL;

  /* Replace the old link array with the new one. */

  if (FSOPntr->parentLinkArray != NULL)
    freeDVP (DiskVolumePntr, FSOPntr->parentLinkArray);
  FSOPntr->parentLinkArray = LinkPntrPntr;

  return true;
}



/******************************************************************************
 * Remove an AGMS link reference from a file system object.  That's all it
 * does, no deallocation, checking validity, locking, updating of indices etc.
 * Assumes that there is only one instance of that link in the array (if there
 * are more then it will remove them too but waste a bit of space and perhaps
 * leave behind an empty array of links rather than no array).  Returns TRUE if
 * the link was removed (works even if we run out of memory), FALSE if it
 * couldn't find it.
 */

static bool RemoveAGMSLinkReferenceFromFSO (
  DiskVolumePointer DiskVolumePntr,
  FileSystemObjectPointer FSOPntr,
  AGMSLinkObjectPointer OldLinkPntr)
{
  AGMSLinkObjectPointer  CurrentLinkPntr;
  AGMSLinkObjectPointer *DestinationLinkPntrPntr;
  bool                   FoundInExistingLinks;
  AGMSLinkObjectPointer *NewArrayLinkPntrPntr;
  int                    NumberOfLinks;
  AGMSLinkObjectPointer *SourceLinkPntrPntr;

  if (OldLinkPntr == NULL)
  {
    dprintf ("ARFS:RemoveAGMSLinkReferenceFromFSO Bug!  Can't remove NULL "
      "from %s.\n", FSOAsString (FSOPntr));
    return false;
  }

  /* Count up the current number of links, and find the one to be removed. */

  NumberOfLinks = 0;
  FoundInExistingLinks = false;
  if (FSOPntr->parentLinkArray != NULL)
  {
    for (SourceLinkPntrPntr = FSOPntr->parentLinkArray;
    *SourceLinkPntrPntr != NULL;
    SourceLinkPntrPntr++)
    {
      if (*SourceLinkPntrPntr == OldLinkPntr)
        FoundInExistingLinks = true;
      NumberOfLinks++;
    }
  }

  if (!FoundInExistingLinks)
    return false; /* Didn't find the link, nothing to do. */

  NumberOfLinks--; /* One less for the one we are removing. */

  if (NumberOfLinks > 0)
  {
    /* There will be some link references remaining after deleting one.
    Allocate the new, smaller, links array and copy the undeleted ones into
    it.  If we run out of memory, do the copy in-place and waste a bit of space
    at the end of the existing array. */

    NewArrayLinkPntrPntr = mallocDVP (DiskVolumePntr,
      (NumberOfLinks + 1) * sizeof (AGMSLinkObjectPointer));
    if (NewArrayLinkPntrPntr != NULL)
      DestinationLinkPntrPntr = NewArrayLinkPntrPntr;
    else
    {
      dprintf ("ARFS:RemoveAGMSLinkReferenceFromFSO while removing "
        "%s from %s AGMS link array, ran out of memory for link "
        "array with %d links.  Will reuse existing link "
        "array, wasting a few bytes.\n",
        FSOAsString (&OldLinkPntr->fso),
        FSOAsString (FSOPntr),
        NumberOfLinks);
      DestinationLinkPntrPntr = FSOPntr->parentLinkArray;
    }
    SourceLinkPntrPntr = FSOPntr->parentLinkArray;

    while ((CurrentLinkPntr = *SourceLinkPntrPntr) != NULL)
    {
      if (CurrentLinkPntr != OldLinkPntr)
        *DestinationLinkPntrPntr++ = CurrentLinkPntr;
      SourceLinkPntrPntr++;
    }
    *DestinationLinkPntrPntr = NULL; /* Mark end of the array with NULL. */

    if (NewArrayLinkPntrPntr != NULL)
    {
      freeDVP (DiskVolumePntr, FSOPntr->parentLinkArray);
      FSOPntr->parentLinkArray = NewArrayLinkPntrPntr;
    }
  }
  else /* No extra directories needed, deallocate old array if any. */
  {
    if (FSOPntr->parentLinkArray != NULL)
      freeDVP (DiskVolumePntr, FSOPntr->parentLinkArray);
    FSOPntr->parentLinkArray = NULL;
  }

  return true;
}



/******************************************************************************
 * Adds a file system object to a directory.  Doesn't do locking etc.  If it is
 * already there, returns an error code.  The target FSO is added to the
 * directory's tree, using its true name.  Then the directory is made the main
 * parent directory of the target FSO (if it already has a main parent, then it
 * isn't fresh and you will get an error message).  Note: you can get weird
 * situations where a file is deleted, but still open, and gets relinked to a
 * directory, kind of undeleting it.  Make sure indexing of attributes
 * exist/don't exist as appropriate.  Returns a negative error code if it
 * fails, 0 if successful.
 */

static status_t AddFreshFSOToDirectory (
  DiskVolumePointer DiskVolumePntr,
  FileSystemObjectPointer AddedFSOPntr,
  DirectoryObjectPointer DirectoryFSOPntr)
{
  AVLDupThingRecord      Key;
  status_t               ReturnCode = -1;

  Key.longStringThing.stringPntr = AddedFSOPntr->trueNamePntr;
  Key.longStringThing.isLongString = true;

  /* Make sure the target is fresh - no current main parent directory.  The
  root of the file system always has NULL as a parent, since its parent is
  outside the file system. */

  if (AddedFSOPntr->mainParentDirectory != NULL ||
  AddedFSOPntr == &DiskVolumePntr->rootFSOPntr->fso)
  {
    dprintf ("ARFS:AddFreshFSOToDirectory %s already has a "
      "main parent of %s, need to have none.\n", FSOAsString (AddedFSOPntr),
      FSOAsString (&AddedFSOPntr->mainParentDirectory->fso));
    ReturnCode = EBADF;
    goto ErrorExit;
  }

  /* See if the name is already in use in the directory.  This is also
  sneakily used by mkdir to see if it is adding a duplicate name. */

  if (AVLDupFind (DirectoryFSOPntr->directory, &Key) != 0)
  {
    dprintf ("ARFS:AddFreshFSOToDirectory name \"%s\" is already in use in "
      "%s.\n", AddedFSOPntr->trueNamePntr,
      FSOAsString (&DirectoryFSOPntr->fso));
    ReturnCode = EEXIST;
    goto ErrorExit;
  }

  /* Add the new name and pointer to the parent directory's list. */

  if (AVLDupAdd (DirectoryFSOPntr->directory, &Key,
  (unsigned int) AddedFSOPntr) < 0)
  {
    dprintf ("ARFS:AddFreshFSOToDirectory unable to add target %s to "
      "%s, out of memory for tree nodes.\n",
      FSOAsString (AddedFSOPntr), FSOAsString (&DirectoryFSOPntr->fso));
    goto ErrorExit;
  }

  /* Make the parent directory the target's main parent directory. */

  AddedFSOPntr->mainParentDirectory = DirectoryFSOPntr;

  ReturnCode = 0;

ErrorExit:
  return ReturnCode;
}



/******************************************************************************
 * Removes a file system object from a directory.  Doesn't do locking etc.
 * Mostly used for undoing partial additions to a directory.  If it isn't
 * there, or the directory isn't the main parent directory then this function
 * returns an error code.  The target FSO is removed from the directory's tree,
 * using its true name.  The main parent directory of the target is NULLified
 * too.
 */

static status_t RemoveFSOFromDirectory (
  DiskVolumePointer DiskVolumePntr,
  FileSystemObjectPointer RemovedFSOPntr,
  DirectoryObjectPointer DirectoryFSOPntr)
{
  AVLDupThingRecord      Key;
  status_t               ReturnCode = -1;

  if (RemovedFSOPntr->mainParentDirectory != DirectoryFSOPntr)
  {
    dprintf ("ARFS:RemoveFSOFromDirectory %s doesn't have %s as its "
      "main parent directory.\n",
      FSOAsString (RemovedFSOPntr),
      FSOAsString (&DirectoryFSOPntr->fso));
    ReturnCode = EINVAL;
    goto ErrorExit;
  }

  /* Delete the name in the directory. */

  Key.longStringThing.stringPntr = RemovedFSOPntr->trueNamePntr;
  Key.longStringThing.isLongString = true;

  if (!AVLDupDelete (DirectoryFSOPntr->directory,
  &Key, (unsigned int) RemovedFSOPntr))
  {
    dprintf ("ARFS:RemoveFSOFromDirectory name \"%s\" is not in "
      "%s.\n", RemovedFSOPntr->trueNamePntr,
      FSOAsString (&DirectoryFSOPntr->fso));
    ReturnCode = ENOENT;
    goto ErrorExit;
  }

  RemovedFSOPntr->mainParentDirectory = NULL;
  ReturnCode = 0;

ErrorExit:
  return ReturnCode;
}



/******************************************************************************
 * Gets the indices directory and does a read lock on it.  If the caller
 * already has gotten and locked AlreadyLockedFSOPntr[AB] and it happens to be
 * the same as the indices directory (or the root directory too), it doesn't do
 * any locking (so that attributes can be added to the indices directory itself
 * without a double-lock deadlock) but still does a get_vnode on the indices
 * directory.  If the directory can't be found, returns NULL.
 */

static DirectoryObjectPointer GetAndLockIndicesDirectory (
  DiskVolumePointer DiskVolumePntr,
  FileSystemObjectPointer AlreadyLockedFSOPntrA,
  FileSystemObjectPointer AlreadyLockedFSOPntrB)
{
  status_t                ErrorCode;
  bool                    GetVNodeDone = false;
  DirectoryObjectPointer  IndicesDirPntr = NULL;
  AVLDupThingRecord       Key;
  bool                    RootDirectoryReadLockDone = false;
  unsigned int            TempUINT;
  void                   *VoidPntr;

  /* Read lock the root directory only while examining its contents. */

  if (AlreadyLockedFSOPntrA != &DiskVolumePntr->rootFSOPntr->fso &&
  AlreadyLockedFSOPntrB != &DiskVolumePntr->rootFSOPntr->fso)
  {
    ErrorCode = ReadLock (&DiskVolumePntr->rootFSOPntr->fso, false);
    if (ErrorCode < 0)
    {
      dprintf ("ARFS:GetAndLockIndicesDirectory: Unable to lock root "
        "directory, code %d.\n", (int) ErrorCode);
      goto ErrorExit;
    }
    RootDirectoryReadLockDone = true;
  }

  Key.longStringThing.stringPntr = (char *) g_IndicesDirectoryName;
  Key.longStringThing.isLongString = true;
  TempUINT = AVLDupFind (DiskVolumePntr->rootFSOPntr->directory, &Key);
  IndicesDirPntr = (DirectoryObjectPointer) TempUINT;

  /* Finished with the root directory, don't need it any more. */

  if (RootDirectoryReadLockDone)
    ReadUnlock (&DiskVolumePntr->rootFSOPntr->fso);

  /* If no such directory.  An allowable not really an error situation. */

  if (IndicesDirPntr == NULL)
    goto ErrorExit;

  /* See if the indices directory is really a directory. */

  ErrorCode = get_vnode (DiskVolumePntr->nameSpaceID,
    (unsigned int) IndicesDirPntr, &VoidPntr);
  if (ErrorCode < 0)
  {
    dprintf ("ARFS:GetAndLockIndicesDirectory: get_vnode failed for %s, "
      "code %d!\n", FSOAsString (&IndicesDirPntr->fso), (int) ErrorCode);
    goto ErrorExit;
  }
  GetVNodeDone = true;

  if (VoidPntr != IndicesDirPntr)
  {
    dprintf ("ARFS:GetAndLockIndicesDirectory: Bug!  get_vnode returned "
      "wrong pointer $%X for %s.\n",
      (unsigned int) VoidPntr, FSOAsString (&IndicesDirPntr->fso));
    goto ErrorExit;
  }

  if (INVALID_FILE_SYSTEM_OBJECT (&IndicesDirPntr->fso))
  {
    dprintf ("ARFS:GetAndLockIndicesDirectory: Bug!  Found an entry for "
      "for \"%s\" in %s, but it is a corrupt %s.\n",
      g_IndicesDirectoryName, FSOAsString (&DiskVolumePntr->rootFSOPntr->fso),
      FSOAsString (&IndicesDirPntr->fso));
    goto ErrorExit;
  }

  if (IndicesDirPntr->fso.objectType != FSO_TYPE_DIRECTORY)
  {
    dprintf ("ARFS:GetAndLockIndicesDirectory: While looking for the indices "
      "directory \"%s\" in %s, found %s, which is not a directory!\n",
      g_IndicesDirectoryName, FSOAsString (&DiskVolumePntr->rootFSOPntr->fso),
      FSOAsString (&IndicesDirPntr->fso));
    goto ErrorExit;
  }

  /* Fail if the user isn't allowed to walk or read the indices directory. */

  ErrorCode =
    PermissionAllowed (DiskVolumePntr, &IndicesDirPntr->fso, X_OK | R_OK);
  if (ErrorCode != 0)
  {
    dprintf ("ARFS:GetAndLockIndicesDirectory: %s doesn't have read or "
      "execute access permission for contents (need the x or r flag).\n",
      FSOAsString (&IndicesDirPntr->fso));
    goto ErrorExit;
  }

  /* Get a read lock on the indices directory, if it isn't already locked. */

  if (AlreadyLockedFSOPntrA != &IndicesDirPntr->fso &&
  AlreadyLockedFSOPntrB != &IndicesDirPntr->fso)
  {
    ErrorCode = ReadLock (&IndicesDirPntr->fso, false);
    if (ErrorCode != 0)
    {
      dprintf ("ARFS:GetAndLockIndicesDirectory: ReadLock failed for %s, "
        "code %d.\n", FSOAsString (&IndicesDirPntr->fso), (int) ErrorCode);
      goto ErrorExit;
    }
  }

  /* Successful!  Leave the get_vnode outstanding, and maybe a read lock. */

  return IndicesDirPntr;

ErrorExit: /* Only errors go through here. */
  if (GetVNodeDone)
    put_vnode (DiskVolumePntr->nameSpaceID, (unsigned int) IndicesDirPntr);
  return NULL;
}



/******************************************************************************
 * Puts the indices directory and undoes the read lock on it.  If the caller
 * already has gotten and locked AlreadyLockedFSOPntr* and it happens to be the
 * same as the indices directory, it doesn't do any unlocking.
 */

static void PutIndicesDirectory (
  DiskVolumePointer DiskVolumePntr,
  DirectoryObjectPointer IndicesDirPntr,
  FileSystemObjectPointer AlreadyLockedFSOPntrA,
  FileSystemObjectPointer AlreadyLockedFSOPntrB)
{
  if (IndicesDirPntr == NULL)
    return;

  if (&IndicesDirPntr->fso != AlreadyLockedFSOPntrA &&
  &IndicesDirPntr->fso != AlreadyLockedFSOPntrB)
    ReadUnlock (&IndicesDirPntr->fso);
  put_vnode (DiskVolumePntr->nameSpaceID, (unsigned int) IndicesDirPntr);
}



/******************************************************************************
 * This is a utility function for finding a particular index object, given the
 * name of the attribute being indexed.  It looks in the given directory for it
 * (if you specified NULL then it finds and uses the standard index directory)
 * and returns the FSOPntr if it finds it, after get_vnodeing it (so that it
 * doesn't get deleted and you can lock it).  Sets the output to NULL and
 * returns an error code if it fails.
 */

static status_t GetIndexFSO (
  DiskVolumePointer DiskVolumePntr,
  DirectoryObjectPointer IndexDirectoryFSOPntr,
  const char *IndexNamePntr,
  IndexObjectPointer *ReturnedIndexFSOPntr)
{
  FileSystemObjectPointer ChildFSOPntr;
  status_t                ErrorCode;
  AVLDupThingRecord       Key;
  DirectoryObjectPointer  ParentFSOPntr;
  status_t                ReturnCode = -1;
  unsigned int            TempUINT;
  void                   *VoidPntr;

  *ReturnedIndexFSOPntr = NULL;

  if (IndexDirectoryFSOPntr != NULL)
    ParentFSOPntr = IndexDirectoryFSOPntr;
  else
  {
    ParentFSOPntr = GetAndLockIndicesDirectory (DiskVolumePntr, NULL, NULL);
    if (ParentFSOPntr == NULL)
    {
      dprintf ("ARFS:GetIndexFSO: Unable to find indices directory.\n");
      ReturnCode = ENOTDIR;
      goto ErrorExit;
    }
  }

  /* Find the named index inside the indices directory. */

  Key.longStringThing.stringPntr = (char *) IndexNamePntr;
  Key.longStringThing.isLongString = true;
  TempUINT = AVLDupFind (ParentFSOPntr->directory, &Key);
  ChildFSOPntr = (FileSystemObjectPointer) TempUINT;

  if (INVALID_FILE_SYSTEM_OBJECT (ChildFSOPntr) ||
  ChildFSOPntr->objectType != FSO_TYPE_INDEX)
  {
    dprintf ("ARFS:GetIndexFSO: unable to find %s index in %s.\n",
      IndexNamePntr, FSOAsString (&ParentFSOPntr->fso));
    ReturnCode = ENOENT;
    goto ErrorExit;
  }

  /* GetVnode the index so that it becomes lockable and doesn't get deleted. */

  ErrorCode = get_vnode (DiskVolumePntr->nameSpaceID,
    (unsigned int) ChildFSOPntr, &VoidPntr);
  if (ErrorCode != 0)
  {
    dprintf ("ARFS:GetIndexFSO: get_vnode failed for "
      "index \"%s\" in %s (trying to load %s), code %d.\n",
      IndexNamePntr, FSOAsString (&ParentFSOPntr->fso),
      FSOAsString (ChildFSOPntr), (int) ErrorCode);
    ReturnCode = ErrorCode;
    goto ErrorExit;
  }

  /* Success! */

  ReturnCode = 0;
  *ReturnedIndexFSOPntr = (IndexObjectPointer) ChildFSOPntr;

ErrorExit:
#if DPRINT_NON_ERROR_MESSAGES
  dprintf ("ARFS:GetIndexFSO exited, looked for \"%s\", return code %d, "
    "returned %s.\n", IndexNamePntr, (int) ReturnCode,
    FSOAsString (&(*ReturnedIndexFSOPntr)->fso));
#endif
  if (IndexDirectoryFSOPntr == NULL && ParentFSOPntr != NULL)
    PutIndicesDirectory (DiskVolumePntr, ParentFSOPntr, NULL, NULL);
  return ReturnCode;
}



/******************************************************************************
 * Data structures used by the query code.  First come the things needed for
 * analysing the query string - basically your standard recursive descent
 * parser with a tokenizer in front.
 *
 * Query implementation: First convert the user's query string into a list of
 * tokens and argument strings.  Then build a parse tree which interprets the
 * tokens as tests to be done in a certain order.  An evaluation subroutine is
 * then used for comparing a candidate file against the parse tree, just
 * reading its attributes as needed by comparison operations.  This is separate
 * from the index system, so we can do queries for attributes which aren't in
 * indices.  Live queries call the evaluation subroutine for any files which
 * change their metadata (name, size, date) and/or attributes and report the
 * file if the evaluation result is different than the evaluation result before
 * the metadata change.  Static queries do an iteration over a particular index
 * and run the resulting files through the evaluation subroutine.  The
 * optimization of static queries is a matter of picking which index and the
 * range limits for the values to iterate over (avoiding the BFS excessive
 * complexity where the index traversal and parse tree traversal are mushed
 * together).
 */

/* These structures represent the user's query string in a more convenient
tokenized format.  White space has been removed, and strings (quoted or not)
extracted and copied.  Other recognized symbols (mostly comparison operators)
are also converted into a single token.  The result of the conversion is a bit
of header info, an array of tokens and some string storage space all in one
record. */

typedef enum TokenCodeEnum
{
  /* The relational operators come first.  The order is defined so that the
  logical opposite of operator X is (TC_LAST_RELATION - X). */
  TC_EQUAL = 0,
  TC_GREATER_THAN,
  TC_LESS_THAN,
  TC_ATTRIBUTE_EXISTS,
  TC_ATTRIBUTE_MISSING,
  TC_GREATER_THAN_OR_EQUAL,
  TC_LESS_THAN_OR_EQUAL,
  TC_NOT_EQUAL,
  TC_LAST_RELATION = TC_NOT_EQUAL,

  /* No particular order for the remaining token codes. */
  TC_STRING,
  TC_LEFT_PARENTHESIS,
  TC_RIGHT_PARENTHESIS,
  TC_AND,
  TC_OR,
  TC_NOT,
  TC_END_OF_QUERY,
  TC_MAX
} TokenCode;


typedef struct TokenStruct
{
  TokenCode tokenCode;
    /* Identifies what the token is. */

  char *stringPntr;
    /* Extra string info associated with the token.  Currently only used by
    TC_STRING for the cleaned up string (quotes removed, escaped quotes
    converted to just a quote).  Points to a string stored somewhere inside the
    TokenizedQueryRecord. */

} TokenRecord, *TokenPointer;


typedef struct TokenizedQueryStruct
{
  int numberOfTokens;
    /* A count of the number of tokens actually in tokenArray. */

  int stringSpaceSize;
    /* Following the full tokenArray (size of the token area actually used is
    numberOfTokens * sizeof (TokenRecord)) is a hunk of memory used for string
    storage by string tokens.  stringSpaceSize is the number of bytes in that
    area (includes NULs at end of string). */

  TokenRecord tokenArray [0];
    /* Marks the start of the array of tokens.  The actual array size is
    specified by numberOfTokens. */

  /* char stringArray [] is a kind of variable position array, it starts at
  address &tokenArray[numberOfTokens] and contains the strings used by the
  string tokens. */

  /* The token array and string data extends past the end of this record, which
  should be allocated to the required size, not just sizeof
  (TokenizedQueryRecord). */

} TokenizedQueryRecord, *TokenizedQueryPointer;


/* The parse tree defines the tests to be done to decide whether a file matches
the query or does not.  A node in the tree defines a logical subpart of the
test.  The query string eventually ends up as a tree of these nodes, with
computational nodes (&& and || operations) having child nodes for their inputs.
Leaf nodes are the actual attribute comparison operations. */

typedef enum ParseNodeOperationEnum
{
  PNO_AND = 0,
  PNO_OR,
  PNO_COMPARE,
  PNO_MAX
} ParseNodeOperation;

typedef struct ParseNodeStruct ParseNodeRecord, *ParseNodePointer;

struct ParseNodeStruct
{
  ParseNodeOperation operation;
    /* The kind of operation this parse node does. */

  union OperationSpecificUnion
  {
    struct ChildrenStruct /* Used by AND and OR operations. */
    {
      ParseNodePointer leftChild;
      ParseNodePointer rightChild;
    } children;

    struct CompareStruct /* Obviously used by COMPARE operations only. */
    {
      TokenCode relation;
        /* Less than, greater than, and so on. */

      const char *attributeNamePntr;
        /* Name of the attribute to test, copied from the query string.
        Actually points to the string stored in the TokenizedQueryRecord. */

      IndexObjectPointer indexFSOPntr;
        /* If there is an index with the same name, a pointer to it will be
        stored here, and the index is left in a get_vnoded, but unlocked state.
        NULL if no matching index exists.  Only used during query
        initialisation (the cached values stored here are used during
        evaluation, thus avoiding deadlock problems). */

      type_code indexValueType;
        /* The type for the value, taken from the index.  Zero if none. */

      int indexValueNormalSize;
        /* The number of bytes used for that type of data, 0 if it is a
        variable length string. */

      bool multipleKeywords;
        /* If the index is a keyword index, then this gets set to TRUE when the
        query is built.  Then later on, when examining files, it can decide to
        split the attribute value into words for the query comparison, and
        consider the comparison result to be true if any of the words match. */

      const char *valueStringPntr;
        /* Value as a string, points to the string stored in the
        TokenizedQueryRecord.  Since the kind of attribute can be specified
        differently for each file, it may need to be converted for every
        comparison. */

      AVLDupThingRecord valueConverted;
        /* The value string converted to an actual comparable value, usually a
        number of some kind.  If it's a string (GetComparisonFunctionForType
        returns zero) then the conversion is done without allocating any extra
        memory (the thing will point to the string in the token). */

      AVLDupComparisonFunctionPointer valueComparisonFunction;
        /* Points to a function which can compare values of type
        valueConvertedToType.  NULL if no such function exists. */

      type_code valueConvertedToType;
        /* The type of thing that the value has been last converted to.
        Initially set to zero to show it hasn't been converted.  If the type
        needed for a comparison hasn't changed, you can reuse valueConverted to
        save a bit of time. */

      int valueNormalSize;
        /* How big is the attribute data that corresponds to
        valueConvertedToType?  This field is the number of bytes used for that
        type of data, 0 if it is a variable length string. */

    } compare;

  } opSpecific;

};


/* This data structure is used for storing all the extra info needed to
describe a query (parse tree, tokens, user parameters, etc).  Note that the
current state (reading position for read_query) is stored in a generic
directory cookie, which has a pointer to this structure. */

struct QueryStateStruct
{
  TokenizedQueryPointer tokenizedQueryPntr;
    /* Points to the data structure storing the tokenized form of the query.
    This also includes a copy of strings derived from parts of the user's query
    (attribute names and comparison values), which the parse tree nodes also
    use. */

  ParseNodePointer parseTreePntr;
    /* Points to the root of the parse tree, or NULL if none. */

  QueryStatePointer nextLiveQueryPntr;
    /* Links together the live queries in the global live query list.  NULL if
    this is the last one in the list. */

  port_id liveQueryPortID;
    /* The port where live query notification messages should be sent.  Set to
    -1 if it isn't a live query. */

  long callbackToken;
    /* The user provided value which is passed back in the notification
    messages for live queries.  It lets them identify which query sent the
    message, or provides some other auxiliary data that they want associated
    with the query. */
};



/******************************************************************************
 * This function does the work of tokenizing a string.  If JustFindSizes is
 * TRUE then it finds out the number of tokens and amount of string space
 * needed - essentially doing the tokenization but not saving any data.  When
 * JustFindSizes is FALSE then it assumes that TokenizedQueryPntr points to a
 * record which has been allocated large enough to hold the results, and that
 * the numberOfTokens field is correct.
 */

static void TokenizeOrCount (
  const char *StringPntr,
  TokenizedQueryPointer TokenizedQueryPntr,
  bool JustFindSizes)
{
  char         Letter;
  char        *StringStoragePntr = NULL;
  TokenPointer TokenStoragePntr = NULL;

  if (JustFindSizes)
  {
    TokenizedQueryPntr->numberOfTokens = 0;
    TokenizedQueryPntr->stringSpaceSize = 0;
  }
  else
  {
    TokenStoragePntr = TokenizedQueryPntr->tokenArray + 0;
    StringStoragePntr = (char *)
      (TokenizedQueryPntr->tokenArray + TokenizedQueryPntr->numberOfTokens);
  }

  while ((Letter = *StringPntr) != 0)
  {
    /* Analyse the next few letters to see what the next token is.  Look for
    exclusive characters, or if it is ambiguous then check in order of longer
    to shorter matches, so we detect a "!=" symbol before we try just a plain
    "!" symbol. */

    if (isspace (Letter))
      StringPntr++; /* Just skip over white space, no token for it. */
    else if (Letter == '&' && StringPntr[1] == '&')
    {
      if (JustFindSizes)
        TokenizedQueryPntr->numberOfTokens++;
      else
        (TokenStoragePntr++)->tokenCode = TC_AND;
      StringPntr += 2;
    }
    else if (Letter == '|' && StringPntr[1] == '|')
    {
      if (JustFindSizes)
        TokenizedQueryPntr->numberOfTokens++;
      else
        (TokenStoragePntr++)->tokenCode = TC_OR;
      StringPntr += 2;
    }
    else if (Letter == '!' && StringPntr[1] == '=')
    {
      if (JustFindSizes)
        TokenizedQueryPntr->numberOfTokens++;
      else
        (TokenStoragePntr++)->tokenCode = TC_NOT_EQUAL;
      StringPntr += 2;
    }
    else if (Letter == '!')
    {
      if (JustFindSizes)
        TokenizedQueryPntr->numberOfTokens++;
      else
        (TokenStoragePntr++)->tokenCode = TC_NOT;
      StringPntr++;
    }
    else if (Letter == '(')
    {
      if (JustFindSizes)
        TokenizedQueryPntr->numberOfTokens++;
      else
        (TokenStoragePntr++)->tokenCode = TC_LEFT_PARENTHESIS;
      StringPntr++;
    }
    else if (Letter == ')')
    {
      if (JustFindSizes)
        TokenizedQueryPntr->numberOfTokens++;
      else
        (TokenStoragePntr++)->tokenCode = TC_RIGHT_PARENTHESIS;
      StringPntr++;
    }
    else if (Letter == '<' && StringPntr[1] == '=')
    {
      if (JustFindSizes)
        TokenizedQueryPntr->numberOfTokens++;
      else
        (TokenStoragePntr++)->tokenCode = TC_LESS_THAN_OR_EQUAL;
      StringPntr += 2;
    }
    else if (Letter == '<' && StringPntr[1] == '>')
    {
      if (JustFindSizes)
        TokenizedQueryPntr->numberOfTokens++;
      else
        (TokenStoragePntr++)->tokenCode = TC_ATTRIBUTE_EXISTS;
      StringPntr += 2;
    }
    else if (Letter == '<')
    {
      if (JustFindSizes)
        TokenizedQueryPntr->numberOfTokens++;
      else
        (TokenStoragePntr++)->tokenCode = TC_LESS_THAN;
      StringPntr++;
    }
    else if (Letter == '>' && StringPntr[1] == '=')
    {
      if (JustFindSizes)
        TokenizedQueryPntr->numberOfTokens++;
      else
        (TokenStoragePntr++)->tokenCode = TC_GREATER_THAN_OR_EQUAL;
      StringPntr += 2;
    }
    else if (Letter == '>' && StringPntr[1] == '<')
    {
      if (JustFindSizes)
        TokenizedQueryPntr->numberOfTokens++;
      else
        (TokenStoragePntr++)->tokenCode = TC_ATTRIBUTE_MISSING;
      StringPntr += 2;
    }
    else if (Letter == '>')
    {
      if (JustFindSizes)
        TokenizedQueryPntr->numberOfTokens++;
      else
        (TokenStoragePntr++)->tokenCode = TC_GREATER_THAN;
      StringPntr++;
    }
    else if (Letter == '=' && StringPntr[1] == '=')
    {
      if (JustFindSizes)
        TokenizedQueryPntr->numberOfTokens++;
      else
        (TokenStoragePntr++)->tokenCode = TC_EQUAL;
      StringPntr += 2;
    }
    else if (Letter == '=')
    {
      if (JustFindSizes)
        TokenizedQueryPntr->numberOfTokens++;
      else
        (TokenStoragePntr++)->tokenCode = TC_EQUAL;
      StringPntr++;
    }
    else if (Letter == '"')
    {
      /* A quoted string.  The contents are everything until the next quote or
      end of string, and convert backslash quote marks to just a quote.  Also
      copies the resulting string into the string storage space. */

      if (JustFindSizes)
        TokenizedQueryPntr->numberOfTokens++;
      else
      {
        TokenStoragePntr->tokenCode = TC_STRING;
        TokenStoragePntr->stringPntr = StringStoragePntr;
        TokenStoragePntr++;
      }

      StringPntr++; /* Skip over the starting quote. */
      while (*StringPntr != '"' && *StringPntr != 0)
      {
        if (*StringPntr == '\\' && StringPntr[1] == '"')
          /* Throw away the backslash escape character, but only if the next
          one is a quote since we have plain backslashes that have to get
          through to pattern matching later on. */
          StringPntr++;

        if (JustFindSizes)
          TokenizedQueryPntr->stringSpaceSize++;
        else
          *StringStoragePntr++ = *StringPntr;
        StringPntr++;
      }

      /* Close off the end of the quoted string. */

      if (JustFindSizes)
        TokenizedQueryPntr->stringSpaceSize++;
      else
        *StringStoragePntr++ = 0;

      if (*StringPntr == '"')
        StringPntr++; /* Skip over the trailing quote. */
    }
    else /* An unquoted string. */
    {
      if (JustFindSizes)
        TokenizedQueryPntr->numberOfTokens++;
      else
      {
        TokenStoragePntr->tokenCode = TC_STRING;
        TokenStoragePntr->stringPntr = StringStoragePntr;
        TokenStoragePntr++;
      }

      /* Accept first letter no matter what it is. */

      if (JustFindSizes)
        TokenizedQueryPntr->stringSpaceSize++;
      else
        *StringStoragePntr++ = *StringPntr;
      StringPntr++;

      /* Absorb all letters until a character which might be part of the next
      symbol is encountered.  Only look ahead at single characters, which is
      what BFS does (rather than looking ahead to see if the "=" is followed by
      another "=" to make "==").  So, stop at ()!|&<>= and white space.  Yes,
      quotes go into the unquoted string. */

      while ((Letter = *StringPntr) != 0)
      {
        if (Letter == '(' || Letter == ')' || Letter == '!' || Letter == '|' ||
        Letter == '&' || Letter == '<' || Letter == '>' || Letter == '=' ||
        isspace (Letter))
          break;
        if (JustFindSizes)
          TokenizedQueryPntr->stringSpaceSize++;
        else
          *StringStoragePntr++ = *StringPntr;
        StringPntr++;
      }

      if (JustFindSizes)
        TokenizedQueryPntr->stringSpaceSize++;
      else
        *StringStoragePntr++ = 0;
    }
  }

  /* Finally stick on the end of query marker, much like an end of string NUL.
  It makes parsing easier (don't need to check token indices). */

  if (JustFindSizes)
    TokenizedQueryPntr->numberOfTokens++;
  else
  {
    TokenStoragePntr->tokenCode = TC_END_OF_QUERY;
    TokenStoragePntr++;
    if (TokenStoragePntr - TokenizedQueryPntr->tokenArray !=
    TokenizedQueryPntr->numberOfTokens)
    {
      dprintf ("ARFS:TokenizeOrCount: Bug!  In second pass, "
        "wrote %d tokens but the array was sized for %d tokens.\n",
        (int) (TokenStoragePntr - TokenizedQueryPntr->tokenArray),
        TokenizedQueryPntr->numberOfTokens);
    }
  }
}



/******************************************************************************
 * Break up a user provided query string into tokens.  The main advantage of
 * doing this as a separate step is to remove white space and process the
 * escape characters all in one spot rather than scattered around the code.
 * Plus the strings are copied into our memory space so that the user can free
 * their query string.  Returns the newly allocated TokenizedQueryRecord (you
 * must free it later), or returns NULL if it ran out of memory.
 */

static TokenizedQueryPointer StringToTokenizedQueryRecord (
  DiskVolumePointer DiskVolumePntr,
  const char *StringPntr)
{
  TokenizedQueryRecord  CountingInfo;
  int                   MemorySize;
  TokenizedQueryPointer TokenizedQueryPntr;

  TokenizeOrCount (StringPntr, &CountingInfo, true /* JustFindSizes */);

  MemorySize = CountingInfo.numberOfTokens * sizeof (TokenRecord) +
    CountingInfo.stringSpaceSize + sizeof (TokenizedQueryRecord);

  TokenizedQueryPntr = mallocDVP (DiskVolumePntr, MemorySize);
  if (TokenizedQueryPntr == NULL)
  {
    dprintf ("ARFS:StringToTokenizedQueryRecord: Unable to allocate %d bytes "
      "for tokenizing query.\n", MemorySize);
    return NULL;
  }

  memset (TokenizedQueryPntr, 0, MemorySize);
  TokenizedQueryPntr->numberOfTokens = CountingInfo.numberOfTokens;
  TokenizedQueryPntr->stringSpaceSize = CountingInfo.stringSpaceSize;

  TokenizeOrCount (StringPntr, TokenizedQueryPntr, false /* JustFindSizes */);

#if DPRINT_NON_ERROR_MESSAGES && 1
  { /* Print out the results of tokenizing for test purposes. */
    int i;
    char TempString [500];

    dprintf ("ARFS:StringToTokenizedQueryRecord: Converted query string to "
      "%d tokens and %d bytes string space, total record size %d.\n",
      TokenizedQueryPntr->numberOfTokens,
      TokenizedQueryPntr->stringSpaceSize, MemorySize);
    for (i = 0; i < TokenizedQueryPntr->numberOfTokens; i++)
    {
      switch (TokenizedQueryPntr->tokenArray[i].tokenCode)
      {
        case TC_STRING:
          if (strlen (TokenizedQueryPntr->tokenArray[i].stringPntr) >
          sizeof (TempString) - 20)
            strcpy (TempString, "String too long to print.");
          else
            sprintf (TempString, "{%s}  String.",
              TokenizedQueryPntr->tokenArray[i].stringPntr);
          break;

        case TC_AND:
          strcpy (TempString, "&&  Logical and."); break;
        case TC_OR:
          strcpy (TempString, "||  Logical or."); break;
        case TC_NOT:
          strcpy (TempString, "!  Logical not."); break;
        case TC_LEFT_PARENTHESIS:
          strcpy (TempString, "(  Left parenthesis."); break;
        case TC_RIGHT_PARENTHESIS:
          strcpy (TempString, ")  Right parenthesis."); break;
        case TC_EQUAL:
          strcpy (TempString, "==  Comparison equal."); break;
        case TC_NOT_EQUAL:
          strcpy (TempString, "!=  Comparison not equal."); break;
        case TC_GREATER_THAN:
          strcpy (TempString, ">  Comparison greater than."); break;
        case TC_LESS_THAN:
          strcpy (TempString, "<  Comparison less than."); break;
        case TC_GREATER_THAN_OR_EQUAL:
          strcpy (TempString, ">=  Comparison greater than or equal."); break;
        case TC_LESS_THAN_OR_EQUAL:
          strcpy (TempString, "<=  Comparison less than or equal."); break;
        case TC_ATTRIBUTE_EXISTS:
          strcpy (TempString, "<>  Attribute exists."); break;
        case TC_ATTRIBUTE_MISSING:
          strcpy (TempString, "><  Attribute missing."); break;
        case TC_END_OF_QUERY:
          strcpy (TempString, "EOQ  End of query."); break;

        default:
          sprintf (TempString, "Unknown token #%d.  Bug!",
            TokenizedQueryPntr->tokenArray[i].tokenCode);
          break;
      }
      dprintf ("ARFS:StringToTokenizedQueryRecord: Token %02d: %s\n",
        i, TempString);
    }
  }
#endif

  return TokenizedQueryPntr;
}



/******************************************************************************
 * Free the parse tree given it's root node.  Safe to pass in NULL.
 */

static void FreeParseTree (
  DiskVolumePointer DiskVolumePntr,
  ParseNodePointer  ParseNodePntr)
{
  if (ParseNodePntr == NULL)
    return;

  switch (ParseNodePntr->operation)
  {
    case PNO_AND:
    case PNO_OR:
      FreeParseTree (DiskVolumePntr,
        ParseNodePntr->opSpecific.children.leftChild);
      ParseNodePntr->opSpecific.children.leftChild = NULL;
      FreeParseTree (DiskVolumePntr,
        ParseNodePntr->opSpecific.children.rightChild);
      ParseNodePntr->opSpecific.children.rightChild = NULL;
      break;

    case PNO_COMPARE:
      if (ParseNodePntr->opSpecific.compare.indexFSOPntr != NULL)
      {
        put_vnode (DiskVolumePntr->nameSpaceID,
          (unsigned int) ParseNodePntr->opSpecific.compare.indexFSOPntr);
        ParseNodePntr->opSpecific.compare.indexFSOPntr = NULL;
      }
      break;

    default:
  }

  ParseNodePntr->operation = PNO_MAX;
  freeDVP (DiskVolumePntr, ParseNodePntr);
}



/******************************************************************************
 * Undo the get_vnode which was done for all indices used by the parse tree.
 */

static void PutVnodeParseTree (
  DiskVolumePointer DiskVolumePntr,
  ParseNodePointer  ParseNodePntr)
{
  if (ParseNodePntr == NULL)
    return;

  switch (ParseNodePntr->operation)
  {
    case PNO_AND:
    case PNO_OR:
      PutVnodeParseTree (DiskVolumePntr,
        ParseNodePntr->opSpecific.children.leftChild);
      PutVnodeParseTree (DiskVolumePntr,
        ParseNodePntr->opSpecific.children.rightChild);
      break;

    case PNO_COMPARE:
      if (ParseNodePntr->opSpecific.compare.indexFSOPntr != NULL)
      {
        put_vnode (DiskVolumePntr->nameSpaceID,
          (unsigned int) ParseNodePntr->opSpecific.compare.indexFSOPntr);
        ParseNodePntr->opSpecific.compare.indexFSOPntr = NULL;
      }
      break;

    default:
  }
}



#if DPRINT_NON_ERROR_MESSAGES
/******************************************************************************
 * Prints the parse tree to the given string.  Returns FALSE if it ran out of
 * buffer space.
 */

static bool PrintParseTree (
  ParseNodePointer ParseTreePntr,
  char *PrintString,
  int BufferSize)
{
  char *StartAddress = PrintString;

  if (ParseTreePntr == NULL)
    return true;

  if (ParseTreePntr->operation == PNO_COMPARE)
  {
    if (BufferSize < strlen (
    ParseTreePntr->opSpecific.compare.attributeNamePntr) +
    ((ParseTreePntr->opSpecific.compare.valueStringPntr == NULL)
    ? 0 : strlen (ParseTreePntr->opSpecific.compare.valueStringPntr)) + 30)
      return false;

    if (ParseTreePntr->opSpecific.compare.relation != TC_ATTRIBUTE_EXISTS &&
    ParseTreePntr->opSpecific.compare.relation != TC_ATTRIBUTE_MISSING)
    {
      sprintf (PrintString, "(\"%s\"%s",
        ParseTreePntr->opSpecific.compare.attributeNamePntr,
        ParseTreePntr->opSpecific.compare.multipleKeywords
          ? "{MultiWord}" : "");
      PrintString += strlen (PrintString);
      switch (ParseTreePntr->opSpecific.compare.relation)
      {
        case TC_EQUAL: strcpy (PrintString, "=="); break;
        case TC_GREATER_THAN: strcpy (PrintString, ">"); break;
        case TC_LESS_THAN: strcpy (PrintString, "<"); break;
        case TC_GREATER_THAN_OR_EQUAL: strcpy (PrintString, ">="); break;
        case TC_LESS_THAN_OR_EQUAL: strcpy (PrintString, "<="); break;
        case TC_NOT_EQUAL: strcpy (PrintString, "!="); break;
        default: strcpy (PrintString, "??"); break;
      }
      PrintString += strlen (PrintString);
      sprintf (PrintString, "\"%s\")",
        ParseTreePntr->opSpecific.compare.valueStringPntr);
    }
    else /* The special attribute name with prefix operators. */
    {
      strcpy (PrintString, "(");
      PrintString += strlen (PrintString);
      switch (ParseTreePntr->opSpecific.compare.relation)
      {
        case TC_ATTRIBUTE_EXISTS: strcpy (PrintString, "<>"); break;
        case TC_ATTRIBUTE_MISSING: strcpy (PrintString, "><"); break;
        default: strcpy (PrintString, "??"); break;
      }
      PrintString += strlen (PrintString);
      sprintf (PrintString, "\"%s\"%s)",
        ParseTreePntr->opSpecific.compare.attributeNamePntr,
        ParseTreePntr->opSpecific.compare.multipleKeywords
          ? "{MultiWord}" : "");
    }
  }
  else
  {
    strcpy (PrintString, "(");
    PrintString += strlen (PrintString);
    if (!PrintParseTree (ParseTreePntr->opSpecific.children.leftChild,
    PrintString, BufferSize - (PrintString - StartAddress)))
      return false;
    PrintString += strlen (PrintString);
    strcpy (PrintString, (ParseTreePntr->operation == PNO_AND) ? "&&" : "||");
    PrintString += strlen (PrintString);
    if (!PrintParseTree (ParseTreePntr->opSpecific.children.rightChild,
    PrintString, BufferSize - (PrintString - StartAddress)))
      return false;
    PrintString += strlen (PrintString);
    strcpy (PrintString, ")");
  }

  return true;
}
#endif /* DPRINT_NON_ERROR_MESSAGES */



/******************************************************************************
 * Negate the logical output of the parse tree.  Use Augustus De Morgan's law
 * about boolean negation to convert ANDs to ORs with the operands negated.
 * Comparision operations are negated by reversing the comparison operation,
 * since we can't reverse the arguments (always the attribute name on the left
 * and test value on the right).
 */

static void NegateParseTree (
  ParseNodePointer TreePntr)
{
  if (TreePntr == NULL)
    return;

  if (TreePntr->operation == PNO_COMPARE)
  {
    TreePntr->opSpecific.compare.relation =
      TC_LAST_RELATION - TreePntr->opSpecific.compare.relation;
    return;
  }

  /* Must be a logical AND or a logical OR operation.  Reverse the operation
  and negate the child trees. */

  if (TreePntr->operation == PNO_AND)
    TreePntr->operation = PNO_OR;
  else
    TreePntr->operation = PNO_AND;

  NegateParseTree (TreePntr->opSpecific.children.leftChild);
  NegateParseTree (TreePntr->opSpecific.children.rightChild);
}



/* Forward reference to a recursively self called function. */

static status_t TokensToParseTree (
  DiskVolumePointer DiskVolumePntr,
  DirectoryObjectPointer IndexDirectoryFSOPntr,
  TokenizedQueryPointer TokenizedQueryPntr,
  int *NextTokenIndexPntr,
  ParseNodePointer *ParseNodePntrPntr);


/******************************************************************************
 * Takes a portion of a tokenized query (starting at index *NextTokenIndexPntr)
 * and builds a parse subtree from it.  This function implements the
 * bottom-level ParseComparison concept.  Returns the usual error codes (out of
 * memory or parse failure or success), the index of the token after the last
 * one used (not modified if it fails), and a pointer to the root node of the
 * parse tree (set to NULL if it failed).
 *
 * <Comparison> ::=
 *   <String> <RelationalOperator> <String> | <UnaryOperator> <String>
 */

static status_t ParseComparison (
  DiskVolumePointer DiskVolumePntr,
  DirectoryObjectPointer IndexDirectoryFSOPntr,
  TokenizedQueryPointer TokenizedQueryPntr,
  int *NextTokenIndexPntr,
  ParseNodePointer *ParseNodePntrPntr)
{
  const char      *AttributeNamePntr;
  TokenCode        ComparisonToken;
  status_t         ErrorCode = 0;
  int              NextIndex = *NextTokenIndexPntr;
  ParseNodePointer ResultTreePntr = NULL;
  const char      *ValueStringPntr;

  if (TokenizedQueryPntr->tokenArray[NextIndex].tokenCode ==
  TC_ATTRIBUTE_EXISTS || TokenizedQueryPntr->tokenArray[NextIndex].tokenCode ==
  TC_ATTRIBUTE_MISSING)
  {
    ComparisonToken = TokenizedQueryPntr->tokenArray[NextIndex].tokenCode;
    if (TokenizedQueryPntr->tokenArray[NextIndex+1].tokenCode != TC_STRING)
    {
      dprintf ("ARFS:ParseComparison: Didn't find the string token "
        "after the unary operator, should be one at index %d.\n",
        NextIndex + 1);
      ErrorCode = EINVAL;
      goto ErrorExit;
    }
    AttributeNamePntr = TokenizedQueryPntr->tokenArray[NextIndex+1].stringPntr;
    ValueStringPntr = NULL;
    NextIndex += 2;
  }
  else /* Look for the usual relational operator between two strings. */
  {
    if (TokenizedQueryPntr->tokenArray[NextIndex].tokenCode != TC_STRING)
    {
      dprintf ("ARFS:ParseComparison: Didn't find the left hand string token "
        "at index %d.\n", NextIndex);
      ErrorCode = EINVAL;
      goto ErrorExit;
    }
    AttributeNamePntr = TokenizedQueryPntr->tokenArray[NextIndex].stringPntr;

    ComparisonToken = TokenizedQueryPntr->tokenArray[NextIndex+1].tokenCode;
    if (ComparisonToken > TC_LAST_RELATION)
    {
      dprintf ("ARFS:ParseComparison: Didn't find a comparison relation token "
        "at index %d.\n", NextIndex + 1);
      ErrorCode = EINVAL;
      goto ErrorExit;
    }

    if (TokenizedQueryPntr->tokenArray[NextIndex+2].tokenCode != TC_STRING)
    {
      dprintf ("ARFS:ParseComparison: Didn't find the right hand string token "
        "at index %d.\n", NextIndex + 2);
      ErrorCode = EINVAL;
      goto ErrorExit;
    }
    ValueStringPntr = TokenizedQueryPntr->tokenArray[NextIndex+2].stringPntr;
    NextIndex += 3;
  }

  /* Make the comparison parse tree node. */

  ResultTreePntr = mallocDVP (DiskVolumePntr, sizeof (ParseNodeRecord));
  if (ResultTreePntr == NULL)
  {
    dprintf ("ARFS:ParseComparison: Out of memory when "
      "making comparison node.\n");
    ErrorCode = ENOMEM;
    goto ErrorExit;
  }

  memset (ResultTreePntr, 0, sizeof (ParseNodeRecord));
  ResultTreePntr->operation = PNO_COMPARE;
  ResultTreePntr->opSpecific.compare.relation = ComparisonToken;
  ResultTreePntr->opSpecific.compare.attributeNamePntr = AttributeNamePntr;
  ErrorCode = GetIndexFSO (DiskVolumePntr, IndexDirectoryFSOPntr,
    AttributeNamePntr, &ResultTreePntr->opSpecific.compare.indexFSOPntr);
  /* If there is no index, the indexFSOPntr field gets set to NULL. */
  if (ResultTreePntr->opSpecific.compare.indexFSOPntr != NULL)
  {
    ResultTreePntr->opSpecific.compare.indexValueType = ResultTreePntr->
      opSpecific.compare.indexFSOPntr->keyType;
    ResultTreePntr->opSpecific.compare.indexValueNormalSize = ResultTreePntr->
      opSpecific.compare.indexFSOPntr->normalAttributeSize;
    ResultTreePntr->opSpecific.compare.multipleKeywords = ResultTreePntr->
      opSpecific.compare.indexFSOPntr->interpretAsMultipleKeywords;
  }
  ResultTreePntr->opSpecific.compare.valueStringPntr = ValueStringPntr;

  ErrorCode = 0;

ErrorExit:
  if (ErrorCode != 0)
  {
    FreeParseTree (DiskVolumePntr, ResultTreePntr);
    *ParseNodePntrPntr = NULL;
  }
  else /* Successful exit.  Next index only set on success. */
  {
    *NextTokenIndexPntr = NextIndex;
    *ParseNodePntrPntr = ResultTreePntr;
  }
  return ErrorCode;
}



/******************************************************************************
 * Takes a portion of a tokenized query (starting at index *NextTokenIndexPntr)
 * and builds a parse subtree from it.  This function implements the low-level
 * Term concept.  Returns the usual error codes (out of memory or parse failure
 * or success), the index of the token after the last one used (not modified if
 * it fails), and a pointer to the root node of the parse tree (set to NULL if
 * it failed).
 *
 * <Term> ::= <Comparison> | "(" <Expression> ( ")" | "\0" ) | "!" <Term>
 */

static status_t ParseTerm (
  DiskVolumePointer DiskVolumePntr,
  DirectoryObjectPointer IndexDirectoryFSOPntr,
  TokenizedQueryPointer TokenizedQueryPntr,
  int *NextTokenIndexPntr,
  ParseNodePointer *ParseNodePntrPntr)
{
  status_t         ErrorCode = 0;
  int              NextIndex = *NextTokenIndexPntr;
  ParseNodePointer ResultTreePntr = NULL;

  /* Look for the low hanging fruit first, as they say.  That's something
  starting with a constant token like a "(" or "!" token. */

  if (TokenizedQueryPntr->tokenArray[NextIndex].tokenCode ==
  TC_LEFT_PARENTHESIS)
  {
    NextIndex++;
    ErrorCode = TokensToParseTree (DiskVolumePntr, IndexDirectoryFSOPntr,
      TokenizedQueryPntr, &NextIndex, &ResultTreePntr);
    if (ErrorCode != 0)
    {
      dprintf ("ARFS:ParseTerm: Didn't find an expression after the '(' "
        "token, starting looking at index %d.\n", NextIndex);
      goto ErrorExit;
    }
    if (TokenizedQueryPntr->tokenArray[NextIndex].tokenCode ==
    TC_RIGHT_PARENTHESIS)
    {
      NextIndex++;
      goto NormalExit;
    }
    if (TokenizedQueryPntr->tokenArray[NextIndex].tokenCode == TC_END_OF_QUERY)
      goto NormalExit; // Note NextIndex is left pointing at the end marker.
    /* No closing parenthesis found. */
    dprintf ("ARFS:ParseTerm: Didn't find a matching closing ')' (or end "
      "of the query which counts as a closing bracket too) after the "
      "expression.  Starting looking at index %d, opening bracket was "
      "found earlier at index %d.\n", NextIndex, *NextTokenIndexPntr);
    ErrorCode = EINVAL;
    goto ErrorExit;
  }

  if (TokenizedQueryPntr->tokenArray[NextIndex].tokenCode == TC_NOT)
  {
    NextIndex++;
    ErrorCode = ParseTerm (DiskVolumePntr, IndexDirectoryFSOPntr,
      TokenizedQueryPntr, &NextIndex, &ResultTreePntr);
    if (ErrorCode != 0)
    {
      dprintf ("ARFS:ParseTerm: Didn't find a term after the '!' "
        "token, starting looking at index %d.\n", NextIndex);
      goto ErrorExit;
    }
    NegateParseTree (ResultTreePntr);
    if (ErrorCode != 0)
      goto ErrorExit;
    goto NormalExit; /* Oh no, it's almost turning into spaghetti goto code! */
  }

  /* It can only be a comparison at this point since there are no alternatives
  left any more. */

  ErrorCode = ParseComparison (DiskVolumePntr, IndexDirectoryFSOPntr,
    TokenizedQueryPntr, &NextIndex, &ResultTreePntr);
  if (ErrorCode != 0)
  {
    dprintf ("ARFS:ParseTerm: Didn't find a comparison equation (and it "
      "can't be anything else but that) at index %d.\n", NextIndex);
    goto ErrorExit;
  }

NormalExit: /* Success */
  ErrorCode = 0;

ErrorExit:
  if (ErrorCode != 0)
  {
    FreeParseTree (DiskVolumePntr, ResultTreePntr);
    *ParseNodePntrPntr = NULL;
  }
  else /* Successful exit.  Next index only set on success. */
  {
    *NextTokenIndexPntr = NextIndex;
    *ParseNodePntrPntr = ResultTreePntr;
  }
  return ErrorCode;
}



/******************************************************************************
 * Takes a portion of a tokenized query (starting at index *NextTokenIndexPntr)
 * and builds a parse subtree from it.  This function implements the mid-level
 * SimplerExpression concept.  Returns the usual error codes (out of memory or
 * parse failure or success), the index of the token after the last one used
 * (not modified if it fails), and a pointer to the root node of the parse tree
 * (set to NULL if it failed).
 *
 * <SimplerExpression> ::= <Term> | <Term> "&&" <SimplerExpression>
 */

static status_t ParseSimplerExpression (
  DiskVolumePointer DiskVolumePntr,
  DirectoryObjectPointer IndexDirectoryFSOPntr,
  TokenizedQueryPointer TokenizedQueryPntr,
  int *NextTokenIndexPntr,
  ParseNodePointer *ParseNodePntrPntr)
{
  status_t         ErrorCode = 0;
  int              NextIndex = *NextTokenIndexPntr;
  ParseNodePointer ResultTreePntr = NULL;
  ParseNodePointer TempTreePntr;

  ErrorCode = ParseTerm (DiskVolumePntr, IndexDirectoryFSOPntr,
    TokenizedQueryPntr, &NextIndex, &ResultTreePntr);
  if (ErrorCode != 0)
  {
    dprintf ("ARFS:ParseSimplerExpression: Couldn't find a term "
      "starting at token %d.\n", NextIndex);
    goto ErrorExit; /* Need at least one term. */
  }

  if (TokenizedQueryPntr->tokenArray[NextIndex].tokenCode != TC_AND)
    goto NormalExit; /* That's it, just one term. */
  NextIndex++;

  /* This is the fancier AND (&&) variety of simple expression.  Save the left
  hand side of the expression as the left child of an AND node. */

  TempTreePntr = mallocDVP (DiskVolumePntr, sizeof (ParseNodeRecord));
  if (TempTreePntr == NULL)
  {
    dprintf ("ARFS:ParseSimplerExpression: Out of memory when "
      "making AND node.\n");
    ErrorCode = ENOMEM;
    goto ErrorExit;
  }

  memset (TempTreePntr, 0, sizeof (ParseNodeRecord));
  TempTreePntr->operation = PNO_AND;
  TempTreePntr->opSpecific.children.leftChild = ResultTreePntr;
  ResultTreePntr = TempTreePntr;

  /* Get the right hand side of the AND, it's another simple expression. */

  ErrorCode = ParseSimplerExpression (DiskVolumePntr, IndexDirectoryFSOPntr,
    TokenizedQueryPntr, &NextIndex, &TempTreePntr);
  if (ErrorCode != 0)
  {
    dprintf ("ARFS:ParseSimplerExpression: Couldn't find a simpler "
      "expression after the AND (&&) token, starting at index %d.\n",
      NextIndex);
    goto ErrorExit;
  }
  ResultTreePntr->opSpecific.children.rightChild = TempTreePntr;

NormalExit: /* Success */
  ErrorCode = 0;

ErrorExit:
  if (ErrorCode != 0)
  {
    FreeParseTree (DiskVolumePntr, ResultTreePntr);
    *ParseNodePntrPntr = NULL;
  }
  else /* Successful exit.  Next index only set on success. */
  {
    *NextTokenIndexPntr = NextIndex;
    *ParseNodePntrPntr = ResultTreePntr;
  }
  return ErrorCode;
}



/******************************************************************************
 * Takes a portion of a tokenized query (starting at index *NextTokenIndexPntr)
 * and builds the parse tree from it.  See the comments before ARFS_open_query
 * for a description of the language parsed.  This function implements the top
 * level Expression concept.  Returns the usual error codes (out of memory or
 * parse failure or success), the index of the token after the last one used
 * (not modified if it fails), and a pointer to the root node of the parse tree
 * (set to NULL if it failed).
 *
 * <Expression> ::= <SimplerExpression> | <SimplerExpression> "||" <Expression>
 */

static status_t TokensToParseTree (
  DiskVolumePointer DiskVolumePntr,
  DirectoryObjectPointer IndexDirectoryFSOPntr,
  TokenizedQueryPointer TokenizedQueryPntr,
  int *NextTokenIndexPntr,
  ParseNodePointer *ParseNodePntrPntr)
{
  status_t         ErrorCode = 0;
  int              NextIndex = *NextTokenIndexPntr;
  ParseNodePointer ResultTreePntr = NULL;
  ParseNodePointer TempTreePntr;

  ErrorCode = ParseSimplerExpression (DiskVolumePntr, IndexDirectoryFSOPntr,
    TokenizedQueryPntr, &NextIndex, &ResultTreePntr);
  if (ErrorCode != 0)
  {
    dprintf ("ARFS:TokensToParseTree: Couldn't find even a simple expression "
      "starting at token %d.\n", NextIndex);
    goto ErrorExit; /* Need at least one simple expression. */
  }

  if (TokenizedQueryPntr->tokenArray[NextIndex].tokenCode != TC_OR)
    goto NormalExit; /* That's it, just one simple expression. */
  NextIndex++;

  /* This is the fancier OR (||) variety of expression.  Save the left hand
  side of the expression as the left child of an OR node. */

  TempTreePntr = mallocDVP (DiskVolumePntr, sizeof (ParseNodeRecord));
  if (TempTreePntr == NULL)
  {
    dprintf ("ARFS:TokensToParseTree: Out of memory when making OR node.\n");
    ErrorCode = ENOMEM;
    goto ErrorExit;
  }

  memset (TempTreePntr, 0, sizeof (ParseNodeRecord));
  TempTreePntr->operation = PNO_OR;
  TempTreePntr->opSpecific.children.leftChild = ResultTreePntr;
  ResultTreePntr = TempTreePntr;

  /* Get the right hand side of the OR token, it is a full expression. */

  ErrorCode = TokensToParseTree (DiskVolumePntr, IndexDirectoryFSOPntr,
    TokenizedQueryPntr, &NextIndex, &TempTreePntr);
  if (ErrorCode != 0)
  {
    dprintf ("ARFS:TokensToParseTree: Couldn't find an expression after the "
      "OR (||) token, starting at index %d.\n", NextIndex);
    goto ErrorExit;
  }
  ResultTreePntr->opSpecific.children.rightChild = TempTreePntr;

NormalExit: /* Success */
  ErrorCode = 0;

ErrorExit:
  if (ErrorCode != 0)
  {
    FreeParseTree (DiskVolumePntr, ResultTreePntr);
    *ParseNodePntrPntr = NULL;
  }
  else /* Successful exit.  Next index only set on success. */
  {
    *NextTokenIndexPntr = NextIndex;
    *ParseNodePntrPntr = ResultTreePntr;
  }
  return ErrorCode;
}



/******************************************************************************
 * Recursively calculates the score of the given parse tree, and also returns
 * the best index (can be NULL if none applicable) and traversal range (lower
 * and upper bounds).  Doesn't do any locking.
 *
 * The score is an evaluation of how much priority to give this branch of the
 * parse tree when evaluating queries and when deciding which index to use for
 * iteration.  Larger numbers mean a higher priority, and the query is more
 * exclusive (returns fewer files).  Maximum score is 1024, minimum is zero
 * (always larger than zero if an index can be used).  1024 is the max since it
 * avoids overflow with 32 bit math calculations with multipliers of over a
 * million; intermediate calculations won't overflow when the RAM disk has a
 * million files in it.  A score of 1024 is awarded to an == comparison to
 * something without wildcards (meaning very few files will match).  For ==
 * with wildcards, score is 32 plus 62 more for each initial non-wildcard
 * character up to a max of 1024 (16 unique characters).  For <, <=, >, >=
 * score is 32 + 992 times the proportion of files not in the index (32 + 992 *
 * (NumberOfFiles - size of index) / NumberOfFiles).  So if all files are in
 * the index, the score is 32, going up to 1024 when the index is more
 * exclusive.  For != it is just 16, since that sort of comparison cuts out
 * very few files (though files without the attribute aren't considered, so the
 * index is still useful).  If the same index is used for both inputs to AND or
 * OR, then the ranges are combined and the score magically adjusted.
 */

static void RecursiveScoreParseTree (
  DiskVolumePointer DiskVolumePntr,
  int NumberOfFilesInWholeVolume,
  ParseNodePointer ParseTreePntr,
  int32 *BestScorePntr,
  IndexObjectPointer *BestIndexFSOPntrPntr,
  AVLIterationRangePointer BestIterationRangePntr)
{
  struct CompareStruct           *ComparePntr;
  AVLDupComparisonFunctionPointer ComparisonFunctionPntr;
  int                             ComparisonResult;
  unsigned char                  *FirstWildCharPntr;
  type_code                       KeyType = 0;
  IndexObjectPointer              OtherIndexFSOPntr;
  AVLIterationRangeRecord         OtherIterationRange;
  uint32                          OtherScore;
  unsigned char                  *StringStart;
  AVLDupThingRecord               SwapThing;
  AVLDupThingRecord               TempThing;

#define SWAP_THINGS(x,y) {SwapThing = x; x = y; y = SwapThing;}

#if DPRINT_NON_ERROR_MESSAGES
  /* Hopefully only one query is opened at a time, otherwise PrintString will
  get overwritten oddly by multiple threads.  But it's just too big to make
  into a local variable in a recursive function. */
  static char                     PrintString [1000];
#endif

  /* NULLify string things so they don't mistakenly get deallocated at exit. */
  TempThing.uint64Thing = 0;
  memset (&OtherIterationRange, 0, sizeof (AVLIterationRangeRecord));
  memset (BestIterationRangePntr, 0, sizeof (AVLIterationRangeRecord));

  *BestScorePntr = 0;
  *BestIndexFSOPntrPntr = NULL;

  if (ParseTreePntr == NULL)
    goto NormalExit;

  if (ParseTreePntr->operation == PNO_COMPARE)
  {
    ComparePntr = &ParseTreePntr->opSpecific.compare;
    KeyType = ComparePntr->indexValueType;

    if (ComparePntr->indexFSOPntr == NULL)
      goto NormalExit; /* Just an unindexed attribute. */

    if (ComparePntr->relation == TC_ATTRIBUTE_MISSING)
      goto NormalExit; /* Can't use an index for this one! */

    if (ComparePntr->relation == TC_NOT_EQUAL)
      *BestScorePntr = 16;
    else if (ComparePntr->relation == TC_EQUAL)
    {
      /* Find the first wildcard character, so we know how many non-wildcards
      are present.  More non-wildcards are better.  Wildcards only used for
      string type comparisons.  Otherwise, it's just numerical equality so it
      should have very few matching values and thus be highly desirable. */

      if (ComparePntr->indexValueNormalSize == 0 /* if it is a string type */)
      {
        FirstWildCharPntr = (unsigned char *)
          IsUTF8Pattern (ComparePntr->valueStringPntr);
        if (FirstWildCharPntr == NULL)
          *BestScorePntr = 1024;
        else
          *BestScorePntr = 32 + 62 *
          (FirstWildCharPntr - (unsigned char *) ComparePntr->valueStringPntr);
      }
      else /* Numerical type equality comparison. */
        *BestScorePntr = 1024;
    }
    else
    {
      /* Score is proportional to the exclusivity of the index for relational
      operations; generally they will include a large part of the index in
      their result set so a smaller index is better. */

      if (NumberOfFilesInWholeVolume <= 0)
        *BestScorePntr = 512;
      else
        *BestScorePntr = 32 + 992 * (NumberOfFilesInWholeVolume -
          AVLDupGetTreeCount (ComparePntr->indexFSOPntr->treePntr)) /
          NumberOfFilesInWholeVolume;
    }
    if (*BestScorePntr < 1)
      *BestScorePntr = 1;
    if (*BestScorePntr > 1024)
      *BestScorePntr = 1024;

    /* Now set the iteration limits for traversing the index. */

    *BestIndexFSOPntrPntr = ComparePntr->indexFSOPntr;
    BestIterationRangePntr->ascendingOrder = true;
    BestIterationRangePntr->infiniteLower = true;
    BestIterationRangePntr->infiniteUpper = true;

    if (ComparePntr->relation == TC_ATTRIBUTE_EXISTS ||
    ComparePntr->relation == TC_NOT_EQUAL)
      goto NormalExit; /* Traverse the whole index. */

    /* Set up the starting/ending specific value of the traversal, based on the
    value in the query.  Allow memory allocation since string limits for
    iteration should not include the wildcards (the upper and lower strings are
    modified versions of the string from the query). */

    if (!AVLDupConvertStringToThing (
    ComparePntr->valueStringPntr,
    &TempThing,
    KeyType,
    DiskVolumePntr->mallocPntr,
    DiskVolumePntr->freePntr,
    DiskVolumePntr->volumePoolID))
    {
      dprintf ("ARFS:RecursiveScoreParseTree: Conversion of \"%s\" to "
        "type %s failed, using zero instead.  Was looking at "
        "attribute \"%s\".\n",
        ComparePntr->valueStringPntr,
        TypeCodeAsString (KeyType),
        ComparePntr->attributeNamePntr);
      TempThing.uint64Thing = 0;
    }

    switch (ComparePntr->relation)
    {
      case TC_GREATER_THAN:
        BestIterationRangePntr->lowerKey = TempThing;
        TempThing.uint64Thing = 0; /* Has been used, no deallocation needed. */
        BestIterationRangePntr->includeThingEqualToLower = false;
        BestIterationRangePntr->infiniteLower = false;
        break;

      case TC_GREATER_THAN_OR_EQUAL:
        BestIterationRangePntr->lowerKey = TempThing;
        TempThing.uint64Thing = 0; /* Has been used, no deallocation needed. */
        BestIterationRangePntr->includeThingEqualToLower = true;
        BestIterationRangePntr->infiniteLower = false;
        break;

      case TC_LESS_THAN:
        BestIterationRangePntr->ascendingOrder = false;
        BestIterationRangePntr->upperKey = TempThing;
        TempThing.uint64Thing = 0; /* Has been used, no deallocation needed. */
        BestIterationRangePntr->includeThingEqualToUpper = false;
        BestIterationRangePntr->infiniteUpper = false;
        break;

      case TC_LESS_THAN_OR_EQUAL:
        BestIterationRangePntr->ascendingOrder = false;
        BestIterationRangePntr->upperKey = TempThing;
        TempThing.uint64Thing = 0; /* Has been used, no deallocation needed. */
        BestIterationRangePntr->includeThingEqualToUpper = true;
        BestIterationRangePntr->infiniteUpper = false;
        break;

      case TC_EQUAL:
        if (!AVLDupCopyThingArray (&BestIterationRangePntr->lowerKey,
        &TempThing, KeyType, 1 /* how many to copy */,
        DiskVolumePntr->mallocPntr, DiskVolumePntr->freePntr,
        DiskVolumePntr->volumePoolID))
        {
          dprintf ("ARFS:RecursiveScoreParseTree: Copying thing of "
            "type %s failed, using zero instead.  Was looking at "
            "attribute \"%s\".\n",
            TypeCodeAsString (KeyType),
            ComparePntr->attributeNamePntr);
          BestIterationRangePntr->lowerKey.uint64Thing = 0;
        }
        BestIterationRangePntr->includeThingEqualToLower = true;
        BestIterationRangePntr->infiniteLower = false;
        BestIterationRangePntr->upperKey = TempThing;
        TempThing.uint64Thing = 0; /* Has been used, no deallocation needed. */
        BestIterationRangePntr->includeThingEqualToUpper = true;
        BestIterationRangePntr->infiniteUpper = false;

        /* Adjust the beginning and ending ranges for string equality to not
        include the wildcard characters. */

        if (ComparePntr->indexValueNormalSize == 0 /* is it a string */)
        {
          /* Truncate the lower string at the start of the pattern part.  So
          "ABC* becomes "ABC". */

          StringStart =
            AVLDupGetStringPntrFromThing (BestIterationRangePntr->lowerKey);
          FirstWildCharPntr = (unsigned char *) IsUTF8Pattern (StringStart);
          if (FirstWildCharPntr != NULL)
          {
            *FirstWildCharPntr = 0;
            if (FirstWildCharPntr <= StringStart)
              BestIterationRangePntr->infiniteLower = true;
          }

          /* The upper string is equal to the truncated lower string plus one,
          as if the string was a series of digits.  "ABC*" becomes "ABD". */

          StringStart =
            AVLDupGetStringPntrFromThing (BestIterationRangePntr->upperKey);
          FirstWildCharPntr = (unsigned char *) IsUTF8Pattern (StringStart);
          if (FirstWildCharPntr != NULL)
          {
            *FirstWildCharPntr-- = 0;
            while (FirstWildCharPntr >= StringStart)
            {
              if (*FirstWildCharPntr != 255)
              {
                *FirstWildCharPntr = (*FirstWildCharPntr) + 1;
                break;
              }
              *FirstWildCharPntr = 0;
              FirstWildCharPntr--;
            }
            if (FirstWildCharPntr < StringStart)
              BestIterationRangePntr->infiniteUpper = true;
          }
        }
        break;

      default:
        dprintf ("ARFS:RecursiveScoreParseTree: Bug!  Unknown comparison "
          "type of %d.\n", (int) ComparePntr->relation);
        *BestScorePntr = 0;
        *BestIndexFSOPntrPntr = NULL;
        goto NormalExit;
    }
    goto NormalExit;
  }

  if (ParseTreePntr->operation != PNO_AND &&
  ParseTreePntr->operation != PNO_OR)
  {
    dprintf ("ARFS:RecursiveScoreParseTree: Bug!  Unknown logic operation "
      "type of %d.\n", (int) ParseTreePntr->operation);
    *BestScorePntr = 0;
    *BestIndexFSOPntrPntr = NULL;
    goto NormalExit;
  }

  /* Evaluate the two subtree options, which both AND and OR have. */

  RecursiveScoreParseTree (DiskVolumePntr, NumberOfFilesInWholeVolume,
    ParseTreePntr->opSpecific.children.leftChild,
    &OtherScore, &OtherIndexFSOPntr, &OtherIterationRange);

  RecursiveScoreParseTree (DiskVolumePntr, NumberOfFilesInWholeVolume,
    ParseTreePntr->opSpecific.children.rightChild,
    BestScorePntr, BestIndexFSOPntrPntr, BestIterationRangePntr);

  /* Merge ranges if the subtrees are using the same index. */

  if (*BestIndexFSOPntrPntr != NULL &&
  *BestIndexFSOPntrPntr == OtherIndexFSOPntr)
  {
    KeyType = (*BestIndexFSOPntrPntr)->keyType;
    ComparisonFunctionPntr = AVLDupGetComparisonFunctionForType (KeyType);
    if (ComparisonFunctionPntr == NULL)
    {
      dprintf ("ARFS:RecursiveScoreParseTree: No comparison function found!  "
        "Using whole range instead of combining ranges.  For %s.\n",
        FSOAsString (&(*BestIndexFSOPntrPntr)->fso));
      AVLDupFreeThingArray (&BestIterationRangePntr->lowerKey, KeyType, 1,
        DiskVolumePntr->freePntr, DiskVolumePntr->volumePoolID);
      AVLDupFreeThingArray (&BestIterationRangePntr->upperKey, KeyType, 1,
        DiskVolumePntr->freePntr, DiskVolumePntr->volumePoolID);
      BestIterationRangePntr->infiniteLower = true;
      BestIterationRangePntr->infiniteUpper = true;
    }
    else
    {
      if (ParseTreePntr->operation == PNO_AND)
      {
        /* Compute the intersection of the two ranges.  New lower bound is the
        larger of the lower bounds, new upper bound is the smaller of the
        uppers.  New score is the larger score, or 1024 if the resulting range
        is empty. */

        if (!OtherIterationRange.infiniteLower)
        {
          if (BestIterationRangePntr->infiniteLower)
          {
            BestIterationRangePntr->infiniteLower = false;
            SWAP_THINGS (BestIterationRangePntr->lowerKey,
              OtherIterationRange.lowerKey);
            BestIterationRangePntr->includeThingEqualToLower =
              OtherIterationRange.includeThingEqualToLower;
          }
          else
          {
            ComparisonResult = ComparisonFunctionPntr (
              &BestIterationRangePntr->lowerKey,
              &OtherIterationRange.lowerKey);
            if (ComparisonResult < 0)
            {
              SWAP_THINGS (BestIterationRangePntr->lowerKey,
                OtherIterationRange.lowerKey);
              BestIterationRangePntr->includeThingEqualToLower =
                OtherIterationRange.includeThingEqualToLower;
            }
            else if (ComparisonResult == 0)
            {
              BestIterationRangePntr->includeThingEqualToLower =
                (BestIterationRangePntr->includeThingEqualToLower &&
                OtherIterationRange.includeThingEqualToLower);
            }
          }
        }

        /* New upper bound is the smaller of the uppers. */

        if (!OtherIterationRange.infiniteUpper)
        {
          if (BestIterationRangePntr->infiniteUpper)
          {
            BestIterationRangePntr->infiniteUpper = false;
            SWAP_THINGS (BestIterationRangePntr->upperKey,
              OtherIterationRange.upperKey);
            BestIterationRangePntr->includeThingEqualToUpper =
              OtherIterationRange.includeThingEqualToUpper;
          }
          else
          {
            ComparisonResult = ComparisonFunctionPntr (
              &BestIterationRangePntr->upperKey,
              &OtherIterationRange.upperKey);
            if (ComparisonResult > 0)
            {
              SWAP_THINGS (BestIterationRangePntr->upperKey,
                OtherIterationRange.upperKey);
              BestIterationRangePntr->includeThingEqualToUpper =
                OtherIterationRange.includeThingEqualToUpper;
            }
            else if (ComparisonResult == 0)
            {
              BestIterationRangePntr->includeThingEqualToUpper =
                (BestIterationRangePntr->includeThingEqualToUpper &&
                OtherIterationRange.includeThingEqualToUpper);
            }
          }
        }

        /* New score is the larger score, or 1024 if the resulting range is
        empty. */

        if (OtherScore > *BestScorePntr)
          *BestScorePntr = OtherScore;
        if (!BestIterationRangePntr->infiniteLower &&
        !BestIterationRangePntr->infiniteUpper &&
        ComparisonFunctionPntr (&BestIterationRangePntr->upperKey,
        &BestIterationRangePntr->lowerKey) < 0)
          *BestScorePntr = 1024;
      }
      else /* OR operation. */
      {
        /* Union of two ranges into a bigger range.  New lower bound is the
        lower of the two.  New upper bound is the higher of the two.  New score
        is the lower score (actually should be less than that). */

        if (!BestIterationRangePntr->infiniteLower)
        {
          if (OtherIterationRange.infiniteLower)
            BestIterationRangePntr->infiniteLower = true;
          else
          {
            ComparisonResult = ComparisonFunctionPntr (
              &BestIterationRangePntr->lowerKey,
              &OtherIterationRange.lowerKey);
            if (ComparisonResult > 0)
            {
              SWAP_THINGS (BestIterationRangePntr->lowerKey,
                OtherIterationRange.lowerKey);
              BestIterationRangePntr->includeThingEqualToLower =
                OtherIterationRange.includeThingEqualToLower;
            }
            else if (ComparisonResult == 0)
            {
              BestIterationRangePntr->includeThingEqualToLower =
                (BestIterationRangePntr->includeThingEqualToLower ||
                OtherIterationRange.includeThingEqualToLower);
            }
          }
        }

        /* New upper bound is the higher of the two. */

        if (!BestIterationRangePntr->infiniteUpper)
        {
          if (OtherIterationRange.infiniteUpper)
            BestIterationRangePntr->infiniteUpper = true;
          else
          {
            ComparisonResult = ComparisonFunctionPntr (
              &BestIterationRangePntr->upperKey,
              &OtherIterationRange.upperKey);
            if (ComparisonResult < 0)
            {
              SWAP_THINGS (BestIterationRangePntr->upperKey,
                OtherIterationRange.upperKey);
              BestIterationRangePntr->includeThingEqualToUpper =
                OtherIterationRange.includeThingEqualToUpper;
            }
            else if (ComparisonResult == 0)
            {
              BestIterationRangePntr->includeThingEqualToUpper =
                (BestIterationRangePntr->includeThingEqualToUpper ||
                OtherIterationRange.includeThingEqualToUpper);
            }
          }
        }

        /* New score is the lower score (actually should be less than that). */

        if (OtherScore < *BestScorePntr)
          *BestScorePntr = OtherScore;
      }
    }
  }
  else /* Different indices to merge, or no indices.  Can't compare keys. */
  {
    /* KeyType = 0; is already true.  Nothing automatically deallocated. */

    if (ParseTreePntr->operation == PNO_AND)
    {
      /* Pick the best (higher score) range for AND. */

      if (OtherScore > *BestScorePntr)
      {
        if (*BestIndexFSOPntrPntr != NULL)
        {
          AVLDupFreeThingArray (&BestIterationRangePntr->lowerKey,
            (*BestIndexFSOPntrPntr)->keyType, 1,
            DiskVolumePntr->freePntr, DiskVolumePntr->volumePoolID);
          AVLDupFreeThingArray (&BestIterationRangePntr->upperKey,
            (*BestIndexFSOPntrPntr)->keyType, 1,
            DiskVolumePntr->freePntr, DiskVolumePntr->volumePoolID);
        }
        *BestScorePntr = OtherScore;
        *BestIndexFSOPntrPntr = OtherIndexFSOPntr;
        *BestIterationRangePntr = OtherIterationRange;
      }
      else
      {
        if (OtherIndexFSOPntr != NULL)
        {
          AVLDupFreeThingArray (&OtherIterationRange.lowerKey,
            OtherIndexFSOPntr->keyType, 1,
            DiskVolumePntr->freePntr, DiskVolumePntr->volumePoolID);
          AVLDupFreeThingArray (&OtherIterationRange.upperKey,
            OtherIndexFSOPntr->keyType, 1,
            DiskVolumePntr->freePntr, DiskVolumePntr->volumePoolID);
        }
      }
    }
    else /* Doing an OR, need to iterate over all files. */
    {
      if (*BestIndexFSOPntrPntr != NULL)
      {
        AVLDupFreeThingArray (&BestIterationRangePntr->lowerKey,
          (*BestIndexFSOPntrPntr)->keyType, 1,
          DiskVolumePntr->freePntr, DiskVolumePntr->volumePoolID);
        AVLDupFreeThingArray (&BestIterationRangePntr->upperKey,
          (*BestIndexFSOPntrPntr)->keyType, 1,
          DiskVolumePntr->freePntr, DiskVolumePntr->volumePoolID);
      }
      if (OtherIndexFSOPntr != NULL)
      {
        AVLDupFreeThingArray (&OtherIterationRange.lowerKey,
          OtherIndexFSOPntr->keyType, 1,
          DiskVolumePntr->freePntr, DiskVolumePntr->volumePoolID);
        AVLDupFreeThingArray (&OtherIterationRange.upperKey,
          OtherIndexFSOPntr->keyType, 1,
          DiskVolumePntr->freePntr, DiskVolumePntr->volumePoolID);
      }

      *BestScorePntr = 0;
      *BestIndexFSOPntrPntr = NULL;
    }
  }

NormalExit:
  if (KeyType != 0)
  {
    /* Free all unused key things. */

    AVLDupFreeThingArray (&TempThing, KeyType, 1 /* how many */,
      DiskVolumePntr->freePntr, DiskVolumePntr->volumePoolID);
    AVLDupFreeThingArray (&OtherIterationRange.lowerKey, KeyType, 1,
      DiskVolumePntr->freePntr, DiskVolumePntr->volumePoolID);
    AVLDupFreeThingArray (&OtherIterationRange.upperKey, KeyType, 1,
      DiskVolumePntr->freePntr, DiskVolumePntr->volumePoolID);
    if (*BestIndexFSOPntrPntr == NULL)
    {
      AVLDupFreeThingArray (&BestIterationRangePntr->lowerKey, KeyType, 1,
        DiskVolumePntr->freePntr, DiskVolumePntr->volumePoolID);
      AVLDupFreeThingArray (&BestIterationRangePntr->upperKey, KeyType, 1,
        DiskVolumePntr->freePntr, DiskVolumePntr->volumePoolID);
    }
  }

#if DPRINT_NON_ERROR_MESSAGES
  /* Print out the score, iteration details, and the subtree. */
  sprintf (PrintString, "ARFS:RecursiveScoreParseTree: Score %ld, choice %s",
    *BestScorePntr, FSOAsString (&(*BestIndexFSOPntrPntr)->fso));
  if (*BestIndexFSOPntrPntr != NULL)
  {
    sprintf (PrintString + strlen (PrintString),
      ", range %scending, lower bound ",
      BestIterationRangePntr->ascendingOrder ? "as" : "des");
    if (BestIterationRangePntr->infiniteLower)
      strcat (PrintString, "-infinity");
    else
    {
      if (!BestIterationRangePntr->includeThingEqualToLower)
        strcat (PrintString, "just after ");
      AVLDupConvertThingToString (&BestIterationRangePntr->lowerKey,
        (*BestIndexFSOPntrPntr)->keyType, PrintString + strlen (PrintString),
        300);
    }
    strcat (PrintString, ", upper ");
    if (BestIterationRangePntr->infiniteUpper)
      strcat (PrintString, "+infinity");
    else
    {
      if (!BestIterationRangePntr->includeThingEqualToUpper)
        strcat (PrintString, "just before ");
      AVLDupConvertThingToString (&BestIterationRangePntr->upperKey,
        (*BestIndexFSOPntrPntr)->keyType, PrintString + strlen (PrintString),
        300);
    }
  }
  strcat (PrintString, ".  Subtree: ");
  if (!PrintParseTree (ParseTreePntr, PrintString + strlen (PrintString),
  sizeof (PrintString) - strlen (PrintString) - 256))
    strcat (PrintString, "...");
  dprintf ("%s\n", PrintString);
#endif
}



/******************************************************************************
 * Finds the best index to traverse over for the query.  Also finds the lower
 * and upper range of values to traverse between.  If no indices are
 * appropriate then uses the "name" index and traverses over everything in it.
 * Changes the pointer pointed to by IndexFSOPntrPntr to point to the index, or
 * NULL if it fails.  Assumes that the indices listed in the parse tree nodes
 * have been get_vnoded (doesn't need them locked, since it just reads the
 * number of items in each one and it doesn't matter if that is slightly
 * incorrect).  Leaves the output index in a get_vnoded state (a get_vnode done
 * in addition to the existing one in the parse tree).  Writes the iteration
 * limits to the given iteration range record.  Returns the usual error codes.
 */

static status_t FindOptimalIndexAndRange (
  DiskVolumePointer DiskVolumePntr,
  DirectoryObjectPointer  IndexDirectoryFSOPntr,
  ParseNodePointer ParseTreePntr,
  IndexObjectPointer *IndexFSOPntrPntr,
  AVLIterationRangePointer IterationRangePntr)
{
  int32              BestScore;
  status_t           ErrorCode;
  IndexObjectPointer NameIndexFSOPntr = NULL;
  int                NumberOfFilesInWholeVolume;
  void              *VoidPntr;

  /* Count the number of files/directories/etc by looking at the "name" index
  size.  It only needs to be approximately correct. */

  ErrorCode = GetIndexFSO (DiskVolumePntr, IndexDirectoryFSOPntr,
    g_NameOfNameIndex, &NameIndexFSOPntr);
  if (ErrorCode != 0)
    return ErrorCode; /* Need to have the name index for several reasons. */

  NumberOfFilesInWholeVolume =
    AVLDupGetTreeCount (NameIndexFSOPntr->treePntr);

  RecursiveScoreParseTree (DiskVolumePntr, NumberOfFilesInWholeVolume,
    ParseTreePntr, &BestScore, IndexFSOPntrPntr, IterationRangePntr);

  if (*IndexFSOPntrPntr == NULL)
  {
    /* No ideal index, so iterate over the whole name index. */

    *IndexFSOPntrPntr = NameIndexFSOPntr;
    NameIndexFSOPntr = NULL;
    memset (IterationRangePntr, 0, sizeof (AVLIterationRangeRecord));
    IterationRangePntr->ascendingOrder = true;
    IterationRangePntr->infiniteLower = true;
    IterationRangePntr->infiniteUpper = true;
  }
  else
  {
    ErrorCode = get_vnode (DiskVolumePntr->nameSpaceID,
      (unsigned int) *IndexFSOPntrPntr, &VoidPntr);
    if (ErrorCode != 0)
    {
      dprintf ("ARFS:FindOptimalIndexAndRange: Can't get_vnode the desired "
        "index (%s), code %d.\n", FSOAsString (&(*IndexFSOPntrPntr)->fso),
        (int) ErrorCode);
    }
  }

  if (NameIndexFSOPntr != NULL)
    put_vnode (DiskVolumePntr->nameSpaceID, (unsigned int) NameIndexFSOPntr);
  return ErrorCode;
}



/******************************************************************************
 * Checks the parse tree to see if the given attribute is used in it.  Returns
 * TRUE if it is used, FALSE if it is absent.  Useful for avoiding a full query
 * evaluation if just one unused attribute is changing.
 */

static bool DoesParseTreeUseAttribute (
  ParseNodePointer ParseNodePntr,
  const char *AttributeName)
{
  if (ParseNodePntr == NULL)
    return false;

  switch (ParseNodePntr->operation)
  {
    case PNO_AND:
    case PNO_OR:
      if (DoesParseTreeUseAttribute (
      ParseNodePntr->opSpecific.children.leftChild, AttributeName))
        return true;
      return DoesParseTreeUseAttribute (
        ParseNodePntr->opSpecific.children.rightChild, AttributeName);

    case PNO_COMPARE:
      return (strcmp (ParseNodePntr->opSpecific.compare.attributeNamePntr,
        AttributeName) == 0);

    default:
      break; /* Unknown parse tree node type. */
  }

  dprintf ("ARFS:DoesParseTreeUseAttribute: Bug!  "
    "Shouldn't get here, corrupt parse tree?\n");
  return false;
}



/******************************************************************************
 * Evaluates the attributes on a file to see if it matches a query (returns
 * TRUE) or doesn't (returns FALSE).  Errors return FALSE too.  Presumably the
 * FSOPntr has been get_vnoded and read locked so that the attributes can be
 * read.  If ExtraAttributeNamePntr is not NULL then it pretends that that
 * attribute is specified by ExtraAttributeDataPntr (NULL to pretend it doesn't
 * exist) rather than the real attribute in FSOPntr.  This function can be
 * extremely recursive (dozens of levels deep) so it's best not to have large
 * local variables, so the two largest work areas are passed in by the caller -
 * being semi-global is OK since they're only used in leaf nodes
 * (comparisions).  They should be declared as uint8 TruncatedKeyword
 * [NAME_MAX]; and PremadeAttributeRecord VirtualAttribute;
 */

static bool DoesFileMatchParseTree (
  FileSystemObjectPointer FSOPntr,
  ParseNodePointer ParseNodePntr,
  const char *ExtraAttributeNamePntr,
  AttributeDataPointer ExtraAttributeDataPntr,
  uint8 *TruncatedKeyword,
  PremadeAttributePointer VirtualAttribute)
{
  AttributeDataPointer            AttributeDataPntr;
  AVLDupThingRecord               AttributeNameThing;
  AVLDupThingRecord               AttributeValueThing;
  struct CompareStruct           *ComparePntr;
  int                             CompareResult;
  int                             KeywordActualSize;
  uint8                          *KeywordsEndPntr;
  uint8                          *KeywordsNextPntr;
  uint8                          *KeywordPntr;
  uint8                           Letter;
  const char                     *PatternErrorMessage;
  unsigned int                    TempUINT;
  int                             TruncatedKeywordSize;

  if (FSOPntr == NULL)
    return false;

  if (ParseNodePntr == NULL)
    return true; /* There's a bit of nothing in everything. */

  switch (ParseNodePntr->operation)
  {
    case PNO_AND:
      if (!DoesFileMatchParseTree (FSOPntr,
      ParseNodePntr->opSpecific.children.leftChild,
      ExtraAttributeNamePntr, ExtraAttributeDataPntr,
      TruncatedKeyword, VirtualAttribute))
        return false;
      return DoesFileMatchParseTree (FSOPntr,
        ParseNodePntr->opSpecific.children.rightChild,
        ExtraAttributeNamePntr, ExtraAttributeDataPntr,
        TruncatedKeyword, VirtualAttribute);


    case PNO_OR:
      if (DoesFileMatchParseTree (FSOPntr,
      ParseNodePntr->opSpecific.children.leftChild,
      ExtraAttributeNamePntr, ExtraAttributeDataPntr,
      TruncatedKeyword, VirtualAttribute))
        return true;
      return DoesFileMatchParseTree (FSOPntr,
        ParseNodePntr->opSpecific.children.rightChild,
        ExtraAttributeNamePntr, ExtraAttributeDataPntr,
        TruncatedKeyword, VirtualAttribute);


    case PNO_COMPARE:
      /* Comparing an attribute, optionally with a given value.  Find the named
      attribute data in the file, or NULL if not there.  Make up virtual
      attribute data for magic index names like name, last_modified etc.  Or
      use the pretend extra attribute if it was specified. */

      ComparePntr = &ParseNodePntr->opSpecific.compare;

      if (ExtraAttributeNamePntr != NULL &&
      strcmp (ComparePntr->attributeNamePntr, ExtraAttributeNamePntr) == 0)
        AttributeDataPntr = ExtraAttributeDataPntr;
      else if (strcmp (ComparePntr->attributeNamePntr, g_NameOfNameIndex) == 0)
      {
        VirtualAttribute->attributeData.type = B_STRING_TYPE;
        VirtualAttribute->attributeData.size =
          strlen (FSOPntr->trueNamePntr) + 1;
        memcpy (VirtualAttribute->storage.nameString,
          FSOPntr->trueNamePntr, VirtualAttribute->attributeData.size);
        AttributeDataPntr = &VirtualAttribute->attributeData;
      }
      else if (strcmp (ComparePntr->attributeNamePntr, g_NameOfSizeIndex) == 0)
      {
        VirtualAttribute->attributeData.size = sizeof (uint64);
        VirtualAttribute->attributeData.type = B_UINT64_TYPE;
        VirtualAttribute->storage.uint64data = GetUserFileSize (FSOPntr);
        AttributeDataPntr = &VirtualAttribute->attributeData;
      }
      else if (strcmp (ComparePntr->attributeNamePntr,
      g_NameOfLastModifiedIndex) == 0)
      {
        VirtualAttribute->attributeData.size = sizeof (int64);
        VirtualAttribute->attributeData.type = B_INT64_TYPE;
        VirtualAttribute->storage.uint64data = FSOPntr->mtime / 1000000;
        /* Tracker expects the date in seconds, sadly not microseconds. */
        AttributeDataPntr = &VirtualAttribute->attributeData;
      }
      else /* Look for a regular attribute with that name. */
      {
        if (FSOPntr->attributesTree == NULL)
          AttributeDataPntr = NULL;
        else
        {
          AttributeNameThing.longStringThing.stringPntr =
            (char *) ComparePntr->attributeNamePntr;
          AttributeNameThing.longStringThing.isLongString = true;
          TempUINT = AVLDupFind (FSOPntr->attributesTree, &AttributeNameThing);
          AttributeDataPntr = (void *) TempUINT;
        }
      }

      if (ComparePntr->relation == TC_ATTRIBUTE_EXISTS)
        return (AttributeDataPntr != NULL);

      if (ComparePntr->relation == TC_ATTRIBUTE_MISSING)
        return (AttributeDataPntr == NULL);

      if (AttributeDataPntr == NULL)
        return false; /* Nothing to compare against, always excluded. */

      /* Prepare the constant Thing used in the comparison, convert it from a
      string to whatever type the attribute has (ignore what type the
      equivalent index is using, if it exists). */

      if (AttributeDataPntr->type != ComparePntr->valueConvertedToType)
      {
        ComparePntr->valueNormalSize =
          AVLDupGetThingTypeSize (AttributeDataPntr->type);
        if (ComparePntr->valueNormalSize < 0)
          dprintf ("ARFS:DoesFileMatchParseTree: Don't know generic "
            "size of type %s.  Will fail all comparisons.  Was testing "
            "attribute \"%s\" of %s.\n",
            TypeCodeAsString (AttributeDataPntr->type),
            ComparePntr->attributeNamePntr, FSOAsString (FSOPntr));

        if (ComparePntr->valueNormalSize == 0 /* a string type of thing */)
        {
          ComparePntr->valueConverted.longStringThing.stringPntr =
            (char *) ComparePntr->valueStringPntr;
          ComparePntr->valueConverted.longStringThing.isLongString = true;
        }
        else if (!AVLDupConvertStringToThing (
        ComparePntr->valueStringPntr,
        &ComparePntr->valueConverted,
        AttributeDataPntr->type,
        NULL /* Malloc function */,
        NULL /* Free function */,
        NULL /* PoolID */))
        {
          dprintf ("ARFS:DoesFileMatchParseTree: Conversion of \"%s\" to "
            "type %s failed, using zero instead.  Was testing "
            "attribute \"%s\" of %s.\n",
            ComparePntr->valueStringPntr,
            TypeCodeAsString (AttributeDataPntr->type),
            ComparePntr->attributeNamePntr, FSOAsString (FSOPntr));
          ComparePntr->valueConverted.uint64Thing = 0;
        }

        /* Set up for comparing the new kind of Thing. */

        ComparePntr->valueComparisonFunction =
          AVLDupGetComparisonFunctionForType (AttributeDataPntr->type);
        if (ComparePntr->valueComparisonFunction == NULL)
          dprintf ("ARFS:DoesFileMatchParseTree: No comparision "
            "function for type %s.  Will fail all comparisons.  Was testing "
            "attribute \"%s\" of %s.\n",
            TypeCodeAsString (AttributeDataPntr->type),
            ComparePntr->attributeNamePntr, FSOAsString (FSOPntr));

        /* Check that the syntax of the pattern is correct, when pattern
        matching is used rather than regular comparisons (patterns are used for
        string equality and inequality). */

        if (ComparePntr->valueNormalSize == 0 &&
        (ComparePntr->relation == TC_EQUAL ||
        ComparePntr->relation == TC_NOT_EQUAL))
        {
          if (!IsValidUTF8Pattern (AVLDupGetStringPntrFromThing (
          ComparePntr->valueConverted), &CompareResult))
          {
            switch (CompareResult)
            {
              case PATTERN_ESC: PatternErrorMessage =
                "pattern has invalid escape ('\' at end of pattern)";
                break;
              case PATTERN_RANGE: PatternErrorMessage =
                "[..] construct has a no end range in a '-' pair (ie [a-])";
                break;
              case PATTERN_CLOSE: PatternErrorMessage =
                "[..] construct has no end bracket (ie [abc-g )";
                break;
              case PATTERN_EMPTY: PatternErrorMessage =
                "[..] construct is empty (ie [])";
                break;
              case PATTERN_UTF8: PatternErrorMessage =
                "mangled UTF-8 encoding encountered in pattern";
                break;
              default: PatternErrorMessage =
                "unknown pattern error";
                break;
            }
            dprintf ("ARFS:DoesFileMatchParseTree: Pattern \"%s\" is bad: "
              "%s.  Will fail all tests using it.  Was testing %s attribute "
              "\"%s\" of %s.\n",
              AVLDupGetStringPntrFromThing (ComparePntr->valueConverted),
              PatternErrorMessage,
              TypeCodeAsString (AttributeDataPntr->type),
              ComparePntr->attributeNamePntr, FSOAsString (FSOPntr));
            ComparePntr->valueComparisonFunction = NULL; /* Turn off tests. */
          }
        }

        ComparePntr->valueConvertedToType = AttributeDataPntr->type;
      }

      if (ComparePntr->valueNormalSize < 0 ||
      ComparePntr->valueComparisonFunction == NULL)
        return false; /* Always fail if we can't compare. */

      /* Prepare a Thing to hold the attribute's data and then test it against
      the constant Thing.  If the attribute is a multiple keyword attribute
      then iterate over all keywords and make a Thing for each one and test
      that Thing.  Return true if any of the keywords matches.  The code for
      iterating over multiple keywords is copied from
      AddOrRemoveAttributeInIndex, so fix that if you change this code. */

      KeywordPntr = AttributeDataPntr->value;
      KeywordsEndPntr = AttributeDataPntr->value + AttributeDataPntr->size;

      while (KeywordPntr < KeywordsEndPntr)
      {
        if (ComparePntr->multipleKeywords)
        {
          if (ComparePntr->valueNormalSize == 0 /* zero for strings */)
          {
            /* Find the start of the keyword.  Skip over leading control
            characters (any UTF-8 code less than 32) and spaces. */

            while (KeywordPntr < KeywordsEndPntr)
            {
              Letter = *KeywordPntr;
              if (Letter > 32 && Letter != ' ')
                break;
              KeywordPntr++;
            }

            /* Find the end of the keyword, and the size of the keyword. */

            KeywordsNextPntr = KeywordPntr;
            while (KeywordsNextPntr < KeywordsEndPntr)
            {
              Letter = *KeywordsNextPntr;
              if (Letter < 32 || Letter == ' ')
                break;
              KeywordsNextPntr++;
            }
          }
          else /* Fixed size multiple keywords. */
          {
            KeywordsNextPntr = KeywordPntr + ComparePntr->valueNormalSize;
            if (KeywordsNextPntr > KeywordsEndPntr)
              KeywordsNextPntr = KeywordsEndPntr; /* Last one is shortened. */
          }
          KeywordActualSize = KeywordsNextPntr - KeywordPntr;
        }
        else /* Whole thing is a keyword. */
        {
          KeywordActualSize = AttributeDataPntr->size;
          KeywordsNextPntr = KeywordPntr + KeywordActualSize;
        }

        /* Ignore keywords which are too short.  This means zero length strings
        in multiple keyword mode (happens for the NUL at the end of the string,
        if it is there) and fixed size keywords less than the fixed size in
        length. */

        if (ComparePntr->valueNormalSize == 0 &&
        ComparePntr->multipleKeywords && KeywordActualSize <= 0)
        {
#if DPRINT_NON_ERROR_MESSAGES
          dprintf ("ARFS:DoesFileMatchParseTree: Skipping empty string "
            "keyword (usually from NUL at end of data) "
            "in attribute \"%s\" of %s.\n",
            ComparePntr->attributeNamePntr, FSOAsString (FSOPntr));
#endif
        }
        else if (ComparePntr->valueNormalSize > 0 &&
        KeywordActualSize < ComparePntr->valueNormalSize)
        {
          dprintf ("ARFS:DoesFileMatchParseTree: Keyword size of %d is too "
            "small for attribute \"%s\" (%s=%d bytes) at offset %d of %d, "
            "skipping comparison for that keyword for %s.\n",
            KeywordActualSize, ComparePntr->attributeNamePntr,
            TypeCodeAsString (AttributeDataPntr->type),
            ComparePntr->valueNormalSize,
            (int) (KeywordPntr - AttributeDataPntr->value),
            (int) AttributeDataPntr->size, FSOAsString (FSOPntr));
        }
        else /* Have keyword data, now is it too long? */
        {
          /* Truncate the keyword to 256 bytes if needed.  Also add a
          terminating NUL to the end of string keywords (otherwise the
          comparisons won't work).  At the end of this, the keyword to be added
          will have been copied to TruncatedKeyword. */

          if (KeywordActualSize >= NAME_MAX)
          {
            memcpy (TruncatedKeyword, KeywordPntr, NAME_MAX);
            if (ComparePntr->valueNormalSize == 0) /* If it's a string. */
              TruncatedKeyword [NAME_MAX - 1] = 0;
            TruncatedKeywordSize = NAME_MAX;
          }
          else /* Keyword data is under the size limit. */
          {
            memcpy (TruncatedKeyword, KeywordPntr, KeywordActualSize);
            if (ComparePntr->valueNormalSize == 0) /* If it's a string. */
            {
              TruncatedKeywordSize = KeywordActualSize;
              if (TruncatedKeywordSize == 0 ||
              TruncatedKeyword [TruncatedKeywordSize - 1] != 0)
              {
                TruncatedKeyword [TruncatedKeywordSize] = 0; /* Add end NUL. */
                TruncatedKeywordSize++;
              }
            }
            else /* A fixed size keyword. */
              TruncatedKeywordSize = KeywordActualSize;
          }

          /* Finally do the actual work.  Make the keyword into a Thing. */

          if (ComparePntr->valueNormalSize == 0)
          {
            AttributeValueThing.longStringThing.stringPntr = TruncatedKeyword;
            AttributeValueThing.longStringThing.isLongString = true;
          }
          else /* A fixed size attribute, up to 8 bytes long. */
          {
            AttributeValueThing.int64Thing = 0; /* Clear unused portion. */
            switch (TruncatedKeywordSize)
            {
              case 0:
                break;
              case 1:
                AttributeValueThing.uint8Thing = * (uint8 *)
                TruncatedKeyword; break;
              case 2: case 3:
                AttributeValueThing.uint16Thing = * (uint16 *)
                TruncatedKeyword; break;
              case 4: case 5: case 6: case 7:
                AttributeValueThing.uint32Thing = * (uint32 *)
                TruncatedKeyword; break;
              case 8: default:
                AttributeValueThing.uint64Thing = * (uint64 *)
                TruncatedKeyword; break;
            }
          }

          /* Now compare it.  String Things and equality tests use pattern
          matching, everything else uses regular difference comparisons
          (effectively a subtraction and testing the sign of the result). */

          if (ComparePntr->valueNormalSize == 0 &&
          (ComparePntr->relation == TC_EQUAL ||
          ComparePntr->relation == TC_NOT_EQUAL))
          {
            CompareResult = ! /* zero means equal so negate */ MatchUTF8 (
              AVLDupGetStringPntrFromThing (ComparePntr->valueConverted),
              AVLDupGetStringPntrFromThing (AttributeValueThing));
          }
          else /* Regular value comparison. */
            CompareResult = ComparePntr->valueComparisonFunction (
              &AttributeValueThing, &ComparePntr->valueConverted);

          switch (ComparePntr->relation)
          {
            case TC_EQUAL:
              if (CompareResult == 0)
                return true;
              break;

            case TC_GREATER_THAN:
              if (CompareResult > 0)
                return true;
              break;

            case TC_LESS_THAN:
              if (CompareResult < 0)
                return true;
              break;

            case TC_GREATER_THAN_OR_EQUAL:
              if (CompareResult >= 0)
                return true;
              break;

            case TC_LESS_THAN_OR_EQUAL:
              if (CompareResult <= 0)
                return true;
              break;

            case TC_NOT_EQUAL:
              if (CompareResult != 0)
                return true;
              break;

            default:
              dprintf ("ARFS:DoesFileMatchParseTree: Bug!  Unknown relation "
                "#%d for comparison of attribute to \"%s\", was testing "
                "attribute \"%s\" of %s.\n",
                (int) ComparePntr->relation, ComparePntr->valueStringPntr,
                ComparePntr->attributeNamePntr, FSOAsString (FSOPntr));
              return false;
          }
        }

        /* Advance to the next keyword. */
        KeywordPntr = KeywordsNextPntr;
      }

      /* None of the keywords matched if we got here. */
      return false;


    default:
      break; /* Unknown parse tree node type. */
  }

  dprintf ("ARFS:DoesFileMatchParseTree: Bug!  "
    "Shouldn't get here, corrupt parse tree?\n");
  return false;
}



/******************************************************************************
 * Adds or removes one key from an index.  AddIt is TRUE to add, FALSE to
 * remove.  IndexPntr identifies the index.  SizeOfDataElement defines the size
 * of the type of data, zero means strings and other values are for fixed size
 * data types.  FSOPntr is the file or other thing which is associated with the
 * key.  KeywordPntr points to the key's data.  KeywordSize is the actual size
 * of the data.  Returns an error if a bug has been detected or if it's out of
 * memory, smaller glitches only get a debug message.
 */

static status_t AddOrRemoveOneKeyInIndex (
  DiskVolumePointer DiskVolumePntr,
  bool AddIt,
  IndexObjectPointer IndexPntr,
  type_code KeyType,
  int SizeOfDataElement,
  FileSystemObjectPointer FSOPntr,
  uint8 *KeywordPntr,
  int KeywordSize)
{
  AVLDupAddReturnCode AVLCode;
  status_t            ErrorCode;
  char                KeyAsString [NAME_MAX];
  int                 KeyAsStringLength;
  status_t            ReturnCode = -1;
  AVLDupThingRecord   TempKey;

  if (KeywordSize <= 0)
  {
    dprintf ("ARFS:AddOrRemoveOneKeyInIndex: Key's data is size %d; "
      "even empty strings have 1 byte.  Thing %s, in %s.\n",
      KeywordSize, FSOAsString (FSOPntr), FSOAsString (&IndexPntr->fso));
    goto ErrorExit;
  }

  if (SizeOfDataElement != 0 && KeywordSize != SizeOfDataElement)
  {
    dprintf ("ARFS:AddOrRemoveOneKeyInIndex: Key's data is size %d "
      "while we were expecting %d.  Thing %s, in %s.\n",
      KeywordSize, SizeOfDataElement,
      FSOAsString (FSOPntr), FSOAsString (&IndexPntr->fso));
    goto ErrorExit;
  }

  /* Convert the attribute value into a key thing. */

  TempKey.int64Thing = 0; /* Clear unused fields for shorter datatypes. */

  switch (SizeOfDataElement)
  {
    case 0: /* A variable length string. */
      TempKey.longStringThing.stringPntr = KeywordPntr;
      TempKey.longStringThing.isLongString = true;
      break;

    case 1:
      TempKey.uint8Thing = * (uint8 *) KeywordPntr;
      break;

    case 2:
      TempKey.uint16Thing = * (uint16 *) KeywordPntr;
      break;

    case 4:
      TempKey.uint32Thing = * (uint32 *) KeywordPntr;
      break;

    case 8:
      TempKey.uint64Thing = * (uint64 *) KeywordPntr;
      break;

    default:
      dprintf ("ARFS:AddOrRemoveOneKeyInIndex: Unhandled data "
        "element size of %d, type %s.  Thing %s in %s.\n",
        SizeOfDataElement, TypeCodeAsString (KeyType),
        FSOAsString (FSOPntr), FSOAsString (&IndexPntr->fso));
      goto ErrorExit;
  }

  /* Get the string version of the key, used in change notification messages
  and the occasional error message.  To avoid problems with old programs that
  get confused when several files in the directory seemingly have the same
  name, we make the names unique by appending the VNodeID in hex. */

  strcpy (KeyAsString, "?");
  AVLDupConvertThingToString (&TempKey, KeyType,
    KeyAsString, sizeof (KeyAsString));
  KeyAsStringLength = strlen (KeyAsString);
  if (KeyAsStringLength + 10 < NAME_MAX)
    sprintf (KeyAsString + KeyAsStringLength, " #%08X",
    (unsigned int) FSOPntr);

  /* Do index processing. */

  if (AddIt)
  {
    /* Add the thing to the index. */

    AVLCode = AVLDupAdd (IndexPntr->treePntr,
      &TempKey, (unsigned int) FSOPntr);

    if (AVLCode < 0)
    {
      dprintf ("ARFS:AddOrRemoveOneKeyInIndex: Out of memory (AVL code %d) "
        "while adding thing %s to %s.\n",
        (int) AVLCode, FSOAsString (FSOPntr), FSOAsString (&IndexPntr->fso));
      ReturnCode = ENOMEM;
      goto ErrorExit;
    }

    if (AVLCode == AVLADD_ALREADY_IN_TREE)
    {
      dprintf ("ARFS:AddOrRemoveOneKeyInIndex: %s key (%s) "
        "is already in %s for %s.\n",
        IndexPntr->interpretAsMultipleKeywords
          ? "You are wasting time with duplicate keys in a multiple "
            "keyword attribute because"
          : "Bug!  It seems that",
        KeyAsString, FSOAsString (&IndexPntr->fso), FSOAsString (FSOPntr));
      if (!IndexPntr->interpretAsMultipleKeywords)
        goto ErrorExit; /* Shouldn't happen with single keyword indices. */
    }

    if (AVLCode == AVLADD_ADDED_A_NODE)
    {
      ErrorCode = notify_listener (B_ENTRY_CREATED,
        DiskVolumePntr->nameSpaceID, (unsigned int) IndexPntr, 0,
        1 /* fake symbolic link code */ + (unsigned int) FSOPntr,
        KeyAsString);
      if (ErrorCode != 0 || DPRINT_NOTIFICATIONS)
      {
        dprintf ("ARFS:AddOrRemoveOneKeyInIndex: B_ENTRY_CREATED "
          "notification for creating %s in %s (key %s), error code %d.\n",
          FSOAsString ((FileSystemObjectPointer)
          (1 /* fake sym link */ + (unsigned int) FSOPntr)),
          FSOAsString (&IndexPntr->fso), KeyAsString, (int) ErrorCode);
      }
    }
  }
  else /* Remove the thing from the index. */
  {
    if (AVLDupDelete (IndexPntr->treePntr,
    &TempKey, (unsigned int) FSOPntr))
    {
      ErrorCode = notify_listener (B_ENTRY_REMOVED,
        DiskVolumePntr->nameSpaceID, (unsigned int) IndexPntr, 0,
        1 /* fake symbolic link code */ + (unsigned int) FSOPntr,
        KeyAsString);
      if (ErrorCode != 0 || DPRINT_NOTIFICATIONS)
      {
        dprintf ("ARFS:AddOrRemoveOneKeyInIndex: B_ENTRY_REMOVED "
          "notification for removing %s from %s (key %s), error code %d.\n",
          FSOAsString ((FileSystemObjectPointer)
          (1 /* fake sym link */ + (unsigned int) FSOPntr)),
          FSOAsString (&IndexPntr->fso), KeyAsString, (int) ErrorCode);
      }
    }
    else /* Delete failed.  Not really a problem, can happen normally. */
    {
      dprintf ("ARFS:AddOrRemoveOneKeyInIndex: Key (%s) "
        "can't be deleted from %s for %s, because it is not there.\n",
        KeyAsString, FSOAsString (&IndexPntr->fso), FSOAsString (FSOPntr));
    }
  }

  ReturnCode = 0;

ErrorExit:
  return ReturnCode;
}



/******************************************************************************
 * Adds or removes attribute contents in an index.  Also updates queries.
 *
 * If AddIt is TRUE it will add them, if it is FALSE they get removed.
 *
 * Assumes the caller has read locked the indices directory, which is specified
 * in IndicesDirectoryPntr (in case we ever have multiple indice directories,
 * possibly with a search path to find the one to use, also can be NULL if
 * there are no indices or if it is one of the standard indices (which we
 * already know the vnode_id of, don't need the .Indices directory)).
 *
 * AttributeNamePntr names the index to use.  It's OK if it doesn't exist, but
 * if it does exist it will get write locked.  In either case any open queries
 * refering to the attribute name will also get updated.
 *
 * For multiple key indices this function will break up the attribute value
 * into single keys and add or remove them all.  The value corresponding to the
 * key(s) will be FSOPntr, the thing that the attribute is attached to.
 * Presumably FSOPntr is read locked.  If the attribute type doesn't match the
 * index type, it will treat it as if it is the index type (mostly for
 * backwards compatability).  Keys which are too long are truncated to the
 * NAME_MAX byte maximum.  Keys which are too short (for fixed size data types)
 * will be ignored.
 */

static status_t AddOrRemoveAttributeInIndex (
  DiskVolumePointer DiskVolumePntr,
  bool AddIt,
  DirectoryObjectPointer IndicesDirectoryPntr,
  const char *AttributeNamePntr,
  AttributeDataPointer AttributeDataPntr,
  FileSystemObjectPointer FSOPntr)
{
  QueryStatePointer      CurrentQueryPntr;
  status_t               ErrorCode;
  int                    i;
  bool                   IndexGetVNodeDone = false;
  IndexObjectPointer     IndexPntr = NULL;
  bool                   IndexWriteLocked = false;
  type_code              KeyType;
  int                    KeywordActualSize;
  uint8                 *KeywordsEndPntr;
  uint8                 *KeywordsNextPntr;
  uint8                 *KeywordPntr;
  uint8                  Letter;
  bool                   MultipleKeywords;
  char                  *NameForNotificationPntr;
  bool                   PostInSet;
  bool                   PreviouslyInSet;
  status_t               ReturnCode = -1;
  int                    SizeOfDataElement;
  AVLDupThingRecord      TempKey;
  unsigned int           TempUINT;
  uint8                  TruncatedKeyword [NAME_MAX];
  int                    TruncatedKeywordSize;
  PremadeAttributeRecord VirtualAttribute;
  void                  *VoidPntr;

  if (FSOPntr->objectType == FSO_TYPE_INDEX)
    return 0; /* Indices don't appear in the indices for simplicity. */
  if (strcmp (AttributeNamePntr, g_NameOfSizeIndex) == 0 &&
  FSOPntr->objectType != FSO_TYPE_FILE)
    return 0; /* Only true files are in the size index for simplicity. */

  /* Send query update notifications for all live queries where changing this
  attribute adds or removes the file from the query result set. */

  ErrorCode = ReadLockLiveQueryList (DiskVolumePntr);
  if (ErrorCode != 0)
  {
    dprintf ("ARFS:AddOrRemoveAttributeInIndex: Unable to lock live "
      "query list, code %d.  Was doing attribute \"%s\" for %s.\n",
      (int) ErrorCode, AttributeNamePntr, FSOAsString (FSOPntr));
    ReturnCode = ErrorCode;
    goto ErrorExit;
  }

  CurrentQueryPntr = DiskVolumePntr->liveQueryListHead;
  while (CurrentQueryPntr != NULL)
  {
    if (DoesParseTreeUseAttribute (CurrentQueryPntr->parseTreePntr,
    AttributeNamePntr))
    {
      PreviouslyInSet = DoesFileMatchParseTree (
        FSOPntr, CurrentQueryPntr->parseTreePntr,
        AttributeNamePntr, AddIt ? NULL : AttributeDataPntr,
        TruncatedKeyword, &VirtualAttribute);

      PostInSet = DoesFileMatchParseTree (
        FSOPntr, CurrentQueryPntr->parseTreePntr,
        AttributeNamePntr, AddIt ? AttributeDataPntr : NULL,
        TruncatedKeyword, &VirtualAttribute);

      if (PreviouslyInSet != PostInSet)
      {
        /* The file name in the FSOPntr may be wrong (the old name) when
        renaming a file.  Use the new name for notification in that case. */
        if (AddIt && AttributeDataPntr->type == B_STRING_TYPE &&
        strcmp (AttributeNamePntr, g_NameOfNameIndex) == 0)
          NameForNotificationPntr = AttributeDataPntr->value;
        else
          NameForNotificationPntr = FSOPntr->trueNamePntr;

        ErrorCode = send_notification (CurrentQueryPntr->liveQueryPortID,
          CurrentQueryPntr->callbackToken, B_QUERY_UPDATE,
          PostInSet ? B_ENTRY_CREATED : B_ENTRY_REMOVED,
          DiskVolumePntr->nameSpaceID,
          0 /* Unused name space ID B */,
          (unsigned int) &FSOPntr->mainParentDirectory->fso,
          0 /* Unused vnode ID B */,
          (unsigned int) FSOPntr,
          NameForNotificationPntr /* Must never be NULL */);

        if (ErrorCode != 0 || DPRINT_NOTIFICATIONS)
        {
          dprintf ("ARFS:AddOrRemoveAttributeInIndex: B_QUERY_UPDATE/%s "
            "notification for %s attribute \"%s\" %s, file name \"%s\", "
            "error code %d.\n",
            PostInSet ? "B_ENTRY_CREATED" : "B_ENTRY_REMOVED",
            FSOAsString (FSOPntr), AttributeNamePntr,
            AddIt ? "addition" : "removal",
            NameForNotificationPntr, (int) ErrorCode);
        }
      }
    }
    CurrentQueryPntr = CurrentQueryPntr->nextLiveQueryPntr;
  }
  ReadUnlockLiveQueryList (DiskVolumePntr);


  /* Look for the index specified by the attribute name.  First try the
  standard indices. */

  IndexPntr = NULL;
  for (i = 0; i < STDINUM_MAX; i++)
  {
    if (strcmp (AttributeNamePntr, g_NamesOfStandardIndices [i]) == 0)
    {
      IndexPntr = (IndexObjectPointer) (unsigned int)
        DiskVolumePntr->standardIndicesVnodes[i];
      break;
    }
  }

  /* Look for the index in the global .Indices directory if it hasn't been
  found yet. */

  if (IndexPntr == NULL && IndicesDirectoryPntr != NULL)
  {
    TempKey.longStringThing.stringPntr = (char *) AttributeNamePntr;
    TempKey.longStringThing.isLongString = true;
    TempUINT = AVLDupFind (IndicesDirectoryPntr->directory, &TempKey);
    IndexPntr = (IndexObjectPointer) TempUINT; /* Can be NULL if none. */
  }

  if (IndexPntr == NULL)
  {
    /* Nothing more to do if there isn't an index for this attribute. */
    ReturnCode = 0;
    goto ErrorExit;
  }

  /* Try to get_vnode the index.  Also check that it is a valid index thing. */

  ErrorCode = get_vnode (DiskVolumePntr->nameSpaceID,
    (unsigned int) IndexPntr, &VoidPntr);
  if (ErrorCode < 0)
  {
    dprintf ("ARFS:AddOrRemoveAttributeInIndex: get_vnode failed for "
      "index \"%s\" in %s (trying to load %s), code %d.\n",
      AttributeNamePntr, FSOAsString (&IndicesDirectoryPntr->fso),
      FSOAsString (&IndexPntr->fso), (int) ErrorCode);
    ReturnCode = ErrorCode;
    goto ErrorExit;
  }
  IndexGetVNodeDone = true;

  if (VoidPntr != IndexPntr)
  {
    dprintf ("ARFS:AddOrRemoveAttributeInIndex: Bug!  get_vnode returned "
      "wrong pointer $%X for %s.\n",
      (unsigned int) VoidPntr, FSOAsString (&IndexPntr->fso));
    goto ErrorExit;
  }

  if (INVALID_FILE_SYSTEM_OBJECT (&IndexPntr->fso))
  {
    dprintf ("ARFS:AddOrRemoveAttributeInIndex: Bug!  Found an index "
      "for \"%s\" in %s, but it is a corrupt %s.\n",
      AttributeNamePntr, FSOAsString (&IndicesDirectoryPntr->fso),
      FSOAsString (&IndexPntr->fso));
    goto ErrorExit;
  }

  if (IndexPntr->fso.objectType != FSO_TYPE_INDEX)
  {
    dprintf ("ARFS:AddOrRemoveAttributeInIndex: While looking for an index "
      "for attribute \"%s\" in %s, found %s, which is not an index!  "
      "Now pretending that no index exists.\n",
      AttributeNamePntr, FSOAsString (&IndicesDirectoryPntr->fso),
      FSOAsString (&IndexPntr->fso));
    put_vnode (DiskVolumePntr->nameSpaceID, (unsigned int) IndexPntr);
    IndexGetVNodeDone = false;
    IndexPntr = NULL;
  }

  /* Write lock the index, we can assume that it's not the same as the already
  write locked FSOPntr since we no longer allow indices to be indexed. */

  ErrorCode = WriteLock (&IndexPntr->fso);
  if (ErrorCode != 0)
  {
    dprintf ("ARFS:AddOrRemoveAttributeInIndex: unable to get a "
      "write lock for %s, code %d.\n",
      FSOAsString (&IndexPntr->fso), (int) ErrorCode);
    ReturnCode = ErrorCode;
    goto ErrorExit;
  }
  IndexWriteLocked = true;

  /* Check the data type of the attribute and the index. */

  MultipleKeywords = IndexPntr->interpretAsMultipleKeywords;
  KeyType = IndexPntr->keyType;
  SizeOfDataElement = IndexPntr->normalAttributeSize;
  if (KeyType != AttributeDataPntr->type &&
  SizeOfDataElement != AVLDupGetThingTypeSize (AttributeDataPntr->type))
  {
    dprintf ("ARFS:AddOrRemoveAttributeInIndex: Data type of "
      "attribute \"%s\" is %s, which isn't even vaguely similar to the "
      "index type of %s (size %d vs %d).  Aborting processing for %s "
      "(even if query processing might work).\n",
      AttributeNamePntr, TypeCodeAsString (AttributeDataPntr->type),
      TypeCodeAsString (KeyType),
      SizeOfDataElement, AVLDupGetThingTypeSize (AttributeDataPntr->type),
      FSOAsString (FSOPntr));
    ReturnCode = ERANGE;
    goto ErrorExit;
  }

  /* Break up the attribute into individual keys and add/remove each one.  But
  only if there is an index and it is a multiple keyword index.  Otherwise the
  whole attribute value is the key to add/remove.  Well, not quite, only the
  first 256 bytes (NAME_MAX = 256 = maximum file name length) gets stored in
  the tree to save space (in case someone went and indexed thumbnail bitmaps or
  something else large).  See also DoesFileMatchParseTree which uses a copy of
  this code, and fix it if you change this. */

  KeywordPntr = AttributeDataPntr->value;
  KeywordsEndPntr = AttributeDataPntr->value + AttributeDataPntr->size;

  while (KeywordPntr < KeywordsEndPntr)
  {
    if (MultipleKeywords)
    {
      if (SizeOfDataElement == 0 /* zero for strings */)
      {
        /* Find the start of the keyword.  Skip over leading control characters
        (any UTF-8 code less than 32) and spaces. */

        while (KeywordPntr < KeywordsEndPntr)
        {
          Letter = *KeywordPntr;
          if (Letter > 32 && Letter != ' ')
            break;
          KeywordPntr++;
        }

        /* Find the end of the keyword, and the size of the keyword. */

        KeywordsNextPntr = KeywordPntr;
        while (KeywordsNextPntr < KeywordsEndPntr)
        {
          Letter = *KeywordsNextPntr;
          if (Letter < 32 || Letter == ' ')
            break;
          KeywordsNextPntr++;
        }
      }
      else /* Fixed size multiple keywords. */
      {
        KeywordsNextPntr = KeywordPntr + SizeOfDataElement;
        if (KeywordsNextPntr > KeywordsEndPntr)
          KeywordsNextPntr = KeywordsEndPntr; /* Last one is shortened. */
      }
      KeywordActualSize = KeywordsNextPntr - KeywordPntr;
    }
    else /* Whole thing is a keyword. */
    {
      KeywordActualSize = AttributeDataPntr->size;
      KeywordsNextPntr = KeywordPntr + KeywordActualSize;
    }

    /* Ignore keywords which are too short.  This means zero length strings in
    multiple keyword mode (happens for the NUL at the end of the string, if it
    is there) and fixed size keywords less than the fixed size in length. */

    if (SizeOfDataElement == 0 && MultipleKeywords && KeywordActualSize <= 0)
    {
#if DPRINT_NON_ERROR_MESSAGES
      dprintf ("ARFS:AddOrRemoveAttributeInIndex: Skipping empty string "
        "keyword (usually from NUL at end of data) "
        "in attribute \"%s\" of %s.\n",
        AttributeNamePntr, FSOAsString (FSOPntr));
#endif
    }
    else if (SizeOfDataElement > 0 && KeywordActualSize < SizeOfDataElement)
    {
      dprintf ("ARFS:AddOrRemoveAttributeInIndex: Keyword size of %d is too "
        "small for attribute \"%s\" (%s=%d bytes) at offset %d of %d, "
        "skipping index processing for that keyword for %s.\n",
        KeywordActualSize, AttributeNamePntr,
        TypeCodeAsString (AttributeDataPntr->type), SizeOfDataElement,
        (int) (KeywordPntr - AttributeDataPntr->value),
        (int) AttributeDataPntr->size, FSOAsString (FSOPntr));
    }
    else /* Have keyword data, now is it too long? */
    {
      /* Truncate the keyword to 256 bytes if needed.  Also add a terminating
      NUL to the end of string keywords (otherwise the comparisons won't work).
      At the end of this, the keyword to be added will have been copied to
      TruncatedKeyword. */

      if (KeywordActualSize >= sizeof (TruncatedKeyword))
      {
        memcpy (TruncatedKeyword, KeywordPntr, sizeof (TruncatedKeyword));
        if (SizeOfDataElement == 0) /* If it is a string keyword. */
          TruncatedKeyword [sizeof (TruncatedKeyword) - 1] = 0;
        TruncatedKeywordSize = sizeof (TruncatedKeyword);
      }
      else /* Keyword data is under the size limit. */
      {
        memcpy (TruncatedKeyword, KeywordPntr, KeywordActualSize);
        if (SizeOfDataElement == 0) /* If it is a string keyword. */
        {
          TruncatedKeywordSize = KeywordActualSize;
          if (TruncatedKeywordSize == 0 ||
          TruncatedKeyword [TruncatedKeywordSize - 1] != 0)
          {
            TruncatedKeyword [TruncatedKeywordSize] = 0; /* Add NUL at end. */
            TruncatedKeywordSize++;
          }
        }
        else /* A fixed size keyword. */
          TruncatedKeywordSize = KeywordActualSize;
      }

      /* Finally do the actual add/remove of the keyword into the index. */

      ErrorCode = AddOrRemoveOneKeyInIndex (DiskVolumePntr, AddIt, IndexPntr,
        KeyType, SizeOfDataElement, FSOPntr,
        TruncatedKeyword, TruncatedKeywordSize);

      if (ErrorCode < 0)
      {
        dprintf ("ARFS:AddOrRemoveAttributeInIndex: Error %d when "
          "%s keyword with size of %d, attribute \"%s\" (%s=%d bytes) "
          "at offset %d of %d, for %s.%s\n",
          (int) ErrorCode, AddIt ? "adding" : "removing",
          KeywordActualSize, AttributeNamePntr,
          TypeCodeAsString (AttributeDataPntr->type), SizeOfDataElement,
          (int) (KeywordPntr - AttributeDataPntr->value),
          (int) AttributeDataPntr->size, FSOAsString (FSOPntr),
          (IndexPntr == NULL) ? ""
          : "  May now be missing entries in the index.");
        if (AddIt) /* Don't care about failed removals. */
        {
          ReturnCode = ErrorCode;
          goto ErrorExit;
        }
      }
    }

    /* Advance to the next keyword. */
    KeywordPntr = KeywordsNextPntr;
  }

  /* Update the date last modified of the index, since it has changed.  Can't
  use ChangeFSODates since that would make it too recursive. */

  IndexPntr->fso.mtime = real_time_clock_usecs ();

  ReturnCode = 0;

ErrorExit:
  if (IndexWriteLocked)
    WriteUnlock (&IndexPntr->fso);
  if (IndexGetVNodeDone)
    put_vnode (DiskVolumePntr->nameSpaceID, (unsigned int) IndexPntr);
  return ReturnCode;
}



/******************************************************************************
 * Add the newly created thing to the standard indices.
 */

static status_t AddFreshThingToStandardIndices (
  DiskVolumePointer DiskVolumePntr,
  FileSystemObjectPointer FSOPntr)
{
  status_t                ErrorCode;
  status_t                ReturnCode = 0;
  PremadeAttributeRecord  TempAttributeData;

  /* Update the name index. */

  TempAttributeData.attributeData.size = strlen (FSOPntr->trueNamePntr) + 1;
  TempAttributeData.attributeData.type = B_STRING_TYPE;
  strcpy (TempAttributeData.storage.nameString, FSOPntr->trueNamePntr);
  ErrorCode = AddOrRemoveAttributeInIndex (DiskVolumePntr,
    true /* false for remove operation */, NULL /* IndicesDirectoryPntr */,
    g_NameOfNameIndex, &TempAttributeData.attributeData, FSOPntr);
  if (ErrorCode < 0) ReturnCode = ErrorCode;

  /* Update the size index. */

  TempAttributeData.attributeData.size = sizeof (uint64);
  TempAttributeData.attributeData.type = B_UINT64_TYPE;
  TempAttributeData.storage.uint64data = 0;
  ErrorCode = AddOrRemoveAttributeInIndex (DiskVolumePntr,
    true, NULL /* IndicesDirectoryPntr */, g_NameOfSizeIndex,
    &TempAttributeData.attributeData, FSOPntr);
  if (ErrorCode < 0) ReturnCode = ErrorCode;

  /* Update the last modified time index. */

  TempAttributeData.attributeData.size = sizeof (uint64);
  TempAttributeData.attributeData.type = B_INT64_TYPE;
  TempAttributeData.storage.uint64data = FSOPntr->mtime;
  ErrorCode = AddOrRemoveAttributeInIndex (DiskVolumePntr,
    true, NULL /* IndicesDirectoryPntr */, g_NameOfLastModifiedIndex,
    &TempAttributeData.attributeData, FSOPntr);
  if (ErrorCode < 0) ReturnCode = ErrorCode;

  return ReturnCode;
}



/******************************************************************************
 * Internal function to remove one attribute from the index as a tree iteration
 * callback.
 */

typedef struct RemoveAttributeArgumentsStruct
{
  DiskVolumePointer diskVolumePntr;
  DirectoryObjectPointer indicesDirectoryPntr;
  FileSystemObjectPointer fsoPntr;
} RemoveAttributeArgumentsRecord, *RemoveAttributeArgumentsPointer;


static bool RemoveOneAttributeFromIndex (
  AVLDupThingConstPointer KeyPntr,
  const unsigned int *ValuePntr,
  void *ExtraData)
{
  RemoveAttributeArgumentsPointer ArgsPntr;
  const char                     *AttributeNamePntr;
  AttributeDataPointer            AttributeDataPntr;
  status_t                        ErrorCode;

  ArgsPntr = ExtraData;
  AttributeNamePntr = AVLDupGetStringPntrFromThing (*KeyPntr);
  AttributeDataPntr = (AttributeDataPointer) *ValuePntr;

  ErrorCode = AddOrRemoveAttributeInIndex (ArgsPntr->diskVolumePntr,
    false /* false to remove things */, ArgsPntr->indicesDirectoryPntr,
    AttributeNamePntr, AttributeDataPntr, ArgsPntr->fsoPntr);

  if (ErrorCode < 0)
  {
    dprintf ("ARFS:RemoveOneAttributeFromIndex: Unable to remove attribute "
      "%s of %s from index, code %d.  Not that it matters much.\n",
      AttributeNamePntr, FSOAsString (ArgsPntr->fsoPntr), (int) ErrorCode);
  }

  return true; /* Continue iterating. */
}



/******************************************************************************
 * Removes the standard virtual attributes from the indices.  File name,
 * modification date and size.  Never gives up, even if errors happen.
 */

static void RemoveStandardVirtualAttributesFromIndices (
  DiskVolumePointer DiskVolumePntr,
  FileSystemObjectPointer FSOPntr)
{
  PremadeAttributeRecord  TempAttributeData;

  TempAttributeData.attributeData.size = strlen (FSOPntr->trueNamePntr) + 1;
  TempAttributeData.attributeData.type = B_STRING_TYPE;
  strcpy (TempAttributeData.storage.nameString, FSOPntr->trueNamePntr);
  AddOrRemoveAttributeInIndex (DiskVolumePntr,
    false /* false for remove operation */, NULL /* IndicesDirPntr */,
    g_NameOfNameIndex, &TempAttributeData.attributeData, FSOPntr);

  TempAttributeData.attributeData.size = sizeof (uint64);
  TempAttributeData.attributeData.type = B_UINT64_TYPE;
  TempAttributeData.storage.uint64data = GetUserFileSize (FSOPntr);
  AddOrRemoveAttributeInIndex (DiskVolumePntr,
    false, NULL /* IndicesDirPntr */, g_NameOfSizeIndex,
    &TempAttributeData.attributeData, FSOPntr);

  TempAttributeData.attributeData.size = sizeof (uint64);
  TempAttributeData.attributeData.type = B_INT64_TYPE;
  TempAttributeData.storage.uint64data = FSOPntr->mtime;
  AddOrRemoveAttributeInIndex (DiskVolumePntr,
    false, NULL /* IndicesDirPntr */, g_NameOfLastModifiedIndex,
    &TempAttributeData.attributeData, FSOPntr);
}



/******************************************************************************
 * Removes all attributes from the indices for the given file.  Doesn't
 * actually delete the attributes, just deindexes them.  Assumes that the file
 * has been read locked (so that nobody changes its list of attributes).
 */

static void RemoveAllAttributesFromIndex (
  DiskVolumePointer DiskVolumePntr,
  FileSystemObjectPointer FSOPntr)
{
  RemoveAttributeArgumentsRecord Arguments;

  RemoveStandardVirtualAttributesFromIndices (DiskVolumePntr, FSOPntr);

  if (FSOPntr->attributesTree != NULL)
  {
    Arguments.diskVolumePntr = DiskVolumePntr;
    Arguments.fsoPntr = FSOPntr;
    Arguments.indicesDirectoryPntr =
      GetAndLockIndicesDirectory (DiskVolumePntr, FSOPntr, NULL);
    if (Arguments.indicesDirectoryPntr == NULL)
      return; /* Nothing to do if no indices in system. */

    AVLDupIterate (FSOPntr->attributesTree, true,
      NULL, 0, 0, false,
      NULL, 0, 0, false,
      RemoveOneAttributeFromIndex, &Arguments);

    PutIndicesDirectory (DiskVolumePntr,
      Arguments.indicesDirectoryPntr, FSOPntr, NULL);
  }
}



/******************************************************************************
 * Updates the selected dates with the specified time (or current time if you
 * specify -1) and updates the corresponding indices.  But doesn't send
 * B_STAT_CHANGED notifications.  Assumes FSOPntr is write locked so that it
 * can change the date fields in it.
 */

static void ChangeFSODates (
  DiskVolumePointer DiskVolumePntr,
  FileSystemObjectPointer FSOPntr,
  DateKindsFlags DateFlags,
  bigtime_t NewDate)
{
  DateKinds              CurrentDateKind;
  DateKindsFlags         CurrentFlag;
  bigtime_t             *DatePntr;
  status_t               ErrorCode;
  PremadeAttributeRecord TempAttributeData;

  if (FSOPntr == NULL)
    return;

  if (NewDate == -1)
    NewDate = real_time_clock_usecs();

  for (CurrentDateKind = 0; CurrentDateKind < DK_MAX; CurrentDateKind++)
  {
    CurrentFlag = (1 << CurrentDateKind);
    if ((DateFlags & CurrentFlag) == 0)
      continue; /* Doesn't want to change this kind of date. */
    switch (CurrentDateKind)
    {
      case DK_ACCESS: DatePntr = &FSOPntr->atime; break;
      case DK_MODIFIED: DatePntr = &FSOPntr->mtime; break;
      case DK_INODE_CHANGED: DatePntr = &FSOPntr->ctime; break;
      case DK_CREATED: DatePntr = &FSOPntr->crtime; break;
      default:
        dprintf ("ARFS:ChangeFSODates: Bug!  No date #%d.\n",
          (int) CurrentDateKind);
        continue;
    }

    /* Update the "last_modified" index, if it was the one changed. */

    if (CurrentDateKind == DK_MODIFIED)
    {
      TempAttributeData.attributeData.size = sizeof (NewDate);
      TempAttributeData.attributeData.type = B_INT64_TYPE;
      TempAttributeData.storage.uint64data = *DatePntr;
      ErrorCode = AddOrRemoveAttributeInIndex (DiskVolumePntr,
        false /* false for remove operation */, NULL /* IndicesDirPntr */,
        g_NameOfLastModifiedIndex, &TempAttributeData.attributeData,
        FSOPntr);
      if (ErrorCode != 0)
        dprintf ("ARFS:ChangeFSODates: Failed to remove old date "
          "%Ld for %s from date index, error %d.\n",
          TempAttributeData.storage.uint64data,
          FSOAsString (FSOPntr), (int) ErrorCode);

      *DatePntr = NewDate;

      TempAttributeData.storage.uint64data = NewDate;
      AddOrRemoveAttributeInIndex (DiskVolumePntr,
        true /* false for remove operation */, NULL /* IndicesDirPntr */,
        g_NameOfLastModifiedIndex, &TempAttributeData.attributeData,
        FSOPntr);
      if (ErrorCode != 0)
        dprintf ("ARFS:ChangeFSODates: Failed to add new date "
          "%Ld for %s to date index, error %d.\n",
          TempAttributeData.storage.uint64data,
          FSOAsString (FSOPntr), (int) ErrorCode);
    }
    *DatePntr = NewDate;
  }
}



/******************************************************************************
 * Updates the name index for the given file system object by removing the old
 * name and adding the new name.  Assumes the index isn't already locked.  If
 * either name is NULL, it skips that part of the operation.  If it fails (due
 * to locking problems or out of memory), it just leaves the index in an
 * inconsistent state.
 */

static void ChangeNameIndexForFSO (
  DiskVolumePointer DiskVolumePntr,
  FileSystemObjectPointer FSOPntr,
  const char *OldName,
  const char *NewName)
{
  status_t               ErrorCode;
  PremadeAttributeRecord TempAttributeData;

  if (FSOPntr == NULL)
    return;

  if (OldName != NULL)
  {
    TempAttributeData.attributeData.size = strlen (OldName) + 1;
    if (TempAttributeData.attributeData.size > NAME_MAX)
      TempAttributeData.attributeData.size = NAME_MAX;
    TempAttributeData.attributeData.type = B_STRING_TYPE;
    memcpy (TempAttributeData.storage.nameString, OldName,
      TempAttributeData.attributeData.size);
    TempAttributeData.storage.nameString[NAME_MAX-1] = 0;

    ErrorCode = AddOrRemoveAttributeInIndex (DiskVolumePntr,
      false /* false for remove operation */, NULL /* IndicesDirPntr */,
      g_NameOfNameIndex, &TempAttributeData.attributeData, FSOPntr);
    if (ErrorCode != 0)
      dprintf ("ARFS:ChangeNameIndexForFSO: Failed to remove old name "
        "\"%s\" for %s from name index, error %d.\n",
        TempAttributeData.storage.nameString,
        FSOAsString (FSOPntr), (int) ErrorCode);
  }

  if (NewName != NULL)
  {
    TempAttributeData.attributeData.size = strlen (NewName) + 1;
    if (TempAttributeData.attributeData.size > NAME_MAX)
      TempAttributeData.attributeData.size = NAME_MAX;
    TempAttributeData.attributeData.type = B_STRING_TYPE;
    memcpy (TempAttributeData.storage.nameString, NewName,
      TempAttributeData.attributeData.size);
    TempAttributeData.storage.nameString[NAME_MAX-1] = 0;

    ErrorCode = AddOrRemoveAttributeInIndex (DiskVolumePntr,
      true /* false for remove operation */, NULL /* IndicesDirPntr */,
      g_NameOfNameIndex, &TempAttributeData.attributeData, FSOPntr);
    if (ErrorCode != 0)
      dprintf ("ARFS:ChangeNameIndexForFSO: Failed to add new name "
        "\"%s\" for %s to name index, error %d.\n",
        TempAttributeData.storage.nameString,
        FSOAsString (FSOPntr), (int) ErrorCode);
  }
}



/******************************************************************************
 * Updates the size index for a file (just real files) if the size has changed
 * from what's in the index.  Also updates the date last modified.  Assumes
 * FilePntr is WriteLocked and nothing else (indices in particular) is.
 */

static void UpdateSizeIndexWithNewFileSize (
  DiskVolumePointer DiskVolumePntr,
  FileObjectPointer FilePntr)
{
  status_t               ErrorCode;
  PremadeAttributeRecord TempAttributeData;

  if (FilePntr->fileSizeInIndex == FilePntr->fileSize)
    return;

  TempAttributeData.attributeData.size = sizeof (uint64);
  TempAttributeData.attributeData.type = B_UINT64_TYPE;
  TempAttributeData.storage.uint64data = FilePntr->fileSizeInIndex;
  ErrorCode = AddOrRemoveAttributeInIndex (DiskVolumePntr,
    false /* false for remove operation */, NULL /* IndicesDirPntr */,
    g_NameOfSizeIndex, &TempAttributeData.attributeData, &FilePntr->fso);
  if (ErrorCode != 0)
    dprintf ("ARFS:UpdateSizeIndexWithNewFileSize: Failed to remove old size "
      "%Ld for %s from size index, error %d.\n",
      TempAttributeData.storage.uint64data,
      FSOAsString (&FilePntr->fso), (int) ErrorCode);

  FilePntr->fileSizeInIndex = FilePntr->fileSize;

  TempAttributeData.storage.uint64data = FilePntr->fileSize;
  ErrorCode = AddOrRemoveAttributeInIndex (DiskVolumePntr,
    true /* false for remove operation */, NULL /* IndicesDirPntr */,
    g_NameOfSizeIndex, &TempAttributeData.attributeData, &FilePntr->fso);
  if (ErrorCode != 0)
    dprintf ("ARFS:UpdateSizeIndexWithNewFileSize: Failed to add new size "
      "%Ld for %s to size index, error %d.\n",
      TempAttributeData.storage.uint64data,
      FSOAsString (&FilePntr->fso), (int) ErrorCode);

  ChangeFSODates (DiskVolumePntr, &FilePntr->fso,
    DK_MODIFIED_FLAG | DK_INODE_CHANGED_FLAG, -1 /* use current time */);
}



/******************************************************************************
 * These are some data structures used by the rename and delete operations.
 *
 * Because the presence of AGMS links allows for cycles in the file name tree
 * (turning it into a directed graph), we need graph traversal algorithms for
 * ensuring that all directories (except the root) always have a valid main
 * parent directory.  Another desired feature is that errors (such as out of
 * memory) during a rename or delete leave the file system unchanged.  This
 * leads to a transaction type of approach, where the code figures out all the
 * work that needs to be done, allocates items needed (such as strings for new
 * names), then goes and implements the changes in a way which avoids doing
 * anything that could fail.  If it fails at any step before the final
 * implementation step, it just frees what it had allocated and exits.
 *
 * Most operations (create, stat, write, etc) don't involve changes that affect
 * more than one or two things in the file system, so they aren't covered here.
 * So only deletion and renaming need this fancy treatment.  Renaming because
 * rename a/b x/y unlinks b from directory a, unlinks y from directory x (maybe
 * deletes it too if it doesn't have any other parents) and adds a link in
 * directory x to b.
 *
 * Deletion (actually, it's more like unlink; deletion of the object doesn't
 * happen until the last parent goes) of a directory can disconnect a
 * ring/cycle of subdirectories from the rest of the file system.  An example
 * of this is a tree of directories A/B/C with an AGMS link named D inside C
 * that goes back to B. B is essentially a subdirectory of directories A and C,
 * with B's main parent directory ".." leading to A and B's extra parent
 * directory "..." leading to C.
 *
 *     A _      A has B as a subdirectory, ordinary (hard link) connection.
 *     |/ \
 *     B  |     B has C as a subdirectory.
 *     |  ^
 *     C  |     C contains an AGMS link named D.
 *     |  ^
 *    (D) |     D is an AGMS link pointing to B, () to show it is an AGMS link.
 *     \__/
 *
 * Now remove the A-B link.  As part of that operation, C is promoted to be the
 * new main parent directory of B, and the now useless AGMS link D is
 * deallocated.
 *
 *   A   _
 *      / \
 *     B  |     B has C as a subdirectory.
 *     |  ^
 *     C  |     C has B as a subdirectory.
 *     \__/
 *
 * B now has C as a parent and C has B as a parent, so they won't get
 * deallocated because they still have parents.  However, there's no way of
 * accessing the B/C directories from anywhere else in the file system, or for
 * getting from B/C to the root via ".." operations.  We want to detect that.
 *
 * As well you can unlink a directory which formerly provided the current path
 * to the root, leaving some other directories without a series of ".."
 * operations leading to the root.  Their main parent directories need to be
 * updated.  For example in directory A add an AGMS link X pointing to
 * directory C and then unlink B from A (cd A ; ln -d B/C X ; rmdir B).  B's
 * main parent becomes C (via AGMS link D which gets merged with B since you
 * can't have an AGMS link as the main parent directory).  Also the main parent
 * of C becomes A (via merging with X), since it has a connection to the root
 * while B just uselessly circles around to C again.  Watch out when merging a
 * directory with an AGMS link - there may be other AGMS links pointing to the
 * former AGMS link, which need to be fixed up to point to the directory.
 * Finally, C used to be inside B but when it switched its main parent
 * directory from B to A, it leaves behind a new AGMS link named Z in directory
 * B.
 *
 * Initial situation:
 *
 *    A   _      A has B as a subdirectory, and contains AGMS link X.
 *    /\ / \
 *  (X) B  |     B has C as a subdirectory.  Main parent is A.
 *    \ |  ^     X is an AGMS link in A which points to C.  Main parent A.
 *      C  |     C contains an AGMS link named D.  Main parent B.
 *      |  ^
 *     (D) |     D is an AGMS link pointing to B.  Main parent C.
 *      \__/
 *
 * After "rmdir B":
 *
 *    A          A just contains directory C.
 *    |   _
 *     \ / \
 *      C  |     C is a directory which contains B.  Main parent A.
 *      |  |
 *      B  |     B is a directory which contains AGMS link Z.
 *      |  ^
 *     (Z) |     Z is an AGMS link to C.
 *      \__/
 *
 *
 * It's actually not that simple, since when something merges with an AGMS
 * link, it takes on the name of the link.  Also, when a placeholder is left
 * behind, it has the old name of the thing that used to be there.  The actual
 * names would be:
 *
 *    A          A just contains directory X (formerly an AGMS link).
 *    |   _
 *     \ / \
 *      X  |     X is a directory (formerly C), contains D.  Main parent is A.
 *      |  |
 *      D  |     D is a directory (formerly B) which contains AGMS link C.
 *      |  ^
 *     (C) |     C is a left behind placeholder AGMS link to X.
 *      \__/
 *
 * /Test:
 * drwxrwxrwx   0 agmsmith agmsmith        1 Mar 17 11:42 .
 * drwxrwxrwx   1 agmsmith agmsmith        0 Mar 17 11:42 ..
 * drwxrwxrwx   1 agmsmith agmsmith        2 Mar 17 11:42 a
 * /Test/a:
 * drwxrwxrwx   1 agmsmith agmsmith        2 Mar 17 11:42 .
 * drwxrwxrwx   0 agmsmith agmsmith        1 Mar 17 11:42 ..
 * drwxrwxrwx   2 agmsmith agmsmith        1 Mar 17 11:42 b
 * lrwxrwxrwx   1 agmsmith agmsmith        1 Mar 17 11:43 x -> /Test/a/b/c
 * /Test/a/b:
 * drwxrwxrwx   2 agmsmith agmsmith        1 Mar 17 11:42 .
 * drwxrwxrwx   1 agmsmith agmsmith        2 Mar 17 11:42 ..
 * lrwxrwxrwx   0 agmsmith agmsmith        2 Mar 17 11:42 ... -> /Test/a/b/c
 * drwxrwxrwx   2 agmsmith agmsmith        1 Mar 17 11:42 c
 * /Test/a/b/c:
 * drwxrwxrwx   2 agmsmith agmsmith        1 Mar 17 11:42 .
 * drwxrwxrwx   2 agmsmith agmsmith        1 Mar 17 11:42 ..
 * lrwxrwxrwx   0 agmsmith agmsmith        2 Mar 17 11:42 ... -> /Test/a
 * lrwxrwxrwx   1 agmsmith agmsmith        1 Mar 17 11:43 d -> /Test/a/b
 *
 * And after rmdir /Test/a/b the listing is:
 *
 * /Test:
 * drwxrwxrwx   0 agmsmith agmsmith        1 Mar 17 11:42 .
 * drwxrwxrwx   1 agmsmith agmsmith        0 Mar 17 11:42 ..
 * drwxrwxrwx   1 agmsmith agmsmith        1 Mar 17 11:42 a
 * /Test/a:
 * drwxrwxrwx   1 agmsmith agmsmith        1 Mar 17 11:42 .
 * drwxrwxrwx   0 agmsmith agmsmith        1 Mar 17 11:42 ..
 * drwxrwxrwx   2 agmsmith agmsmith        1 Mar 17 11:42 x
 * /Test/a/x:
 * drwxrwxrwx   2 agmsmith agmsmith        1 Mar 17 11:42 .
 * drwxrwxrwx   1 agmsmith agmsmith        1 Mar 17 11:42 ..
 * lrwxrwxrwx   0 agmsmith agmsmith        2 Mar 17 11:42 ... -> /Test/a/x/d
 * drwxrwxrwx   1 agmsmith agmsmith        1 Mar 17 11:42 d
 * /Test/a/x/d:
 * drwxrwxrwx   1 agmsmith agmsmith        1 Mar 17 11:42 .
 * drwxrwxrwx   2 agmsmith agmsmith        1 Mar 17 11:42 ..
 * lrwxrwxrwx   1 agmsmith agmsmith        1 Mar 17 11:44 c -> /Test/a/x
 *
 *
 * The detection of cycles and fixing up of main parent directories is done in
 * three steps:
 *
 * Step 1.
 *   Find all children of the unlinked objects.  This is done by having a list
 *   of objects needing to be examined, initialised with the one or two objects
 *   being unlinked.  Then while the pending examination list isn't empty, an
 *   object is removed from the pending list and added to the list of children
 *   found, and its children (just directory or AGMS link type children,
 *   including ones through a removed link but not through the new added link)
 *   are added to the pending examination list.
 *
 * Step 2.
 *   Find the new main parent directory or AGMS link for each affected item.
 *   Start by adding all children found in step 1 to an orphan list.  Then
 *   while the orphan list isn't empty, repeatedly scan through it for an
 *   object that has some parent directory (excluding parents through removed
 *   links but including parents through the new added link) or parent AGMS
 *   link which isn't in also in the orphan list.  Mark that parent as the new
 *   main parent and remove the object from the orphan list (yes, other orphans
 *   can now use the object as a main parent if they are children), and resume
 *   scanning at the next orphan.  If the scan gets to the end of the list,
 *   restart scanning at the beginning.  If the scan goes from beginning to end
 *   of the non-empty list without finding any parents then we have a
 *   disconnected cycle, and an error code needs to be returned to the user.
 *
 * Step 3.
 *   Change the main parents to the new choices.  For each item being
 *   re-parented, create a new AGMS link to the item in the former main parent
 *   directory.  If an AGMS link is the new main parent, merge with the AGMS
 *   link (replacing the AGMS link in effect, references to the old AGMS link
 *   are changed to point to the re-parented item, and the re-parented item
 *   takes on the name of the old AGMS link).
 *
 * In this code the list of children is implemented as an AVL tree of
 * FSOWorkRecords, so that we can quickly test to see if a child has been added
 * already and quickly locate it by VNodeID.  The pending lists are implemented
 * as a doubly linked list reusing the same FSOWorkRecords.  Later on when the
 * child computations have been done, other FSOs that need work are added to
 * the tree (so that we don't lock a node twice by accident), then the actual
 * implementation step involves iterating through all nodes in the tree to do
 * the work marked in each one.
 *
 * So, let's walk through a renaming operation, renaming A/X to B/Y.  A user
 * level rename is roughly implemented as unlinking X from A, unlinking Y from
 * B and deleting Y if it has no more parents, and adding X to directory B
 * under its new name of "Y".  For degenerate cases some of those operations
 * are left out.
 *
 * In more detail: Allocate new true name string containing "Y".  Set up the
 * state with an empty work tree, empty pending list.  Record the objects
 * involved in the unlink of A/X and B/Y and addition of the B/X link in main
 * state record.  Add the unlinked objects X and Y to the pending list (adding
 * includes calling get_vnode, read locking them and adding them to the pending
 * list).  Check that the deleted thing Y won't end up as a non-empty directory
 * with no parents.  Recursively find all children of the pending list objects
 * and add them to the tree.  Find new main parent directories/links for
 * everything in tree and record them for later use.  Create temporary AGMS
 * link objects for each former parent directory.  Add directory A to the work
 * tree (just so that it gets read locked), check that it has write permission.
 * Add directory B to the tree, check for write permission.  Upgrade B to a
 * write lock.  If Y doesn't exist in B, add a new entry for X to directory B
 * using X's new name of "Y".  If Y exists but has no AGMS links, it will get
 * deleted later.  If Y does exist and has an AGMS link then Y will merge with
 * one of its parent AGMS links (picked arbitrarily, call it the mergee) which
 * will disappear and Y will appear in its place.  The AGMS links which pointed
 * to the mergee have to be changed to point to Y too.  So, start by adding the
 * parent AGMS links of the mergee to the work tree (so they get locked).  Make
 * a temporary combined AGMS link array containing all of the AGMS links
 * pointing at Y plus the AGMS links pointing at the mergee.  Convert read
 * locks to write locks for everything in the tree which will be modified and
 * isn't already a write lock.  Failures can't occur after this point (no more
 * allocations).  Remove X from parent directory A's contents.  Change X's true
 * name to previously allocated string.  Set X's parent directory to be B.
 * Overwrite Y's AVL node value in directory B with X's address (or remove the
 * AVL node if there is no X).  If no AGMS links to Y exist, delete
 * (remove_vnode) Y and if Y itself was an AGMS link, remove it from the
 * target's list of parent AGMS links too.  Otherwise, merge Y and one of its
 * AGMS links (Y takes over mergee's true name string (swap them), set parent
 * directory of Y to be mergee's parent directory, overwrite mergee's entry in
 * mergee's parent directory with Y's address, mark mergee for deletion
 * (remove_vnode)).  Release locks, put_vnode, free old true name string, free
 * tree.  Done.
 *
 * Degenerate case where A/X is the same as B/Y: do nothing, return success.
 *
 * Degenerate case where the name doesn't change, just moving a file between
 * directories: treat like full case (wastefully reallocates the true name).
 *
 * Degenerate case where directory A and B are the same, just doing a name
 * change: same as full case but watch out for removing the newly added
 * directory entry with the same name.  Well, now that we overwrite the
 * directory entries, that isn't a problem.
 *
 * Degenerate case where an AGMS link to Y is also in B: should be OK.  But if
 * it has the same name too?  Test by renaming A/X to A/Y where Y was an AGMS
 * link to X.
 *
 * Test overwriting an AGMS link which refers to the mergee.
 */

/* This structure is used to keep track of the work involved during a Rename or
Delete operation that affects a particular File System Object. */

typedef struct RDWorkFSOStruct RDWorkFSORecord, *RDWorkFSOPointer;

struct RDWorkFSOStruct {
  FileSystemObjectPointer fsoPntr;
    /* The file system object under consideration.  Usually this will be a
    descendant of a directory which is having a link removed, and thus may need
    its main parent directory changed so that the direct path to the root
    exists.  Or it could be a parent directory or AGMS link which needs
    modification.  Or it could be the actual file system object being renamed
    or deleted. */

  RDWorkFSOPointer nextPendingPntr;
  RDWorkFSOPointer previousPendingPntr;
    /* These two pointers implement a doubly linked list of pending work.  The
    list is used for several different purposes - finding children and finding
    orphans.  FSOWorkRecords in the list are also in the tree, ones that are
    not in the list have garbage values in these pointers.  The first item has
    previousPendingList NULL and the last has nextPendingList NULL.  The tail
    and head of the list are identified in the RDWorkStateRecord. */

  bool getNodeDone;
    /* TRUE if a call to get_vnode has been done, and thus on exit a call to
    put_vnode needs to be done. */

  bool gotReadLock;
    /* TRUE if we have a read lock to this file system object.  Does not
    include former read locks which have been converted to write locks. */

  bool gotWriteLock;
    /* TRUE if we have a write lock to the file system object.  This includes
    former read locks which have been converted to write locks. */

  bool eventuallyWriteLock;
    /* Set to TRUE if this is the work unit for something which will be
    modified.  A final spin through the work tree before modifications start
    will be done to write lock everything with this flag set.  It includes:
    former main parent directories (write locked so that a new AGMS link can
    replace the former file that was in the directory), the source and
    destination directories in a rename/delete, the renamed or deleted item,
    the overwritten item, and items which have a new main parent. */

  bool orphan;
    /* TRUE means that the file system object needs to find a new main parent
    directory.  FALSE means that we have found the new parent or decided to
    keep the existing one.  TRUE also means that this is on the orphans list of
    pending work items.  Only used during orphan computation. */

  FileSystemObjectPointer newMainParent;
    /* Indirectly identifies the new main parent directory or AGMS link for the
    object.  Will be NULL if no change in the parent is needed.  If it's a
    directory then that's the new main parent.  If it's an AGMS link then the
    new main parent will be the directory holding the specified AGMS link.
    That AGMS link will be merged with this object as part of the change in
    main parent directory.  Also, the previous main parent directory will get a
    new AGMS link placeholder for this object (unless it is a deleted link). */

  AGMSLinkObjectPointer replacementAGMSLinkForOldMainParent;
    /* If needed, a new AGMS link object is allocated to act as a placeholder
    in the former main parent directory, pointing to this object.  NULL if not
    allocated or if the changes have been committed. */

  AGMSLinkObjectPointer *arrayOfMergedParentLinks;
    /* When merging with an AGMS link, the other AGMS links pointing to both
    this object and the mergee are combined and stored in this NULL terminated
    array, which later will become the parentLinkArray of this object.  NULL if
    the array isn't allocated or the changes have been committed. */
};


typedef struct RDWorkStateStruct
{
  /* The following fields are input arguments, set by the caller of
  RenameOrDelete(). */

  DiskVolumePointer diskVolumePntr;
    /* A copy of the disk volume pointer to cut down on the number of arguments
    passed. */

  DirectoryObjectPointer sourceDirectoryPntr;
    /* This is the directory from which the renamed item comes from.  Or NULL
    if just doing a delete.  It will be get_vnode'd and locked (so the caller
    shouldn't lock it because double locking doesn't work). */

  DirectoryObjectPointer destinationDirectoryPntr;
    /* This is the directory where the renamed item will move to.  It's also
    the directory where the replaced item will be deleted from.  When just
    doing a delete, it's the directory of the deleted item.  It will be
    get_vnode'd and locked (so the caller shouldn't lock it because double
    locking doesn't work). */

  const char *oldNamePntr;
    /* This is the old name of the renamed item, as it is known in the source
    directory.  It will be used to find the item in the directory.  NULL if
    just doing a delete.*/

  const char *newNamePntr;
    /* The user provided new name of the renamed item.  An existing item with
    this name in the destination directory will get unlinked.  It's also the
    name of the item to delete if you are just doing a deletion.  This string
    is copied when making the new true name. */

  bool rmdirNotUnlink;
    /* TRUE to behave like rmdir, which will only delete directories.  FALSE to
    behave like unlink which only deletes files.  Ignored during renames. */

  /* The following fields are for internal use.  Initialise them to zero before
  calling RenameOrDelete().  Generally, once the items refered to by these
  variables get used (the file system changes are committed) then they get
  cleared back to NULL or false.  That way if an error happens before the
  commit, the cleanup code will deallocate them, but if it completed
  successfully, the same cleanup code will see NULLs and not do the
  deallocations. */

  FileSystemObjectPointer renamedFSOPntr;
    /* This is the item to be renamed.  oldNamePntr is used to find it in the
    source directory.  It will be removed from the source directory and added
    to the destination directory, possibly replacing something there with the
    misfortune of having the same name as newNamePntr.  NULL if not doing a
    rename. */

  char *originalTrueNamePntr;
    /* Points to the true name string of the renamed item before it was
    renamed.  Will be deallocated on exit if it isn't NULL. */

  char *newTrueNamePntr;
    /* Points to a newly allocated true name string for the renamed item after
    it is renamed (copied from the user provided newNamePntr).  Will be
    deallocated on exit if it isn't NULL. */

  FileSystemObjectPointer deletedFSOPntr;
   /* Usually it's the file which is being overwritten by a rename operation in
   the destination directory, or it can be the file which is being deleted in a
   plain delete operation.  Determined by looking for name newNamePntr in
   directory destinationDirectoryPntr.  After unlinking, the object will
   actually get deleted if it doesn't have any more parent directories.  NULL
   if nothing is getting deleted (the ordinary case for rename). */

  AVLDupTreePointer workFSOTree;
    /* This tree contains all the file system objects affected by the rename or
    delete operation.  The keys are pointers to the FSO records (same as the
    VNodeIDs) and the values are RDWorkFSOPointers to temporarily allocated
    work state info about each file.  That way we can quickly add files for
    processing or determine that they already have been processed. */

  RDWorkFSOPointer headOfPendingList;
  RDWorkFSOPointer tailOfPendingList;
    /* These point to the head and tail nodes in the list of pending work,
    which is used during various different algorithms.  NULL if the list is
    empty.  The nodes are a subset of the nodes stored in the tree. */

  bool addedNewDirectoryEntry;
    /* TRUE if a new directory entry with name newNamePntr and target
    renamedFSOPntr was created in directory destinationDirectoryPntr as part of
    the rename operation.  FALSE if we aren't doing a rename (no need for a new
    entry) or if the changes have been committed (so that the cleanup code
    doesn't delete the new directory entry). */

} RDWorkStateRecord, *RDWorkStatePointer;



#if DPRINT_NON_ERROR_MESSAGES
/******************************************************************************
 * Internal tree traversal callback function for debugging.  Lists the contents
 * of the work tree.  Key is NULL if being called outside of a tree traversal.
 */

static bool RDWorkListOneFSO (
  AVLDupThingConstPointer KeyPntr,
  const unsigned int *ValuePntr,
  void *ExtraData)
{
  AGMSLinkObjectPointer *MergedParentPntrPntr;
  char                  *StringPntr;
  char                   TempString [2048];
  RDWorkFSOPointer       WorkPntr = (void *) *ValuePntr;

  if (KeyPntr != NULL && KeyPntr->int32Thing != (int) WorkPntr->fsoPntr)
  {
    dprintf ("ARFS:RDWorkListOneFSO   Corrupt work unit, key is for %s "
      "while fsoPntr is %s.\n",
      FSOAsString ((FileSystemObjectPointer) KeyPntr->int32Thing),
      FSOAsString (WorkPntr->fsoPntr));
    return true; /* Continue iteration but don't do anything about this one. */
  }

  sprintf (TempString, "ARFS:RDWorkListOneFSO   Work $%X = %s%s%s%s%s%s",
    (unsigned int) WorkPntr,
    FSOAsString (WorkPntr->fsoPntr),
    WorkPntr->getNodeDone ? ", GotVNode" : "",
    WorkPntr->gotReadLock ? ", ReadLock" : "",
    WorkPntr->gotWriteLock ? ", WriteLock" : "",
    WorkPntr->eventuallyWriteLock ? ", EventuallyWriteLock" : "",
    WorkPntr->orphan ? ", Orphan" : "");
  StringPntr = TempString + strlen (TempString);

  if (WorkPntr->newMainParent != NULL)
    sprintf (StringPntr, ", future parent %s",
    FSOAsString (WorkPntr->newMainParent));
  StringPntr += strlen (StringPntr);

  if (WorkPntr->replacementAGMSLinkForOldMainParent != NULL)
    sprintf (StringPntr, ", stand-in link $%X",
    (int) WorkPntr->replacementAGMSLinkForOldMainParent);
  StringPntr += strlen (StringPntr);

  if (WorkPntr->arrayOfMergedParentLinks != NULL)
  {
    sprintf (StringPntr, ", array of merged parent links $%X (",
      (int) WorkPntr->arrayOfMergedParentLinks);
    StringPntr += strlen (StringPntr);
    MergedParentPntrPntr = WorkPntr->arrayOfMergedParentLinks;
    while (*MergedParentPntrPntr != NULL &&
    StringPntr < TempString + (sizeof (TempString) - 30))
    {
      sprintf (StringPntr, "$%X \"%s\"",
        (int) *MergedParentPntrPntr,
        (*MergedParentPntrPntr)->fso.trueNamePntr);
      StringPntr += strlen (StringPntr);

      MergedParentPntrPntr++;

      if (*MergedParentPntrPntr != NULL)
      {
        strcat (StringPntr, ", ");
        StringPntr += strlen (StringPntr);
      }
    }
    if (StringPntr >= TempString + (sizeof (TempString) - 275))
      strcat (StringPntr, "...");
    strcat (StringPntr, ")");
    StringPntr += strlen (StringPntr);
  }

  dprintf ("%s.\n", TempString);

  return true; /* Continue iterating. */
}
#endif /* DPRINT_NON_ERROR_MESSAGES */



#if DPRINT_NON_ERROR_MESSAGES
/******************************************************************************
 * Debugging function to list the contents of the tree.
 */

static void RDWorkListAll (
  RDWorkStatePointer StatePntr)
{
  /* Iterate through all work items and list them. */

  if (StatePntr->workFSOTree != NULL)
  {
    dprintf ("ARFS:RDWorkListOneFSO Listing of %d work units:\n",
      AVLDupGetTreeCount (StatePntr->workFSOTree));

    AVLDupIterate (StatePntr->workFSOTree, true,
      NULL, 0, 0, false,
      NULL, 0, 0, false,
      RDWorkListOneFSO, StatePntr);
  }
}
#endif /* DPRINT_NON_ERROR_MESSAGES */



/******************************************************************************
 * Add a work item to the tail of the pending work list.  Assumes arguments are
 * valid and that the work item isn't already in the list.
 */

static void AddToTailOfPendingList (
  RDWorkStatePointer StatePntr,
  RDWorkFSOPointer WorkPntr)
{
  if (StatePntr->tailOfPendingList == NULL)
  {
    /* Adding the first work unit to an empty pending list. */
    StatePntr->tailOfPendingList = WorkPntr;
    StatePntr->headOfPendingList = WorkPntr;
    WorkPntr->nextPendingPntr = NULL;
    WorkPntr->previousPendingPntr = NULL;
  }
  else /* Add to the tail of the non-empty list. */
  {
    WorkPntr->previousPendingPntr = StatePntr->tailOfPendingList;
    WorkPntr->nextPendingPntr = NULL;
    StatePntr->tailOfPendingList->nextPendingPntr = WorkPntr;
    StatePntr->tailOfPendingList = WorkPntr;
  }
}



/******************************************************************************
 * Removes a work item from the head of the pending work list.  Assumes
 * arguments are valid.  Returns NULL if the list was empty.
 */

static RDWorkFSOPointer RemoveFromHeadOfPendingList (
  RDWorkStatePointer StatePntr)
{
  RDWorkFSOPointer WorkPntr;

  WorkPntr = StatePntr->headOfPendingList;

  if (WorkPntr != NULL)
  {
    /* Remove the work item from the list. */
    StatePntr->headOfPendingList = WorkPntr->nextPendingPntr;
    if (StatePntr->headOfPendingList == NULL)
      StatePntr->tailOfPendingList = NULL; /* List is now empty. */
    else
      StatePntr->headOfPendingList->previousPendingPntr = NULL;
    WorkPntr->nextPendingPntr = NULL;
    WorkPntr->previousPendingPntr = NULL;
  }

  return WorkPntr;
}



/******************************************************************************
 * Removes a work item from the middle of the pending work list.  Assumes
 * arguments are valid and the item is in the list.
 */

static void RemoveFromMiddleOfPendingList (
  RDWorkStatePointer StatePntr,
  RDWorkFSOPointer WorkPntr)
{
  /* Adjust previous node to point to next one. */

  if (WorkPntr->previousPendingPntr == NULL)
    StatePntr->headOfPendingList = WorkPntr->nextPendingPntr;
  else
    WorkPntr->previousPendingPntr->nextPendingPntr =
    WorkPntr->nextPendingPntr;

  /* Adjust next node to point backwards to the right place. */

  if (WorkPntr->nextPendingPntr == NULL)
    StatePntr->tailOfPendingList = WorkPntr->previousPendingPntr;
  else
    WorkPntr->nextPendingPntr->previousPendingPntr =
    WorkPntr->previousPendingPntr;

  WorkPntr->nextPendingPntr = NULL;
  WorkPntr->previousPendingPntr = NULL;
}



/******************************************************************************
 * Look for an existing work item for the given file system object.  Returns a
 * pointer to it if it was found, or NULL if it isn't in the work tree.
 */

static RDWorkFSOPointer GetFSOWorkItem (
  RDWorkStatePointer StatePntr,
  FileSystemObjectPointer FSOPntr)
{
  AVLDupThingRecord Key;
  unsigned int      TempUINT;

  Key.int32Thing = (int) FSOPntr;
  TempUINT = AVLDupFind (StatePntr->workFSOTree, &Key);
  return (RDWorkFSOPointer) TempUINT;
}



/******************************************************************************
 * Create a new work item for the given file system object, doesn't check to
 * see if it is already there.  Also does get_vnode and read locking (and marks
 * the work item as needing them to be undone upon exit).  Sets the user
 * provided pointer in *NewWorkPntrPntr to point to the new work item if
 * successful and returns zero.  Returns a negative error code and sets the
 * provided pointer to NULL if it fails.  You can specify NULL for
 * NewWorkPntrPntr if you're not interested in the new work item.
 */

static status_t MakeFSOWorkItemBlindly (
  RDWorkStatePointer StatePntr,
  FileSystemObjectPointer FSOPntr,
  RDWorkFSOPointer *NewWorkPntrPntr)
{
  status_t          ErrorCode;
  bool              GetVNodeDone = false;
  AVLDupThingRecord Key;
  bool              ReadLockDone = false;
  status_t          ReturnCode = -1;
  vnode_id          VNodeID = 0;
  void             *VoidPntr = 0;
  RDWorkFSOPointer  WorkPntr = NULL;

  /* Do the get_vnode. */

  VNodeID = (unsigned int) FSOPntr;

  ErrorCode = get_vnode (StatePntr->diskVolumePntr->nameSpaceID,
    VNodeID, &VoidPntr);
  if (ErrorCode < 0)
  {
    dprintf ("ARFS:MakeWorkItem get_vnode failed for %s, code %d!\n",
      FSOAsString (FSOPntr), (int) ErrorCode);
    ReturnCode = ErrorCode;
    goto ErrorExit;
  }
  GetVNodeDone = true;
  if (VoidPntr != FSOPntr)
  {
    dprintf ("ARFS:MakeWorkItem get_vnode returned an incorrect private "
      "pointer for %s.\n", FSOAsString (FSOPntr));
    goto ErrorExit;
  }

  /* Do the read lock. */

  ErrorCode = ReadLock (FSOPntr, false);
  if (ErrorCode != 0)
  {
    dprintf ("ARFS:MakeWorkItem ReadLock failed for %s, code %d.\n",
      FSOAsString (FSOPntr), (int) ErrorCode);
    ReturnCode = ErrorCode;
    goto ErrorExit;
  }
  ReadLockDone = true;

  /* Allocate and initialise the work record. */

  WorkPntr = mallocDVP (StatePntr->diskVolumePntr, sizeof (RDWorkFSORecord));
  if (WorkPntr == NULL)
  {
    dprintf ("ARFS:MakeWorkItem Out of memory when allocating work record "
      "for %s.\n", FSOAsString (FSOPntr));
    ReturnCode = ENOMEM;
    goto ErrorExit;
  }

  memset (WorkPntr, 0, sizeof (RDWorkFSORecord));
  WorkPntr->fsoPntr = FSOPntr;
  WorkPntr->getNodeDone = true;
  WorkPntr->gotReadLock = true;

  /* Try adding it to the tree.  The key is the VNodeID==FSOPntr, the value is
  the pointer to the new work record. */

  Key.int32Thing = (int) FSOPntr;
  ErrorCode =
    AVLDupAdd (StatePntr->workFSOTree, &Key, (unsigned int) WorkPntr);
  if (ErrorCode != AVLADD_ADDED_A_NODE)
  {
    dprintf ("ARFS:MakeWorkItem Unable to add work record to tree, code %d, "
      "usually out of memory is the cause, or an added twice bug.  For work "
      "record for %s.\n", (int) ErrorCode, FSOAsString (FSOPntr));
    ReturnCode = ENOMEM; /* AVLDupAdd error codes are specific to itself. */
    goto ErrorExit;
  }

  /* Successful.  Return results. */

  if (NewWorkPntrPntr != NULL)
    *NewWorkPntrPntr = WorkPntr;
  return 0;

ErrorExit: /* Only errors handled past here. */
  if (WorkPntr != NULL)
    freeDVP (StatePntr->diskVolumePntr, WorkPntr);
  if (ReadLockDone)
    ReadUnlock (FSOPntr);
  if (GetVNodeDone)
    put_vnode (StatePntr->diskVolumePntr->nameSpaceID, VNodeID);
  if (NewWorkPntrPntr != NULL)
    *NewWorkPntrPntr = NULL;
  return ReturnCode;
}



/******************************************************************************
 * Create a new work item, but check first that it doesn't already exist for
 * the given file system object.
 */

static status_t MakeFSOWorkItemSafely (
  RDWorkStatePointer StatePntr,
  FileSystemObjectPointer FSOPntr,
  RDWorkFSOPointer *NewWorkPntrPntr)
{
  RDWorkFSOPointer  WorkPntr;

  /* If there's already a work unit, just use it. */

  WorkPntr = GetFSOWorkItem (StatePntr, FSOPntr);
  if (WorkPntr != NULL)
  {
    if (NewWorkPntrPntr != NULL)
      *NewWorkPntrPntr = WorkPntr;
    return 0;
  }

  return MakeFSOWorkItemBlindly (StatePntr, FSOPntr, NewWorkPntrPntr);
}



/******************************************************************************
 * Makes sure that a work item ends up being write locked.  Returns zero on
 * success, or a negative error code if a deadlock happens, or if other things
 * go wrong.
 */

static status_t WriteLockWorkItem (
  RDWorkFSOPointer WorkPntr,
  RDWorkStatePointer StatePntr)
{
  status_t         ErrorCode;

  if (WorkPntr->gotWriteLock)
    return 0; /* Already write locked.  Success. */

  if (WorkPntr->gotReadLock)
  {
    ErrorCode = ConvertReadLockToWriteLock (WorkPntr->fsoPntr);
    if (ErrorCode != 0)
    {
      dprintf ("ARFS:WriteLockWorkItem Can't convert read to write "
        "for %s.\n", FSOAsString (WorkPntr->fsoPntr));
      return ErrorCode;
    }

    WorkPntr->gotWriteLock = true;
    WorkPntr->gotReadLock = false;
    return 0;
  }

  /* Not locked at all yet. */

  ErrorCode = WriteLock (WorkPntr->fsoPntr);
  if (ErrorCode != 0)
  {
    dprintf ("ARFS:WriteLockWorkItem Can't write lock %s.\n",
      FSOAsString (WorkPntr->fsoPntr));
    return ErrorCode;
  }

  WorkPntr->gotWriteLock = true;
  return 0;
}



/******************************************************************************
 * Find the work item in the work tree for the given file system object and
 * make sure that they end up being write locked.  Returns zero on success, or
 * a negative error code if a deadlock happens, or if other things go wrong.
 */

static status_t WriteLockWorkItemByFSO (
  FileSystemObjectPointer FSOPntr,
  RDWorkStatePointer StatePntr)
{
  RDWorkFSOPointer WorkPntr;

  WorkPntr = GetFSOWorkItem (StatePntr, FSOPntr);
  if (WorkPntr == NULL)
  {
    dprintf ("ARFS:WriteLockWorkItemByFSO Can't find work unit for %s.\n",
      FSOAsString (FSOPntr));
    return ENOENT;
  }

  return WriteLockWorkItem (WorkPntr, StatePntr);
}



/******************************************************************************
 * Internal tree traversal callback function to clean up one work unit.  Also
 * callable to clean up a single work unit outside the tree (KeyPntr specified
 * as NULL in that case).  It undoes locks, calls put_vnode if needed, deletes
 * the replacement AGMS link if it wasn't used, frees the
 * arrayOfMergedParentLinks if it wasn't used, and deallocates the work record
 * itself (but not the tree node that points to the work record).
 */

static bool RDWorkCleanupForOneFSO (
  AVLDupThingConstPointer KeyPntr,
  const unsigned int *ValuePntr,
  void *ExtraData)
{
  status_t           ErrorCode;
  RDWorkStatePointer StatePntr = ExtraData;
  RDWorkFSOPointer   WorkPntr = (RDWorkFSOPointer) *ValuePntr;

  if (KeyPntr != NULL && KeyPntr->int32Thing != (int) WorkPntr->fsoPntr)
  {
    dprintf ("ARFS:RDWorkCleanupForOneFSO Corrupt work unit, key is for %s "
      "while fsoPntr is %s.  Skipping cleanup of this item, leaving it "
      "locked etc.\n",
      FSOAsString ((FileSystemObjectPointer) KeyPntr->int32Thing),
      FSOAsString (WorkPntr->fsoPntr));
    return true; /* Continue iteration but don't do anything about this one. */
  }

  if (WorkPntr->gotWriteLock)
  {
    ErrorCode = WriteUnlock (WorkPntr->fsoPntr);
    if (ErrorCode < 0)
    {
      dprintf ("ARFS:RDWorkCleanupForOneFSO Unable to free write lock "
        "for %s, code %d.  Ignoring the problem.\n",
        FSOAsString (WorkPntr->fsoPntr), (int) ErrorCode);
    }
  }

  if (WorkPntr->gotReadLock)
  {
    ErrorCode = ReadUnlock (WorkPntr->fsoPntr);
    if (ErrorCode < 0)
    {
      dprintf ("ARFS:RDWorkCleanupForOneFSO Unable to free read lock "
        "for %s, code %d.  Ignoring the problem.\n",
        FSOAsString (WorkPntr->fsoPntr), (int) ErrorCode);
    }
  }

  if (WorkPntr->getNodeDone)
  {
    ErrorCode = put_vnode (StatePntr->diskVolumePntr->nameSpaceID,
      (unsigned int) WorkPntr->fsoPntr);
    if (ErrorCode < 0)
    {
      dprintf ("ARFS:RDWorkCleanupForOneFSO Unable to put_vnode "
        "for %s, code %d.  Ignoring the problem.\n",
        FSOAsString (WorkPntr->fsoPntr), (int) ErrorCode);
    }
  }

  if (WorkPntr->replacementAGMSLinkForOldMainParent != NULL)
  {
    ErrorCode = DeallocateAnyFSO (StatePntr->diskVolumePntr,
      &WorkPntr->replacementAGMSLinkForOldMainParent->fso);
    if (ErrorCode < 0)
    {
      dprintf ("ARFS:RDWorkCleanupForOneFSO Unable to deallocate "
        "replacement link %s, code %d.  Ignoring the problem.\n",
        FSOAsString (&WorkPntr->replacementAGMSLinkForOldMainParent->fso),
        (int) ErrorCode);
    }
  }

  if (WorkPntr->arrayOfMergedParentLinks != NULL)
  {
    freeDVP (StatePntr->diskVolumePntr, WorkPntr->arrayOfMergedParentLinks);
  }

  /* Deallocate the work record, fill it with obvious bad values so that buggy
  code gets revealed faster. */

  memset (WorkPntr, -1, sizeof (RDWorkFSORecord));
  freeDVP (StatePntr->diskVolumePntr, WorkPntr);

  return true; /* Continue iterating. */
}



/******************************************************************************
 * Cleanup for all work done as part of a successful or an aborted attempt.
 * Cleans up the individual FSO work items and the global state too.
 */

static void RDWorkCleanup (RDWorkStatePointer StatePntr)
{
  AVLDupThingRecord Key;

  /* While things are still locked, undo changes made during an aborted
  attempt.  Remove a new but unused directory entry. */

  if (StatePntr->addedNewDirectoryEntry)
  {
    Key.longStringThing.stringPntr = (char *) StatePntr->newNamePntr;
    Key.longStringThing.isLongString = true;
    if (!AVLDupDelete (StatePntr->destinationDirectoryPntr->directory,
    &Key, (unsigned int) StatePntr->renamedFSOPntr))
    {
      dprintf ("ARFS:RDWorkCleanup Bug!  Unable to delete the key (%s) and "
        "value $%X for the removed %s from the destination %s.\n",
        Key.longStringThing.stringPntr,
        (unsigned int) StatePntr->renamedFSOPntr,
        FSOAsString (StatePntr->renamedFSOPntr),
        FSOAsString (&StatePntr->destinationDirectoryPntr->fso));
    }
    StatePntr->addedNewDirectoryEntry = false;
  }

  /* Iterate through all work items and free them. */

  if (StatePntr->workFSOTree != NULL)
  {
    AVLDupIterate (StatePntr->workFSOTree, true,
      NULL, 0, 0, false,
      NULL, 0, 0, false,
      RDWorkCleanupForOneFSO, StatePntr);
  }

  /* Global state cleanup.  Deallocate the work tree data structures. */

  AVLDupFreeTree (StatePntr->workFSOTree);
  StatePntr->workFSOTree = NULL;

  /* Deallocate strings which weren't used, or aren't needed any more. */

  if (StatePntr->newTrueNamePntr != NULL)
  {
    freeDVP (StatePntr->diskVolumePntr, StatePntr->newTrueNamePntr);
    StatePntr->newTrueNamePntr = NULL;
  }

  if (StatePntr->originalTrueNamePntr != NULL)
  {
    freeDVP (StatePntr->diskVolumePntr, StatePntr->originalTrueNamePntr);
    StatePntr->originalTrueNamePntr = NULL;
  }
}



/******************************************************************************
 * An internal callback function used during iteration over a directory's
 * contents that adds all child directories and child AGMS links (they are kind
 * of like a directory with one entry) to the list of work items to be
 * processed.  The key is a name string in the directory (or NULL if called to
 * process the target of an AGMS link), the value will be the FSO pointer
 * associated with the name.  Ignores files since they don't have children and
 * thus processing them for updating the main parent directories would be a big
 * waste of time (and make the system run out of semaphores if you have a
 * directory with thousands of files).  Symbolic links are also ignored.
 * Subdirectories already in the work tree are also skipped over.  Returns TRUE
 * to continue iterating, FALSE if something goes horribly wrong that would
 * affect the integrity of the results (such as running out of memory to add
 * the new work unit for the child directory).
 */

static bool RDWorkAddChildItems (
  AVLDupThingConstPointer KeyPntr,
  const unsigned int *ValuePntr,
  void *ExtraData)
{
  FileSystemObjectPointer ChildFSOPntr = (FileSystemObjectPointer) *ValuePntr;
  status_t                ErrorCode;
  RDWorkStatePointer      StatePntr = ExtraData;
  RDWorkFSOPointer        WorkPntr;

  if (INVALID_FILE_SYSTEM_OBJECT (ChildFSOPntr))
  {
    dprintf ("ARFS:RDWorkAddChildItems Corrupt directory entry %s.\n",
      FSOAsString (ChildFSOPntr));
    return false;
  }

  if (ChildFSOPntr->objectType != FSO_TYPE_DIRECTORY &&
  ChildFSOPntr->objectType != FSO_TYPE_AGMS_LINK)
    return true; /* Skip over non-directories. */

  WorkPntr = GetFSOWorkItem (StatePntr, ChildFSOPntr);
  if (WorkPntr != NULL)
    return true; /* Skip subdirectories/AGMS links already in the work tree. */

  ErrorCode = MakeFSOWorkItemBlindly (StatePntr, ChildFSOPntr, &WorkPntr);
  if (ErrorCode != 0)
  {
    dprintf ("ARFS:RDWorkAddChildItems Unable to create work record "
      "for %s.\n", FSOAsString (ChildFSOPntr));
    return false;
  }

  AddToTailOfPendingList (StatePntr, WorkPntr);
  return true; /* Continue iterating. */
}



/******************************************************************************
 * This function finds all affected children of the items in the pending list.
 * Child files are ignored since their main parent directory is always some
 * containing directory, which will eventually be on a valid path to the root.
 * Symbolic links and indices are ignored too; they are as dumb as files.  AGMS
 * links are included because they are effectively a directory which always has
 * one entry.  A non-recursive approach is used to avoid running out of stack
 * space (the worst case is at most 512 directory levels deep - due to the
 * maximum path length being 1024).
 */

static status_t RDFindDescendantChildren (
  RDWorkStatePointer StatePntr)
{
  FileSystemObjectPointer FSOPntr;
  status_t                ReturnCode = -1;
  RDWorkFSOPointer        WorkPntr;

  while ((WorkPntr = RemoveFromHeadOfPendingList (StatePntr)) != NULL)
  {
    /* Ignore files, symbolic links and indices when looking for children. */

    FSOPntr = WorkPntr->fsoPntr;

    if (INVALID_FILE_SYSTEM_OBJECT (FSOPntr))
    {
      dprintf ("ARFS:RDFindDescendantChildren Corrupt data for %s.\n",
        FSOAsString (FSOPntr));
      ReturnCode = EINVAL;
      goto ErrorExit;
    }

    if (FSOPntr->objectType == FSO_TYPE_AGMS_LINK)
    {
      if (!RDWorkAddChildItems (NULL,
      (unsigned int *) &((AGMSLinkObjectPointer) FSOPntr)->targetFSOPntr,
      StatePntr))
      {
        dprintf ("ARFS:RDFindDescendantChildren Problems when adding children "
          "of %s.\n", FSOAsString (FSOPntr));
        goto ErrorExit;
      }
    }
    else if (FSOPntr->objectType == FSO_TYPE_DIRECTORY)
    {
      /* Add all relevant things in the directory to the work collection. */

      if (!AVLDupIterate (((DirectoryObjectPointer) FSOPntr)->directory, true,
      NULL, 0, 0, false,
      NULL, 0, 0, false,
      RDWorkAddChildItems, StatePntr))
      {
        dprintf ("ARFS:RDFindDescendantChildren Problems when adding children "
          "of %s.\n", FSOAsString (FSOPntr));
        goto ErrorExit;
      }
    }
  }

  return 0; /* Success! */

ErrorExit:
  return ReturnCode;
}



/******************************************************************************
 * An internal callback function for use when iterating over the work items
 * tree that adds a work item to the pending list as a new orphan.
 */

static bool RDWorkAddOrphanToPendingList (
  AVLDupThingConstPointer KeyPntr,
  const unsigned int *ValuePntr,
  void *ExtraData)
{
  RDWorkStatePointer      StatePntr = ExtraData;
  RDWorkFSOPointer        WorkPntr = (RDWorkFSOPointer) *ValuePntr;

  if (WorkPntr == NULL || KeyPntr->int32Thing != (int) WorkPntr->fsoPntr)
  {
    dprintf ("ARFS:RDWorkAddOrphanToPendingList Corrupt data detected!\n");
    return false;
  }

  AddToTailOfPendingList (StatePntr, WorkPntr);
  WorkPntr->orphan = true;

  return true; /* Continue iterating. */
}



/******************************************************************************
 * An internal function for determining if a given CandidateParentFSOPntr
 * directory or AGMS link is useable as a main parent directory for the FSO
 * WorkPntr refers to.  That means that it has a direct upwards path to the
 * root of the file system (you can get to the top by following only ".." links
 * - the AGMS links will be merged with the objects so that after further
 * processing ".." points to the main directory of the former AGMS link).  In
 * practice, that means that it can't be an orphan directory (orphans haven't
 * had their correct main parent set yet) - it either has to be marked as not
 * an orphan, or not on the list of directories being affected by the
 * rename/delete, or the root directory (which has NULL as a main parent
 * directory - conceptually the mount point directory is its "..").  Includes
 * the modifications to the file system listed in the current state (pretend
 * that the rename or delete has been done).  Returns TRUE if it is a useable
 * main parent directory, FALSE if it isn't.
 */

static bool RDIsValidMainParent (
  RDWorkStatePointer StatePntr,
  RDWorkFSOPointer WorkPntr,
  FileSystemObjectPointer CandidateParentFSOPntr)
{
  RDWorkFSOPointer  CandidateWorkPntr;

  /* The root directory's main parent is always NULL (representing the mount
  point directory). */

  if (WorkPntr->fsoPntr == &StatePntr->diskVolumePntr->rootFSOPntr->fso)
  {
    if (CandidateParentFSOPntr == NULL)
    {
#if DPRINT_NON_ERROR_MESSAGES
      dprintf ("ARFS:RDIsValidMainParent Candidate parent %s for root %s "
        "is valid by definition.\n",
        FSOAsString (CandidateParentFSOPntr),
        FSOAsString (WorkPntr->fsoPntr));
#endif
      return true;
    }

    /* Subdirectories can't become the main parent of the root, otherwise
    things will get weird.  Normally this code should never get executed,
    unless the root's main parent somehow becomes non-NULL. */

    dprintf ("ARFS:RDIsValidMainParent Bug!  Candidate parent %s for root %s "
      "is not NULL.\n", FSOAsString (CandidateParentFSOPntr),
      FSOAsString (WorkPntr->fsoPntr));
    return false;
  }

  if (INVALID_FILE_SYSTEM_OBJECT (CandidateParentFSOPntr) ||
  (CandidateParentFSOPntr->objectType != FSO_TYPE_DIRECTORY &&
  CandidateParentFSOPntr->objectType != FSO_TYPE_AGMS_LINK))
  {
    dprintf ("ARFS:RDIsValidMainParent Bug! Candidate %s isn't valid!\n",
      FSOAsString (CandidateParentFSOPntr));
    return false;
  }

  /* Exclude parent directories that are reached through deleted links.  They
  will be removed from the list of parents later on.  Well, unless it's also an
  added parent as part of the rename. */

  if ((CandidateParentFSOPntr == &StatePntr->sourceDirectoryPntr->fso &&
  WorkPntr->fsoPntr == StatePntr->renamedFSOPntr) ||
  (CandidateParentFSOPntr == &StatePntr->destinationDirectoryPntr->fso &&
  WorkPntr->fsoPntr == StatePntr->deletedFSOPntr))
  {
    /* This is one of the pretend-it-is-cut links.  If it is added back later
    then it isn't really cut. */

    if (CandidateParentFSOPntr != &StatePntr->destinationDirectoryPntr->fso ||
    WorkPntr->fsoPntr != StatePntr->renamedFSOPntr)
    {
#if DPRINT_NON_ERROR_MESSAGES
      dprintf ("ARFS:RDIsValidMainParent Candidate parent %s for %s "
        "uses a deleted link, so it's not valid.\n",
        FSOAsString (CandidateParentFSOPntr),
        FSOAsString (WorkPntr->fsoPntr));
#endif
      return false; /* This will be a deleted link, can't use it. */
    }
#if DPRINT_NON_ERROR_MESSAGES
    dprintf ("ARFS:RDIsValidMainParent Candidate parent %s for %s "
      "uses a deleted link and an added link, no decision yet.\n",
      FSOAsString (CandidateParentFSOPntr),
      FSOAsString (WorkPntr->fsoPntr));
#endif
  }

  /* Check and see if the given directory/link is outside the work tree. */

  CandidateWorkPntr = GetFSOWorkItem (StatePntr, CandidateParentFSOPntr);
  if (CandidateWorkPntr == NULL)
  {
    /* Directories outside the items being worked on are always presumed valid.
    If some other thread is moving them around, it will fix them up and keep
    them valid.  If it's being deleted, it would have to write-lock the child
    when it takes the candidate directory out of the child's list of parent
    directories, and we've already got the child read-locked, so that won't be
    a problem. */

#if DPRINT_NON_ERROR_MESSAGES
    dprintf ("ARFS:RDIsValidMainParent Candidate parent %s for %s "
      "is not in tree, so it is valid.\n",
      FSOAsString (CandidateParentFSOPntr),
      FSOAsString (WorkPntr->fsoPntr));
#endif
    return true;
  }

  if (CandidateWorkPntr->orphan)
  {
#if DPRINT_NON_ERROR_MESSAGES
    dprintf ("ARFS:RDIsValidMainParent Candidate parent %s for %s "
      "is in tree, but it is an orphan, so it is not valid.\n",
      FSOAsString (CandidateParentFSOPntr),
      FSOAsString (WorkPntr->fsoPntr));
#endif
    return false; /* An orphan can't be a valid parent directory. */
  }

#if DPRINT_NON_ERROR_MESSAGES
  dprintf ("ARFS:RDIsValidMainParent Candidate parent %s for %s "
    "is in tree, and is not an orphan, so it is valid.\n",
    FSOAsString (CandidateParentFSOPntr),
    FSOAsString (WorkPntr->fsoPntr));
#endif
  return true;
}



/******************************************************************************
 * For all items in the work tree, try to find a new main parent directory or
 * parent AGMS link (which will later be merged).  Return a failure code if a
 * cycle is found, which leaves some items without a path to the root of the
 * file system (except for the case where the only orphaned FSO is the one
 * being deleted).  Pretend that the rename or delete mentioned in the state
 * record have taken place when considering possible main parents.
 */

static status_t RDFindMainParents (
  RDWorkStatePointer StatePntr)
{
  RDWorkFSOPointer        AlreadySeenWorkPntr;
  AGMSLinkObjectPointer  *ExtraLinkPntrPntr;
  RDWorkFSOPointer        NextWorkPntr;
  status_t                ReturnCode = -1;
  bool                    SeenTwice;
  RDWorkFSOPointer        WorkPntr;

  if (StatePntr == NULL || StatePntr->workFSOTree == NULL)
  {
    dprintf ("ARFS:RDFindMainParents Called with NULL arguments!\n");
    goto ErrorExit;
  }

  /* Put all items in the work tree into the pending list.  The pending list is
  now a list of orphans which need to find a main parent directory. */

  StatePntr->headOfPendingList = NULL;
  StatePntr->tailOfPendingList = NULL;
  if (!AVLDupIterate (StatePntr->workFSOTree, true,
  NULL, 0, 0, false,
  NULL, 0, 0, false,
  RDWorkAddOrphanToPendingList, StatePntr))
  {
    dprintf ("ARFS:RDFindMainParents Corruption detected while "
      "initialising orphan list.\n");
    goto ErrorExit;
  }

  /* Process the orphans.  Repeatedly cycle through the orphans list, finding
  and removing ones which have acceptable main parents.  Stop when there are no
  more or if we do a complete iteration through the list without fixing any
  orphans. */

  WorkPntr = StatePntr->headOfPendingList;
  AlreadySeenWorkPntr = WorkPntr;
  SeenTwice = false;
  while (WorkPntr != NULL)
  {
    /* See if the current file system object has a valid main parent.  This is
    a parent directory or AGMS link which is a former orphan or a parent
    directory/link which is outside the list of affected directories.  Imagine
    that the rename or delete operation has been completed when considering
    possible parents.  First see if the old main parent directory can be used.
    */

    if (!WorkPntr->orphan)
    {
      dprintf ("ARFS:RDFindMainParents Bug!  Items in orphan list like "
        "%s should have the orphan flag true!\n",
        FSOAsString (WorkPntr->fsoPntr));
      goto ErrorExit;
    }

    if (RDIsValidMainParent (StatePntr, WorkPntr,
    &WorkPntr->fsoPntr->mainParentDirectory->fso))
    {
      WorkPntr->orphan = false;
      WorkPntr->newMainParent = NULL; /* No change needed. */
    }

    /* See if we can use the to-be-added parent directory (pretend it will be a
    parent), if we still need a parent directory for this file. */

    if (WorkPntr->orphan && WorkPntr->fsoPntr == StatePntr->renamedFSOPntr)
    {
      if (RDIsValidMainParent (StatePntr, WorkPntr,
      &StatePntr->destinationDirectoryPntr->fso))
      {
        WorkPntr->orphan = false;
        WorkPntr->newMainParent = &StatePntr->destinationDirectoryPntr->fso;
      }
    }

    /* See if we can use any of the AGMS links pointing to this object as
    parents, if needed. */

    if (WorkPntr->orphan && NULL !=
    (ExtraLinkPntrPntr = WorkPntr->fsoPntr->parentLinkArray))
    {
      while (*ExtraLinkPntrPntr != NULL) /* Scan through AGMS links. */
      {
        if (RDIsValidMainParent (StatePntr, WorkPntr,
        &(*ExtraLinkPntrPntr)->fso))
        {
          WorkPntr->orphan = false;
          WorkPntr->newMainParent = &(*ExtraLinkPntrPntr)->fso;
          break;
        }
        ExtraLinkPntrPntr++;
      }
    }

    /* Get the next work unit pointer, before removing the current one from the
    list and thus invalidating its next pointer.  Can be NULL if this is the
    work unit at the tail of the list. */

    NextWorkPntr = WorkPntr->nextPendingPntr;

    if (!WorkPntr->orphan)
    {
      /* No longer an orphan.  Remove it from the orphan list. */

#if DPRINT_NON_ERROR_MESSAGES
      dprintf ("ARFS:RDFindMainParents New main parent of %s will be %s, "
        "was %s.\n",
        FSOAsString (WorkPntr->fsoPntr),
        (WorkPntr->newMainParent == NULL) ?
          "unchanged" : FSOAsString (WorkPntr->newMainParent),
        FSOAsString (&WorkPntr->fsoPntr->mainParentDirectory->fso));
#endif
      RemoveFromMiddleOfPendingList (StatePntr, WorkPntr);
      AlreadySeenWorkPntr = NextWorkPntr; /* Note, may be NULL. */
      SeenTwice = false;
    }
    else /* Still an orphan.  Have we already seen this one? */
    {
      if (WorkPntr == AlreadySeenWorkPntr)
      {
        if (SeenTwice)
          break; /* Processed the whole list without finding more orphans. */
        else
          SeenTwice = true;
      }
    }

    /* When the end of the pending list is reached, cycle around back to the
    beginning. */

    WorkPntr = NextWorkPntr;
    if (WorkPntr == NULL)
      WorkPntr = StatePntr->headOfPendingList;
    if (AlreadySeenWorkPntr == NULL)
      AlreadySeenWorkPntr = WorkPntr;
  } /* End while more orphans to process. */


  /* See if there are any orphans left over.  There can be at most just the FSO
  which is being deleted, any others mean that there would be an isolated cycle
  in the tree after the rename/delete operation. */

  if (StatePntr->headOfPendingList != NULL)
  {
    WorkPntr = StatePntr->headOfPendingList;
    if (WorkPntr->fsoPntr != StatePntr->deletedFSOPntr ||
    WorkPntr->nextPendingPntr != NULL)
    {
      dprintf ("ARFS:RDFindMainParents There would be some orphaned files, "
        "with no link back to the root (usually due to a disconnected "
        "cycle).  The first one is %s, and %s.\n",
        FSOAsString (StatePntr->headOfPendingList->fsoPntr),
        (WorkPntr->nextPendingPntr) ?
          "there are more" : "that's all");
      ReturnCode = ELOOP; /* Not quite the right code, but close enough. */
      goto ErrorExit;
    }
  }

  return 0; /* Success! */

ErrorExit:
  return ReturnCode;
}



/******************************************************************************
 * An internal iteration callback function that adds all work units which will
 * be changing parents (plus the added and deleted items, even if they aren't
 * changing parents) to the pending list (so that the related parent
 * directories and AGMS links can later be added to the tree without messing up
 * the current iteration).  Returns TRUE if successful, FALSE if it found some
 * corrupt data.
 */

static bool RDWorkAddThingsChangingParentsToPendingList (
  AVLDupThingConstPointer KeyPntr,
  const unsigned int *ValuePntr,
  void *ExtraData)
{
  RDWorkStatePointer      StatePntr = ExtraData;
  RDWorkFSOPointer        WorkPntr = (RDWorkFSOPointer) *ValuePntr;

  if (WorkPntr == NULL || KeyPntr->int32Thing != (int) WorkPntr->fsoPntr)
  {
    dprintf ("ARFS:RDWorkAddThingsChangingParentsToPendingList "
      "Corrupt data detected!\n");
    return false;
  }

  if (WorkPntr->newMainParent == NULL /* Not changing main parent */ &&
  WorkPntr->fsoPntr != StatePntr->renamedFSOPntr &&
  WorkPntr->fsoPntr != StatePntr->deletedFSOPntr)
    return true; /* Continue iterating, didn't add it to list. */

  if (WorkPntr->fsoPntr->mainParentDirectory == NULL)
  {
    dprintf ("ARFS:RDWorkAddThingsChangingParentsToPendingList Bug!  "
      "Trying to change the main parent of %s, which has a NULL main "
      "parent (usually the root).\n",
      FSOAsString (WorkPntr->fsoPntr));
    return false;
  }

  AddToTailOfPendingList (StatePntr, WorkPntr);

  return true; /* Continue iterating. */
}



/******************************************************************************
 * Prepare things for items which will be moving to a new main parent directory
 * (and for the renamed and deleted items too, even if they aren't moving).
 * The moving items are often morphing too - merging with a former AGMS link
 * and leaving behind a placeholder AGMS link in their old main parent
 * directory.  This includes creating (if not already in the tree) and marking
 * work units for: the item moving, the old parent directory, the new parent
 * directory, the AGMS link it is merging with (the mergee), all the AGMS links
 * referencing the mergee.  It also includes creating a new AGMS link
 * placeholder in the old parent directory (if the item isn't being deleted
 * from the old directory).  Plus it needs a new array of references for the
 * moving item (combining its existing references and the references to the
 * mergee).  Returns zero if successful, a negative error code if the
 * allocations failed or something else went wrong.
 */

static status_t RDPrepareMovedItemsAndReferences (
  RDWorkStatePointer StatePntr)
{
  AGMSLinkObjectPointer  *CombinedLinkPntrPntr;
  status_t                ErrorCode;
  AGMSLinkObjectPointer  *ExtraLinkPntrPntr;
  AGMSLinkObjectPointer   MergeeAGMSLinkPntr;
  AGMSLinkObjectPointer   NewLinkPntr;
  DirectoryObjectPointer  NewParentDirectoryPntr;
  RDWorkFSOPointer        NewWorkPntr;
  int                     NumberOfReferencesFromMorpher;
  int                     NumberOfReferencesFromMergee;
  int                     NumberOfReferencesTotal;
  DirectoryObjectPointer  OldParentDirectoryPntr;
  status_t                ReturnCode = -1;
  RDWorkFSOPointer        WorkPntr;

  /* Add work units of the moved things to pending list for further processing
  (making new work units and adding them to the tree can't be done inside the
  tree traversal).  Also adds the renamed and deleted items, even if they
  aren't changing parents. */

  StatePntr->headOfPendingList = NULL;
  StatePntr->tailOfPendingList = NULL;
  if (!AVLDupIterate (StatePntr->workFSOTree, true,
  NULL, 0, 0, false,
  NULL, 0, 0, false,
  RDWorkAddThingsChangingParentsToPendingList, StatePntr))
  {
    dprintf ("ARFS:RDPrepareMovedItemsAndReferences Corrupted data.\n");
    goto ErrorExit;
  }

  /* Create a work unit for the morphing items, old parent directories, old
  AGMS links and new parent directories, so that we can read lock them now and
  later write lock them for morphing.  Make a placeholder new AGMS link too
  (but no work unit for it since it's invisible to other threads).  Also, for
  each morphing thing, make a list of its new parent links by including the new
  AGMS link placeholder in the old directory, the parentLinkArray from the
  merging AGMS link (if any) and the parentLinkArray from the morphing item,
  but exclude the mergee AGMS link since it is vanishing. */

  while ((WorkPntr = RemoveFromHeadOfPendingList (StatePntr)) != NULL)
  {
    /* Mark the item which is morphing as needing a write lock later. */

    WorkPntr->eventuallyWriteLock = true;

    /* Make the placeholder AGMS link which will get left behind in the old
    main parent directory.  Don't leave behind a placeholder for the renamed
    item or for the deleted item. */

    OldParentDirectoryPntr = WorkPntr->fsoPntr->mainParentDirectory;

    if (WorkPntr->fsoPntr != StatePntr->renamedFSOPntr &&
    WorkPntr->fsoPntr != StatePntr->deletedFSOPntr)
    {
      NewLinkPntr = AllocateAGMSLinkFSO (StatePntr->diskVolumePntr,
        WorkPntr->fsoPntr->trueNamePntr,
        WorkPntr->fsoPntr /* TargetFSOPntr */);
      if (NewLinkPntr == NULL)
      {
        dprintf ("ARFS:RDPrepareMovedItemsAndReferences Can't make AGMS link "
          "targetting %s.\n", FSOAsString (WorkPntr->fsoPntr));
        ReturnCode = ENOMEM;
        goto ErrorExit;
      }
      NewLinkPntr->fso.mainParentDirectory = OldParentDirectoryPntr;
      NewLinkPntr->fso.uid = WorkPntr->fsoPntr->uid;
      NewLinkPntr->fso.gid = WorkPntr->fsoPntr->gid;
      NewLinkPntr->fso.modeBits |=
        S_IRWXU | S_IRWXG | S_IRWXO; /* BFS uses all permission bits set. */

      WorkPntr->replacementAGMSLinkForOldMainParent = NewLinkPntr;

      /* Create a work unit (if one doesn't already exist) for the former main
      parent directory of the morphing item and mark it as needing a future
      write lock.  Theoretically the user would need write permission to the
      directory, but since we're kind of just changing something already in it,
      no check is done. */

      ErrorCode = MakeFSOWorkItemSafely (StatePntr,
        &OldParentDirectoryPntr->fso, &NewWorkPntr);
      if (ErrorCode != 0)
      {
        dprintf ("ARFS:RDPrepareMovedItemsAndReferences Unable to create "
          "work unit for former main parent %s of %s.\n",
          FSOAsString (&OldParentDirectoryPntr->fso),
          FSOAsString (WorkPntr->fsoPntr));
        ReturnCode = ErrorCode;
        goto ErrorExit;
      }
      NewWorkPntr->eventuallyWriteLock = true;
    }

    MergeeAGMSLinkPntr = NULL;
    NewParentDirectoryPntr = NULL;
    NumberOfReferencesFromMergee = 0;
    NumberOfReferencesFromMorpher = 0;

    if (WorkPntr->newMainParent != NULL &&
    WorkPntr->newMainParent->objectType == FSO_TYPE_AGMS_LINK)
    {
      /* Also need to eventually write lock the mergee AGMS link which will be
      merged with the morphing file. */

      MergeeAGMSLinkPntr = (AGMSLinkObjectPointer) WorkPntr->newMainParent;

      ErrorCode = MakeFSOWorkItemSafely (StatePntr,
        &MergeeAGMSLinkPntr->fso, &NewWorkPntr);
      if (ErrorCode != 0)
      {
        dprintf ("ARFS:RDPrepareMovedItemsAndReferences Unable to create work "
          "unit for %s being merged with %s.\n",
          FSOAsString (&MergeeAGMSLinkPntr->fso),
          FSOAsString (WorkPntr->fsoPntr));
        ReturnCode = ErrorCode;
        goto ErrorExit;
      }
      NewWorkPntr->eventuallyWriteLock = true;

      /* And need to modify all the AGMS links which point to the disappearing
      mergee link.  Also count them up.  Pretend that none of them will be
      deleted (which can happen if renaming overwrites an existing AGMS link to
      the thing being renamed). */

      ExtraLinkPntrPntr = MergeeAGMSLinkPntr->fso.parentLinkArray;
      if (ExtraLinkPntrPntr != NULL)
      {
        while (*ExtraLinkPntrPntr != NULL)
        {
          ErrorCode = MakeFSOWorkItemSafely (StatePntr,
            &(*ExtraLinkPntrPntr)->fso, &NewWorkPntr);
          if (ErrorCode != 0)
          {
            dprintf ("ARFS:RDPrepareMovedItemsAndReferences Unable to create "
              "work unit for %s which references the mergee %s.\n",
              FSOAsString (&(*ExtraLinkPntrPntr)->fso),
              FSOAsString (&MergeeAGMSLinkPntr->fso));
            ReturnCode = ErrorCode;
            goto ErrorExit;
          }
          NewWorkPntr->eventuallyWriteLock = true;

          NumberOfReferencesFromMergee++;
          ExtraLinkPntrPntr++;
        }
      }
      NewParentDirectoryPntr = MergeeAGMSLinkPntr->fso.mainParentDirectory;
    }
    else if (WorkPntr->newMainParent != NULL &&
    WorkPntr->newMainParent->objectType == FSO_TYPE_DIRECTORY)
    {
      NewParentDirectoryPntr =
      (DirectoryObjectPointer) WorkPntr->newMainParent;
    }

    /* Also need to eventually write lock the new parent directory.  It is
    either specified directly or is the parent of the merging AGMS link.  NULL
    if this is the deleted thing and it is going to nowhere or if this is a
    rename within one directory (it's been already marked for locking). */

    if (NewParentDirectoryPntr != NULL)
    {
      ErrorCode = MakeFSOWorkItemSafely (StatePntr,
        &NewParentDirectoryPntr->fso, &NewWorkPntr);
      if (ErrorCode != 0)
      {
        dprintf ("ARFS:RDPrepareMovedItemsAndReferences Unable to create work "
          "unit for new parent %s of former %s.\n",
          FSOAsString (&NewParentDirectoryPntr->fso),
          FSOAsString (WorkPntr->fsoPntr));
        ReturnCode = ErrorCode;
        goto ErrorExit;
      }
      NewWorkPntr->eventuallyWriteLock = true;
    }

    /* Also need to eventually write lock the target of an AGMS link which is
    actually being deleted (no new parent).  Need to write lock it so that the
    former AGMS link can be removed from the target's parent links array. */

    if (WorkPntr->fsoPntr == StatePntr->deletedFSOPntr &&
    NewParentDirectoryPntr == NULL &&
    WorkPntr->fsoPntr->objectType == FSO_TYPE_AGMS_LINK)
    {
      ErrorCode = MakeFSOWorkItemSafely (StatePntr,
        ((AGMSLinkObjectPointer) WorkPntr->fsoPntr)->targetFSOPntr,
        &NewWorkPntr);
      if (ErrorCode != 0)
      {
        dprintf ("ARFS:RDPrepareMovedItemsAndReferences Unable to create work "
          "unit for %s, which is the target of the to-be-deleted %s.\n",
          FSOAsString (
          ((AGMSLinkObjectPointer) WorkPntr->fsoPntr)->targetFSOPntr),
          FSOAsString (WorkPntr->fsoPntr));
        ReturnCode = ErrorCode;
        goto ErrorExit;
      }
      NewWorkPntr->eventuallyWriteLock = true;
    }

    /* Make an array listing all the references to the moved item.  This
    includes its current references, minus the mergee link, plus the
    placeholder link left behind in the old main parent directory, plus all the
    references to the mergee.  Start by counting the existing ones. */

    ExtraLinkPntrPntr = WorkPntr->fsoPntr->parentLinkArray;
    if (ExtraLinkPntrPntr != NULL)
    {
      while (*ExtraLinkPntrPntr != NULL)
      {
        if (*ExtraLinkPntrPntr != MergeeAGMSLinkPntr)
          NumberOfReferencesFromMorpher++;
        ExtraLinkPntrPntr++;
      }
    }

    NumberOfReferencesTotal =
      NumberOfReferencesFromMorpher + NumberOfReferencesFromMergee;
    if (WorkPntr->replacementAGMSLinkForOldMainParent != NULL)
      NumberOfReferencesTotal++;

    if (NumberOfReferencesTotal > 0)
    {
      WorkPntr->arrayOfMergedParentLinks =
        mallocDVP (StatePntr->diskVolumePntr,
        sizeof (AGMSLinkObjectPointer) * (NumberOfReferencesTotal + 1));
      if (WorkPntr->arrayOfMergedParentLinks == NULL)
      {
        dprintf ("ARFS:RDPrepareMovedItemsAndReferences Unable to allocate "
          "memory for %d references array for morphing %s.\n",
          NumberOfReferencesTotal, FSOAsString (WorkPntr->fsoPntr));
        ReturnCode = ENOMEM;
        goto ErrorExit;
      }

      /* Add the references from the morpher and the mergee to the new list. */

      CombinedLinkPntrPntr = WorkPntr->arrayOfMergedParentLinks;

      if ((ExtraLinkPntrPntr = WorkPntr->fsoPntr->parentLinkArray) != NULL)
      {
        while (*ExtraLinkPntrPntr != NULL)
        {
          if (*ExtraLinkPntrPntr != MergeeAGMSLinkPntr)
            *(CombinedLinkPntrPntr++) = *ExtraLinkPntrPntr;
          ExtraLinkPntrPntr++;
        }
      }

      if (MergeeAGMSLinkPntr != NULL &&
      (ExtraLinkPntrPntr = MergeeAGMSLinkPntr->fso.parentLinkArray) != NULL)
      {
        while (*ExtraLinkPntrPntr != NULL)
          *(CombinedLinkPntrPntr++) = *(ExtraLinkPntrPntr++);
      }

      if (WorkPntr->replacementAGMSLinkForOldMainParent != NULL)
        *(CombinedLinkPntrPntr++) =
        WorkPntr->replacementAGMSLinkForOldMainParent;

      *CombinedLinkPntrPntr = NULL; /* End of the list marker. */
    }
  }

  ReturnCode = 0; /* Success! */

ErrorExit:
  return ReturnCode;
}



/******************************************************************************
 * An internal iteration callback function that tries to get a write lock to
 * all things which need a write lock.  Returns TRUE if successful, FALSE if it
 * was unable to do a write lock.
 */

static bool RDWorkWriteLockThingsNeedingChanges (
  AVLDupThingConstPointer KeyPntr,
  const unsigned int *ValuePntr,
  void *ExtraData)
{
  RDWorkStatePointer      StatePntr = ExtraData;
  RDWorkFSOPointer        WorkPntr = (RDWorkFSOPointer) *ValuePntr;

  if (WorkPntr == NULL || KeyPntr->int32Thing != (int) WorkPntr->fsoPntr)
  {
    dprintf ("ARFS:RDWorkWriteLockThingsNeedingChanges "
      "Corrupt data detected!\n");
    return false;
  }

  if (!WorkPntr->eventuallyWriteLock || WorkPntr->gotWriteLock)
    return true; /* Nothing needs to be done for this one. */

  if (WriteLockWorkItem (WorkPntr, StatePntr) != 0)
    return false; /* An error has occured, stop iterating. */

  return true; /* Continue iterating. */
}



/******************************************************************************
 * An internal iteration callback function that moves a file system object
 * (call it the "merging" object) from its current parent directory to a new
 * parent directory.  If present, an AGMS link that refered to it (call it the
 * "mergee" object) from the new main parent directory gets deallocated and
 * replaced by the merging object (the merging object also takes on the name of
 * the AGMS link it replaces).  In the old main parent directory, a new AGMS
 * link is installed (if it has already been created) to act as a placeholder
 * for the moved object.  All existing AGMS links which pointed to the mergee
 * get changed to point to the merging object.  Returns TRUE if successful,
 * FALSE if something went wrong and the iteration should halt.
 */

static bool RDWorkChangeMainParent (
  AVLDupThingConstPointer KeyPntr,
  const unsigned int *ValuePntr,
  void *ExtraData)
{
  status_t                ErrorCode;
  AVLDupThingRecord       Key;
  AGMSLinkObjectPointer   LinkPntr;
  AGMSLinkObjectPointer  *LinkPntrPntr;
  RDWorkFSOPointer        LinkTargetWorkPntr;
  RDWorkFSOPointer        LinkWorkPntr;
  AGMSLinkObjectPointer   MergeeAGMSLinkPntr;
  FileSystemObjectPointer MergingFSOPntr;
  DirectoryObjectPointer  NewParentDirPntr;
  DirectoryObjectPointer  OldParentDirPntr;
  RDWorkStatePointer      StatePntr = ExtraData;
  char                   *StringPntr;
  RDWorkFSOPointer        WorkPntr = (RDWorkFSOPointer) *ValuePntr;

  MergingFSOPntr = WorkPntr->fsoPntr;

  if (WorkPntr->newMainParent == NULL &&
  MergingFSOPntr != StatePntr->renamedFSOPntr &&
  MergingFSOPntr != StatePntr->deletedFSOPntr)
    return true; /* No change in main parent needed for this one. */

  if (!WorkPntr->gotWriteLock)
  {
    dprintf ("ARFS:RDWorkChangeMainParent Bug!  %s isn't write locked.\n",
      FSOAsString (MergingFSOPntr));
    return false;
  }

  /* Prepare by finding old and new parent directories, and mergee too. */

  OldParentDirPntr = MergingFSOPntr->mainParentDirectory;
  if (OldParentDirPntr == NULL)
  {
    dprintf ("ARFS:RDWorkChangeMainParent Bug!  %s doesn't have a "
      "parent directory (and the root shouldn't be changing).\n",
      FSOAsString (MergingFSOPntr));
    return false;
  }

  NewParentDirPntr = NULL;
  MergeeAGMSLinkPntr = NULL;
  if (WorkPntr->newMainParent == NULL)
  {
  }
  else if (WorkPntr->newMainParent->objectType == FSO_TYPE_DIRECTORY)
    NewParentDirPntr = (DirectoryObjectPointer) WorkPntr->newMainParent;
  else if (WorkPntr->newMainParent->objectType == FSO_TYPE_AGMS_LINK)
  {
    MergeeAGMSLinkPntr = (AGMSLinkObjectPointer) WorkPntr->newMainParent;
    NewParentDirPntr = MergeeAGMSLinkPntr->fso.mainParentDirectory;
  }

  if (NewParentDirPntr == NULL &&
  MergingFSOPntr != StatePntr->deletedFSOPntr &&
  MergingFSOPntr != StatePntr->renamedFSOPntr)
  {
    dprintf ("ARFS:RDWorkChangeMainParent Bug!  New main parent is "
      "NULL for %s, but it is not the deleted %s or renamed %s.\n",
      FSOAsString (MergingFSOPntr),
      FSOAsString (StatePntr->deletedFSOPntr),
      FSOAsString (StatePntr->renamedFSOPntr));
    return false;
  }

#if DPRINT_NON_ERROR_MESSAGES
  dprintf ("ARFS:RDWorkChangeMainParent Moving %s%s from "
    "old parent %s to new parent %s, merging with %s.\n",
    FSOAsString (MergingFSOPntr),
    (MergingFSOPntr == StatePntr->deletedFSOPntr)
      ? " (the deleted thing)"
      : ((MergingFSOPntr == StatePntr->renamedFSOPntr)
        ? " (the renamed thing)"
        : ""),
    FSOAsString (&OldParentDirPntr->fso),
    FSOAsString (&NewParentDirPntr->fso),
    FSOAsString (&MergeeAGMSLinkPntr->fso));
#endif

  /* Update name and date indices and send the various notification messages
  now, before the data gets used up or is in an inconsistent state.  There will
  be a B_ENTRY_CREATED for a new left-behind AGMS link.  A B_ENTRY_REMOVED for
  a mergee.  B_ENTRY_MOVED for a directory or file moving to a new parent.
  Plus of course, B_ENTRY_MOVED for the renamed file (even if it is just
  changing its name) and a B_ENTRY_MOVED or B_ENTRY_REMOVED for the deleted
  file (if it takes over the mergee or just gets deleted). */

  if (MergeeAGMSLinkPntr != NULL)
  {
    ErrorCode = notify_listener (B_ENTRY_REMOVED,
      StatePntr->diskVolumePntr->nameSpaceID,
      (unsigned int) NewParentDirPntr, 0,
      (unsigned int) MergeeAGMSLinkPntr,
      MergeeAGMSLinkPntr->fso.trueNamePntr);
    if (ErrorCode != 0 || DPRINT_NOTIFICATIONS)
    {
      dprintf ("ARFS:RDWorkChangeMainParent: B_ENTRY_REMOVED notification for "
        "removing mergee %s in %s, error code %d.\n",
        FSOAsString (&MergeeAGMSLinkPntr->fso),
        FSOAsString (&NewParentDirPntr->fso),
        (int) ErrorCode);
    }
  }

  if (NewParentDirPntr == NULL)
  {
    /* Not moving, so must be deleting, or just changing the name. */

    if (MergingFSOPntr == StatePntr->renamedFSOPntr)
    {
      ChangeFSODates (StatePntr->diskVolumePntr, MergingFSOPntr,
        DK_INODE_CHANGED_FLAG, -1);
      ChangeFSODates (StatePntr->diskVolumePntr, &OldParentDirPntr->fso,
        DK_INODE_CHANGED_FLAG | DK_MODIFIED, -1);
      ChangeNameIndexForFSO (StatePntr->diskVolumePntr, MergingFSOPntr,
        MergingFSOPntr->trueNamePntr, StatePntr->newNamePntr);

      ErrorCode = notify_listener (B_ENTRY_MOVED,
        StatePntr->diskVolumePntr->nameSpaceID,
        (unsigned int) OldParentDirPntr,
        (unsigned int) OldParentDirPntr,
        (unsigned int) MergingFSOPntr,
        StatePntr->newNamePntr);
      if (ErrorCode != 0 || DPRINT_NOTIFICATIONS)
      {
        dprintf ("ARFS:RDWorkChangeMainParent: B_ENTRY_MOVED notification for "
          "name change of %s in %s, new name \"%s\", error code %d.\n",
          FSOAsString (MergingFSOPntr),
          FSOAsString (&OldParentDirPntr->fso),
          StatePntr->newNamePntr,
          (int) ErrorCode);
      }
    }
    else /* Deleted thing really being deleted. */
    {
      ChangeFSODates (StatePntr->diskVolumePntr, &OldParentDirPntr->fso,
        DK_INODE_CHANGED_FLAG | DK_MODIFIED, -1);
      ErrorCode = notify_listener (B_ENTRY_REMOVED,
        StatePntr->diskVolumePntr->nameSpaceID,
        (unsigned int) OldParentDirPntr, 0,
        (unsigned int) MergingFSOPntr,
        MergingFSOPntr->trueNamePntr);
      if (ErrorCode != 0 || DPRINT_NOTIFICATIONS)
      {
        dprintf ("ARFS:RDWorkChangeMainParent: B_ENTRY_REMOVED notification "
          "for actually removing deleted %s from %s, error code %d.\n",
          FSOAsString (MergingFSOPntr),
          FSOAsString (&OldParentDirPntr->fso),
          (int) ErrorCode);
      }
    }
  }
  else /* Moving to a new directory.  Can include deleted or renamed file. */
  {
    ChangeFSODates (StatePntr->diskVolumePntr, &OldParentDirPntr->fso,
      DK_INODE_CHANGED_FLAG | DK_MODIFIED, -1);
    ChangeFSODates (StatePntr->diskVolumePntr, &NewParentDirPntr->fso,
      DK_INODE_CHANGED_FLAG | DK_MODIFIED, -1);
    ChangeFSODates (StatePntr->diskVolumePntr, MergingFSOPntr,
      DK_INODE_CHANGED_FLAG, -1);
    if (MergingFSOPntr == StatePntr->renamedFSOPntr)
      ChangeNameIndexForFSO (StatePntr->diskVolumePntr, MergingFSOPntr,
        MergingFSOPntr->trueNamePntr, StatePntr->newNamePntr);
    else if (MergeeAGMSLinkPntr != NULL)
      ChangeNameIndexForFSO (StatePntr->diskVolumePntr, MergingFSOPntr,
        MergingFSOPntr->trueNamePntr, MergeeAGMSLinkPntr->fso.trueNamePntr);

    ErrorCode = notify_listener (B_ENTRY_MOVED,
      StatePntr->diskVolumePntr->nameSpaceID,
      (unsigned int) OldParentDirPntr,
      (unsigned int) NewParentDirPntr,
      (unsigned int) MergingFSOPntr,
      (MergingFSOPntr == StatePntr->renamedFSOPntr)
        ? StatePntr->newNamePntr
        : ((MergeeAGMSLinkPntr == NULL)
          ? MergingFSOPntr->trueNamePntr
          : MergeeAGMSLinkPntr->fso.trueNamePntr));
    if (ErrorCode != 0 || DPRINT_NOTIFICATIONS)
    {
      dprintf ("ARFS:RDWorkChangeMainParent: B_ENTRY_MOVED notification for "
        "movement of %s from %s to %s, new name \"%s\", error code %d.\n",
        FSOAsString (MergingFSOPntr),
        FSOAsString (&OldParentDirPntr->fso),
        FSOAsString (&NewParentDirPntr->fso),
        (MergingFSOPntr == StatePntr->renamedFSOPntr)
          ? StatePntr->newNamePntr
          : ((MergeeAGMSLinkPntr == NULL)
            ? MergingFSOPntr->trueNamePntr
            : MergeeAGMSLinkPntr->fso.trueNamePntr),
        (int) ErrorCode);
    }
  }

  if (WorkPntr->replacementAGMSLinkForOldMainParent != NULL)
  {
    ErrorCode = notify_listener (B_ENTRY_CREATED,
      StatePntr->diskVolumePntr->nameSpaceID,
      (unsigned int) OldParentDirPntr, 0,
      (unsigned int) WorkPntr->replacementAGMSLinkForOldMainParent,
      WorkPntr->replacementAGMSLinkForOldMainParent->fso.trueNamePntr);
    if (ErrorCode != 0 || DPRINT_NOTIFICATIONS)
    {
      dprintf ("ARFS:RDWorkChangeMainParent: B_ENTRY_CREATED notification for "
        "creating placeholder %s in %s, error code %d.\n",
        FSOAsString (&WorkPntr->replacementAGMSLinkForOldMainParent->fso),
        FSOAsString (&OldParentDirPntr->fso),
        (int) ErrorCode);
    }
  }

  /* Fix up the the old parent directory.  If a replacement AGMS link exists
  (allocated earlier) then change the directory entry for the merging object to
  point to the placeholder AGMS link.  If no AGMS link was prepared, just
  remove the merging object from the directory. */

  Key.longStringThing.stringPntr = MergingFSOPntr->trueNamePntr;
  Key.longStringThing.isLongString = true;
  if (WorkPntr->replacementAGMSLinkForOldMainParent != NULL)
  {
    if (OldParentDirPntr !=
    WorkPntr->replacementAGMSLinkForOldMainParent->fso.mainParentDirectory)
    {
      dprintf ("ARFS:RDWorkChangeMainParent Bug!  Placeholder %s has "
        "parent %s, but it should be %s.\n",
        FSOAsString (&WorkPntr->replacementAGMSLinkForOldMainParent->fso),
        FSOAsString (&WorkPntr->replacementAGMSLinkForOldMainParent->fso.
          mainParentDirectory->fso),
        FSOAsString (&OldParentDirPntr->fso));
      return false;
    }
    if (!AVLDupChangeValueInPlace (OldParentDirPntr->directory, &Key,
    (unsigned int) MergingFSOPntr /* old value */,
    (unsigned int) WorkPntr->replacementAGMSLinkForOldMainParent /* new */))
    {
      dprintf ("ARFS:RDWorkChangeMainParent Bug!  Unable to overwrite "
        "old parent directory (%s) entry for old %s with "
        "new placeholder %s.\n",
        FSOAsString (&OldParentDirPntr->fso),
        FSOAsString (MergingFSOPntr),
        FSOAsString (&WorkPntr->replacementAGMSLinkForOldMainParent->fso));
      return false;
    }
    WorkPntr->replacementAGMSLinkForOldMainParent = NULL; /* Used up now. */
  }
  else /* Just remove the object, no replacement placeholder. */
  {
    if (!AVLDupDelete (OldParentDirPntr->directory,
    &Key, (unsigned int) MergingFSOPntr))
    {
      dprintf ("ARFS:RDWorkChangeMainParent Bug!  Unable to remove %s "
        "from old parent %s.\n",
        FSOAsString (MergingFSOPntr),
        FSOAsString (&OldParentDirPntr->fso));
      return false;
    }
  }

  /* Fix up the new parent directory.  Overwrite an existing directory entry
  for the mergee AGMS link, but it can also be a new entry allocated earlier if
  there is no mergee.  Watch out for the deleted object, which may have none or
  may be merging.  Also watch out for the renamed object which may or may not
  be moving to a new directory. */

  if (MergeeAGMSLinkPntr != NULL) /* Overwriting existing new parent entry. */
  {
    Key.longStringThing.stringPntr = MergeeAGMSLinkPntr->fso.trueNamePntr;
    Key.longStringThing.isLongString = true;
    if (!AVLDupChangeValueInPlace (NewParentDirPntr->directory, &Key,
    (unsigned int) MergeeAGMSLinkPntr /* old value */,
    (unsigned int) MergingFSOPntr /* new value */))
    {
      dprintf ("ARFS:RDWorkChangeMainParent Bug!  Unable to overwrite "
        "new parent directory (%s) entry for old mergee %s with "
        "new merging %s.\n",
        FSOAsString (&NewParentDirPntr->fso),
        FSOAsString (&MergeeAGMSLinkPntr->fso),
        FSOAsString (MergingFSOPntr));
      return false;
    }
  }
  else /* Simple move to new parent directory or to nowhere (NULL). */
  {
    if (MergingFSOPntr == StatePntr->deletedFSOPntr)
    {
      if (NewParentDirPntr != NULL)
      {
        dprintf ("ARFS:RDWorkChangeMainParent Bug!  The deleted %s is moving "
          "to a new parent %s.  It should only move to nowhere (NULL) "
          "or merge.\n",
          FSOAsString (MergingFSOPntr),
          FSOAsString (&NewParentDirPntr->fso));
        return false;
      }
    }
    else if (MergingFSOPntr == StatePntr->renamedFSOPntr)
    {
      if (NewParentDirPntr != NULL &&
      NewParentDirPntr != StatePntr->destinationDirectoryPntr)
      {
        dprintf ("ARFS:RDWorkChangeMainParent Bug!  Renamed %s is moving to "
          "a new parent %s, but it should be moving to %s.\n",
          FSOAsString (MergingFSOPntr),
          FSOAsString (&NewParentDirPntr->fso),
          FSOAsString (&StatePntr->destinationDirectoryPntr->fso));
        return false;
      }
      if (!StatePntr->addedNewDirectoryEntry)
      {
        dprintf ("ARFS:RDWorkChangeMainParent Bug!  Renamed %s is moving to "
          "a new parent %s (can be NULL if same parent), but no directory "
          "entry has been pre-created.\n",
          FSOAsString (MergingFSOPntr),
          FSOAsString (&NewParentDirPntr->fso));
        return false;
      }
      StatePntr->addedNewDirectoryEntry = false; /* Use up the new entry. */
    }
    else /* Not the renamed or deleted thing. */
    {
      dprintf ("ARFS:RDWorkChangeMainParent Bug!  %s is moving to "
        "a new parent %s, but it isn't the renamed %s or the "
        "deleted %s.\n",
        FSOAsString (MergingFSOPntr),
        FSOAsString (&NewParentDirPntr->fso),
        FSOAsString (StatePntr->renamedFSOPntr),
        FSOAsString (StatePntr->deletedFSOPntr));
      return false;
    }
  }

  /* Adjust all AGMS links that used to point to the mergee to point to the
  object merging with the mergee.  Of course, with BeOS, it's hard to make
  links to a link so this code may never get executed.  Well, that was until I
  added the llink function. */

  if (MergeeAGMSLinkPntr != NULL)
  {
    LinkPntrPntr = MergeeAGMSLinkPntr->fso.parentLinkArray;
    if (LinkPntrPntr != NULL)
    {
      while ((LinkPntr = *LinkPntrPntr) != NULL)
      {
        /* Make sure that the refering link has been write-locked. */

        LinkWorkPntr = GetFSOWorkItem (StatePntr, &LinkPntr->fso);
        if (LinkWorkPntr == NULL || !LinkWorkPntr->gotWriteLock)
        {
          dprintf ("ARFS:RDWorkChangeMainParent Bug!  %s, which points to the "
            "mergee %s, does not have a write locked work unit, so can't "
            "change it to point to the merger %s.\n",
            FSOAsString (&LinkPntr->fso),
            FSOAsString (&MergeeAGMSLinkPntr->fso),
            FSOAsString (MergingFSOPntr));
          return false;
        }

        if (LinkPntr->targetFSOPntr != &MergeeAGMSLinkPntr->fso)
        {
          dprintf ("ARFS:RDWorkChangeMainParent Bug!  %s, which was supposed "
            "to point to the mergee %s, does not.\n",
            FSOAsString (&LinkPntr->fso),
            FSOAsString (&MergeeAGMSLinkPntr->fso));
          return false;
        }

        /* Change it to refer to the merging object instead of the mergee. */

        LinkPntr->targetFSOPntr = MergingFSOPntr;

        ChangeFSODates (StatePntr->diskVolumePntr, &LinkPntr->fso,
          DK_INODE_CHANGED_FLAG | DK_MODIFIED, -1);

        LinkPntrPntr++; /* Advance to next element in parentLinkArray. */
      }
    }
  }

  /* Update the parent links array for the merging object to be the combined
  list from the mergee and the merger's existing references, plus the
  placeholder link if it exists. */

  if (MergingFSOPntr->parentLinkArray != NULL)
    freeDVP (StatePntr->diskVolumePntr, MergingFSOPntr->parentLinkArray);
  MergingFSOPntr->parentLinkArray = WorkPntr->arrayOfMergedParentLinks;
  WorkPntr->arrayOfMergedParentLinks = NULL; /* It's been used up now. */

  /* And the new main parent pointer too, if it is changing.  Will be NULL if
  it is staying the same (a rename that just changes the name) or the thing is
  being deleted. */

  if (NewParentDirPntr != NULL)
    MergingFSOPntr->mainParentDirectory = NewParentDirPntr;

  /* Change the name of the merging object to be the same as the mergee it is
  replacing (it is reusing the same directory entry in the ordered tree, so it
  has to keep the same name).  Do this last to avoid confusing debug output
  with the wrong names.  Also delete the mergee from the file system after
  swapping name strings. */

  if (MergeeAGMSLinkPntr != NULL)
  {
    if (MergingFSOPntr == StatePntr->renamedFSOPntr)
    {
      dprintf ("ARFS:RDWorkChangeMainParent Bug!  Merging %s with an existing "
        "mergee %s, but it is the renamed object, which never merges.\n",
        FSOAsString (MergingFSOPntr),
        FSOAsString (&MergeeAGMSLinkPntr->fso));
      return false;
    }

    /* Remove the name index entry for the mergee now, since it will have the
    wrong name (all dashes) later on when it actually gets deleted. */

    ChangeNameIndexForFSO (StatePntr->diskVolumePntr, &MergeeAGMSLinkPntr->fso,
      MergeeAGMSLinkPntr->fso.trueNamePntr, NULL);

    /* Swap the name strings. */

    StringPntr = MergingFSOPntr->trueNamePntr;
    MergingFSOPntr->trueNamePntr = MergeeAGMSLinkPntr->fso.trueNamePntr;
    MergeeAGMSLinkPntr->fso.trueNamePntr = StringPntr;
    while (*StringPntr)
      *StringPntr++ = '-'; /* Change disappearing name to avoid confusion. */

    ErrorCode = remove_vnode (StatePntr->diskVolumePntr->nameSpaceID,
      (unsigned int) MergeeAGMSLinkPntr);
    if (ErrorCode != 0)
    {
      dprintf ("ARFS:RDWorkChangeMainParent Bug!  remove_vnode failed with "
        "code %d when removing the mergee %s, merger is %s.\n",
        (int) ErrorCode,
        FSOAsString (&MergeeAGMSLinkPntr->fso),
        FSOAsString (MergingFSOPntr));
      return false;
    }
  }
  else if (MergingFSOPntr == StatePntr->deletedFSOPntr)
  {
    /* Actually have to delete the deleted thing, because it didn't have a
    mergee to save its life.  If we're deleting an AGMS link, then the target's
    list of parents also needs to be fixed up.  Other deleted things don't have
    dependencies (only delete empty directories). */

    if (MergingFSOPntr->objectType == FSO_TYPE_AGMS_LINK)
    {
      LinkPntr = (AGMSLinkObjectPointer) MergingFSOPntr;
      LinkTargetWorkPntr = GetFSOWorkItem (StatePntr, LinkPntr->targetFSOPntr);
      if (LinkTargetWorkPntr == NULL || !LinkTargetWorkPntr->gotWriteLock)
      {
        dprintf ("ARFS:RDWorkChangeMainParent Bug!  Deleting %s, but target "
          "%s doesn't have a write locked work unit.\n",
          FSOAsString (MergingFSOPntr),
          FSOAsString (LinkPntr->targetFSOPntr));
        return false;
      }
      if (!RemoveAGMSLinkReferenceFromFSO (StatePntr->diskVolumePntr,
      LinkPntr->targetFSOPntr, LinkPntr))
      {
        dprintf ("ARFS:RDWorkChangeMainParent Bug!  Unable to remove "
          "to-be-deleted %s from list of parents in %s.\n",
          FSOAsString (MergingFSOPntr),
          FSOAsString (LinkPntr->targetFSOPntr));
        return false;
      }
      ChangeFSODates (StatePntr->diskVolumePntr, LinkPntr->targetFSOPntr,
        DK_INODE_CHANGED_FLAG, -1);
      LinkPntr->targetFSOPntr = NULL;
    }

    ErrorCode = remove_vnode (StatePntr->diskVolumePntr->nameSpaceID,
      (unsigned int) MergingFSOPntr);
    if (ErrorCode != 0)
    {
      dprintf ("ARFS:RDWorkChangeMainParent Bug!  remove_vnode failed with "
        "code %d when removing the deleted object %s.\n",
        (int) ErrorCode,
        FSOAsString (MergingFSOPntr));
      return false;
    }
  }
  else if (MergingFSOPntr == StatePntr->renamedFSOPntr)
  {
    /* Change the renamed thing's true name to the new name.  Non-NULL string
    pointers in the state record will get deallocated on exit. */

    if (StatePntr->newTrueNamePntr == NULL)
    {
      dprintf ("ARFS:RDWorkChangeMainParent Bug!  New name string for the "
        "renamed %s isn't there!\n",
        FSOAsString (MergingFSOPntr));
      return false;
    }
    StatePntr->originalTrueNamePntr = MergingFSOPntr->trueNamePntr;
    MergingFSOPntr->trueNamePntr = StatePntr->newTrueNamePntr;
    StatePntr->newTrueNamePntr = NULL;
  }
  else
  {
    dprintf ("ARFS:RDWorkChangeMainParent Bug!  Not correctly processing "
      "%s.\n", FSOAsString (MergingFSOPntr));
    return false;
  }

  return true;
}



/******************************************************************************
 * Process a rename or delete.  The caller has set up a RDWorkStateRecord with
 * the parameters of the rename or delete operation, from diskVolumePntr on
 * down (set the unused and other fields to zero).  This function either
 * successfully does the operation and returns zero or apparently does nothing
 * and returns a negative error code.
 */

static status_t RenameOrDelete (
  RDWorkStatePointer StatePntr)
{
  bool                    DestinationDirectoryGetVNodeDone = false;
  bool                    DestinationDirectoryReadLockDone = false;
  status_t                ErrorCode;
  AVLDupThingRecord       Key;
  int                     NewNameLength;
  status_t                ReturnCode = -1;
  bool                    SourceDirectoryGetVNodeDone = false;
  bool                    SourceDirectoryReadLockDone = false;
  char                   *StringPntr;
  unsigned int            TempUINT;
  RDWorkFSOPointer        WorkPntr;
  void                   *VoidPntr;

  /* Some validity checks, just in case. */

  if (StatePntr == NULL ||

  INVALID_DISK_VOLUME (StatePntr->diskVolumePntr) ||

  /* Make sure the source directory and file name are valid, if used. */
  (StatePntr->sourceDirectoryPntr != NULL &&
  (StatePntr->oldNamePntr == NULL ||
  INVALID_FILE_SYSTEM_OBJECT (&StatePntr->sourceDirectoryPntr->fso) ||
  StatePntr->sourceDirectoryPntr->fso.objectType != FSO_TYPE_DIRECTORY)) ||

  /* Check the destination directory and file name, they are always used. */
  INVALID_FILE_SYSTEM_OBJECT (&StatePntr->destinationDirectoryPntr->fso) ||
  StatePntr->destinationDirectoryPntr->fso.objectType != FSO_TYPE_DIRECTORY ||
  StatePntr->newNamePntr == NULL)
  {
    dprintf ("ARFS:RenameOrDelete Insane arguments provided by caller.\n");
    return EINVAL;
  }

  /* Reject names consisting of all dots, like ".." or "...." or "", which we
  reserve for assorted kinds of parent directories that are automatically
  generated by the file system. */

  if (StatePntr->sourceDirectoryPntr != NULL)
  {
    for (StringPntr = (char *) StatePntr->oldNamePntr;
    *StringPntr != 0; StringPntr++)
      if (*StringPntr != '.')
        break;
    if (*StringPntr == 0)
    {
      dprintf ("ARFS:RenameOrDelete invalid source \"%s\" is all periods, "
        "which are reserved for parent directory names.\n",
        StatePntr->oldNamePntr);
      ReturnCode = EINVAL;
      goto ErrorExit;
    }
  }

  for (StringPntr = (char *) StatePntr->newNamePntr;
  *StringPntr != 0; StringPntr++)
    if (*StringPntr != '.')
      break;
  if (*StringPntr == 0)
  {
    dprintf ("ARFS:RenameOrDelete: invalid input newname \"%s\" is all "
      "periods, which are reserved for parent directory names.\n",
      StatePntr->newNamePntr);
    ReturnCode = EINVAL;
    goto ErrorExit;
  }

  /* Create an empty work tree. */

  StatePntr->workFSOTree = AVLDupAllocTree (B_INT32_TYPE,
    StatePntr->diskVolumePntr->mallocPntr,
    StatePntr->diskVolumePntr->freePntr,
    StatePntr->diskVolumePntr->volumePoolID);

  if (StatePntr->workFSOTree == NULL)
  {
    dprintf ("ARFS:RenameOrDelete Unable to create work item tree.\n");
    ReturnCode = ENOMEM;
    goto ErrorExit;
  }

  /* Find the source file, but only if renaming is being done.  Also getvnode
  the source directory and read lock it so that it doesn't change between the
  time we're looking at it and later when it gets changed. */

  if (StatePntr->sourceDirectoryPntr != NULL)
  {
    ErrorCode = get_vnode (StatePntr->diskVolumePntr->nameSpaceID,
      (unsigned int) StatePntr->sourceDirectoryPntr, &VoidPntr);
    if (ErrorCode == 0)
      SourceDirectoryGetVNodeDone = true;
    else
      ReturnCode = ErrorCode;
    if (ErrorCode != 0 || VoidPntr != StatePntr->sourceDirectoryPntr)
    {
      dprintf ("ARFS:RenameOrDelete get_vnode didn't work, code %d, "
        "for source %s.\n", (int) ErrorCode,
        FSOAsString (&StatePntr->sourceDirectoryPntr->fso));
      goto ErrorExit;
    }

    ErrorCode = ReadLock (&StatePntr->sourceDirectoryPntr->fso, false);
    if (ErrorCode != 0)
    {
      dprintf ("ARFS:RenameOrDelete unable to get read lock for source %s, "
        "error %d.\n", FSOAsString (&StatePntr->sourceDirectoryPntr->fso),
        (int) ErrorCode);
      ReturnCode = ErrorCode;
      goto ErrorExit;
    }
    SourceDirectoryReadLockDone = true;

    /* Fail if the user isn't allowed to write this directory. */

    ErrorCode = PermissionAllowed (StatePntr->diskVolumePntr,
      &StatePntr->sourceDirectoryPntr->fso, W_OK);
    if (ErrorCode != 0)
    {
      dprintf ("ARFS:RenameOrDelete Don't have write access permission for "
        "source %s (need the w flag).\n",
        FSOAsString (&StatePntr->sourceDirectoryPntr->fso));
      ReturnCode = ErrorCode;
      goto ErrorExit;
    }

    /* Look for the source file name in the directory. */

    Key.longStringThing.stringPntr = (char *) StatePntr->oldNamePntr;
    Key.longStringThing.isLongString = true;
    TempUINT = AVLDupFind (StatePntr->sourceDirectoryPntr->directory, &Key);
    StatePntr->renamedFSOPntr = (FileSystemObjectPointer) TempUINT;
    if (StatePntr->renamedFSOPntr == NULL)
    {
      dprintf ("ARFS:RenameOrDelete Can't find name \"%s\" in source %s.\n",
        StatePntr->oldNamePntr,
        FSOAsString (&StatePntr->sourceDirectoryPntr->fso));
      ReturnCode = ENOENT;
      goto ErrorExit;
    }

    /* Make a new true name string for the renamed file, which will be used to
    replace the old true name later on. */

    NewNameLength = strlen (StatePntr->newNamePntr);
    if (NewNameLength <= 0 || NewNameLength >= NAME_MAX)
    {
      dprintf ("ARFS:RenameOrDelete New name length is %d, which is not "
        "between 1 and %d.\n", NewNameLength, (int) NAME_MAX - 1);
      ReturnCode = ENAMETOOLONG;
      goto ErrorExit;
    }
    StatePntr->newTrueNamePntr =
      mallocDVP (StatePntr->diskVolumePntr, NewNameLength + 1);
    if (StatePntr->newTrueNamePntr == NULL)
    {
      dprintf ("ARFS:RenameOrDelete Unable to allocate string copy of "
        "new name \"%s\".\n", StatePntr->newNamePntr);
      ReturnCode = ENOMEM;
      goto ErrorExit;
    }
    strcpy (StatePntr->newTrueNamePntr, StatePntr->newNamePntr);

    /* Add the source file to the tree of children which may need new parent
    directories - since it will be moving and might need a new main parent
    directory. */

    ErrorCode = MakeFSOWorkItemBlindly (StatePntr,
      StatePntr->renamedFSOPntr, &WorkPntr);
    if (ErrorCode != 0)
    {
      dprintf ("ARFS:RenameOrDelete Unable to add %s found in source %s, "
        "to work tree.\n", FSOAsString (StatePntr->renamedFSOPntr),
        FSOAsString (&StatePntr->sourceDirectoryPntr->fso));
      ReturnCode = ErrorCode;
      goto ErrorExit;
    }
    WorkPntr->eventuallyWriteLock = true;
    AddToTailOfPendingList (StatePntr, WorkPntr);
  }

  /* Prepare to access the destination directory, don't need to lock it if it
  is the same as the source one, which is already locked. */

  if (StatePntr->destinationDirectoryPntr != StatePntr->sourceDirectoryPntr)
  {
    ErrorCode = get_vnode (StatePntr->diskVolumePntr->nameSpaceID,
      (unsigned int) StatePntr->destinationDirectoryPntr, &VoidPntr);
    if (ErrorCode == 0)
      DestinationDirectoryGetVNodeDone = true;
    else
      ReturnCode = ErrorCode;
    if (ErrorCode != 0 || VoidPntr != StatePntr->destinationDirectoryPntr)
    {
      dprintf ("ARFS:RenameOrDelete get_vnode didn't work, code %d, "
        "for destination %s.\n", (int) ErrorCode,
        FSOAsString (&StatePntr->destinationDirectoryPntr->fso));
      goto ErrorExit;
    }

    ErrorCode = ReadLock (&StatePntr->destinationDirectoryPntr->fso, false);
    if (ErrorCode != 0)
    {
      dprintf ("ARFS:RenameOrDelete unable to get read lock for "
        "destination %s, error %d.\n",
        FSOAsString (&StatePntr->destinationDirectoryPntr->fso),
        (int) ErrorCode);
      ReturnCode = ErrorCode;
      goto ErrorExit;
    }
    DestinationDirectoryReadLockDone = true;
  }

  /* Need write access to the destination directory. */

  ErrorCode = PermissionAllowed (StatePntr->diskVolumePntr,
    &StatePntr->destinationDirectoryPntr->fso, W_OK);
  if (ErrorCode != 0)
  {
    dprintf ("ARFS:RenameOrDelete Don't have write access permission for "
      "destination %s (need the w flag).\n",
      FSOAsString (&StatePntr->destinationDirectoryPntr->fso));
    ReturnCode = ErrorCode;
    goto ErrorExit;
  }

  /* Look for the deleted/overwritten file name in destination directory.  It's
  OK to not find it if a rename is being done. */

  Key.longStringThing.stringPntr = (char *) StatePntr->newNamePntr;
  Key.longStringThing.isLongString = true;
  TempUINT = AVLDupFind (StatePntr->destinationDirectoryPntr->directory, &Key);
  StatePntr->deletedFSOPntr = (FileSystemObjectPointer) TempUINT;

  /* Special behavior for deleting files and directories. */

  if (StatePntr->sourceDirectoryPntr == NULL /* doing a delete, not rename */)
  {
    if (StatePntr->deletedFSOPntr == NULL)
    {
      dprintf ("ARFS:RenameOrDelete Can't find to-be-deleted name \"%s\" in "
        "%s.\n", StatePntr->newNamePntr,
        FSOAsString (&StatePntr->destinationDirectoryPntr->fso));
      ReturnCode = ENOENT;
      goto ErrorExit;
    }

    /* Implement the antique unlink or rmdir behaviour as requested.  Indices
    are a lot like files (no real child objects) but appear to the user as a
    directory. */

    if (StatePntr->rmdirNotUnlink) /* If doing a rmdir. */
    {
      if (StatePntr->deletedFSOPntr->objectType != FSO_TYPE_DIRECTORY &&
      StatePntr->deletedFSOPntr->objectType != FSO_TYPE_INDEX)
      {
        dprintf ("ARFS:RenameOrDelete rmdir behaviour - can't delete a "
          "non-directory, like %s.\n",
          FSOAsString (StatePntr->deletedFSOPntr));
        ReturnCode = ENOTDIR;
        goto ErrorExit;
      }
    }
    else /* Doing an unlink. */
    {
      if (StatePntr->deletedFSOPntr->objectType == FSO_TYPE_DIRECTORY ||
      StatePntr->deletedFSOPntr->objectType == FSO_TYPE_INDEX)
      {
        dprintf ("ARFS:RenameOrDelete unlink behaviour - can't delete a "
          "directory, like %s.\n",
          FSOAsString (StatePntr->deletedFSOPntr));
        ReturnCode = EISDIR;
        goto ErrorExit;
      }
    }
  }

  /* Check that the source and destination aren't the same - renaming a file to
  have the same name as it already has, or something similarly confusing.
  Would cause problems if it was added to the tree twice. */

  if (StatePntr->deletedFSOPntr == StatePntr->renamedFSOPntr)
  {
    dprintf ("ARFS:RenameOrDelete Source and destination are the same file, "
      "renaming %s as \"%s\" in %s?  Not a problem, will do nothing.\n",
      FSOAsString (StatePntr->deletedFSOPntr), StatePntr->newNamePntr,
      FSOAsString (&StatePntr->destinationDirectoryPntr->fso));
    ReturnCode = 0; /* Success, in a lazy way. */
    goto ErrorExit;
  }

  /* Add the overwritten/deleted file to the tree of children, if it exists. */

  if (StatePntr->deletedFSOPntr != NULL)
  {
    ErrorCode = MakeFSOWorkItemBlindly (StatePntr,
      StatePntr->deletedFSOPntr, &WorkPntr);
    if (ErrorCode != 0)
    {
      dprintf ("ARFS:RenameOrDelete Unable to add destination item %s "
        "to work tree.\n", FSOAsString (StatePntr->deletedFSOPntr));
      ReturnCode = ErrorCode;
      goto ErrorExit;
    }
    WorkPntr->eventuallyWriteLock = true;
    AddToTailOfPendingList (StatePntr, WorkPntr);
  }

  /* Check that we aren't trying to delete a non-empty directory.  It's OK to
  "delete" it if there is more than one parent directory since in that case it
  only gets unlinked from one parent, leaving it undeleted.  Well, we could
  allow deleting non-empty directories by deleting all things on the orphan
  list, but that would annoy users who have gotten used to the lack of a
  recursive delete.  It's OK to delete an index since the children aren't
  really there (no parent links in the children to worry about either). */

  if (StatePntr->deletedFSOPntr != NULL &&
  StatePntr->deletedFSOPntr->objectType == FSO_TYPE_DIRECTORY &&
  StatePntr->deletedFSOPntr->parentLinkArray == NULL)
  {
    /* Got a directory about to be really deleted.  Check that it's empty. */

    if (0 != AVLDupGetTreeCount (
    ((DirectoryObjectPointer) StatePntr->deletedFSOPntr)->directory))
    {
      dprintf ("ARFS:RenameOrDelete Would actually delete a non-empty %s.\n",
        FSOAsString (StatePntr->deletedFSOPntr));
      ReturnCode = ENOTEMPTY;
      goto ErrorExit;
    }
  }

  /* Find all descendants of the items about to be unlinked.  Their main parent
  directory could be affected by the change. */

  ErrorCode = RDFindDescendantChildren (StatePntr);
  if (ErrorCode < 0)
  {
    dprintf ("ARFS:RenameOrDelete Unable to find children of "
      "deleted things, code %d.\n", (int) ErrorCode);
    ReturnCode = ErrorCode;
    goto ErrorExit;
  }

  /* Recompute the new main parent for the affected items.  Can fail if there
  is a disconnected cycle in the tree. */

  ErrorCode = RDFindMainParents (StatePntr);

#if 0 && DPRINT_NON_ERROR_MESSAGES
  dprintf ("ARFS:RenameOrDelete After finding main parents:\n");
  RDWorkListAll (StatePntr);
#endif

  if (ErrorCode < 0)
  {
    dprintf ("ARFS:RenameOrDelete Unable to fix up main parent of all "
      "affected children, code %d.\n", (int) ErrorCode);
    ReturnCode = ErrorCode;
    goto ErrorExit;
  }

  /* Put the parent directories of source and destination into the tree for
  more convenient processing, if they aren't already duplicated in there (note
  that they may be in the tree, and thus have an extra read lock and getvnode).
  Can't let duplicates exist when we need to write lock (since you can't have
  any extra read-locks when doing the write-lock). */

  if (StatePntr->sourceDirectoryPntr != NULL)
  {
    ErrorCode = MakeFSOWorkItemSafely (StatePntr,
      &StatePntr->sourceDirectoryPntr->fso, &WorkPntr);
    if (ErrorCode != 0)
    {
      dprintf ("ARFS:RenameOrDelete Unable to add work unit to "
        "tree, code %d for source %s.\n", (int) ErrorCode,
        FSOAsString (&StatePntr->sourceDirectoryPntr->fso));
      ReturnCode = ErrorCode;
      goto ErrorExit;
    }
    WorkPntr->eventuallyWriteLock = true;
    ReadUnlock (&StatePntr->sourceDirectoryPntr->fso);
    SourceDirectoryReadLockDone = false;
    put_vnode (StatePntr->diskVolumePntr->nameSpaceID,
      (unsigned int) StatePntr->sourceDirectoryPntr);
    SourceDirectoryGetVNodeDone = false;
  }

  if (StatePntr->destinationDirectoryPntr != StatePntr->sourceDirectoryPntr)
  {
    ErrorCode = MakeFSOWorkItemSafely (StatePntr,
      &StatePntr->destinationDirectoryPntr->fso, &WorkPntr);
    if (ErrorCode != 0)
    {
      dprintf ("ARFS:RenameOrDelete Unable to add work unit to "
        "tree, code %d for destination %s.\n", (int) ErrorCode,
        FSOAsString (&StatePntr->destinationDirectoryPntr->fso));
      ReturnCode = ErrorCode;
      goto ErrorExit;
    }
    WorkPntr->eventuallyWriteLock = true;
    ReadUnlock (&StatePntr->destinationDirectoryPntr->fso);
    DestinationDirectoryReadLockDone = false;
    put_vnode (StatePntr->diskVolumePntr->nameSpaceID,
      (unsigned int) StatePntr->destinationDirectoryPntr);
    DestinationDirectoryGetVNodeDone = false;
  }

  /* Do all the allocations and many other preparations needed for moving items
  to new parent directories. */

  ErrorCode = RDPrepareMovedItemsAndReferences (StatePntr);
  if (ErrorCode != 0)
  {
    dprintf ("ARFS:RenameOrDelete Unable to prepare items which are changing "
      "their main parent directory.  Code %d.\n", (int) ErrorCode);
    ReturnCode = ErrorCode;
    goto ErrorExit;
  }

#if DPRINT_NON_ERROR_MESSAGES
  dprintf ("ARFS:RenameOrDelete After RDPrepareMovedItemsAndReferences:\n");
  RDWorkListAll (StatePntr);
#endif

  /* Upgrade the destination directory to a write lock. */

  ErrorCode = WriteLockWorkItemByFSO (
    &StatePntr->destinationDirectoryPntr->fso, StatePntr);
  if (ErrorCode != 0)
  {
    dprintf ("ARFS:RenameOrDelete Can't get write lock to destination %s.\n",
      FSOAsString (&StatePntr->destinationDirectoryPntr->fso));
    ReturnCode = ErrorCode;
    goto ErrorExit;
  }

  /* Add a new entry in the destination directory for the renamed file, using
  the new name (if we are doing renaming, there can temporarily be two entries
  in the directory with the same name). */

  if (StatePntr->sourceDirectoryPntr != NULL)
  {
    Key.longStringThing.stringPntr = StatePntr->newTrueNamePntr;
    Key.longStringThing.isLongString = true;
    ErrorCode = AVLDupAdd (StatePntr->destinationDirectoryPntr->directory,
      &Key, (unsigned int) StatePntr->renamedFSOPntr);
    if (ErrorCode != AVLADD_ADDED_A_NODE)
    {
      dprintf ("ARFS:RenameOrDelete unable to add name \"%s\" to "
        "%s, tree code: %d.\n", Key.longStringThing.stringPntr,
        FSOAsString (&StatePntr->destinationDirectoryPntr->fso),
        (int) ErrorCode);
      ReturnCode = ENOMEM;
      goto ErrorExit;
    }
    StatePntr->addedNewDirectoryEntry = true;
  }

  /* All allocations have been done, so nothing can fail now other than due to
  bugs or deadlocks.  Get write locks to everything which is being changed. */

  if (!AVLDupIterate (StatePntr->workFSOTree, true,
  NULL, 0, 0, false,
  NULL, 0, 0, false,
  RDWorkWriteLockThingsNeedingChanges, StatePntr))
  {
    dprintf ("ARFS:RenameOrDelete unable to get write locks for "
      "everything which is changing.\n");
    ReturnCode = EDEADLK; /* Usually a deadlock. */
    goto ErrorExit;
  }

  /* Deal with files that are changing to a new main parent (remove from old
  parent directory, add to new, redirect AGMS links).  This also includes the
  deleted / overwritten file which is changing to have no main parent (if it
  has no AGMS links), and the renamed file too. */

  if (!AVLDupIterate (StatePntr->workFSOTree, true,
  NULL, 0, 0, false,
  NULL, 0, 0, false,
  RDWorkChangeMainParent, StatePntr))
  {
    dprintf ("ARFS:RenameOrDelete Bug!  Unable to change main parent "
      "directories.\n");
    goto ErrorExit;
  }

  /* Some final sanity checks. */

  if (StatePntr->addedNewDirectoryEntry)
  {
    dprintf ("ARFS:RenameOrDelete New directory entry for renamed "
      "%s wasn't used up.\n",
      FSOAsString (StatePntr->renamedFSOPntr));
    goto ErrorExit;
  }

  if (StatePntr->newTrueNamePntr)
  {
    dprintf ("ARFS:RenameOrDelete New true name for renamed "
      "%s wasn't used up.\n",
      FSOAsString (StatePntr->renamedFSOPntr));
    goto ErrorExit;
  }

  /* Successful! */

  ReturnCode = 0;

ErrorExit:
  if (DestinationDirectoryReadLockDone)
    ReadUnlock (&StatePntr->destinationDirectoryPntr->fso);
  if (DestinationDirectoryGetVNodeDone)
    put_vnode (StatePntr->diskVolumePntr->nameSpaceID,
    (unsigned int) StatePntr->destinationDirectoryPntr);
  if (SourceDirectoryReadLockDone)
    ReadUnlock (&StatePntr->sourceDirectoryPntr->fso);
  if (SourceDirectoryGetVNodeDone)
    put_vnode (StatePntr->diskVolumePntr->nameSpaceID,
    (unsigned int) StatePntr->sourceDirectoryPntr);
  RDWorkCleanup (StatePntr);
  return ReturnCode;
}



/******************************************************************************
 * Doea the RenameOrDelete operation, and if it fails due to deadlock, delays a
 * while and then tries again.
 */

static status_t RenameOrDeleteWithRetry (
  RDWorkStatePointer StatePntr)
{
  status_t          ErrorCode;
  int               RetryCount;
  unsigned int      SavedRandomState = 0;
  RDWorkStateRecord StateCopy;

  RetryCount = 1;
  do
  {
    if (RetryCount != 1)
      dprintf ("ARFS:RenameOrDeleteWithRetry: Trying again, "
        "attempt #%d of 5.\n", RetryCount);

    StateCopy = *StatePntr;
    ErrorCode = RenameOrDelete (&StateCopy);
    if (ErrorCode != EDEADLK)
      return ErrorCode;

    RandomDelay (&SavedRandomState);

  } while (++RetryCount <= 5);

  return EDEADLK;
}



/******************************************************************************
 * This structure is used for passing arguments and returned results between
 * various directory listing functions and the tree traversal iterator they
 * use.
 */

typedef enum TreeToDirEntKindEnum
{
  TTDEK_REGULAR_FSO = 0,
  TTDEK_FAKE_SYMBOLIC_LINK_TO_FSO,
  TTDEK_ATTRIBUTE_NAME,
  TTDEK_INDICES_ONLY,
  TTDEK_QUERY,
  TTDEK_MAX
} TreeToDirEntKind;


typedef struct TreeToDirEntArgumentsStruct
{
  DiskVolumePointer diskVolumePntr;
    /* Useful for many purposes. */

  FileSystemObjectPointer parentFSOPntr;
    /* The file system object which will be given as the parent directory for
    the items listed.  Not necessarily a directory - could be anything if we
    are listing attributes.  For queries, the parent is the object's actual
    parent, so it varies from dirent to dirent. */

  TreeToDirEntKind kindOfDirEnt;
    /* Which variation of the DirEnt should be created.  Is it going to make a
    DirEnt for a regular file, a DirEnt for an attribute name or should the
    DirEnt reference a fake symbolic link which references the actual thing? */

  status_t returnCode;
    /* The error code to return to the user if something goes wrong.  The
    actual traversal routines just return true or false, so we need this extra
    bit of info.  Initialised to zero. */

  ParseNodePointer parseTreePntr;
    /* Points to the information needed to do the query tests, for queries. */

  type_code keyType;
    /* What kind of thing the key is.  Needed so that it can be converted to a
    readable string (should be the same kind as the keys in the tree, if there
    is a tree).  Usually B_STRING_TYPE. */

  AVLDupThingConstPointer lastKeyReturnedPntr;
    /* Set to point to the key (actually in one of the tree nodes) that
    corresponds to the last entry added to the list.  Note that the key could
    get deallocated if the tree is changed, making the pointer invalid. */

  unsigned int lastValueReturned;
    /* The value (FSO pointer) involved with the last directory entry returned
    to the user.  Used along with lastKeyReturnedPntr to find the place where
    the listing is resumed.  Note that it is the actual value from the tree,
    not the value the user sees (the user's value may be a modified value with
    flag bits set to make it a fake symbolic link). */

  unsigned int numberOfItemsRemaining;
    /* Current number of items to add to the buffer, often initialised to 1 by
    callers which can't handle multiple dentries.  Decremented as items are
    added. */

  unsigned int numberOfItemsListed;
    /* Number of items actually put into the buffer.  If zero, then no items
    were found or the buffer could be too small - see the returnCode field to
    determine the cause.  Initialise to zero. */

  uint8 *bufferPntr;
    /* Points to the current position being filled in the output buffer,
    initialise to the start of the buffer. */

  unsigned int bufferSizeRemaining;
    /* Number of bytes remaining in the buffer.  Initialise it to the buffer
    size, decreased as items are added. */

} TreeToDirEntArgumentsRecord, *TreeToDirEntArgumentsPointer;



/******************************************************************************
 * This callback is used when iterating through an AVLDupTree to generate
 * dirent records for each item found in the tree.  The name placed in the
 * dirent is the key in the tree (converted to a string, since the key could be
 * an integer or float as well as strings).  The value is usually a pointer to
 * a file system object, but can also be a pointer to attribute data or an
 * index's fake symbolic link.  May also be called manually to add one item at
 * a time.  ExtraData should be a TreeToDentryArgumentsPointer.  Returns TRUE
 * if it successfully added a dirent.  Returns FALSE and sets the returnCode if
 * something went wrong (usually because output buffer is full).
 */

static bool AddDirEnt (
  AVLDupThingConstPointer KeyPntr,
  const unsigned int *ValuePntr,
  void *ExtraData)
{
  TreeToDirEntArgumentsPointer ArgsPntr = ExtraData;
  bool                         ContinueIterating = true;
  dirent_t                    *DirEntPntr;
  status_t                     ErrorCode = -1;
  FileSystemObjectPointer      FSOPntr = (FileSystemObjectPointer) *ValuePntr;
  bool                         GetVnodeDone = false;
  unsigned int                 NameLength;
  char                         NameString [NAME_MAX];
  bool                         ReadLockDone = false;
  unsigned int                 RecordLength;
  uint8                        TruncatedKeyword [NAME_MAX];
  PremadeAttributeRecord       VirtualAttribute;
  void                        *VoidPntr;

  if (KeyPntr == NULL || ValuePntr == NULL || ExtraData == NULL)
  {
    dprintf ("ARFS:AddDirEnt: Bug!  Bad arguments.\n");
    return false;
  }

  if (ArgsPntr->numberOfItemsRemaining <= 0)
    return false; /* No work to do, output quota reached, halt iteration. */

  if (ArgsPntr->kindOfDirEnt == TTDEK_INDICES_ONLY)
  {
    if (INVALID_FILE_SYSTEM_OBJECT (FSOPntr) ||
    FSOPntr->objectType != FSO_TYPE_INDEX)
      goto ErrorExit; /* When listing indices, skip non-index things. */
  }
  else if (ArgsPntr->kindOfDirEnt == TTDEK_QUERY)
  {
    /* It's a query listing.  Need to read-lock the object and then compare its
    attributes with the parse tree. */

    if (INVALID_FILE_SYSTEM_OBJECT (FSOPntr))
    {
      dprintf ("ARFS:AddDirEnt: Bug!  Skipping corrupt value in %s.\n",
        FSOAsString (ArgsPntr->parentFSOPntr));
      goto ErrorExit;
    }

    ErrorCode = get_vnode (ArgsPntr->diskVolumePntr->nameSpaceID,
      (unsigned int) FSOPntr, &VoidPntr);
    if (ErrorCode != 0)
    {
      dprintf ("ARFS:AddDirEnt: get_vnode failed while trying to examine "
        "%s for a query, code %d.  Skipping it.\n", FSOAsString (FSOPntr),
        (int) ErrorCode);
      goto ErrorExit;
    }
    GetVnodeDone = true;

    /* To avoid deadlocks with other threads writelocking objects we are
    examining and then trying to use the index, the readlock is tenatively
    attempted, and if it fails, the query unlocks everything and pauses for a
    while. */

    ErrorCode = ReadLock (FSOPntr, true /* Tentative */);
    if (ErrorCode != 0)
    {
      if (ErrorCode == EDEADLK)
        ContinueIterating = false;
      else /* A real error.  Ignore this file and continue on. */
        dprintf ("ARFS:AddDirEnt: ReadLock failed while trying to examine "
          "%s for a query, code %d.  Skipping it.\n", FSOAsString (FSOPntr),
          (int) ErrorCode);
      goto ErrorExit;
    }
    ReadLockDone = true;

    if (!DoesFileMatchParseTree (FSOPntr, ArgsPntr->parseTreePntr,
    NULL, NULL, TruncatedKeyword, &VirtualAttribute))
      goto ErrorExit; /* Skip files that don't match the query. */
  }

  /* Get the name of the item from the key (except for queries which use the
  file's true name).  If the key is a number or date, convert into a string.
  May need to convert '/' into some other character to avoid invalid names, but
  that's one for the todo list. */

  if (ArgsPntr->kindOfDirEnt == TTDEK_QUERY)
    strcpy (NameString, FSOPntr->trueNamePntr);
  else if (!AVLDupConvertThingToString (KeyPntr, ArgsPntr->keyType,
  NameString, sizeof (NameString)))
  {
    dprintf ("ARFS:AddDirEnt: Bug!  Unable to convert key to a string for "
      "the thing name.\n");
    ErrorCode = -1;
    ContinueIterating = false;
    goto ErrorExit;
  }

  /* Index entries are mangled to make them unique by appending the VNodeID of
  the object to the name, as a hex string.  Otherwise several files with the
  same file size, or name, or date, etc, would all have the same name, which
  confuses old programs. */

  if (ArgsPntr->kindOfDirEnt == TTDEK_FAKE_SYMBOLIC_LINK_TO_FSO)
  {
    NameLength = strlen (NameString);
    if (NameLength + 10 < NAME_MAX)
      sprintf (NameString + NameLength, " #%08X", (unsigned int) FSOPntr);
  }

  /* Find space for the dirent in the output buffer. */

  NameLength = strlen (NameString);
  RecordLength = sizeof (dirent_t) + NameLength; /* Total record size. */
  if (RecordLength > ArgsPntr->bufferSizeRemaining)
  {
    /* This name won't fit in the buffer.  Either the buffer is full or the
    user gave us a buffer which was way too small. */

    if (ArgsPntr->numberOfItemsListed == 0)
    {
      dprintf ("ARFS:AddDirEnt: Buffer for dirent is too small, while "
        "preparing \"%s\".  Got %d bytes of data, but buffer is %d bytes "
        "big.\n", NameString, RecordLength, ArgsPntr->bufferSizeRemaining);
      ErrorCode = EINVAL;
      ContinueIterating = false;
      goto ErrorExit;
    }

    /* Buffer has stuff in it, and is merely full.  This also means not
    updating the lastKeyReturnedPntr for the current key since we want to show
    that one to the user next time. */

    ErrorCode = 0;
    ContinueIterating = false;
    goto ErrorExit;
  }

  /* Write the DirEnt contents to the buffer. */

  DirEntPntr = (dirent_t *) ArgsPntr->bufferPntr;

  /* Set the parent directory reference. */

  DirEntPntr->d_pdev = ArgsPntr->diskVolumePntr->nameSpaceID;
  if (ArgsPntr->kindOfDirEnt != TTDEK_QUERY)
    DirEntPntr->d_pino = (unsigned int) ArgsPntr->parentFSOPntr;
  else /* Queries use the file's actual parent directory for the parent. */
    DirEntPntr->d_pino = (unsigned int) FSOPntr->mainParentDirectory;

  /* Set the file reference. */

  if (ArgsPntr->kindOfDirEnt == TTDEK_FAKE_SYMBOLIC_LINK_TO_FSO)
  {
    /* Make the vnode number odd to turn it into a fake symbolic link to the
    real thing.  The low 3 bits are a special code, 0 means a normal item, 1
    means a fake symbolic link. */
    DirEntPntr->d_dev = ArgsPntr->diskVolumePntr->nameSpaceID;
    DirEntPntr->d_ino = 1 + *ValuePntr;
  }
  else if (ArgsPntr->kindOfDirEnt == TTDEK_ATTRIBUTE_NAME)
  {
    /* Attribute dirents just use the name, the rest isn't used. */
    DirEntPntr->d_dev = 0;
    DirEntPntr->d_ino = 0;
  }
  else /* A DirEnt which just refers to a regular file system object. */
  {
    DirEntPntr->d_dev = ArgsPntr->diskVolumePntr->nameSpaceID;
    DirEntPntr->d_ino = *ValuePntr;
  }
  strcpy (DirEntPntr->d_name, NameString);
  DirEntPntr->d_reclen = RecordLength;

#if DPRINT_NON_ERROR_MESSAGES
  dprintf ("ARFS:AddDirEnt: Adding dirent for vnode $%016LX, "
    "name \"%s\", actual %s, d_reclen:%d.\n",
    DirEntPntr->d_ino, DirEntPntr->d_name,
    FSOAsString ((FileSystemObjectPointer) (unsigned int) DirEntPntr->d_ino),
    (int) DirEntPntr->d_reclen);
#endif

  ArgsPntr->numberOfItemsListed++;
  ArgsPntr->numberOfItemsRemaining--;
  ArgsPntr->bufferSizeRemaining -= RecordLength;
  ArgsPntr->bufferPntr += RecordLength;
  ArgsPntr->lastKeyReturnedPntr = KeyPntr;
  ArgsPntr->lastValueReturned = *ValuePntr;

ErrorExit:
  if (ReadLockDone)
    ReadUnlock (FSOPntr);
  if (GetVnodeDone)
    put_vnode (ArgsPntr->diskVolumePntr->nameSpaceID, (unsigned int) FSOPntr);
  if (!ContinueIterating) /* If halting, save the error code for the caller. */
    ArgsPntr->returnCode = ErrorCode;
  return ContinueIterating;
}



/******************************************************************************
 * Start listing the contents of a directory or other directory-like thing
 * (DirectoryVariation tells you which one it is).  After verifying that it
 * exists and is the right kind of file system object, a cookie is allocated
 * and initialised to keep track of the progress of listing the items in the
 * thing.  See ARFS_opendir for more documentation.
 */

status_t GenericOpenDir (
  void *ns,
  void *node,
  void **cookie,
  GenericDirVariations DirectoryVariation,
  QueryStatePointer QueryPntr,
  AVLIterationRangePointer IterationRangePntr)
{
  GenericDirCookiePointer CookiePntr;
  DiskVolumePointer       DiskVolumePntr = ns;
  status_t                ErrorCode;
  FileSystemObjectPointer FSOPntr = node;
  bool                    ReadLockDone = false;
  status_t                ReturnCode = -1;
  GenericDirCookieRecord  TempCookie;
  char                    TempString [B_OS_NAME_LENGTH];

  memset (&TempCookie, 0, sizeof (TempCookie));
  TempCookie.mutexSemID = -1; /* In case an early error exit happens. */

#if 0
  dprintf ("ARFS:GenericOpenDir entered, NS:$%X, node:$%X, "
    "cookie:$%X, variation:%d.\n",
    (unsigned int) ns, (unsigned int) node, (unsigned int) cookie,
    (int) DirectoryVariation);
#endif

  if (cookie == NULL)
  {
    dprintf ("ARFS:GenericOpenDir: NULL cookie pointer to pointer.\n");
    goto ErrorExit;
  }

  if (INVALID_DISK_VOLUME (DiskVolumePntr))
  {
    dprintf ("ARFS:GenericOpenDir: Invalid input disk volume pointer $%X.\n",
      (int) DiskVolumePntr);
    goto ErrorExit;
  }

  if (INVALID_FILE_SYSTEM_OBJECT (FSOPntr))
  {
    dprintf ("ARFS:GenericOpenDir: Invalid input FSO pointer $%X.\n",
      (unsigned int) FSOPntr);
    goto ErrorExit;
  }

  /* Get a read lock on the directory / thing, so that permissions can be
  safely read.  Not really needed. */

  ErrorCode = ReadLock (FSOPntr, false);
  if (ErrorCode != 0)
  {
    dprintf ("ARFS:GenericOpenDir: unable to get read lock for %s.\n",
      FSOAsString (FSOPntr));
    ReturnCode = ErrorCode;
    goto ErrorExit;
  }
  ReadLockDone = true;

  /* Fail if the user isn't allowed to read this thing. */

  ErrorCode = PermissionAllowed (DiskVolumePntr, FSOPntr, R_OK);
  if (ErrorCode != 0)
  {
    dprintf ("ARFS:GenericOpenDir: %s doesn't have read access permission "
      "for contents (need the r flag).\n", FSOAsString (FSOPntr));
    ReturnCode = ErrorCode;
    goto ErrorExit;
  }

  /* Set up things in a cookie stored on the stack.  If things work out, it
  will be copied to some dynamically allocated memory and returned to the user
  as the cookie. */

  TempCookie.fsoPntr = FSOPntr;
  TempCookie.listingVariety = DirectoryVariation;
  TempCookie.queryPntr = QueryPntr;
  TempCookie.directoryOpen = true;
  if (IterationRangePntr == NULL)
  {
    /* Set up the range to include the whole tree. */
    TempCookie.iterationRange.ascendingOrder = true;
    TempCookie.iterationRange.infiniteLower = true;
    TempCookie.iterationRange.infiniteUpper = true;
  }
  else /* Range provided by caller. */
    TempCookie.iterationRange = *IterationRangePntr;

  /* Set up the mutual exclusion semaphore in the cookie. */

  strcpy (TempString, "DirCookie ");
  strncat (TempString, FSOPntr->trueNamePntr,
    B_OS_NAME_LENGTH - strlen (TempString) - 1);
  TempCookie.mutexSemID = create_sem (1 /* Initial count */, TempString);
  if (TempCookie.mutexSemID < 0)
  {
    dprintf ("ARFS:GenericOpenDir: Unable to create semaphore for cookie "
      "for %s, error code %d!\n",
      FSOAsString (FSOPntr), (int) TempCookie.mutexSemID);
    ReturnCode = TempCookie.mutexSemID;
    goto ErrorExit;
  }

  /* Verify that the object is the right kind for the type of thing we are
  listing.  Also set up the other fields in the cookie. */

  switch (DirectoryVariation)
  {
    case GDV_PLAIN_DIRECTORY:
      if (FSOPntr->objectType != FSO_TYPE_DIRECTORY &&
      FSOPntr->objectType != FSO_TYPE_INDEX)
      {
        dprintf ("ARFS:GenericOpenDir: %s isn't a "
          "directory or index object!\n", FSOAsString (FSOPntr));
        ReturnCode = ENOTDIR;
        goto ErrorExit;
      }
      if (FSOPntr->objectType == FSO_TYPE_INDEX)
        TempCookie.keyType = ((IndexObjectPointer) FSOPntr)->keyType;
      else /* Plain directory, keys are just the file names. */
        TempCookie.keyType = B_STRING_TYPE;
      break;

    case GDV_ATTRIBUTE_LISTING:
      TempCookie.nextParentDirIndex = -1;
      TempCookie.keyType = B_STRING_TYPE;
      break;

    case GDV_INDEX_CONTENTS:
      if (FSOPntr->objectType != FSO_TYPE_DIRECTORY)
      {
        dprintf ("ARFS:GenericOpenDir: %s (the indices directory) isn't a "
          "directory!\n", FSOAsString (FSOPntr));
        ReturnCode = ENOTDIR;
        goto ErrorExit;
      }
      TempCookie.nextParentDirIndex = -1;
      TempCookie.keyType = B_STRING_TYPE;
      break;

    case GDV_QUERY:
      if (FSOPntr->objectType != FSO_TYPE_INDEX)
      {
        dprintf ("ARFS:GenericOpenDir: %s (the main index in the query) isn't "
          "an index!\n", FSOAsString (FSOPntr));
        ReturnCode = ENOTDIR;
        goto ErrorExit;
      }
      TempCookie.nextParentDirIndex = -1;
      TempCookie.keyType = ((IndexObjectPointer) FSOPntr)->keyType;
      break;

    default:
      dprintf ("ARFS:GenericOpenDir: Bug!  Requested directory variation %d "
        "isn't a known one.  Was doing %s.\n",
        (int) DirectoryVariation, FSOAsString (FSOPntr));
      goto ErrorExit;
  }

  /* Allocate a new cookie to track the progress of listing the directory.  The
  user will get the pointer to the cookie as their handle to it. */

  CookiePntr = mallocDVP (DiskVolumePntr, sizeof (TempCookie));
  if (CookiePntr == NULL)
  {
    dprintf ("ARFS:GenericOpenDir: unable to create cookie when opening "
      "%s, out of memory.\n", FSOAsString (FSOPntr));
    ReturnCode = ENOMEM;
    goto ErrorExit;
  }
  memcpy (CookiePntr, &TempCookie, sizeof (TempCookie));
  *cookie = CookiePntr;

  TempCookie.mutexSemID = -1; /* Semaphore doesn't need deallocation now. */
  ReturnCode = 0;

#if DPRINT_NON_ERROR_MESSAGES
  dprintf ("ARFS:GenericOpenDir successful for %s, cookie:$%X (#%d:%s).\n",
    FSOAsString (FSOPntr), (unsigned int) CookiePntr,
    CookiePntr->nextParentDirIndex, CookiePntr->lastName);
#endif

ErrorExit:
#if DPRINT_NON_ERROR_MESSAGES
  dprintf ("ARFS:GenericOpenDir exited, return code %d.\n", (int) ReturnCode);
#endif
  if (TempCookie.mutexSemID >= 0)
    delete_sem (TempCookie.mutexSemID);
  if (ReadLockDone)
    ReadUnlock (FSOPntr);
  return ReturnCode;
}



/******************************************************************************
 * Closing the directory just stops any further threads from accessing the
 * directory.  Don't deallocate the cookie just yet, there could be other
 * threads busy reading at the moment.
 */

status_t GenericCloseDir (
  void *ns,
  void *node,
  void *cookie)
{
  bool                    CookieLockDone = false;
  GenericDirCookiePointer CookiePntr = cookie;
  DiskVolumePointer       DiskVolumePntr = ns;
  status_t                ErrorCode;
  FileSystemObjectPointer FSOPntr = node;
  status_t                ReturnCode = -1;

#if 0
  dprintf ("ARFS:GenericCloseDir entered, NS:$%X, node:$%X, cookie:$%X.\n",
    (unsigned int) ns, (unsigned int) node, (unsigned int) cookie);
#endif

  if (INVALID_DISK_VOLUME (DiskVolumePntr))
  {
    dprintf ("ARFS:GenericCloseDir: invalid input disk volume pointer $%X.\n",
      (unsigned int) DiskVolumePntr);
    goto ErrorExit;
  }

  if (INVALID_FILE_SYSTEM_OBJECT (FSOPntr))
  {
    dprintf ("ARFS:GenericCloseDir: invalid input FSO pointer $%X.\n",
      (unsigned int) FSOPntr);
    goto ErrorExit;
  }

  if (CookiePntr == NULL || CookiePntr->fsoPntr != FSOPntr)
  {
    dprintf ("ARFS:GenericCloseDir: %s doesn't match cookie!\n",
      FSOAsString (FSOPntr));
    ReturnCode = EINVAL;
    goto ErrorExit;
  }

  ErrorCode = acquire_sem_etc (CookiePntr->mutexSemID, 1 /* Count */,
    B_CAN_INTERRUPT /* flags */, 0 /* unused timeout value */);
  if (ErrorCode != 0)
  {
    dprintf ("ARFS:GenericCloseDir: failed to aquire cookie "
      "semaphore for %s, semaphoreID is %d, error %d.\n",
      FSOAsString (FSOPntr), (int) CookiePntr->mutexSemID, (int) ErrorCode);
    ReturnCode = ErrorCode;
    goto ErrorExit;
  }
  CookieLockDone = true;

  if (!CookiePntr->directoryOpen)
  {
    dprintf ("ARFS:GenericCloseDir: %s was closed earlier, no further "
      "operations allowed.\n", FSOAsString (FSOPntr));
    ReturnCode = EBADF;
    goto ErrorExit;
  }

  /* Finally, do the actual work.  No locking needed (not reading or writing
  the FSO, just the cookie and in a multitasking safe manner since it's so
  simple). */

  CookiePntr->directoryOpen = false;
  ReturnCode = 0;

#if DPRINT_NON_ERROR_MESSAGES
  dprintf ("ARFS:GenericCloseDir successful for %s, cookie:$%X (#%d:%s).\n",
    FSOAsString (FSOPntr), (unsigned int) CookiePntr,
    CookiePntr->nextParentDirIndex, CookiePntr->lastName);
#endif

ErrorExit:
#if DPRINT_NON_ERROR_MESSAGES
  dprintf ("ARFS:GenericCloseDir exited, return code %d.\n",
    (int) ReturnCode);
#endif
  if (CookieLockDone)
    release_sem (CookiePntr->mutexSemID);
  return ReturnCode;
}



/******************************************************************************
 * Sets the directory state back to the first thing in the directory, so the
 * user can re-read the directory from the beginning.  Even works for queries,
 * though the BeOS API doesn't have a way of accessing this functionality.
 */

status_t GenericRewindDir (
  void *ns,
  void *node,
  void *cookie)
{
  bool                    CookieLockDone = false;
  GenericDirCookiePointer CookiePntr = cookie;
  DiskVolumePointer       DiskVolumePntr = ns;
  status_t                ErrorCode;
  FileSystemObjectPointer FSOPntr = node;
  status_t                ReturnCode = -1;

#if 0
  dprintf ("ARFS:GenericRewindDir entered, NS:$%X, node:$%X, cookie:$%X.\n",
    (unsigned int) ns, (unsigned int) node, (unsigned int) cookie);
#endif

  if (INVALID_DISK_VOLUME (DiskVolumePntr))
  {
    dprintf ("ARFS:GenericRewindDir: invalid input disk volume pointer $%X.\n",
      (int) DiskVolumePntr);
    goto ErrorExit;
  }

  if (INVALID_FILE_SYSTEM_OBJECT (FSOPntr))
  {
    dprintf ("ARFS:GenericRewindDir: invalid input FSO pointer $%X.\n",
      (unsigned int) FSOPntr);
    goto ErrorExit;
  }

  if (CookiePntr == NULL || CookiePntr->fsoPntr != FSOPntr)
  {
    dprintf ("ARFS:GenericRewindDir: %s doesn't match cookie!\n",
      FSOAsString (FSOPntr));
    ReturnCode = EINVAL;
    goto ErrorExit;
  }

  ErrorCode = acquire_sem_etc (CookiePntr->mutexSemID, 1 /* Count */,
    B_CAN_INTERRUPT /* flags */, 0 /* unused timeout value */);
  if (ErrorCode != 0)
  {
    dprintf ("ARFS:GenericRewindDir: failed to aquire cookie "
      "semaphore for %s, semaphoreID is %d, error %d.\n",
      FSOAsString (FSOPntr), (int) CookiePntr->mutexSemID, (int) ErrorCode);
    ReturnCode = ErrorCode;
    goto ErrorExit;
  }
  CookieLockDone = true;

  if (!CookiePntr->directoryOpen)
  {
    dprintf ("ARFS:GenericRewindDir: %s was closed earlier, no further "
      "operations allowed.\n", FSOAsString (FSOPntr));
    ReturnCode = EBADF;
    goto ErrorExit;
  }

  /* Finally, do the actual work. */

  if (CookiePntr->listingVariety == GDV_PLAIN_DIRECTORY)
    CookiePntr->nextParentDirIndex = 0;
  else /* All others don't start by listing the parent directories. */
    CookiePntr->nextParentDirIndex = -1;
  strcpy (CookiePntr->lastName, "Rewound...");
  ReturnCode = 0;

#if DPRINT_NON_ERROR_MESSAGES
  dprintf ("ARFS:GenericRewindDir successful for %s, cookie:$%X (#%d:%s).\n",
    FSOAsString (FSOPntr), (unsigned int) CookiePntr,
    CookiePntr->nextParentDirIndex, CookiePntr->lastName);
#endif

ErrorExit:
#if DPRINT_NON_ERROR_MESSAGES
  dprintf ("ARFS:GenericRewindDir exited, return code %d.\n",
    (int) ReturnCode);
#endif
  if (CookieLockDone)
    release_sem (CookiePntr->mutexSemID);
  return ReturnCode;
}



/******************************************************************************
 * Do the actual listing of a directory or thing.  Adds one or more directory
 * entries to the supplied output buffer and updates the cookie state so that
 * the next call to this function will resume where the previous one left off.
 */

status_t GenericReadDir (
  void *ns,
  void *node,
  void *cookie,
  long *num,
  struct dirent *buf,
  size_t bufsize)
{
  bool                        CookieLockDone = false;
  GenericDirCookiePointer     CookiePntr = cookie;
  TreeToDirEntArgumentsRecord DirEntArgs;
  DiskVolumePointer           DiskVolumePntr = ns;
  status_t                    ErrorCode;
  FileSystemObjectPointer     FSOPntr = node;
  AGMSLinkObjectPointer      *ExtraLinkPntrPntr;
  AVLDupThingRecord           Key;
  const char                 *NamePntr;
  bool                        ReadLockDone = false;
  int                         RemainingIndex;
  status_t                    ReturnCode = -1;
  unsigned int                TempUINT;
  AVLDupTreePointer           TreePntr;
  vnode_id                    VNodeID;

#if 0
  dprintf ("ARFS:GenericReadDir entered, "
    "NS:$%X, node:$%X, cookie:$%X, *num:%d, buf:$%X, bufsize:%d.\n",
    (unsigned int) ns, (unsigned int) node, (unsigned int) cookie,
    (int) *num, (unsigned int) buf, (int) bufsize);
#endif

  if (buf == NULL || num == NULL || bufsize < sizeof (dirent_t) || *num <= 0)
  {
    dprintf ("ARFS:GenericReadDir: NULL pointers or other "
      "silly input values.\n");
    goto ErrorExit;
  }

  if (INVALID_DISK_VOLUME (DiskVolumePntr))
  {
    dprintf ("ARFS:GenericReadDir: invalid input disk volume pointer $%X.\n",
      (int) DiskVolumePntr);
    goto ErrorExit;
  }

  if (INVALID_FILE_SYSTEM_OBJECT (FSOPntr))
  {
    dprintf ("ARFS:GenericReadDir: invalid input FSO pointer $%X.\n",
      (unsigned int) FSOPntr);
    goto ErrorExit;
  }

  if (CookiePntr == NULL || CookiePntr->fsoPntr != FSOPntr)
  {
    dprintf ("ARFS:GenericReadDir: %s doesn't match cookie!\n",
      FSOAsString (FSOPntr));
    ReturnCode = EINVAL;
    goto ErrorExit;
  }

  ErrorCode = acquire_sem_etc (CookiePntr->mutexSemID, 1 /* Count */,
    B_CAN_INTERRUPT /* flags */, 0 /* unused timeout value */);
  if (ErrorCode != 0)
  {
    dprintf ("ARFS:GenericReadDir: failed to aquire cookie "
      "semaphore for %s, semaphoreID is %d, error %d.\n",
      FSOAsString (FSOPntr), (int) CookiePntr->mutexSemID, (int) ErrorCode);
    ReturnCode = ErrorCode;
    goto ErrorExit;
  }
  CookieLockDone = true;

  if (!CookiePntr->directoryOpen)
  {
    dprintf ("ARFS:GenericReadDir: %s was closed earlier, no further "
      "operations allowed.\n", FSOAsString (FSOPntr));
    ReturnCode = EBADF;
    goto ErrorExit;
  }

  ErrorCode = ReadLock (FSOPntr, false);
  if (ErrorCode != 0)
  {
    dprintf ("ARFS:GenericReadDir: unable to get read lock for %s.\n",
      FSOAsString (FSOPntr));
    ReturnCode = ErrorCode;
    goto ErrorExit;
  }
  ReadLockDone = true;

  memset (&DirEntArgs, 0, sizeof (DirEntArgs));
  DirEntArgs.diskVolumePntr = DiskVolumePntr;
  DirEntArgs.parentFSOPntr = FSOPntr;
  DirEntArgs.numberOfItemsRemaining = *num;
  DirEntArgs.bufferPntr = (uint8 *) buf;
  DirEntArgs.bufferSizeRemaining = bufsize;

  /* This while loop generates directory entries for the ".", "..", "..." and
  so on entries, corresponding to non-negative values of nextParentDirIndex. */

  DirEntArgs.keyType = B_STRING_TYPE;
  DirEntArgs.kindOfDirEnt = TTDEK_REGULAR_FSO;

  while (CookiePntr->nextParentDirIndex >= 0)
  {
    VNodeID = 0; /* Will be set to parent directory or thing, or left zero. */

    /* Generate the name for the next parent directory, "." for index 0, ".."
    for index 1, "..." for index 2 and so on. */

    NamePntr = g_MaxDotName +
      (NAME_MAX - 2 - CookiePntr->nextParentDirIndex);

    if (CookiePntr->nextParentDirIndex == 0) /* "." */
      VNodeID = (unsigned int) FSOPntr;
    else if (CookiePntr->nextParentDirIndex == 1) /* ".." */
    {
      if (FSOPntr == &DiskVolumePntr->rootFSOPntr->fso)
        /* Dummy .. entry for the file system root, BeOS might replace it
        with the actual mount point parent directory, which is in some
        other file system.  Internally NULL is its parent, which we can't
        use here, so just use itself as the root's parent. */
        VNodeID = (unsigned int) FSOPntr;
      else
        VNodeID = (unsigned int) FSOPntr->mainParentDirectory;
    }
    else if (CookiePntr->nextParentDirIndex < NAME_MAX - 2)
    { /* Count 2 to 253 are one of the AGMS link reference directories. */
      ExtraLinkPntrPntr = FSOPntr->parentLinkArray;
      if (ExtraLinkPntrPntr != NULL)
      {
        RemainingIndex = CookiePntr->nextParentDirIndex - 2;
        while (RemainingIndex > 0 && *ExtraLinkPntrPntr != NULL)
        {
          ExtraLinkPntrPntr++;
          RemainingIndex--;
        }
        if (*ExtraLinkPntrPntr != NULL)
        {
          /* Find the parent directory of the AGMS link, and return a
          fake symbolic link reference to it. */

          if (VALID_FILE_SYSTEM_OBJECT (&(*ExtraLinkPntrPntr)->fso))
            VNodeID = (unsigned int)
              (*ExtraLinkPntrPntr)->fso.mainParentDirectory;
          else /* Bad AGMS link.  Point it at the root directory. */
            VNodeID = (unsigned int) DiskVolumePntr->rootFSOPntr;

          /* Make it odd to turn it into a fake symbolic link.  The low 3 bits
          are a special code, 0 means a normal item, 1 means a fake symbolic
          link. */

          VNodeID += 1;
        }
      }
    }
    else /* nextParentDirIndex 254 and larger. */
    {
      /* Counts of 254 and higher refer to parent directories with dot names
      that are too long to fit in the 256 byte maximum name length.  Since they
      are rare, it's too much bother to list them. */
    }

    if (VNodeID == 0)
    {
      /* No more parents, switch to listing the files in the directory. */
      CookiePntr->nextParentDirIndex = -1; /* This also stops the loop. */
    }
    else /* Found a parent directory, attempt to add a DirEnt for it. */
    {
      Key.longStringThing.stringPntr = (char *) NamePntr;
      Key.longStringThing.isLongString = true;
      TempUINT = VNodeID;
      if (!AddDirEnt (&Key, &TempUINT, &DirEntArgs))
        goto NormalExit; /* Unable to add it (may be an error or maybe not). */
      CookiePntr->nextParentDirIndex++; /* Done, advance to next. */
    }
  }

  /* Switch to generating directory entries for the tree of things in the
  directory / index / thing.  These can have a different key type (not
  necessarily a string) and the values listed may be listed differently than
  the "..." entries (as fake symbolic links etc). */

  DirEntArgs.keyType = CookiePntr->keyType;

  if (CookiePntr->listingVariety == GDV_PLAIN_DIRECTORY)
  {
    if (FSOPntr->objectType == FSO_TYPE_INDEX)
    {
      TreePntr = ((IndexObjectPointer) FSOPntr)->treePntr;
      DirEntArgs.kindOfDirEnt = TTDEK_FAKE_SYMBOLIC_LINK_TO_FSO;
    }
    else
      TreePntr = ((DirectoryObjectPointer) FSOPntr)->directory;
  }
  else if (CookiePntr->listingVariety == GDV_ATTRIBUTE_LISTING)
  {
    TreePntr = FSOPntr->attributesTree; /* Can safely be NULL. */
    DirEntArgs.kindOfDirEnt = TTDEK_ATTRIBUTE_NAME;
  }
  else if (CookiePntr->listingVariety == GDV_INDEX_CONTENTS)
  {
    TreePntr = ((DirectoryObjectPointer) FSOPntr)->directory;
    DirEntArgs.kindOfDirEnt = TTDEK_INDICES_ONLY;
  }
  else if (CookiePntr->listingVariety == GDV_QUERY)
  {
    TreePntr = ((IndexObjectPointer) FSOPntr)->treePntr;
    DirEntArgs.kindOfDirEnt = TTDEK_QUERY;
    DirEntArgs.parseTreePntr = CookiePntr->queryPntr->parseTreePntr;
  }
  else
  {
    dprintf ("ARFS:GenericReadDir: Bug!  Unknown directory listing "
      "variation %d for %s.\n",
      (int) CookiePntr->listingVariety, FSOAsString (FSOPntr));
    goto ErrorExit;
  }

  /* All parent directories have been done, now append dirents for the files
  listed in the directory (or whatever the things are inside whatever the
  container is).  The tree traversal starts just after the last name used and
  continues on in ascending order until an error happens or until the buffer is
  full or the requested number of dirents have been written or there are no
  more left in the tree. */

  DirEntArgs.lastKeyReturnedPntr = NULL;

  if (CookiePntr->nextParentDirIndex >= -1)
  {
    /* Start a new iteration at the beginning, whatever direction it is in. */

    AVLDupIterate (TreePntr,
      CookiePntr->iterationRange.ascendingOrder,
      CookiePntr->iterationRange.infiniteLower
        ? NULL /* Start the iteration at the beginning. */
        : &CookiePntr->iterationRange.lowerKey /* Start at or after a key */,
      CookiePntr->iterationRange.includeThingEqualToLower ? -1 : +1,
      0 /* lower value compared with this, not used here */,
      false /* include equal to key/value irrelevant, value is +-infinity */,
      CookiePntr->iterationRange.infiniteUpper
        ? NULL /* End the iteration at the end. */
        : &CookiePntr->iterationRange.upperKey /* End at or before a key */,
      CookiePntr->iterationRange.includeThingEqualToUpper ? +1 : -1,
      0 /* upper value compared with this, not used here */,
      false /* include equal to key/value irrelevant, value is +-infinity */,
      AddDirEnt /* callback function to use */,
      &DirEntArgs /* passed as extradata to the callback */);
  }
  else /* Continue an existing iteration. */
  {
    if (CookiePntr->iterationRange.ascendingOrder)
    {
      /* Ascending order.  Change the lower bound to the point where it should
      restart the iteration.  The upper bound is as before. */

      AVLDupIterate (TreePntr,
        CookiePntr->iterationRange.ascendingOrder,
        &CookiePntr->lastKey /* resume iteration just after the last key */,
        0 /* zero means also compare lower value to given value */,
        CookiePntr->lastValue /* lower value compared with this */,
        false /* true to include things equal to key/value */,
        CookiePntr->iterationRange.infiniteUpper
          ? NULL /* End the iteration at the end. */
          : &CookiePntr->iterationRange.upperKey /* End at or before a key */,
        CookiePntr->iterationRange.includeThingEqualToUpper ? +1 : -1,
        0 /* upper value compared with this, not used here */,
        false /* include equal to key/value irrelevant, value is +-infinity */,
        AddDirEnt /* callback function to use */,
        &DirEntArgs /* passed as extradata to the callback */);
    }
    else /* Descending order, starts at upper bound, lower bound unchanged. */
    {
      AVLDupIterate (TreePntr,
        CookiePntr->iterationRange.ascendingOrder,
        CookiePntr->iterationRange.infiniteLower
          ? NULL /* End the iteration at the beginning. */
          : &CookiePntr->iterationRange.lowerKey /* End at or after a key */,
        CookiePntr->iterationRange.includeThingEqualToLower ? -1 : +1,
        0 /* lower value compared with this, not used here */,
        false /* include equal to key/value irrelevant, value is +-infinity */,
        &CookiePntr->lastKey /* resume iteration just after the last key */,
        0 /* zero means also compare upper value to given value */,
        CookiePntr->lastValue /* upper value compared with this */,
        false /* true to include things equal to key/value */,
        AddDirEnt /* callback function to use */,
        &DirEntArgs /* passed as extradata to the callback */);
    }
  }

  if (DirEntArgs.lastKeyReturnedPntr != NULL)
  {
    /* Set nextParentDirIndex so that next time the iteration will resume just
    after lastKey/lastValue. */

    CookiePntr->nextParentDirIndex = -2;

    /* Copy the last key returned to our cookie.  Can't just save a pointer to
    it since the tree node and key could be deleted by some other thread.  To
    cut down on memory allocations from using the generic thing copying
    routines to copy strings, copy a string type key to the string buffer in
    the cookie.  If the keys aren't strings, then a printable version of the
    key is used in lastName just to make debugging easier. */

    AVLDupConvertThingToString (DirEntArgs.lastKeyReturnedPntr,
      CookiePntr->keyType, CookiePntr->lastName,
      sizeof (CookiePntr->lastName));

    if (CookiePntr->keyType == B_STRING_TYPE ||
    CookiePntr->keyType == B_MIME_STRING_TYPE)
    {
      CookiePntr->lastKey.longStringThing.stringPntr = CookiePntr->lastName;
      CookiePntr->lastKey.longStringThing.isLongString = true;
    }
    else /* Not a string thing, copy the key thing generically. */
    {
      /* Free the old last key.  Safe to use on the initial zero key. */

      AVLDupFreeThingArray (&CookiePntr->lastKey, CookiePntr->keyType,
        1 /* array size */, DiskVolumePntr->freePntr,
        DiskVolumePntr->volumePoolID);

      /* Copy the new last key into the cookie. */

      if (!AVLDupCopyThingArray (&CookiePntr->lastKey,
      DirEntArgs.lastKeyReturnedPntr, CookiePntr->keyType, 1 /* array size */,
      DiskVolumePntr->mallocPntr, DiskVolumePntr->freePntr,
      DiskVolumePntr->volumePoolID))
      {
        dprintf ("ARFS:GenericReadDir: Ran out of memory while trying to "
          "copy a key for %s, type %s, cookie:$%X (#%d:%s)\n",
          FSOAsString (FSOPntr), TypeCodeAsString (CookiePntr->keyType),
          (unsigned int) CookiePntr, CookiePntr->nextParentDirIndex,
          CookiePntr->lastName);
        ReturnCode = ENOMEM;
        goto ErrorExit;
      }
    }

    CookiePntr->lastValue = DirEntArgs.lastValueReturned;
  }

NormalExit:
  ChangeFSODates (DiskVolumePntr, FSOPntr, DK_ACCESS_FLAG, -1);
  *num = DirEntArgs.numberOfItemsListed;
  ReturnCode = DirEntArgs.returnCode;

#if DPRINT_NON_ERROR_MESSAGES
  dprintf ("ARFS:GenericReadDir successful for %s, cookie:$%X (#%d:%s), "
    "number of entries returned:%ld.\n",
    FSOAsString (FSOPntr), (unsigned int) CookiePntr,
    CookiePntr->nextParentDirIndex, CookiePntr->lastName, *num);
#endif

ErrorExit:
#if DPRINT_NON_ERROR_MESSAGES
  dprintf ("ARFS:GenericReadDir exited, return code %d.\n", (int) ReturnCode);
#endif
  if (ReadLockDone)
    ReadUnlock (FSOPntr);
  if (CookieLockDone)
    release_sem (CookiePntr->mutexSemID);
  return ReturnCode;
}



/******************************************************************************
 * The last thread has finished reading the directory, you can now really
 * close it.  Deallocate the cookie and otherwise clean up here.
 */

status_t GenericFreeDirCookie (
  void *ns,
  void *node,
  void *cookie)
{
  GenericDirCookiePointer CookiePntr = cookie;
  DiskVolumePointer       DiskVolumePntr = ns;
  status_t                ErrorCode;
  FileSystemObjectPointer FSOPntr = node;
  status_t                ReturnCode = -1;

#if 0
  dprintf ("ARFS:GenericFreeDirCookie entered, "
    "NS:$%X, node:$%X, cookie:$%X.\n",
    (unsigned int) ns, (unsigned int) node, (unsigned int) cookie);
#endif

  if (INVALID_DISK_VOLUME (DiskVolumePntr))
  {
    dprintf ("ARFS:GenericFreeDirCookie: invalid input disk volume "
      "pointer $%X.\n", (int) DiskVolumePntr);
    goto ErrorExit;
  }

  if (INVALID_FILE_SYSTEM_OBJECT (FSOPntr))
  {
    dprintf ("ARFS:GenericFreeDirCookie: invalid input FSO pointer $%X.\n",
      (unsigned int) FSOPntr);
    goto ErrorExit;
  }

  if (CookiePntr == NULL || CookiePntr->fsoPntr != FSOPntr)
  {
    dprintf ("ARFS:GenericFreeDirCookie: %s doesn't match cookie!\n",
      FSOAsString (FSOPntr));
    ReturnCode = EINVAL;
    goto ErrorExit;
  }

  if (CookiePntr->directoryOpen)
  {
    /* Just issue a warning, and continue to free the cookie. */
    dprintf ("ARFS:GenericFreeDirCookie: Bug!  %s was not closed!\n",
      FSOAsString (FSOPntr));
  }

#if DPRINT_NON_ERROR_MESSAGES
  dprintf ("ARFS:GenericFreeDirCookie working on %s, cookie:$%X (#%d:%s).\n",
    FSOAsString (FSOPntr), (unsigned int) CookiePntr,
    CookiePntr->nextParentDirIndex, CookiePntr->lastName);
#endif

  /* Finally, deallocate the semaphore, key and cookie itself. */

  if (CookiePntr->mutexSemID >= 0)
  {
    ErrorCode = delete_sem (CookiePntr->mutexSemID);
    if (ErrorCode < 0)
    {
      dprintf ("ARFS:GenericFreeDirCookie: Unable to delete semaphore %d "
        "in cookie for %s, code %d!\n",
        (int) CookiePntr->mutexSemID, FSOAsString (FSOPntr), (int) ErrorCode);
      /* Don't really care if it fails so no error returned to user. */
    }
    CookiePntr->mutexSemID = -1;
  }

  /* As a special case for listing the indices directory, the outstanding
  get_vnode needs to be undone. */

  if (CookiePntr->listingVariety == GDV_INDEX_CONTENTS)
    put_vnode (DiskVolumePntr->nameSpaceID,
    (unsigned int) CookiePntr->fsoPntr);

  /* Another special case is for queries.  Need to deallocate the query data
  structures and undo the get_vnode of the index. */

  if (CookiePntr->queryPntr != NULL)
  {
    AVLDupFreeThingArray (&CookiePntr->iterationRange.lowerKey,
      CookiePntr->keyType, 1 /* how many */,
      DiskVolumePntr->freePntr, DiskVolumePntr->volumePoolID);
    AVLDupFreeThingArray (&CookiePntr->iterationRange.upperKey,
      CookiePntr->keyType, 1 /* how many */,
      DiskVolumePntr->freePntr, DiskVolumePntr->volumePoolID);
    put_vnode (DiskVolumePntr->nameSpaceID,
      (unsigned int) CookiePntr->fsoPntr);
    if (CookiePntr->queryPntr->tokenizedQueryPntr != NULL)
      freeDVP (DiskVolumePntr, CookiePntr->queryPntr->tokenizedQueryPntr);
    CookiePntr->queryPntr->tokenizedQueryPntr = NULL;
    if (CookiePntr->queryPntr->parseTreePntr != NULL)
      FreeParseTree (DiskVolumePntr, CookiePntr->queryPntr->parseTreePntr);
    CookiePntr->queryPntr->parseTreePntr = NULL;
    freeDVP (DiskVolumePntr, CookiePntr->queryPntr);
    CookiePntr->queryPntr = NULL;
  }

  if (CookiePntr->keyType != B_STRING_TYPE &&
  CookiePntr->keyType != B_MIME_STRING_TYPE)
  {
    AVLDupFreeThingArray (&CookiePntr->lastKey, CookiePntr->keyType, 1,
      DiskVolumePntr->freePntr, DiskVolumePntr->volumePoolID);
  }
  CookiePntr->fsoPntr = NULL; /* Make it an invalid cookie. */
  freeDVP (DiskVolumePntr, CookiePntr);

  ReturnCode = 0;

ErrorExit:
#if DPRINT_NON_ERROR_MESSAGES
  dprintf ("ARFS:GenericFreeDirCookie exited, return code %d.\n",
    (int) ReturnCode);
#endif
  return ReturnCode;
}



/******************************************************************************
 * This call from the OS converts a vnode_id into actual data stored in memory,
 * in your shadow portion of a vnode.  The ns pointer identifies the file
 * system / name space instance (you returned it in the mount call, probably as
 * a pointer to your file system's semi-global data).  Besides returning this
 * node thing (could be a number or anything that fits in a pointer sized 32
 * bits) that gets saved in the OS's internal vnode record, it also implies
 * that someone is using that file/directory, until write_vnode (or
 * remove_vnode) is called to tell you that the OS has finished using it.
 * Typically your code will read the catalog info on the file/directory and
 * store it in freshly allocated memory, returning a pointer to that memory.
 * If you are doing file level multitasking locking, this is a good place to
 * create a semaphore to protect that file, naturally storing it in your shadow
 * vnode data for easy access in other file operations.  If you want to do the
 * same read_vnode operation yourself, don't call your read_vnode function,
 * instead call get_vnode, and the OS will call read_vnode if it doesn't
 * already have that vnode loaded.  Finally, the "r" field is a boolean flag to
 * signal a re-entrant call so that you know it was called from a function
 * which your file system called (often from your walk function calling the
 * OS's get_vnode which then calls your read_vnode), in case that affects your
 * usage of semaphores or other multitasking locks.  Note that read_vnode is
 * single threaded with respect to a particular vnode_id, so you don't have to
 * worry about two threads trying to load the same vnode_id at the same time.
 * Returns the usual error codes.
 */

int ARFS_read_vnode (
  void *ns,
  vnode_id vnid,
  char r,
  void **node)
{
  DiskVolumePointer       DiskVolumePntr = ns;
  status_t                ErrorCode;
  int                     FlagCode;
  FileSystemObjectPointer FSOPntr;
  bool                    IsFakeSymbolicLink = false;
  status_t                ReturnCode = -1;

#if DPRINT_NON_ERROR_MESSAGES
  dprintf ("ARFS_read_vnode entered, NS:$%X, VNodeID:$%016LX, "
    "Reentrant:%d\n", (unsigned int) ns, vnid, (int) r);
#endif

  if (INVALID_DISK_VOLUME (DiskVolumePntr))
  {
    dprintf ("ARFS_read_vnode: invalid input disk volume pointer $%X.\n",
      (unsigned int) DiskVolumePntr);
    goto ErrorExit;
  }

  if ((((unsigned long long) vnid) >> 32) != 0)
  {
    dprintf ("ARFS_read_vnode: "
      "Invalid VNodeID because high 32 bits aren't zero!\n");
    goto ErrorExit;
  }

  /* Low 3 bits of the vnodeid are flags, since the vnodeid is actually a
  pointer and the memory allocation system always allocates addresses that are
  multiples of 8.  A flag of zero means that this is an ordinary file system
  object pointer.  A flag of 1 means that this is a fake symbolic link to the
  file system object.  2 to 7 aren't used yet.  Strip off the flag bits (set
  them to zero) to get the true file system object pointer. */

  FlagCode = ((unsigned int) vnid & 7);
  if (FlagCode == 1)
  {
    IsFakeSymbolicLink = true;
    FSOPntr = (void *) ((unsigned int) vnid - 1);
  }
  else if (FlagCode == 0)
    FSOPntr = (void *) (unsigned int) vnid;
  else /* Unknown flag code. */
  {
    dprintf ("ARFS_read_vnode: Unknown flag code %d, low 3 bits aren't "
      "0 or 1 in vnodeid $%016LX.\n", FlagCode, vnid);
    goto ErrorExit;
  }

  if (INVALID_FILE_SYSTEM_OBJECT (FSOPntr))
  {
    dprintf ("ARFS_read_vnode: Invalid VNodeID because $%016LX doesn't "
      "contain a pointer to a valid file system object!\n", vnid);
    ReturnCode = EBADF;
    goto ErrorExit;
  }

#if DPRINT_NON_ERROR_MESSAGES
  dprintf ("ARFS_read_vnode operating on %s.\n",
    FSOAsString ((FileSystemObjectPointer) (unsigned int) vnid));
#endif

  /* Activate this file system object - just allocate the semaphore if it isn't
  already there (it should not be).  Not needed for fake symbolic links, which
  don't really exist (aren't lockable) and which keep on working (returning
  error codes but not crashing) even if the referenced object disappears. */

  if (!IsFakeSymbolicLink)
  {
    ErrorCode = ActivateFSO (FSOPntr);
    if (ErrorCode < 0)
    {
      dprintf ("ARFS_read_vnode: Unable to activate %s, "
        "code %d!\n", FSOAsString (FSOPntr), (int) ErrorCode);
      ReturnCode = ErrorCode;
      goto ErrorExit;
    }
  }

  *node = (void *) (unsigned int) vnid; /* Preserve flag bits. */
  ReturnCode = 0;

ErrorExit:
#if DPRINT_NON_ERROR_MESSAGES
  dprintf ("ARFS_read_vnode for $%016LX exited, return code %d, "
    "returned node $%X.\n", vnid, (int) ReturnCode, (unsigned int) *node);
#endif
  return ReturnCode;
}



/******************************************************************************
 * The OS calls this when it has finished using the vnode (nobody is using it
 * (vnode reference count is zero) and the OS wants to free up its vnode
 * record, this can include parent directories of an open file).  To force
 * flushing of vnodes, do an "ls -R" on some other file system which has a lot
 * of files, as there is a limited number of active vnodes (hopefully less than
 * the number of semaphores if you allocate one semaphore per active shadow
 * vnode).  See also remove_vnode, which gets called instead of this when
 * deleting a file.  You can now write the file info back to disk, or leave it
 * in your cache (the book recommends that you don't change the disk data here,
 * guess they want you to do it at the actual time of the main operations).  In
 * any case, this is a cue to deallocate memory and delete semaphores you
 * allocated in read_vnode.  The r flag tells you if this is a re-entrant call
 * (one of your file system functions called something in the OS which then
 * called your write_vnode).  Single threaded with respect to a particular
 * vnode_id.  Returns the usual error codes.
 */

int ARFS_write_vnode (
  void *ns,
  void *node,
  char r)
{
  DiskVolumePointer       DiskVolumePntr = ns;
  status_t                ErrorCode;
  FileSystemObjectPointer FSOPntr = node;
  status_t                ReturnCode = -1;

#if DPRINT_NON_ERROR_MESSAGES
  dprintf ("ARFS_write_vnode entered, NS:$%X, Node:$%X, Reentrant:%d\n",
    (unsigned int) ns, (unsigned int) node, (int) r);
#endif

  if (INVALID_DISK_VOLUME (DiskVolumePntr))
  {
    dprintf ("ARFS_write_vnode: invalid input disk volume pointer $%X.\n",
      (unsigned int) DiskVolumePntr);
    goto ErrorExit;
  }

  if (((unsigned int) FSOPntr & 7) != 0)
  {
    /* Fake symbolic links and other odd things don't have a real file system
    object, so they're automatically done, whether they are valid or not. */

#if DPRINT_NON_ERROR_MESSAGES
    dprintf ("ARFS_write_vnode Nothing to do for fake %s.\n",
      FSOAsString (FSOPntr));
#endif
    ReturnCode = 0; /* Success. */
    goto ErrorExit;
  }

  if (INVALID_FILE_SYSTEM_OBJECT (FSOPntr))
  {
    dprintf ("ARFS_write_vnode: Invalid node pointer because $%X isn't "
      "a pointer to a valid file system object!\n", (unsigned int) FSOPntr);
    goto ErrorExit;
  }

#if DPRINT_NON_ERROR_MESSAGES
  dprintf ("ARFS_write_vnode operating on %s.\n", FSOAsString (FSOPntr));
#endif

  /* Deallocate the semaphore and other things used by live FSOs. */

  ErrorCode = DeactivateFSO (FSOPntr);
  if (ErrorCode < 0)
  {
    dprintf ("ARFS_write_vnode: Unable to deactivate %s, "
      "code %d!\n", FSOAsString (FSOPntr), (int) ErrorCode);
    ReturnCode = ErrorCode;
    goto ErrorExit;
  }

  /* If this is a file, then we know nobody has it open, so pack memory.  Can
  assume nobody else is accessing it since the vnode is locked internally by
  the OS. */

  if (FSOPntr->objectType == FSO_TYPE_FILE)
    PackFileMemory (DiskVolumePntr, (FileObjectPointer) FSOPntr);

  ReturnCode = 0;

ErrorExit:
#if DPRINT_NON_ERROR_MESSAGES
  dprintf ("ARFS_write_vnode exited, return code %d.\n", (int) ReturnCode);
#endif
  return ReturnCode;
}



/******************************************************************************
 * When everybody has finished using a deleted vnode, this gets called, instead
 * of write_vnode.  Your file system should then free all resources and also
 * remove the file/directory/thing (disk blocks etc) from your file system.  As
 * usual, the r flag is set for re-entrant calls and it returns an error code.
 */

int ARFS_remove_vnode (
  void *ns,
  void *node,
  char r)
{
  DiskVolumePointer       DiskVolumePntr = ns;
  status_t                ErrorCode;
  FileSystemObjectPointer FSOPntr = node;
  status_t                ReturnCode = -1;

#if DPRINT_NON_ERROR_MESSAGES
  dprintf ("ARFS_remove_vnode entered, NS:$%X, Node:$%X, Reentrant:%d\n",
    (unsigned int) ns, (unsigned int) node, (int) r);
#endif

  if (INVALID_DISK_VOLUME (DiskVolumePntr))
  {
    dprintf ("ARFS_remove_vnode: invalid input disk volume pointer $%X.\n",
      (unsigned int) DiskVolumePntr);
    goto ErrorExit;
  }

  if (((unsigned int) node & 7) != 0)
  {
    /* Fake symbolic links and other odd things don't have a real file system
    object, so they're automatically done, whether they are valid or not. */

    ReturnCode = 0; /* Success. */
    goto ErrorExit;
  }

  /* Note that this can happen when ARFS_create partly makes a FSO. */

  if (INVALID_FILE_SYSTEM_OBJECT (FSOPntr))
  {
    dprintf ("ARFS_remove_vnode: Invalid node pointer because $%X isn't "
      "a pointer to a valid file system object!\n", (unsigned int) FSOPntr);
    goto ErrorExit;
  }

#if DPRINT_NON_ERROR_MESSAGES
  dprintf ("ARFS_remove_vnode operating on %s.\n", FSOAsString (FSOPntr));
#endif

  ReturnCode = 0; /* Run through all remaining cleanup without aborting. */

  RemoveAllAttributesFromIndex (DiskVolumePntr, FSOPntr);

  ErrorCode = DeactivateFSO (FSOPntr);
  if (ErrorCode < 0) ReturnCode = ErrorCode;

  ErrorCode = DeallocateAnyFSO (DiskVolumePntr, FSOPntr);
  if (ErrorCode < 0) ReturnCode = ErrorCode;

ErrorExit:
#if DPRINT_NON_ERROR_MESSAGES
  dprintf ("ARFS_remove_vnode exited, return code %d.\n", (int) ReturnCode);
#endif
  return ReturnCode;
}



/******************************************************************************
 * This is used for traversing down a path string to get the actual files
 * (vnodes) the user wants.  Given a directory (base is your shadow vnode
 * directory info data that's associated with the directory vnode) and a file
 * name (in file, with '/' characters already removed), see if you can find
 * that file (or directory or whatnot) inside the base directory.  You also
 * have to handle requests for "." and maybe even "..".  I incorrectly assumed
 * that the vnodes of all parent directories up to the root level had all been
 * get_vnode'd (otherwise where did it get the value for base?), but testing
 * reveals that they won't stay in memory for the duration (they get flushed
 * with write_vnode if BeOS needs vnodes for other operations).  Return the
 * vnode_id in vnid if you found it, and you also need to do a get_vnode on the
 * found file too (otherwise you'll get a kernel error).  The OS will do the
 * corresponding put_vnode later, when it needs to recycle the vnode for other
 * purposes.  If the name doesn't exist in the directory, return ENOENT.  If
 * you find a symbolic link instead of a file, and newpath isn't NULL (if it is
 * NULL, treat the symbolic link like a regular file), copy the symbolic link's
 * string to newpath, using the new_path function to do the string
 * copy/allocate (check if it fails before assigning the result), and don't do
 * get_vnode (or undo it with put_vnode if you did it).  BeOS seems to use walk
 * in pairs to find subdirectories, once with the name in the parent directory
 * and once with "." as the name in the vnode it found in the first walk - I
 * think it is to verify the directory access permissions, walking the "." file
 * tests the permissions of that particular directory.
 */

int ARFS_walk (
  void *ns,
  void *base,
  const char *file,
  char **newpath,
  vnode_id *vnid)
{
  FileSystemObjectPointer ActualFSOPntr;
  DiskVolumePointer       DiskVolumePntr = ns;
  int                     DotCount;
  status_t                ErrorCode;
  AGMSLinkObjectPointer  *ExtraLinkPntrPntr;
  bool                    GetVNodeDone = false;
  IndexObjectPointer      IndexPntr;
  bool                    IsFakeSymbolicLink = false;
  AVLDupThingRecord       Key;
  int                     KeyStringLength;
  const char             *NamePntr;
  FileSystemObjectPointer ParentFSOPntr = base;
  char                   *PathNamePntr;
  bool                    ReadLockDone = false;
  status_t                ReturnCode = -1;
  vnode_id                ReturnVNodeID = 0;
  char                    SymbolicPath [PATH_MAX];
  char                    TempString [NAME_MAX];
  unsigned int            TempUINT;
  const char             *VNodeHexStringPntr;
  vnode_id                VNodeFromHexString;
  void                   *VoidPntr;

#if DPRINT_NON_ERROR_MESSAGES
  dprintf ("ARFS_walk entered, ns:$%X, base:$%X, file:\"%s\", newpath:$%X, "
    "*newpath:$%X, vnid:$%X.\n",
    (unsigned int) ns, (unsigned int) base, file, (unsigned int) newpath,
    (newpath != NULL) ? (unsigned int) *newpath : -1, (unsigned int) vnid);
#endif

  if (INVALID_DISK_VOLUME (DiskVolumePntr))
  {
    dprintf ("ARFS_walk: invalid input disk volume pointer $%X.\n",
      (unsigned int) DiskVolumePntr);
    goto ErrorExit;
  }

  if (INVALID_FILE_SYSTEM_OBJECT (ParentFSOPntr))
  {
    dprintf ("ARFS_walk: invalid input base directory FSO pointer $%X.\n",
      (unsigned int) ParentFSOPntr);
    goto ErrorExit;
  }

  if (ParentFSOPntr->objectType != FSO_TYPE_DIRECTORY &&
  ParentFSOPntr->objectType != FSO_TYPE_INDEX)
  {
    dprintf ("ARFS_walk: base %s isn't a directory or index object!\n",
      FSOAsString (ParentFSOPntr));
    ReturnCode = ENOTDIR;
    goto ErrorExit;
  }

  ErrorCode = ReadLock (ParentFSOPntr, false);
  if (ErrorCode != 0)
  {
    dprintf ("ARFS_walk: unable to get read lock for %s, error %d.\n",
      FSOAsString (ParentFSOPntr), (int) ErrorCode);
    ReturnCode = ErrorCode;
    goto ErrorExit;
  }
  ReadLockDone = true;

  /* Fail if the user isn't allowed to walk this directory.  Normally you need
  the x (execute) permission to be able to use something in a directory
  (assuming you know its name - the r (read) permission is needed to list a
  directory's contents).  However, to get a directory listing of a directory
  that only has r (read) permission, the system does a Walk to get that
  directory's vnode, so we need to allow read access walkers too. */

  ErrorCode =
    PermissionAllowed (DiskVolumePntr, ParentFSOPntr, X_OK | R_OK);
  if (ErrorCode != 0)
  {
    dprintf ("ARFS_walk %s doesn't have access permission for contents (need "
      "the r or x flag).\n", FSOAsString (ParentFSOPntr));
    ReturnCode = ErrorCode;
    goto ErrorExit;
  }

  /* Look for parent directory names, consisting of a sequence of periods. */

  DotCount = 0;
  for (NamePntr = file; *NamePntr != 0; NamePntr++)
  {
    if (*NamePntr != '.')
      break;
    DotCount++;
  }

  ReturnVNodeID = 0;
  if (*NamePntr == 0) /* File name is all periods or an empty string. */
  {
    if (DotCount == 1) /* "." for this directory. */
      ReturnVNodeID = (unsigned int) ParentFSOPntr;
    else if (DotCount == 2) /* ".." for the main parent directory. */
      ReturnVNodeID = (unsigned int) ParentFSOPntr->mainParentDirectory;
    else if (DotCount >= 3) /* "..." and up for other parent directories. */
    {
      ExtraLinkPntrPntr = ParentFSOPntr->parentLinkArray;
      if (ExtraLinkPntrPntr != NULL)
      {
        DotCount -= 3;
        while (DotCount > 0 && *ExtraLinkPntrPntr != NULL)
        {
          DotCount--;
          ExtraLinkPntrPntr++;
        }
        if (*ExtraLinkPntrPntr != NULL)
        {
          /* Have an AGMS link, return a fake symbolic link pointing at its
          parent directory.  Stick in a validity check here because we are
          doing a dereference (normally we don't check, and let the next call
          to the file system validate the FSO pointer, which will prevent
          dereferencing of bad memory addresses). */

          if (VALID_FILE_SYSTEM_OBJECT (&(*ExtraLinkPntrPntr)->fso) &&
          (*ExtraLinkPntrPntr)->fso.objectType == FSO_TYPE_AGMS_LINK)
          {
            IsFakeSymbolicLink = true;
            ReturnVNodeID = 1 /* Flag bits of 1 for fake symbolic link */ +
              (unsigned int) (*ExtraLinkPntrPntr)->fso.mainParentDirectory;
          }
        }
      }
    }
  }
  else /* Not a magic dot name, search the directory or index listing. */
  {
    if (ParentFSOPntr->objectType == FSO_TYPE_INDEX)
    {
      /* Convert the file name into a key of the appropriate datatype.  That
      can get as complex as converting a date string into the number of seconds
      since the start of time.  To make the names unique (for stupid software
      which doesn't expect several entries all with the same name, such as
      files with the same size in the "size" index), the vnode_id is present
      after the key.  The format is "key-as-string #Vnode", with a space and
      octothorpe between the key and vnode as a hex number.  However, for
      really long file names, it isn't present (and watch out for really long
      file names that happen to have a # in them). */

      IndexPntr = (IndexObjectPointer) ParentFSOPntr;

      VNodeFromHexString = 0;
      VNodeHexStringPntr = strrchr (file, '#');
      if (VNodeHexStringPntr != NULL &&
      VNodeHexStringPntr > file && VNodeHexStringPntr[-1] == ' ')
      {
        VNodeHexStringPntr++; /* Point past the " #" marker. */
        VNodeFromHexString = strtoull (VNodeHexStringPntr,
          NULL, 16 /* base */); /* Leaves it zero if it isn't a number. */
      }
      if (VNodeFromHexString == 0)
        VNodeHexStringPntr = NULL;
      if (VNodeHexStringPntr == NULL)
        KeyStringLength = strlen (file);
      else /* Length not including the " #" and the rest. */
        KeyStringLength = (VNodeHexStringPntr - file) - 2;
      if (KeyStringLength >= NAME_MAX)
        KeyStringLength = NAME_MAX - 1;
      memcpy (TempString, file, KeyStringLength);
      TempString [KeyStringLength] = 0;

      if (!AVLDupConvertStringToThing (TempString, &Key, IndexPntr->keyType,
      DiskVolumePntr->mallocPntr, DiskVolumePntr->freePntr,
      DiskVolumePntr->volumePoolID))
      {
        dprintf ("ARFS_walk Can't convert name \"%s\" into a key of type %s "
          "for %s.\n", TempString, TypeCodeAsString (IndexPntr->keyType),
          FSOAsString (ParentFSOPntr));
        goto ErrorExit;
      }

      if (VNodeFromHexString != 0)
      {
        if (AVLDupKeyAndValueExists (IndexPntr->treePntr, &Key,
        VNodeFromHexString))
          TempUINT = VNodeFromHexString;
        else
          TempUINT = 0;
      }
      else /* No clue as to which vnode it is.  Just pick the first one that
      matches the key. */
        TempUINT = AVLDupFind (IndexPntr->treePntr, &Key);

      AVLDupFreeThingArray (&Key, IndexPntr->keyType, 1,
        DiskVolumePntr->freePntr, DiskVolumePntr->volumePoolID);

      ActualFSOPntr = (FileSystemObjectPointer) TempUINT;

      /* The things in an index are all fake symbolic links to the actual
      things, since the index isn't actually a parent directory of the things
      inside it.  So set the low 3 bits of the VNode to 1 to mark it as a fake
      symbolic link. */

      if (ActualFSOPntr != NULL)
      {
        IsFakeSymbolicLink = true;
        ReturnVNodeID = 1 + (unsigned int) ActualFSOPntr;
      }
      else /* Didn't find the key in the index. */
        ReturnVNodeID = 0;
    }
    else /* Must be a regular directory.  Key is a simple string. */
    {
      Key.longStringThing.stringPntr = (char *) file;
      Key.longStringThing.isLongString = true;
      TempUINT = AVLDupFind (
        ((DirectoryObjectPointer) ParentFSOPntr)->directory, &Key);
      ActualFSOPntr = (FileSystemObjectPointer) TempUINT;
      ReturnVNodeID = (unsigned int) ActualFSOPntr;
    }
  }

  if (ReturnVNodeID == 0) /* Didn't find it, return an error. */
  {
#if DPRINT_NON_ERROR_MESSAGES
    dprintf ("ARFS_walk: Didn't find name \"%s\" in %s.\n",
      file, FSOAsString (ParentFSOPntr));
#endif
    ReturnCode = ENOENT;
    goto ErrorExit;
  }

  /* Get the shadow vnode for the VNodeID.  This may make a call to our
  read_vnode function (with the recursive flag set) if the VNode isn't
  currently loaded.  The get_vnode is left outstanding, the OS will do the
  put_vnode later when it feels like it.  Do this even for fake symbolic links
  so that the OS builds its vnode record even though we don't have a shadow
  vnode of our own for it. */

  VoidPntr = NULL;
  ErrorCode =
    get_vnode (DiskVolumePntr->nameSpaceID, ReturnVNodeID, &VoidPntr);
  if (ErrorCode < 0)
  {
    dprintf ("ARFS_walk: get_vnode failed for VNodeID $%016LX, code %d!\n",
      ReturnVNodeID, (int) ErrorCode);
    ReturnCode = ErrorCode;
    goto ErrorExit;
  }
  GetVNodeDone = true;
  if (IsFakeSymbolicLink)
    ActualFSOPntr = (void *) ((unsigned int) VoidPntr - 1);
  else
    ActualFSOPntr = VoidPntr;

  if (INVALID_FILE_SYSTEM_OBJECT (ActualFSOPntr))
  {
    dprintf ("ARFS_walk: Invalid walking FSO $%X from get_vnode's $%X.\n",
      (unsigned int) ActualFSOPntr, (unsigned int) VoidPntr);
    goto ErrorExit;
  }

#if DPRINT_NON_ERROR_MESSAGES
  dprintf ("ARFS_walk: get_vnode gave us %s.\n",
    FSOAsString ((FileSystemObjectPointer) VoidPntr));
#endif

  /* See if this item should be represented as a symbolic link.  First, the OS
  has to be asking for the symbolic link path (newpath not NULL).  Then the
  item has to be a real symbolic link or a fake symbolic link. */

  if (newpath != NULL && (IsFakeSymbolicLink ||
  ActualFSOPntr->objectType == FSO_TYPE_AGMS_LINK ||
  ActualFSOPntr->objectType == FSO_TYPE_SYMBOLIC_LINK))
  {
    if (IsFakeSymbolicLink)
    {
      ErrorCode = GetTruePath (DiskVolumePntr, ActualFSOPntr,
        SymbolicPath, sizeof (SymbolicPath));
      if (ErrorCode != 0)
      {
        dprintf ("ARFS_walk: Unable to get true path to %s for making "
          "fake symbolic link.\n",
          FSOAsString ((FileSystemObjectPointer) VoidPntr));
        ReturnCode = ErrorCode;
        goto ErrorExit;
      }
      PathNamePntr = SymbolicPath;
    }
    else if (ActualFSOPntr->objectType == FSO_TYPE_AGMS_LINK)
    {
      ErrorCode = GetTruePath (DiskVolumePntr,
        ((AGMSLinkObjectPointer) ActualFSOPntr)->targetFSOPntr,
        SymbolicPath, sizeof (SymbolicPath));
      if (ErrorCode != 0)
      {
        dprintf ("ARFS_walk: Unable to get true path to target of %s "
          "while making pseudo-symbolic link for it.\n",
          FSOAsString ((FileSystemObjectPointer) VoidPntr));
        ReturnCode = ErrorCode;
        goto ErrorExit;
      }
      PathNamePntr = SymbolicPath;
    }
    else
    {
      PathNamePntr =
        ((SymbolicLinkObjectPointer) ActualFSOPntr)->symbolicPathPntr;
    }

    /* Now have the symbolic path name in PathNamePntr. */

    ErrorCode = new_path (PathNamePntr, newpath);
    if (ErrorCode != 0)
    {
      dprintf ("ARFS_walk: error %d while copying symbolic link path string "
        "\"%s\" for target %s.\n", (int) ErrorCode, PathNamePntr,
        FSOAsString ((FileSystemObjectPointer) VoidPntr));
      ReturnCode = ErrorCode;
      goto ErrorExit;
    }

#if DPRINT_NON_ERROR_MESSAGES
    dprintf ("ARFS_walk: Prepared symbolic path of \"%s\" for %s.\n",
      PathNamePntr,
      FSOAsString ((FileSystemObjectPointer) VoidPntr));
#endif

    /* The OS only wanted the symbolic link path, undo the get_vnode. */

    put_vnode (DiskVolumePntr->nameSpaceID, ReturnVNodeID);
    GetVNodeDone = false;
  }

  /* Success! */

  ReturnCode = 0;
  if (vnid != NULL)
    *vnid = ReturnVNodeID;

ErrorExit:
#if DPRINT_NON_ERROR_MESSAGES
  dprintf ("ARFS_walk exited, return code %d, returned VNodeID $%016LX.\n",
    (int) ReturnCode, ReturnVNodeID);
#endif
  if (ReadLockDone)
    ReadUnlock (ParentFSOPntr);
  if (GetVNodeDone && ReturnCode < 0) /* Undo get_vnode done in error. */
    put_vnode (DiskVolumePntr->nameSpaceID, ReturnVNodeID);
  return ReturnCode;
}



/******************************************************************************
 * Check security permissions to see if the caller is allowed to do the
 * specified operations (mode flags R_OK, W_OK, X_OK and F_OK) on that
 * particular node (the Posix standard says to check that all things in the
 * path pass the permission test, but we only get one node to check so that's
 * it).  A mode of F_OK (0) just tests if the thing exists.  The others check
 * for the various read/write/execute permissions (using the calling thread's
 * user and group settings).  Of course, with BeOS using the root user (user
 * #0) as the default, it automatically gets read and write permission to
 * everything.  Note that read only file systems will always fail W_OK, even
 * for the root user.  Combinations of flags check that any one or more of the
 * flagged permissions is met.  See the Posix access() function for more info,
 * on how it should be done, not how I do it.  Returns 0 for access allowed, or
 * an error code for access denied.  Only used by user programs, the OS uses
 * secure_vnode for its security checks, and you can always add your own checks
 * to the individual file manipulation functions.
 */

int ARFS_access (
  void *ns,
  void *node,
  int mode)
{
  DiskVolumePointer       DiskVolumePntr = ns;
  FileSystemObjectPointer FSOPntr = node;
  status_t                ReturnCode = -1;

#if DPRINT_NON_ERROR_MESSAGES
  dprintf ("ARFS_access entered, NS:$%X, node:$%X, mode:%d./$%X.\n",
    (unsigned int) ns, (unsigned int) node, mode, mode);
#endif

  if (INVALID_DISK_VOLUME (DiskVolumePntr))
  {
    dprintf ("ARFS_access: invalid input disk volume pointer $%X.\n",
      (unsigned int) DiskVolumePntr);
    goto ErrorExit;
  }

  if (INVALID_FILE_SYSTEM_OBJECT (FSOPntr))
  {
    dprintf ("ARFS_access: invalid input FSO pointer $%X.\n",
      (unsigned int) FSOPntr);
    goto ErrorExit;
  }

  /* Check permissions, don't need to lock the FSO since it only involves
  reading the mode flags and user/group IDs. */

  ReturnCode = PermissionAllowed (DiskVolumePntr, FSOPntr, mode);
  if (ReturnCode != 0)
  {
    dprintf ("ARFS_access: Don't have permission $%X to access %s\n",
      (int) mode, FSOAsString (FSOPntr));
  }

ErrorExit:
#if DPRINT_NON_ERROR_MESSAGES
  dprintf ("ARFS_access exited, return code %d.\n", (int) ReturnCode);
#endif
  return ReturnCode;
}



/******************************************************************************
 * Create a new file, much like the open() function with the create flag
 * specified (I suspect that you'll never see open() called with the create
 * flag), but it also returns a vnode_id in addition to the open file cookie.
 * It can get called with an existing file if the user calls create with an
 * existing file name.  You should use get_vnode to get the existing file in
 * that case, and reuse it (other programs may have it open too).  The dir
 * specifies the shadow vnode portion of the vnode for the parent directory.
 * The name is the name of the new file, the VFS layer filters out names which
 * are too long (more than 256 bytes), but doesn't filter out names like "..".
 * The omode argument specifies the open flags, much like open(), which
 * specifies if it should fail if the file exists (or is a directory), or if
 * the existing file should be truncated, or left as is.  The PFSDWTBFS book
 * specifies the details.  Perms contains the new file access permission bits
 * (rwxrwxrwx etc), which you need to modify by clearing the bits specified by
 * the caller's umask setting (invert the mask and AND it), well, actually BeOS
 * does this for you before passing in the permissions.  The userid of the new
 * file is presumably the same as the calling process, while the group is
 * copied from the directory which it is created in.  Besides returning the new
 * vnode_id, also return a cookie to the newly open file when done.  If you
 * actually created the file, you also have to use new_vnode too (and activate
 * your shadow vnode as if your read_vnode had been called) since having a
 * cookie without an OS vnode plus your shadow vnode is kind of weird - yup,
 * without it you get a kernel error: KERN 'sh'[1929]: KERNEL: FS agmsrfs0:
 * CREATE(/Test/Junk, 0601): vn (2826a0b0) NULL!.  There may be a race
 * condition with other programs trying to open the file when they see it
 * appearing in an index (like the index of file names), so block other threads
 * in read_vnode until it is ready, or don't let anyone know the new vnode_id
 * until it is ready, then call notify_listener() and do the index stuff if you
 * have indices to update.
 */

int ARFS_create (
  void *ns,
  void *dir,
  const char *name,
  int omode,
  int perms,
  vnode_id *vnid,
  void **cookie)
{
  bool                    ActivationDone = false;
  DirectoryObjectPointer  DirectoryPntr = dir;
  DiskVolumePointer       DiskVolumePntr = ns;
  status_t                ErrorCode;
  FileSystemObjectPointer ExistingFSOPntr;
  FileCookiePointer       FileCookiePntr;
  bool                    GetNodeDone = false;
  AVLDupThingRecord       Key;
  FileObjectPointer       NewFilePntr = NULL;
  vnode_id                NewVNodeID = 0;
  status_t                ReturnCode = -1;
  const char             *StringPntr;
  void                   *TargetNode = NULL;
  unsigned int            TempUINT;
  bool                    VNodeCreated = false;
  bool                    WriteLockBaseDirDone = false;

#if DPRINT_NON_ERROR_MESSAGES
  dprintf ("ARFS_create entered, NS:$%X, parent:%X, name:\"%s\", "
    "omode:$%X, perms:$%X.\n",
    (unsigned int) ns, (unsigned int) dir, name, omode, perms);
#endif

  /* Validate the name of the new file. */

  if (name == NULL)
  {
    dprintf ("ARFS_create: invalid input NULL name.\n");
    goto ErrorExit;
  }

  if (strlen (name) >= NAME_MAX)
  {
    dprintf ("ARFS_create: invalid input name \"%s\" is too long, "
      "maximum is %d bytes total (including NUL at end).\n",
      name, NAME_MAX);
    ReturnCode = ENAMETOOLONG;
    goto ErrorExit;
  }

  /* Reject names consisting of all dots, like ".." or "...." or "", which we
  reserve for assorted kinds of parent directories. */

  for (StringPntr = name; *StringPntr != 0; StringPntr++)
    if (*StringPntr != '.')
      break;
  if (*StringPntr == 0)
  {
    dprintf ("ARFS_create: invalid input name \"%s\" is all periods, which "
      "are reserved for parent directory names.\n", name);
    ReturnCode = EINVAL;
    goto ErrorExit;
  }

  if (INVALID_DISK_VOLUME (DiskVolumePntr))
  {
    dprintf ("ARFS_create: invalid input disk volume pointer $%X.\n",
      (unsigned int) DiskVolumePntr);
    goto ErrorExit;
  }

  if (INVALID_FILE_SYSTEM_OBJECT (&DirectoryPntr->fso))
  {
    dprintf ("ARFS_create: invalid input base directory $%X.\n",
      (unsigned int) DirectoryPntr);
    goto ErrorExit;
  }

  if (DirectoryPntr->fso.objectType != FSO_TYPE_DIRECTORY)
  {
    dprintf ("ARFS_create: base directory (%s) isn't a directory object!\n",
      FSOAsString (&DirectoryPntr->fso));
    ReturnCode = ENOTDIR;
    goto ErrorExit;
  }

  ErrorCode = WriteLock (&DirectoryPntr->fso);
  if (ErrorCode != 0)
  {
    dprintf ("ARFS_create: unable to get write lock for %s, code %d.\n",
      FSOAsString (&DirectoryPntr->fso), (int) ErrorCode);
    ReturnCode = ErrorCode;
    goto ErrorExit;
  }
  WriteLockBaseDirDone = true;

  /* See if the given file name already exists. */

  Key.longStringThing.stringPntr = (char *) name;
  Key.longStringThing.isLongString = true;
  TempUINT = AVLDupFind (DirectoryPntr->directory, &Key);
  ExistingFSOPntr = (FileSystemObjectPointer) TempUINT;
  if (ExistingFSOPntr == NULL)
  {
    /* The name doesn't exist.  Have to create a new one.  Fail if the user
    isn't allowed to write to the parent directory.  This also tests for a
    read-only file system. */

    ErrorCode = PermissionAllowed (DiskVolumePntr, &DirectoryPntr->fso, W_OK);
    if (ErrorCode != 0)
    {
      dprintf ("ARFS_create: %s doesn't have write access permission "
        "for contents (need the w flag).\n",
        FSOAsString (&DirectoryPntr->fso));
      ReturnCode = ErrorCode;
      goto ErrorExit;
    }

    /* Create the file.  It's always a file. */

    NewFilePntr = AllocateFileFSO (DiskVolumePntr, name);
    if (NewFilePntr == NULL)
    {
      dprintf ("ARFS_create: unable to allocate new file FSO when "
        "making file \"%s\" in parent %s.\n",
        name, FSOAsString (&DirectoryPntr->fso));
      ReturnCode = ENOMEM;
      goto ErrorExit;
    }

    NewFilePntr->fso.modeBits |=
      (perms & S_IUMSK /* only allow user setable bits */);

    NewFilePntr->fso.gid = DirectoryPntr->fso.gid;

    /* Activate the new file, as if read_vnode had been done to it.  The OS
    assumes this, and will call write_vnode without any previous read_vnode. */

    ErrorCode = ActivateFSO (&NewFilePntr->fso);
    if (ErrorCode < 0)
    {
      dprintf ("ARFS_create: ActivateFSO failed, error code %d.\n",
        (int) ErrorCode);
      ReturnCode = ErrorCode;
      goto ErrorExit;
    }
    ActivationDone = true;

    /* Make the matching vnode in the OS. */

    NewVNodeID = (unsigned int) &NewFilePntr->fso;
    ErrorCode =
      new_vnode (DiskVolumePntr->nameSpaceID, NewVNodeID, &NewFilePntr->fso);
    if (ErrorCode < 0)
    {
      dprintf ("ARFS_create: new_vnode failed, the OS doesn't want to know "
        "about %s.\n", name);
      ReturnCode = ErrorCode;
      goto ErrorExit;
    }
    VNodeCreated = true;

    /* Hook the file into the directory structure.  Do this last since there is
    no easy way of undoing it in case of an error. */

    ErrorCode = AddFreshFSOToDirectory (DiskVolumePntr,
      &NewFilePntr->fso, DirectoryPntr);
    if (ErrorCode != 0)
    {
      dprintf ("ARFS_create: unable to add new file \"%s\" to parent %s, "
        "code %d.\n", name, FSOAsString (&DirectoryPntr->fso),
        (int) ErrorCode);
      ReturnCode = ErrorCode;
      goto ErrorExit;
    }
    ChangeFSODates (DiskVolumePntr, &DirectoryPntr->fso, DK_MODIFIED_FLAG, -1);

    /* The rest of the code uses ExistingFSOPntr, since NewFilePntr has to
    remain NULL in the case where an existing file is found, otherwise error
    exits could delete an existing file. */

    ExistingFSOPntr = &NewFilePntr->fso;

#if DPRINT_NON_ERROR_MESSAGES
    dprintf ("ARFS_create: Successfully made %s in %s.\n",
      FSOAsString (ExistingFSOPntr), FSOAsString (&DirectoryPntr->fso));
#endif
  }
  else /* Something with the specified name already exists. */
  {
    /* Verify that exclusive create isn't needed. */

    if (omode & O_EXCL)
    {
      dprintf ("ARFS_create: O_EXCL specified and %s already exists.\n",
        FSOAsString (ExistingFSOPntr));
      ReturnCode = EEXIST;
      goto ErrorExit;
    }

    /* Make sure that it is a file.  Could be a directory or symbolic link. */

    if (ExistingFSOPntr->objectType != FSO_TYPE_FILE)
    {
      dprintf ("ARFS_create: %s exists and is not a file.\n",
        FSOAsString (ExistingFSOPntr));
      ReturnCode = EISDIR; /* Unfortunately no special error for symlinks. */
      goto ErrorExit;
    }

    /* Load the vnode for the existing file. */

    NewVNodeID = (unsigned int) ExistingFSOPntr;
    ErrorCode =
      get_vnode (DiskVolumePntr->nameSpaceID, NewVNodeID, &TargetNode);
    if (ErrorCode < 0)
    {
      dprintf ("ARFS_create: get_vnode failed for VNodeID $%016LX, code %d!\n",
        NewVNodeID, (int) ErrorCode);
      ReturnCode = ErrorCode;
      goto ErrorExit;
    }
    GetNodeDone = true;

    if (ExistingFSOPntr != TargetNode)
    {
      dprintf ("ARFS_create: Bug!  Target FSO not same as VNode.\n");
      goto ErrorExit;
    }

#if DPRINT_NON_ERROR_MESSAGES
    dprintf ("ARFS_create: Successfully found existing %s in %s.\n",
      FSOAsString (ExistingFSOPntr), FSOAsString (&DirectoryPntr->fso));
#endif
  }

  /* Open the file now, to get the cookie and do truncation etc, reusing the
  regular open code to avoid bugs.  Note that we keep the write lock on the
  directory so that other programs are prevented (ARFS_walk is blocked by the
  directory write lock, though links through other directories and the indices
  may avoid that) from prematurely (due to the change notification message and
  associated task switch) opening the new file (would cause our open to fail if
  they got an exclusive lock on the file).  Fortunately the Open function
  doesn't lock the directory so we don't have to worry about recursive locking
  problems. */

  ErrorCode =
    ARFS_open (ns, ExistingFSOPntr, omode, (void **) &FileCookiePntr);
  if (ErrorCode != 0)
  {
    dprintf ("ARFS_create: unable to re-entrantly open file %s, code %d.\n",
      FSOAsString (ExistingFSOPntr), (int) ErrorCode);
    ReturnCode = ErrorCode;
    goto ErrorExit;
  }

  /* Successful, wrap it up.  Tell the rest of the world about it if a new file
  was actually created. */

  if (NewFilePntr != NULL)
  {
    /* Create entries in the standard indices for the new file.  This also
    triggers queries watching for those items, which is why we do it after the
    file has been opened (so some other thread doesn't notice the new file and
    delete it before we have opened it). */

    AddFreshThingToStandardIndices (DiskVolumePntr, ExistingFSOPntr);

    /* Finally do the generic notification. */

    ErrorCode = notify_listener (B_ENTRY_CREATED, DiskVolumePntr->nameSpaceID,
      (unsigned int) DirectoryPntr, 0, (unsigned int) ExistingFSOPntr,
      ExistingFSOPntr->trueNamePntr);
    if (ErrorCode != 0 || DPRINT_NOTIFICATIONS)
    {
      dprintf ("ARFS_create: B_ENTRY_CREATED notification for creating %s "
        "in %s, error code %d.\n",
        FSOAsString (ExistingFSOPntr), FSOAsString (&DirectoryPntr->fso),
        (int) ErrorCode);
    }
  }

  *cookie = FileCookiePntr;
  *vnid = NewVNodeID;
  ReturnCode = 0;
  goto NormalExit;

ErrorExit:
  if (ActivationDone)
    DeactivateFSO (&NewFilePntr->fso);
  if (NewFilePntr != NULL)
    DeallocateAnyFSO (DiskVolumePntr, &NewFilePntr->fso);
  if (VNodeCreated) /* Note: OS will call our remove_vnode with invalid FSO. */
    remove_vnode (DiskVolumePntr->nameSpaceID, NewVNodeID);
  if (GetNodeDone)
    put_vnode (DiskVolumePntr->nameSpaceID, NewVNodeID);
NormalExit:
#if DPRINT_NON_ERROR_MESSAGES
  dprintf ("ARFS_create exited, return code %d.\n", (int) ReturnCode);
#endif
  if (WriteLockBaseDirDone)
    WriteUnlock (&DirectoryPntr->fso);
  return ReturnCode;
}



/******************************************************************************
 * Make a new directory.  If you don't create them on the fly, also add entries
 * for "." and ".." to the new directory.  No vnode_id is returned, so you only
 * have to update your data structures, and call notify_listener(), etc.  No
 * need for a call to new_vnode.  The man pages for mkdir say: The directory
 * path is created with the access permissions specified by perms and
 * restricted by the the umask of the calling process.  The directory's owner
 * ID is set to the process's effective user ID.  The directory's group ID is
 * set to that of the parent directory in which it is created.
 */

int ARFS_mkdir (
  void *ns,
  void *dir,
  const char *name,
  int perms)
{
  FileSystemObjectPointer BaseFSOPntr = dir;
  DirectoryObjectPointer  DirectoryPntr = NULL;
  DiskVolumePointer       DiskVolumePntr = ns;
  status_t                ErrorCode;
  DirectoryObjectPointer  NewDirectoryPntr = NULL;
  status_t                ReturnCode = -1;
  const char             *StringPntr;
  bool                    WriteLockedDirectory = false;

#if DPRINT_NON_ERROR_MESSAGES
  dprintf ("ARFS_mkdir entered, NS:$%X, parent:$%X, name:\"%s\", "
    "perms:$%X.\n",
    (unsigned int) ns, (unsigned int) dir, name, (unsigned int) perms);
#endif

  if (name == NULL)
  {
    dprintf ("ARFS_mkdir: invalid input NULL name.\n");
    goto ErrorExit;
  }

  if (strlen (name) >= NAME_MAX)
  {
    dprintf ("ARFS_mkdir: invalid input name \"%s\" is too long, "
      "maximum is %d bytes total (including NUL at end).\n",
      name, NAME_MAX);
    ReturnCode = ENAMETOOLONG;
    goto ErrorExit;
  }

  /* Reject names consisting of all dots, like ".." or "...." or "", which we
  reserve for assorted kinds of parent directories. */

  for (StringPntr = name; *StringPntr != 0; StringPntr++)
    if (*StringPntr != '.')
      break;
  if (*StringPntr == 0)
  {
    dprintf ("ARFS_mkdir: invalid input name \"%s\" is all periods, which "
      "are reserved for parent directory names.\n", name);
    ReturnCode = EINVAL;
    goto ErrorExit;
  }

  if (INVALID_DISK_VOLUME (DiskVolumePntr))
  {
    dprintf ("ARFS_mkdir: invalid input disk volume pointer $%X.\n",
      (unsigned int) DiskVolumePntr);
    goto ErrorExit;
  }

  if (INVALID_FILE_SYSTEM_OBJECT (BaseFSOPntr))
  {
    dprintf ("ARFS_mkdir: invalid input dir FSO pointer $%X.\n",
      (unsigned int) BaseFSOPntr);
    goto ErrorExit;
  }

  if (BaseFSOPntr->objectType != FSO_TYPE_DIRECTORY)
  {
    dprintf ("ARFS_mkdir: base %s isn't a directory object!\n",
      FSOAsString (BaseFSOPntr));
    ReturnCode = ENOTDIR;
    goto ErrorExit;
  }
  DirectoryPntr = (DirectoryObjectPointer) BaseFSOPntr;

  ErrorCode = WriteLock (&DirectoryPntr->fso);
  if (ErrorCode != 0)
  {
    dprintf ("ARFS_mkdir unable to get write lock for %s, code %d.\n",
      FSOAsString (&DirectoryPntr->fso), (int) ErrorCode);
    ReturnCode = ErrorCode;
    goto ErrorExit;
  }
  WriteLockedDirectory = true;

  /* Fail if the user isn't allowed to write to this directory. */

  ErrorCode = PermissionAllowed (DiskVolumePntr, &DirectoryPntr->fso, W_OK);
  if (ErrorCode != 0)
  {
    dprintf ("ARFS_mkdir %s doesn't have write access permission "
      "for contents (need the w flag).\n", FSOAsString (&DirectoryPntr->fso));
    ReturnCode = ErrorCode;
    goto ErrorExit;
  }

  NewDirectoryPntr = AllocateDirectoryFSO (DiskVolumePntr, name);
  if (NewDirectoryPntr == NULL)
  {
    dprintf ("ARFS_mkdir: unable to allocate new subdirectory FSO when "
      "making directory \"%s\" in parent %s.\n",
      name, FSOAsString (&DirectoryPntr->fso));
    ReturnCode = ENOMEM;
    goto ErrorExit;
  }

  NewDirectoryPntr->fso.modeBits |=
    (perms & S_IUMSK /* only allow user setable bits */);

  NewDirectoryPntr->fso.gid = DirectoryPntr->fso.gid;

  /* Attempting to add it to the directory will detect duplicate names, as well
  as other error codes for the usual reasons. */

  ErrorCode = AddFreshFSOToDirectory (DiskVolumePntr,
    &NewDirectoryPntr->fso, DirectoryPntr);
  if (ErrorCode != 0)
  {
    dprintf ("ARFS_mkdir: unable to add new directory \"%s\" to parent %s, "
      "code %d.\n", name, FSOAsString (&DirectoryPntr->fso),
      (int) ErrorCode);
    ReturnCode = ErrorCode;
    goto ErrorExit;
  }
  ChangeFSODates (DiskVolumePntr, &DirectoryPntr->fso, DK_MODIFIED_FLAG, -1);

#if DPRINT_NON_ERROR_MESSAGES
  dprintf ("ARFS_mkdir: Created %s in %s.\n",
    FSOAsString (&NewDirectoryPntr->fso),
    FSOAsString (&DirectoryPntr->fso));
#endif

  ErrorCode = notify_listener (B_ENTRY_CREATED, DiskVolumePntr->nameSpaceID,
    (unsigned int) DirectoryPntr, 0, (unsigned int) NewDirectoryPntr,
    NewDirectoryPntr->fso.trueNamePntr);
  if (ErrorCode != 0 || DPRINT_NOTIFICATIONS)
  {
    dprintf ("ARFS_mkdir: B_ENTRY_CREATED notification for creating %s in %s, "
      "error code %d.\n",
      FSOAsString (&NewDirectoryPntr->fso), FSOAsString (&DirectoryPntr->fso),
      (int) ErrorCode);
  }

  AddFreshThingToStandardIndices (DiskVolumePntr, &NewDirectoryPntr->fso);

  ReturnCode = 0;
  goto NormalExit;

ErrorExit:
  if (NewDirectoryPntr != NULL)
    DeallocateAnyFSO (DiskVolumePntr, &NewDirectoryPntr->fso);
NormalExit:
#if DPRINT_NON_ERROR_MESSAGES
  dprintf ("ARFS_mkdir exited, return code %d.\n", (int) ReturnCode);
#endif
  if (WriteLockedDirectory)
    WriteUnlock (&DirectoryPntr->fso);
  return ReturnCode;
}



/******************************************************************************
 * Create a new symbolic link thing.  Add the name to the given shadow vnode
 * directory.  The name contains the path string, which is used during walk()
 * to redirect to the place path names, so save it in association with the new
 * symbolic link's name.
 */

int ARFS_symlink (
  void *ns,
  void *dir,
  const char *name,
  const char *path)
{
  DirectoryObjectPointer     DirectoryPntr = dir;
  DiskVolumePointer          DiskVolumePntr = ns;
  status_t                   ErrorCode;
  SymbolicLinkObjectPointer  NewSymlinkPntr = NULL;
  status_t                   ReturnCode = -1;
  const char                *StringPntr;
  bool                       WriteLockedDirectory = false;

#if DPRINT_NON_ERROR_MESSAGES
  dprintf ("ARFS_symlink entered, NS:$%X, parent:$%X, "
    "name:\"%s\", path:\"%s\".\n",
    (unsigned int) ns, (unsigned int) dir, name, path);
#endif

  if (name == NULL || path == NULL)
  {
    dprintf ("ARFS_symlink: invalid input NULL name or path.\n");
    goto ErrorExit;
  }

  if (strlen (name) >= NAME_MAX)
  {
    dprintf ("ARFS_symlink: invalid input name \"%s\" is too long, "
      "maximum is %d bytes total (including NUL at end).\n",
      name, NAME_MAX);
    ReturnCode = ENAMETOOLONG;
    goto ErrorExit;
  }

  /* Reject names consisting of all dots, like ".." or "...." or "", which we
  reserve for assorted kinds of parent directories. */

  for (StringPntr = name; *StringPntr != 0; StringPntr++)
    if (*StringPntr != '.')
      break;
  if (*StringPntr == 0)
  {
    dprintf ("ARFS_symlink: invalid input name \"%s\" is all periods, which "
      "are reserved for parent directory names.\n", name);
    ReturnCode = EINVAL;
    goto ErrorExit;
  }

  if (INVALID_DISK_VOLUME (DiskVolumePntr))
  {
    dprintf ("ARFS_symlink: invalid input disk volume pointer $%X.\n",
      (unsigned int) DiskVolumePntr);
    goto ErrorExit;
  }

  if (INVALID_FILE_SYSTEM_OBJECT (&DirectoryPntr->fso))
  {
    dprintf ("ARFS_symlink: invalid input dir pointer $%X.\n",
      (unsigned int) DirectoryPntr);
    goto ErrorExit;
  }

  if (DirectoryPntr->fso.objectType != FSO_TYPE_DIRECTORY)
  {
    dprintf ("ARFS_symlink: base %s isn't a directory object!\n",
      FSOAsString (&DirectoryPntr->fso));
    ReturnCode = ENOTDIR;
    goto ErrorExit;
  }

  ErrorCode = WriteLock (&DirectoryPntr->fso);
  if (ErrorCode != 0)
  {
    dprintf ("ARFS_symlink unable to get write lock for %s, code %d.\n",
      FSOAsString (&DirectoryPntr->fso), (int) ErrorCode);
    ReturnCode = ErrorCode;
    goto ErrorExit;
  }
  WriteLockedDirectory = true;

  /* Fail if the user isn't allowed to write to this directory. */

  ErrorCode = PermissionAllowed (DiskVolumePntr, &DirectoryPntr->fso, W_OK);
  if (ErrorCode != 0)
  {
    dprintf ("ARFS_symlink %s doesn't have write access permission "
      "for contents (need the w flag).\n", FSOAsString (&DirectoryPntr->fso));
    ReturnCode = ErrorCode;
    goto ErrorExit;
  }

  NewSymlinkPntr = AllocateSymbolicLinkFSO (DiskVolumePntr, name, path);
  if (NewSymlinkPntr == NULL)
  {
    dprintf ("ARFS_symlink: unable to allocate new symbolic link FSO when "
      "making symlink \"%s\" in parent %s.\n",
      name, FSOAsString (&DirectoryPntr->fso));
    ReturnCode = ENOMEM;
    goto ErrorExit;
  }

  NewSymlinkPntr->fso.gid = DirectoryPntr->fso.gid;
  NewSymlinkPntr->fso.modeBits |=
    S_IRWXU | S_IRWXG | S_IRWXO; /* BFS uses all permission bits set. */

  /* Attempting to add it to the directory will detect duplicate names, as well
  as other error codes for the usual reasons. */

  ErrorCode = AddFreshFSOToDirectory (DiskVolumePntr,
    &NewSymlinkPntr->fso, DirectoryPntr);
  if (ErrorCode != 0)
  {
    dprintf ("ARFS_symlink: unable to add new symlink \"%s\" to parent %s, "
      "code %d.\n", name, FSOAsString (&DirectoryPntr->fso),
      (int) ErrorCode);
    ReturnCode = ErrorCode;
    goto ErrorExit;
  }
  ChangeFSODates (DiskVolumePntr, &DirectoryPntr->fso, DK_MODIFIED_FLAG, -1);

#if DPRINT_NON_ERROR_MESSAGES
  dprintf ("ARFS_symlink: Created %s in %s.\n",
    FSOAsString (&NewSymlinkPntr->fso),
    FSOAsString (&DirectoryPntr->fso));
#endif

  ErrorCode = notify_listener (B_ENTRY_CREATED, DiskVolumePntr->nameSpaceID,
    (unsigned int) DirectoryPntr, 0, (unsigned int) NewSymlinkPntr,
    NewSymlinkPntr->fso.trueNamePntr);
  if (ErrorCode != 0 || DPRINT_NOTIFICATIONS)
  {
    dprintf ("ARFS_symlink: B_ENTRY_CREATED notification for creating "
      "%s in %s, error code %d.\n",
      FSOAsString (&NewSymlinkPntr->fso), FSOAsString (&DirectoryPntr->fso),
      (int) ErrorCode);
  }

  AddFreshThingToStandardIndices (DiskVolumePntr, &NewSymlinkPntr->fso);

  ReturnCode = 0;
  goto NormalExit;

ErrorExit:
  if (NewSymlinkPntr != NULL)
    DeallocateAnyFSO (DiskVolumePntr, &NewSymlinkPntr->fso);
NormalExit:
#if DPRINT_NON_ERROR_MESSAGES
  dprintf ("ARFS_symlink exited, return code %d.\n", (int) ReturnCode);
#endif
  if (WriteLockedDirectory)
    WriteUnlock (&DirectoryPntr->fso);
  return ReturnCode;
}



/******************************************************************************
 * Make a hard link to a file/directory/thing.  Basically this puts it into the
 * given directory, using the given name, possibly giving it more than one
 * parent directory and multiple names.  Often it is a good idea to have a
 * reference counter in your node data so that a thing only gets deleted when
 * it has been removed from all directories it was in.  As you can expect, node
 * is your shadow vnode of the thing.  Remember to add it to the relevant
 * indices (the new name, and all the attributes if this is the first link -
 * can happen if someone deletes the file then relinks it).
 */

int ARFS_link (
  void *ns,
  void *dir,
  const char *name,
  void *node)
{
  DirectoryObjectPointer     DirectoryPntr = dir;
  DiskVolumePointer          DiskVolumePntr = ns;
  status_t                   ErrorCode;
  bool                       NewAGMSLinkAddedToDirectory = false;
  AGMSLinkObjectPointer      NewAGMSLinkPntr = NULL;
  status_t                   ReturnCode = -1;
  const char                *StringPntr;
  FileSystemObjectPointer    TargetFSOPntr = node;
  bool                       WriteLockedDirectory = false;
  bool                       WriteLockedTarget = false;

#if DPRINT_NON_ERROR_MESSAGES
  dprintf ("ARFS_link entered, ns:$%X, dir:$%X, name:\"%s\", node:$%X.\n",
    (unsigned int) ns, (unsigned int) dir, name, (unsigned int) node);
#endif

  if (name == NULL)
  {
    dprintf ("ARFS_link: invalid input NULL name.\n");
    goto ErrorExit;
  }

  if (strlen (name) >= NAME_MAX)
  {
    dprintf ("ARFS_link: invalid input name \"%s\" is too long, "
      "maximum is %d bytes total (including NUL at end).\n",
      name, NAME_MAX);
    ReturnCode = ENAMETOOLONG;
    goto ErrorExit;
  }

  /* Reject names consisting of all dots, like ".." or "...." or "", which we
  reserve for assorted kinds of parent directories. */

  for (StringPntr = name; *StringPntr != 0; StringPntr++)
    if (*StringPntr != '.')
      break;
  if (*StringPntr == 0)
  {
    dprintf ("ARFS_link: invalid input name \"%s\" is all periods, which "
      "are reserved for parent directory names.\n", name);
    ReturnCode = EINVAL;
    goto ErrorExit;
  }

  if (INVALID_DISK_VOLUME (DiskVolumePntr))
  {
    dprintf ("ARFS_link: invalid input disk volume pointer $%X.\n",
      (unsigned int) DiskVolumePntr);
    goto ErrorExit;
  }

  if (INVALID_FILE_SYSTEM_OBJECT (&DirectoryPntr->fso))
  {
    dprintf ("ARFS_link: invalid input dir pointer $%X.\n",
      (unsigned int) DirectoryPntr);
    goto ErrorExit;
  }

  if (INVALID_FILE_SYSTEM_OBJECT (TargetFSOPntr))
  {
    dprintf ("ARFS_link: invalid target file system object pointer $%X.\n",
      (unsigned int) TargetFSOPntr);
    goto ErrorExit;
  }

  if (DirectoryPntr->fso.objectType != FSO_TYPE_DIRECTORY)
  {
    dprintf ("ARFS_link: base %s isn't a directory object!\n",
      FSOAsString (&DirectoryPntr->fso));
    ReturnCode = ENOTDIR;
    goto ErrorExit;
  }

  ErrorCode = WriteLock (&DirectoryPntr->fso);
  if (ErrorCode != 0)
  {
    dprintf ("ARFS_link unable to get write lock for %s, code %d.\n",
      FSOAsString (&DirectoryPntr->fso), (int) ErrorCode);
    ReturnCode = ErrorCode;
    goto ErrorExit;
  }
  WriteLockedDirectory = true;

  /* Don't need to write lock the target if it is also the directory the link
  is being made in.  Links to their own directory aren't very useful, but they
  are there for completeness. */

  if (TargetFSOPntr != &DirectoryPntr->fso)
  {
    ErrorCode = WriteLock (TargetFSOPntr);
    if (ErrorCode != 0)
    {
      dprintf ("ARFS_link unable to get write lock for %s, code %d.\n",
        FSOAsString (TargetFSOPntr), (int) ErrorCode);
      ReturnCode = ErrorCode;
      goto ErrorExit;
    }
    WriteLockedTarget = true;
  }

  /* Fail if the user isn't allowed to write to this directory. */

  ErrorCode = PermissionAllowed (DiskVolumePntr, &DirectoryPntr->fso, W_OK);
  if (ErrorCode != 0)
  {
    dprintf ("ARFS_link: %s doesn't have write access permission "
      "for contents (need the w flag).\n", FSOAsString (&DirectoryPntr->fso));
    ReturnCode = ErrorCode;
    goto ErrorExit;
  }

  NewAGMSLinkPntr = AllocateAGMSLinkFSO (DiskVolumePntr, name, TargetFSOPntr);
  if (NewAGMSLinkPntr == NULL)
  {
    dprintf ("ARFS_link: unable to allocate new AGMS link FSO when "
      "making link \"%s\" in parent %s.\n",
      name, FSOAsString (&DirectoryPntr->fso));
    ReturnCode = ENOMEM;
    goto ErrorExit;
  }

  NewAGMSLinkPntr->fso.gid = DirectoryPntr->fso.gid;
  NewAGMSLinkPntr->fso.modeBits |=
    S_IRWXU | S_IRWXG | S_IRWXO; /* BFS uses all permission bits set. */

  /* Attempting to add it to the directory will detect duplicate names, as well
  as other error codes for the usual reasons. */

  ErrorCode = AddFreshFSOToDirectory (DiskVolumePntr,
    &NewAGMSLinkPntr->fso, DirectoryPntr);
  if (ErrorCode < 0)
  {
    dprintf ("ARFS_link: unable to add link name \"%s\" to parent %s, "
      "code %d.\n", name, FSOAsString (&DirectoryPntr->fso),
      (int) ErrorCode);
    ReturnCode = ErrorCode;
    goto ErrorExit;
  }
  ChangeFSODates (DiskVolumePntr, &DirectoryPntr->fso, DK_MODIFIED_FLAG, -1);
  NewAGMSLinkAddedToDirectory = true;

  if (!AddAGMSLinkReferenceToFSO (DiskVolumePntr,
  TargetFSOPntr, NewAGMSLinkPntr))
  {
    dprintf ("ARFS_link: Unable to add back reference to %s for %s.\n",
      FSOAsString (TargetFSOPntr), FSOAsString (&NewAGMSLinkPntr->fso));
    goto ErrorExit;
  }

#if DPRINT_NON_ERROR_MESSAGES
  dprintf ("ARFS_link: Created link to %s under name \"%s\" in %s.\n",
    FSOAsString (TargetFSOPntr), name,
    FSOAsString (&DirectoryPntr->fso));
#endif

  ErrorCode = notify_listener (B_ENTRY_CREATED, DiskVolumePntr->nameSpaceID,
    (unsigned int) DirectoryPntr, 0, (unsigned int) NewAGMSLinkPntr,
    NewAGMSLinkPntr->fso.trueNamePntr);
  if (ErrorCode != 0 || DPRINT_NOTIFICATIONS)
  {
    dprintf ("ARFS_link: B_ENTRY_CREATED notification for creating %s in %s, "
      "error code %d.\n",
      FSOAsString (&NewAGMSLinkPntr->fso), FSOAsString (&DirectoryPntr->fso),
      (int) ErrorCode);
  }

  AddFreshThingToStandardIndices (DiskVolumePntr, &NewAGMSLinkPntr->fso);

  NewAGMSLinkPntr = NULL; /* So it doesn't get mistakenly deallocated. */
  NewAGMSLinkAddedToDirectory = false;
  ReturnCode = 0;

ErrorExit:
#if DPRINT_NON_ERROR_MESSAGES
  dprintf ("ARFS_link exited, return code %d.\n", (int) ReturnCode);
#endif
  if (NewAGMSLinkAddedToDirectory)
    RemoveFSOFromDirectory (DiskVolumePntr,
    &NewAGMSLinkPntr->fso, DirectoryPntr);
  if (NewAGMSLinkPntr != NULL)
    DeallocateAnyFSO (DiskVolumePntr, &NewAGMSLinkPntr->fso);
  if (WriteLockedTarget)
    WriteUnlock (TargetFSOPntr);
  if (WriteLockedDirectory)
    WriteUnlock (&DirectoryPntr->fso);
  return ReturnCode;
}



/******************************************************************************
 * Rename the file (or other thing) from the old directory (given its shadow
 * vnode), with the old name, to a new directory (may be the same as the old)
 * with a new name.  Must be done atomically, either it succeeds or it returns
 * an error code and does nothing.  If the new name exists, it also deletes the
 * existing file in the new directory as well as doing the replacement.  This
 * is useful for replacing a file with a new version while ensuring that there
 * is always a file available there, as far as other threads can tell.  Avoid
 * cyclical loops - don't rename a directory to be one of its own descendants.
 * The POSIX standard says you can't replace a file with a directory and vice
 * versa, but BFS does allow that.  And of course update indices etc.
 */

int ARFS_rename (
  void *ns,
  void *olddir,
  const char *oldname,
  void *newdir,
  const char *newname)
{
  status_t          ReturnCode;
  RDWorkStateRecord State;

#if DPRINT_NON_ERROR_MESSAGES
  dprintf ("ARFS_rename entered, ns:$%X, olddir:$%X, oldname:\"%s\", "
    "newdir:$%X, newname:\"%s\".\n",
    (unsigned int) ns, (unsigned int) olddir, oldname,
    (unsigned int) newdir, newname);
#endif

  memset (&State, 0, sizeof (State));
  State.diskVolumePntr = ns;
  State.sourceDirectoryPntr = olddir;
  State.destinationDirectoryPntr = newdir;
  State.oldNamePntr = oldname;
  State.newNamePntr = newname;

  ReturnCode = RenameOrDeleteWithRetry (&State);

#if DPRINT_NON_ERROR_MESSAGES
  dprintf ("ARFS_rename exited, return code %d.\n", (int) ReturnCode);
#endif
  return ReturnCode;
}



/******************************************************************************
 * Remove the file or symbolic link (not directory) with the given name from
 * the given directory (and take that name out of the index if you allow
 * different names in different parent directories).  If it is listed in zero
 * directories after that (if you are using hard links it could be in several
 * directories), then also delete it from the file system.  Normally the final
 * delete is done by using get_vnode, taking it out of the remaining indices,
 * then calling the OS's remove_vnode() (confusingly named the same as your
 * remove_vnode () hook function) which sets the delete flag on the vnode, then
 * calling put_vnode.  When all the users of the vnode have called put_vnode,
 * the OS sees the delete flag and will call your remove_vnode to actually
 * deallocate the file (rather than using write_vnode), so all you should be
 * doing here is taking it out of the directory.  That's why the actual
 * deletion of the file itself isn't done here, as there may still be open file
 * handles using the file, even though it isn't listed in any directory.
 */

int ARFS_unlink (
  void *ns,
  void *dir,
  const char *name)
{
  status_t          ReturnCode;
  RDWorkStateRecord State;

#if DPRINT_NON_ERROR_MESSAGES
  dprintf ("ARFS_unlink entered, ns:$%X, dir:$%X, name:\"%s\".\n",
    (unsigned int) ns, (unsigned int) dir, name);
#endif

  memset (&State, 0, sizeof (State));
  State.diskVolumePntr = ns;
  State.destinationDirectoryPntr = dir;
  State.newNamePntr = name;
  State.rmdirNotUnlink = false;

  ReturnCode = RenameOrDeleteWithRetry (&State);

#if DPRINT_NON_ERROR_MESSAGES
  dprintf ("ARFS_unlink exited, return code %d.\n", (int) ReturnCode);
#endif
  return ReturnCode;
}



/******************************************************************************
 * Remove the directory with the given name from the given directory.  If the
 * directory isn't empty (or the named thing isn't a directory), return an
 * error code and do nothing.  Otherwise similar to unlink.
 */

int ARFS_rmdir (
  void *ns,
  void *dir,
  const char *name)
{
  status_t          ReturnCode;
  RDWorkStateRecord State;

#if DPRINT_NON_ERROR_MESSAGES
  dprintf ("ARFS_rmdir entered, ns:$%X, dir:$%X, name:\"%s\".\n",
    (unsigned int) ns, (unsigned int) dir, name);
#endif

  memset (&State, 0, sizeof (State));
  State.diskVolumePntr = ns;
  State.destinationDirectoryPntr = dir;
  State.newNamePntr = name;
  State.rmdirNotUnlink = true;

  ReturnCode = RenameOrDeleteWithRetry (&State);

#if DPRINT_NON_ERROR_MESSAGES
  dprintf ("ARFS_rmdir exited, return code %d.\n", (int) ReturnCode);
#endif
  return ReturnCode;
}



/******************************************************************************
 * Get the path string out of the given symbolic link and copy it into the
 * user's buffer, as much of it that fits.  This seems to duplicate some of the
 * functionality of Walk, but the default internal implementation just returns
 * an error rather than using Walk to get the info.  Presumably bufsize is
 * always specified (kernel bug if NULL), and on entry contains the size of the
 * user's buffer, and on exit the number of bytes required to hold the full
 * symbolic string.  Note that if the buffer is too small, it still returns the
 * number of bytes for the full string, but only writes the amount that fits.
 * buf points to the user's buffer (allowed to be NULL if size is zero).  The
 * string written to the buffer doesn't include a trailing NUL (and the size
 * returned reflects that).  Of course, the maximum buffer size needed will be
 * the maximum path length (1024 bytes in BeOS).
 */

int ARFS_readlink (
  void *ns,
  void *node,
  char *buf,
  size_t *bufsize)
{
  DiskVolumePointer         DiskVolumePntr = ns;
  status_t                  ErrorCode;
  bool                      IsFakeSymbolicLink = false;
  FileSystemObjectPointer   LinkFSOPntr = node;
  size_t                    ProvidedBufSize;
  bool                      ReadLockDone = false;
  status_t                  ReturnCode = -1;
  size_t                    ReturnBufSize = -1;
  char                     *SymbolicPathPntr;
  char                      TempPathString [PATH_MAX];

#if DPRINT_NON_ERROR_MESSAGES
  dprintf ("ARFS_readlink entered, ns:$%X, node:$%X, buf:$%X, bufsize:$%X, "
    "*bufsize:%ld.\n",
    (unsigned int) ns, (unsigned int) node,
    (unsigned int) buf, (unsigned int) bufsize,
    (bufsize != NULL) ? *bufsize : (size_t) -1);
#endif

  if (bufsize == NULL)
  {
    dprintf ("ARFS_readlink: NULL bufsize pointer, kernel bug!\n");
    goto ErrorExit;
  }
  ProvidedBufSize = *bufsize;

  if (ProvidedBufSize > 0 && buf == NULL)
  {
    dprintf ("ARFS_readlink: NULL user buffer, and size is not zero.\n");
    ReturnCode = EFAULT;
    goto ErrorExit;
  }

  if (INVALID_DISK_VOLUME (DiskVolumePntr))
  {
    dprintf ("ARFS_readlink: invalid input disk volume pointer $%X.\n",
      (unsigned int) DiskVolumePntr);
    goto ErrorExit;
  }

  /* This function presents symbolic links for real symbolic links, AGMS links
  and fake symbolic links. */

  if ((7 & (unsigned int) LinkFSOPntr) == 1) /* Low 3 bits == 1 for fake. */
  {
    IsFakeSymbolicLink = true;
    LinkFSOPntr = (void *) ((unsigned int) LinkFSOPntr - 1);
  }

  if (INVALID_FILE_SYSTEM_OBJECT (LinkFSOPntr))
  {
    dprintf ("ARFS_readlink: invalid input FSO pointer $%X.\n",
      (unsigned int) LinkFSOPntr);
    goto ErrorExit;
  }

  if (!IsFakeSymbolicLink)
  {
    ErrorCode = ReadLock (LinkFSOPntr, false);
    if (ErrorCode != 0)
    {
      dprintf ("ARFS_readlink unable to get read lock for %s.\n",
        FSOAsString (LinkFSOPntr));
      ReturnCode = ErrorCode;
      goto ErrorExit;
    }
    ReadLockDone = true;
  }

  /* Would check the permissions to see about getting info on this symlink here
  in the code, but that would normally be done in the parent directory, so we
  presume you are allowed to find out about the file if you got the vnode. */

  /* Now get the symbolic string. */

  if (IsFakeSymbolicLink)
  {
    ErrorCode = GetTruePath (DiskVolumePntr, LinkFSOPntr,
      TempPathString, sizeof (TempPathString));
    if (ErrorCode != 0)
    {
      dprintf ("ARFS_readlink: Unable to get true path to %s.\n",
        FSOAsString ((FileSystemObjectPointer) node));
      ReturnCode = ErrorCode;
      goto ErrorExit;
    }
    SymbolicPathPntr = TempPathString;
  }
  else if (LinkFSOPntr->objectType == FSO_TYPE_SYMBOLIC_LINK)
  {
    SymbolicPathPntr =
      ((SymbolicLinkObjectPointer) LinkFSOPntr)->symbolicPathPntr;
  }
  else if (LinkFSOPntr->objectType == FSO_TYPE_AGMS_LINK)
  {
    ErrorCode = GetTruePath (DiskVolumePntr,
      ((AGMSLinkObjectPointer) LinkFSOPntr)->targetFSOPntr,
      TempPathString, sizeof (TempPathString));
    if (ErrorCode != 0)
    {
      dprintf ("ARFS_readlink: Unable to get true path via %s to %s.\n",
        FSOAsString (LinkFSOPntr),
        FSOAsString (((AGMSLinkObjectPointer) LinkFSOPntr)->targetFSOPntr));
      ReturnCode = ErrorCode;
      goto ErrorExit;
    }
    SymbolicPathPntr = TempPathString;
  }
  else
  {
    dprintf ("ARFS_readlink: Meaningless to ask for the symbolic "
      "link from %s.\n", FSOAsString (LinkFSOPntr));
    goto ErrorExit;
  }

  ReturnBufSize = strlen (SymbolicPathPntr) /* note it doesn't include NUL */;
  if (ProvidedBufSize > 0 && ReturnBufSize > 0)
  {
    if (ProvidedBufSize >= ReturnBufSize)
      memcpy (buf, SymbolicPathPntr, ReturnBufSize);
    else
      memcpy (buf, SymbolicPathPntr, ProvidedBufSize);
  }

  ReturnCode = 0;

#if DPRINT_NON_ERROR_MESSAGES
  dprintf ("ARFS_readlink successful for %s, returning %ld as *bufsize, "
    "string \"%s\".\n",
    FSOAsString ((FileSystemObjectPointer) node),
    ReturnBufSize, SymbolicPathPntr);
#endif

ErrorExit:
#if DPRINT_NON_ERROR_MESSAGES
  dprintf ("ARFS_readlink exited, return code %d.\n", (int) ReturnCode);
#endif
  if (ReadLockDone)
    ReadUnlock (LinkFSOPntr);
  if (bufsize != NULL)
    *bufsize = ((ReturnCode == 0) ? ReturnBufSize : -1);
  return ReturnCode;
}



/******************************************************************************
 * Starts examining a directory.  The node pointer identifies the directory
 * (it's the 32 bit shadow vnode pointer your file system provided during
 * read_vnode).  The cookie holds a pointer to a pointer, which you set to
 * point to your private directory state data (usually you allocate it as part
 * of your opendir, containing a counter to keep track of where you are in the
 * directory) or some other 32 bit value (like an array index).  The OS will
 * keep the cookie as part of its open directory status, and pass it back to
 * you whenever the user wants more directory information (usually the user
 * calls the Posix functions opendir, readdir, closedir, rewinddir).
 * Incidentally, the Posix functions also seem to open the same directory as a
 * file (so that nobody deletes it while in use) and close it as part of
 * closedir.  So, it looks like you don't need to worry about the directory
 * disappearing while it's open, though the contents may change.  The function
 * returns the usual error codes.
 */

int ARFS_opendir (
  void *ns,
  void *node,
  void **cookie)
{
#if DPRINT_NON_ERROR_MESSAGES
  dprintf ("ARFS_opendir entered, NS:$%X, node:$%X.\n",
    (unsigned int) ns, (unsigned int) node);
#endif

  return GenericOpenDir (ns, node, cookie, GDV_PLAIN_DIRECTORY, NULL, NULL);
}



/******************************************************************************
 * Stop any further threads from accessing the directory.  Don't deallocate the
 * cookie just yet, there could be other threads busy reading at the moment.
 * In fact, you should delete semaphores or otherwise unblock any threads
 * waiting to read this directory via this cookie (and have them return with an
 * error code), and return an error for any further attempts to use the same
 * cookie.
 */

int ARFS_closedir (
  void *ns,
  void *node,
  void *cookie)
{
#if DPRINT_NON_ERROR_MESSAGES
  dprintf ("ARFS_closedir entered, NS:$%X, node:$%X, cookie:$%X.\n",
    (unsigned int) ns, (unsigned int) node, (unsigned int) cookie);
#endif

  return GenericCloseDir (ns, node, cookie);
}



/******************************************************************************
 * Sets the directory state back to the first thing in the directory, so the
 * user can re-read the directory from the beginning.
 */

int ARFS_rewinddir (
  void *ns,
  void *node,
  void *cookie)
{
#if DPRINT_NON_ERROR_MESSAGES
  dprintf ("ARFS_rewinddir entered, NS:$%X, node:$%X, cookie:$%X.\n",
    (unsigned int) ns, (unsigned int) node, (unsigned int) cookie);
#endif

  return GenericRewindDir (ns, node, cookie);
}



/******************************************************************************
 * Read info about one or more things (files or subdirs or ?) in the directory
 * and advance the directory state to the appropriate next thing in the
 * directory, so that the next readdir will continue where this one left off.
 * Keep in mind that the directory contents may change between readdir calls.
 * Note that the dirent structures returned in the buffer are variable sized
 * (the file name at the end of the structure is as big as needed, or
 * optionally padded larger), so use the buffer size in combination with the
 * input value from *num to decide how many can be read, or if you want to keep
 * it simple, just read one (that's what BFS currently does).  As noted by S&W,
 * you need to fill in more than just the name; the Tracker uses the
 * d_dev/d_ino fields for displaying icons.  You can also fill in the parent
 * info in d_pdev/d_pino if you want to (assuming that there is one parent -
 * currently unimplemented hardlinks would add more parents - but then the
 * directory you are reading would be the obvious parent to use).  You
 * shouldn't call get_vnode to set up the shadow vnode for things in the
 * directory, unless you need the shadow vnode for your own purposes (then you
 * must call put_vnode later to balance it out).  The d_reclen field in the
 * dirent should contain the length of the entire record, including the name
 * and trailing NUL and padding space (if you have any) up to the next record
 * (so you can add it to the start of the record pointer to get the next
 * record).  But the file system examples I've seen use the length of the name,
 * or the length of the name plus one, or other numbers.  Fortunately all those
 * examples only return 1 record at a time, so nobody needs their buggy
 * d_reclen.  You should return "." and ".." entries too, and the root
 * directory's ".." is best set to itself (the OS will ignore your ".." and use
 * the mount point directory's parent when walking, but just in case...).
 * Actually, you'll see the OS do a walk and rstat on the ".." name after your
 * readdir returns.  Returns an error code if it fails.  Also you should set
 * *num to the number of entries it read, which can include zero (don't return
 * an ENOENT error when the end of the directory is reached) or less than the
 * requested number.  Perhaps return EINVAL/B_BAD_VALUE if the buffer is too
 * small to hold at least one file's dirent.
 *
 * Here's the famous dirent structure, note the open ended array at the end:
 *
 *   typedef struct dirent {
 *     dev_t          d_dev;
 *     dev_t          d_pdev;
 *     ino_t          d_ino;
 *     ino_t          d_pino;
 *     unsigned short d_reclen;
 *     char           d_name[1];
 *   } dirent_t;
 */

int ARFS_readdir (
  void *ns,
  void *node,
  void *cookie,
  long *num,
  struct dirent *buf,
  size_t bufsize)
{
#if DPRINT_NON_ERROR_MESSAGES
  dprintf ("ARFS_readdir entered, "
    "NS:$%X, node:$%X, cookie:$%X, *num:%d, buf:$%X, bufsize:%d.\n",
    (unsigned int) ns, (unsigned int) node, (unsigned int) cookie,
    (int) *num, (unsigned int) buf, (int) bufsize);
#endif

  return GenericReadDir (ns, node, cookie, num, buf, bufsize);
}



/******************************************************************************
 * The last thread has finished reading the directory, you can now really
 * close it.  Deallocate the cookie and otherwise clean up here.
 */

int ARFS_free_dircookie (
  void *ns,
  void *node,
  void *cookie)
{
#if DPRINT_NON_ERROR_MESSAGES
  dprintf ("ARFS_free_dircookie entered, NS:$%X, node:$%X, cookie:$%X.\n",
    (unsigned int) ns, (unsigned int) node, (unsigned int) cookie);
#endif

  return GenericFreeDirCookie (ns, node, cookie);
}



/******************************************************************************
 * Opens an existing file or directory or other thing, using your shadow
 * portion of the vnode, which read_node earlier was used to find (thus the
 * file/dir/thing must exist for read_node to work, see the create() function
 * for making new files).  The open mode flags in omode specify the type of
 * open (B_READ_ONLY/O_RDONLY, B_WRITE_ONLY/O_WRONLY, B_READ_WRITE/O_RDWR,
 * B_FAIL_IF_EXISTS/O_EXCL, B_CREATE_FILE/O_CREAT, B_ERASE_FILE/O_TRUNC,
 * B_OPEN_AT_END/O_APPEND).  You shouldn't see the O_CREAT flag; the create()
 * function will get called by the OS instead of open().  If you can't open the
 * file in the requested mode, return an error, B_PERMISSION_DENIED/EACCES is
 * good for most situations.  As usual, allocate a cookie structure with state
 * information about the file (like the open mode - so you can detect attempts
 * to write to a read-only file, though the FSIL sometimes does it for you) and
 * return a pointer to it in *cookie.  You may also want to do your own
 * security checking here too, if secure_vnode isn't working.  Note that
 * opening directories isn't too useful (the FreeBSD man pages say you should
 * get an error if opening a directory for write access but BeOS allows this,
 * and BNode/BEntry seems to reattempt an open with read access if write
 * fails), but user programs will need to do it in order to use fsync,
 * attribute operations and some other file handle operations (though read or
 * write should fail when used on non-files) or merely to get exclusive access
 * to the directory, or to prevent it from being deleted.  Same thing goes for
 * symbolic links and other non-file objects (the BeOS "Tracker" file browser
 * tries to open everything, possibly to look for icon attributes, and locks up
 * for some operations if it can't open the thing).
 */

int ARFS_open (
  void *ns,
  void *node,
  int omode,
  void **cookie)
{
  DiskVolumePointer       DiskVolumePntr = ns;
  status_t                ErrorCode;
  FileCookiePointer       FileCookiePntr = NULL;
  FileSystemObjectPointer FSOPntr = node;
  bool                    GotReadLock = false;
  bool                    GotWriteLock = false;
  bool                    IsFakeSymbolicLink = false;
  status_t                ReturnCode = -1;
  char                    TempString [B_OS_NAME_LENGTH];

#if DPRINT_NON_ERROR_MESSAGES
  dprintf ("ARFS_open entered, NS:$%X, node:$%X, openmode:$%X.\n",
    (int) ns, (int) node, omode);
#endif

  if (INVALID_DISK_VOLUME (DiskVolumePntr))
  {
    dprintf ("ARFS_open: invalid input disk volume pointer $%X.\n",
      (unsigned int) DiskVolumePntr);
    goto ErrorExit;
  }

  if ((7 & (unsigned int) FSOPntr) == 1)
  {
    IsFakeSymbolicLink = true;
    FSOPntr = (void *) (((unsigned int) FSOPntr) - 1);
  }

  if (INVALID_FILE_SYSTEM_OBJECT (FSOPntr))
  {
    dprintf ("ARFS_open: invalid input FSO $%X.\n",
      (unsigned int) node);
    goto ErrorExit;
  }

  /* Get a lock on the thing FSO (but not on fake objects, which are
  unlockable).  If we are truncating the file, then we also need to modify the
  file info and thus use a write-lock. */

  if (!IsFakeSymbolicLink)
  {
    if (omode & O_TRUNC)
    {
      ErrorCode = WriteLock (FSOPntr);
      if (ErrorCode != 0)
      {
        dprintf ("ARFS_open: unable to get write lock for %s.\n",
          FSOAsString (FSOPntr));
        ReturnCode = ErrorCode;
        goto ErrorExit;
      }
      GotWriteLock = true;
    }
    else /* Ordinary opening, file FSO doesn't need modification. */
    {
      ErrorCode = ReadLock (FSOPntr, false);
      if (ErrorCode != 0)
      {
        dprintf ("ARFS_open unable to get read lock for %s.\n",
          FSOAsString (FSOPntr));
        ReturnCode = ErrorCode;
        goto ErrorExit;
      }
      GotReadLock = true;
    }
  }

  /* Check if the caller wants to open the thing for writing.  Fake objects
  don't have permission bits, so just assume that they're always writeable.*/

  if (!IsFakeSymbolicLink &&
  ((omode & O_RWMASK) != O_RDONLY || (omode & O_TRUNC)))
  {
    /* User wants to write to this thing, or possibly its attributes. */

    ErrorCode = PermissionAllowed (DiskVolumePntr, FSOPntr, W_OK);
    if (ErrorCode != 0)
    {
      dprintf ("ARFS_open %s doesn't have required write access permission "
        "for contents (need the w flag).\n",
        FSOAsString (FSOPntr));
      ReturnCode = ErrorCode;
      goto ErrorExit;
    }
  }

  /* If opening requires read access, check for permission.  Fake objects don't
  have permission bits, so just assume that they're always readable. */

  if (!IsFakeSymbolicLink && ((omode & O_RWMASK) != O_WRONLY))
  {
    ErrorCode = PermissionAllowed (DiskVolumePntr, FSOPntr, R_OK);
    if (ErrorCode != 0)
    {
      dprintf ("ARFS_open %s doesn't have required read access permission "
        "for contents (need the r flag).\n",
        FSOAsString (FSOPntr));
      ReturnCode = ErrorCode;
      goto ErrorExit;
    }
  }

  /* Make the cookie. */

  FileCookiePntr = mallocDVP (DiskVolumePntr, sizeof (FileCookieRecord));
  if (FileCookiePntr == NULL)
  {
    dprintf ("ARFS_open: unable to create cookie when opening %s, "
      "out of memory.\n", FSOAsString (FSOPntr));
    ReturnCode = ENOMEM;
    goto ErrorExit;
  }

  /* Set up the mutual exclusion semaphore in the cookie. */

  strcpy (TempString, "FileCookie ");
  strncat (TempString, FSOPntr->trueNamePntr,
    B_OS_NAME_LENGTH - strlen (TempString) - 1);
  FileCookiePntr->mutexSemID = create_sem (1 /* Initial count */, TempString);
  if (FileCookiePntr->mutexSemID < 0)
  {
    dprintf ("ARFS_open: Unable to create semaphore for cookie "
      "for %s, error code %d!\n",
      FSOAsString (FSOPntr), (int) FileCookiePntr->mutexSemID);
    ReturnCode = FileCookiePntr->mutexSemID;
    goto ErrorExit;
  }

  /* Save the true file pointer (including the low order flag bits for fake
  items) as part of the cookie. */

  FileCookiePntr->fsoPntr = node;
  FileCookiePntr->fileIsOpen = true;

  /* Note that only files can have TRUE values for the read and/or write
  allowed flags.  However, you can open a directory, symbolic link, fake
  symbolic link, AGMS link etc, though not actually read or write from it (so
  that a file handle can reference a directory or other thing, which Tracker
  seems to want to do).  In the non-file cases the cookie has both the read and
  write flags turned off. */

  if (IsFakeSymbolicLink || FSOPntr->objectType != FSO_TYPE_FILE)
  {
    FileCookiePntr->readingAllowed = false;
    FileCookiePntr->writingAllowed = false;
    FileCookiePntr->appendMode = false;
  }
  else /* A real genuine data file. */
  {
    FileCookiePntr->readingAllowed = ((omode & O_RWMASK) != O_WRONLY);
    FileCookiePntr->writingAllowed = ((omode & O_RWMASK) != O_RDONLY);
    FileCookiePntr->appendMode = ((omode & O_APPEND) != 0);
  }

  /* Do the file truncation. */

  if ((omode & O_TRUNC) && !IsFakeSymbolicLink &&
  FSOPntr->objectType == FSO_TYPE_FILE)
  {
    DeallocateFileContents (DiskVolumePntr, (FileObjectPointer) FSOPntr);
    ((FileObjectPointer) FSOPntr)->sizeChangedButNotificationNotSent = true;
  }

  /* Successful, wrap it up. */

#if DPRINT_NON_ERROR_MESSAGES
  dprintf ("ARFS_open: Successfully opened %s.  "
    "Cookie $%X, readable:%d, writeable:%d, append:%d.\n",
    FSOAsString ((FileSystemObjectPointer) node),
    (int) FileCookiePntr, (int) FileCookiePntr->readingAllowed,
    (int) FileCookiePntr->writingAllowed, (int) FileCookiePntr->appendMode);
#endif

  *cookie = FileCookiePntr;
  ReturnCode = 0;
  goto NormalExit;

ErrorExit:
  if (FileCookiePntr != NULL)
  {
    if (FileCookiePntr->mutexSemID >= 0)
      delete_sem (FileCookiePntr->mutexSemID);
    freeDVP (DiskVolumePntr, FileCookiePntr);
  }
NormalExit:
#if DPRINT_NON_ERROR_MESSAGES
  dprintf ("ARFS_open exited, return code %d.\n", (int) ReturnCode);
#endif
  if (GotReadLock)
    ReadUnlock (FSOPntr);
  if (GotWriteLock)
    WriteUnlock (FSOPntr);
  return ReturnCode;
}



/******************************************************************************
 * Prepare to close the file.  Other threads may be busy reading it at this
 * moment, so don't deallocate the cookie yet or deallocate stuff the other
 * threads might need.  Still, you might want to bump off other threads which
 * are busy waiting for IO (delete their semaphore or something, forcing them
 * to return an error code to the caller).  Further read/writes should return
 * an error like B_FILE_ERROR/EBADF.
 */

int ARFS_close (void *ns, void *node, void *cookie)
{
  bool                    CookieLockDone = false;
  DiskVolumePointer       DiskVolumePntr = ns;
  status_t                ErrorCode;
  FileCookiePointer       FileCookiePntr = cookie;
  FileSystemObjectPointer FSOPntr = node;
  bool                    IsFakeSymbolicLink = false;
  status_t                ReturnCode = -1;

#if DPRINT_NON_ERROR_MESSAGES
  dprintf ("ARFS_close entered, NS:$%X, node:$%X, cookie:$%X.\n",
    (unsigned int) ns, (unsigned int) node, (unsigned int) cookie);
#endif

  if (INVALID_DISK_VOLUME (DiskVolumePntr))
  {
    dprintf ("ARFS_close: invalid input disk volume pointer $%X.\n",
      (unsigned int) DiskVolumePntr);
    goto ErrorExit;
  }

  if ((7 & (unsigned int) FSOPntr) == 1)
    IsFakeSymbolicLink = true; /* Don't care if target is valid or not. */
  else
  {
    if (INVALID_FILE_SYSTEM_OBJECT (FSOPntr))
    {
      dprintf ("ARFS_close: invalid input FSO pointer $%X.\n",
        (unsigned int) FSOPntr);
      goto ErrorExit;
    }
  }

  if (FileCookiePntr == NULL || FileCookiePntr->fsoPntr != FSOPntr)
  {
    dprintf ("ARFS_close: %s doesn't match cookie!\n",
      FSOAsString (FSOPntr));
    ReturnCode = EINVAL;
    goto ErrorExit;
  }

  /* Get the cookie lock here to make sure no other threads using the same file
  handle change the open flag. */

  ErrorCode = acquire_sem_etc (FileCookiePntr->mutexSemID, 1 /* Count */,
    B_CAN_INTERRUPT /* flags */, 0 /* unused timeout value */);
  if (ErrorCode != 0)
  {
    dprintf ("ARFS_close: failed to aquire cookie "
      "semaphore for %s, semaphoreID is %d, error %d.\n",
      FSOAsString (FSOPntr), (int) FileCookiePntr->mutexSemID,
      (int) ErrorCode);
    ReturnCode = ErrorCode;
    goto ErrorExit;
  }
  CookieLockDone = true;

  if (!FileCookiePntr->fileIsOpen)
  {
    dprintf ("ARFS_close: %s was closed earlier, no further operations "
      "allowed.\n", FSOAsString (FSOPntr));
    ReturnCode = EBADF;
    goto ErrorExit;
  }

  /* Finally, do the actual work. */

  FileCookiePntr->fileIsOpen = false;
  ReturnCode = 0;

#if DPRINT_NON_ERROR_MESSAGES
  dprintf ("ARFS_close successful for %s, cookie:$%X.\n",
    FSOAsString (FSOPntr), (unsigned int) FileCookiePntr);
#endif

ErrorExit:
#if DPRINT_NON_ERROR_MESSAGES
  dprintf ("ARFS_close exited, return code %d.\n", (int) ReturnCode);
#endif
  if (CookieLockDone)
    release_sem (FileCookiePntr->mutexSemID);
  return ReturnCode;
}



/******************************************************************************
 * Nobody is using the file now, it's safe to really close it.  Send the final
 * file size change notification message if needed.  Deallocate your cookie
 * too.  Single threaded with respect to the particular cookie.
 */

int ARFS_free_cookie (void *ns, void *node, void *cookie)
{
  bigtime_t               CurrentTime;
  DiskVolumePointer       DiskVolumePntr = ns;
  status_t                ErrorCode;
  FileCookiePointer       FileCookiePntr = cookie;
  FileObjectPointer       FilePntr;
  FileSystemObjectPointer FSOPntr = node;
  bool                    GotWriteLock = false;
  bool                    IsFakeSymbolicLink = false;
  status_t                ReturnCode = -1;

#if DPRINT_NON_ERROR_MESSAGES
  dprintf ("ARFS_free_cookie entered, NS:$%08X, node:%d.\n",
    (int) ns, (int) node);
#endif

  if (INVALID_DISK_VOLUME (DiskVolumePntr))
  {
    dprintf ("ARFS_free_cookie: invalid input disk volume pointer $%X.\n",
      (unsigned int) DiskVolumePntr);
    goto ErrorExit;
  }

  if ((7 & (unsigned int) FSOPntr) == 1)
    IsFakeSymbolicLink = true; /* Don't care if target is valid or not. */
  else
  {
    if (INVALID_FILE_SYSTEM_OBJECT (FSOPntr))
    {
      dprintf ("ARFS_free_cookie: invalid input FSO pointer $%X.\n",
        (unsigned int) FSOPntr);
      goto ErrorExit;
    }
  }

  if (FileCookiePntr == NULL || FileCookiePntr->fsoPntr != FSOPntr)
  {
    dprintf ("ARFS_free_cookie: %s doesn't match cookie!\n",
      FSOAsString (FSOPntr));
    ReturnCode = EINVAL;
    goto ErrorExit;
  }

  if (FileCookiePntr->fileIsOpen)
  {
    /* Just issue a warning, and continue to free the cookie. */
    dprintf ("ARFS_free_cookie: Bug!  %s is still open, should be closed!\n",
      FSOAsString (FSOPntr));
  }

  /* Send the file size change notification when closing a data file, and do a
  final index and date stamp update.  Note that we only write lock the file if
  we have to, so we can safely update the size notification fields, without
  some other thread changing the size or otherwise adding a slight
  inconsistency. */

  if (!IsFakeSymbolicLink && FSOPntr->objectType == FSO_TYPE_FILE)
  {
    FilePntr = (FileObjectPointer) FSOPntr;
    ErrorCode = WriteLock (FSOPntr);
    if (ErrorCode != 0)
    {
      dprintf ("ARFS_free_cookie: Unable to get write lock on %s, so won't "
        "send B_STAT_CHANGED notification or update \"size\" index.\n",
        FSOAsString (FSOPntr));
    }
    else
    {
      GotWriteLock = true;
      if (FilePntr->fileActuallyRead)
      {
        ChangeFSODates (DiskVolumePntr, FSOPntr, DK_ACCESS_FLAG, -1);
        FilePntr->fileActuallyRead = false;
      }
      if (FilePntr->sizeChangedButNotificationNotSent)
      {
        CurrentTime = system_time ();
        UpdateSizeIndexWithNewFileSize (DiskVolumePntr, FilePntr);
        ErrorCode = notify_listener (B_STAT_CHANGED,
          DiskVolumePntr->nameSpaceID, 0, 0, (unsigned int) FilePntr, NULL);
        if (ErrorCode != 0 || DPRINT_NOTIFICATIONS)
        {
          dprintf ("ARFS_free_cookie: B_STAT_CHANGED notification "
            "for %s, error code %d.\n",
            FSOAsString (FSOPntr), (int) ErrorCode);
        }
        FilePntr->timeOfLastSizeChangeNotification = CurrentTime;
        FilePntr->sizeChangedButNotificationNotSent = false;
      }
    }
  }

#if DPRINT_NON_ERROR_MESSAGES
  dprintf ("ARFS_free_cookie successful for %s, cookie:$%X.\n",
    FSOAsString (FSOPntr), (unsigned int) FileCookiePntr);
#endif

  /* Finally, deallocate the cookie. */

  delete_sem (FileCookiePntr->mutexSemID);
  FileCookiePntr->mutexSemID = -1;
  FileCookiePntr->fsoPntr = NULL; /* Make it an invalid cookie. */
  freeDVP (DiskVolumePntr, FileCookiePntr);

  ReturnCode = 0;

ErrorExit:
#if DPRINT_NON_ERROR_MESSAGES
  dprintf ("ARFS_free_cookie exited, return code %d.\n", (int) ReturnCode);
#endif
  if (GotWriteLock)
    WriteUnlock (FSOPntr);
  return ReturnCode;
}



/******************************************************************************
 * Read data from the file.  If the thing is a directory, return an error.  As
 * the OS is tracking individual open files and their positions, you get the
 * position to read as one of the arguments.  The OS also tracks read only and
 * write only modes, and will reject user attempts to do inappropriate reads or
 * writes, so you never see them.  The rest of the arguments are as you would
 * expect, with the cookie from the open, the node being your shadow vnode, and
 * so on.  See the user level read() for an explanation of how it works.  If
 * the user opened the file with the append option, read at the provided
 * position, not the end.  Returns a negative error code on failure, or 0 if
 * successful, plus it returns the number of bytes read in *len (can be zero if
 * attempting to read at end of file).
 */

int ARFS_read (
  void *ns,
  void *node,
  void *cookie,
  off_t pos,
  void *buf,
  size_t *len)
{
  size_t                  AmountRead = 0;
  uint32                  BufferSizeRemaining;
  bool                    CookieLockDone = false;
  DiskVolumePointer       DiskVolumePntr = ns;
  status_t                ErrorCode;
  FileCookiePointer       FileCookiePntr = cookie;
  uint32                  FileOffset;
  FileObjectPointer       FilePntr;
  FileSystemObjectPointer FSOPntr = node;
  bool                    GotReadLock = false;
  bool                    IsFakeSymbolicLink = false;
  status_t                ReturnCode = -1;
  uint32                  StorageAvailable;
  uint32                  StorageCopied;
  uint8                  *StoragePntr;
  uint8                  *UserPntr;

#if DPRINT_READ_AND_WRITE_OPERATIONS
  dprintf ("ARFS_read entered, NS:$%X, node:$%X, cookie:$%X, "
    "pos:%Ld, buf:$%X, *len:%u.\n",
    (int) ns, (int) node, (int) cookie,
    pos, (unsigned int) buf, (unsigned int) *len);
#endif

  if (INVALID_DISK_VOLUME (DiskVolumePntr))
  {
    dprintf ("ARFS_read: invalid input disk volume pointer $%X.\n",
      (unsigned int) DiskVolumePntr);
    goto ErrorExit;
  }

  if ((7 & (unsigned int) FSOPntr) == 1)
    IsFakeSymbolicLink = true; /* Don't care if target is valid or not. */
  else
  {
    if (INVALID_FILE_SYSTEM_OBJECT (FSOPntr))
    {
      dprintf ("ARFS_read: invalid input FSO pointer $%X.\n",
        (unsigned int) FSOPntr);
      goto ErrorExit;
    }
  }

  if (FileCookiePntr == NULL || FileCookiePntr->fsoPntr != FSOPntr)
  {
    dprintf ("ARFS_read: %s doesn't match cookie!\n",
      FSOAsString (FSOPntr));
    ReturnCode = EINVAL;
    goto ErrorExit;
  }

  /* Get the cookie lock here to make sure no other threads using the same file
  handle change the open flag. */

  ErrorCode = acquire_sem_etc (FileCookiePntr->mutexSemID, 1 /* Count */,
    B_CAN_INTERRUPT /* flags */, 0 /* unused timeout value */);
  if (ErrorCode != 0)
  {
    dprintf ("ARFS_read: failed to aquire cookie "
      "semaphore for %s, semaphoreID is %d, error %d.\n",
      FSOAsString (FSOPntr), (int) FileCookiePntr->mutexSemID,
      (int) ErrorCode);
    ReturnCode = ErrorCode;
    goto ErrorExit;
  }
  CookieLockDone = true;

  if (!FileCookiePntr->fileIsOpen)
  {
    dprintf ("ARFS_read: %s has been closed previously!\n",
      FSOAsString (FSOPntr));
    goto ErrorExit;
  }

  /* Also need to readlock the file so that file size and contents don't change
  while reading it. */

  if (!IsFakeSymbolicLink)
  {
    ErrorCode = ReadLock (FSOPntr, false);
    if (ErrorCode != 0)
    {
      dprintf ("ARFS_read: Unable to get read lock on %s!\n",
        FSOAsString (FSOPntr));
      ReturnCode = ErrorCode;
      goto ErrorExit;
    }
    GotReadLock = true;
  }

  if (!FileCookiePntr->readingAllowed)
  {
    if (IsFakeSymbolicLink || FSOPntr->objectType != FSO_TYPE_FILE)
    {
      dprintf ("ARFS_read: %s is not a file, can't read from it.\n",
        FSOAsString (FSOPntr));
      ReturnCode = EISDIR;
    }
    else
    {
      dprintf ("ARFS_read: %s was not opened for reading.\n",
        FSOAsString (FSOPntr));
      ReturnCode = EACCES;
    }
    goto ErrorExit;
  }

  if (FSOPntr->objectType != FSO_TYPE_FILE)
  {
    dprintf ("ARFS_read: Bug!  %s is not a file.\n", FSOAsString (FSOPntr));
    goto ErrorExit;
  }

  /* Have a real file, now read the requested data.  Return no error and zero
  bytes read if at end of file. */

  ReturnCode = 0;
  FilePntr = (FileObjectPointer) FSOPntr;
  if (pos < 0) pos = 0; /* Fix up invalid negative file position. */
  if (pos > 0x70000000) pos = 0x70000000; /* Avoid number overflow. */
  if (pos < FilePntr->fileSize) /* If not past end of file, then read data. */
  {
    FileOffset = pos;
    BufferSizeRemaining = *len;
    UserPntr = buf;
    if (BufferSizeRemaining > FilePntr->fileSize - FileOffset)
      BufferSizeRemaining = FilePntr->fileSize - FileOffset;

    while (BufferSizeRemaining > 0)
    {
      StoragePntr =
        GetFileBufferAtOffset (FilePntr, FileOffset, &StorageAvailable);
      if (StoragePntr == NULL)
      {
        dprintf ("ARFS_read: Error while reading data for %s, at offset %u.\n",
          FSOAsString (&FilePntr->fso), (unsigned int) FileOffset);
        if (AmountRead == 0)
          ReturnCode = EIO; /* Allow partial reads, error if nothing read. */
        break;
      }
      StorageCopied = StorageAvailable;
      if (StorageCopied > BufferSizeRemaining)
        StorageCopied = BufferSizeRemaining;

      memcpy (UserPntr, StoragePntr, StorageCopied);

      UserPntr += StorageCopied;
      FileOffset += StorageCopied;
      AmountRead += StorageCopied;
      BufferSizeRemaining -= StorageCopied;
    }
  }
  FilePntr->fileActuallyRead = true;

#if DPRINT_READ_AND_WRITE_OPERATIONS
  dprintf ("ARFS_read: Read %d bytes from %s.\n", (int) AmountRead,
    FSOAsString (&FilePntr->fso));
#endif

ErrorExit:
  *len = AmountRead;
#if DPRINT_READ_AND_WRITE_OPERATIONS
  dprintf ("ARFS_read exited, return code %d.\n", (int) ReturnCode);
#endif
  if (GotReadLock)
    ReadUnlock (FSOPntr);
  if (CookieLockDone)
    release_sem (FileCookiePntr->mutexSemID);
  return ReturnCode;
}



/******************************************************************************
 * Reserve some space out of the system wide maximum total file size limit.
 * Returns the amount requested, or less if less is available.  The amount
 * reserved is taken off the total file size, so if you don't use it all, put
 * back the file size which you haven't used (just atomic_add it to the
 * totalFileSize; you can't use this function with negative values).  Returns 0
 * if it fails.
 */

static uint32 ReserveTotalFileSize (
  DiskVolumePointer DiskVolumePntr,
  uint32 AmountRequested)
{
  int32 AmountRequestedAsInt32;
  int32 AmountReserved;
  int32 NewTotal;

  AmountRequestedAsInt32 = AmountRequested;
  if (AmountRequestedAsInt32 <= 0)
    return 0; /* Number too big for an int32 or is zero. */

  while (true)
  {
    AmountReserved = DiskVolumePntr->maxTotalFileSize -
      DiskVolumePntr->totalFileSize;
    if (AmountReserved > AmountRequestedAsInt32)
      AmountReserved = AmountRequestedAsInt32;
    if (AmountReserved < 0)
      AmountReserved = 0;
    if (AmountReserved == 0)
      break; /* No space currently available, or it is overbooked. */

    NewTotal = atomic_add (&DiskVolumePntr->totalFileSize, AmountReserved);
    if (NewTotal <= DiskVolumePntr->maxTotalFileSize)
      break; /* Successfully got what we wanted. */

    /* Oops, went over the maximum (some other thread snuck in and grabbed some
    space while we were calculating).  Try again. */

    atomic_add (&DiskVolumePntr->totalFileSize, -AmountReserved);
  }

  return AmountReserved;
}



/******************************************************************************
 * Write data to a file.  See read.  Note that for append mode (file opened
 * with O_APPEND), ignore the position the OS gives you (which is an incorrect
 * value anyways - the OS thinks you are writing at the start of the file) and
 * always write at the end (use suitable thread locking so that the active
 * thread gets the real current end of the file position and finishes the write
 * before other theads try to write).  This is what BFS does.  Oddly enough,
 * the read position that the OS uses in append mode is also off, starting at
 * zero (the beginning) and advancing for both writes and reads, making it
 * useless since the actual written data is at the end of the file.  Users will
 * have to do lseek after writing to avoid confusion about the read position.
 * Anyway, this way O_APPEND guarantees that the user's writes will be at the
 * end, even if several threads are writing at the same time (useful for log
 * files where the order in the file should correspond to the to the
 * chronological order of writes by the threads).
 */

int ARFS_write (
  void *ns,
  void *node,
  void *cookie,
  off_t pos,
  const void *buf,
  size_t *len)
{
  uint32                  AmountReserved;
  size_t                  AmountWritten = 0;
  uint32                  BufferSizeRemaining;
  bool                    CookieLockDone = false;
  bigtime_t               CurrentTime;
  DiskVolumePointer       DiskVolumePntr = ns;
  status_t                ErrorCode;
  uint32                  EstimatedNewFileSize;
  FileCookiePointer       FileCookiePntr = cookie;
  uint32                  FileOffset;
  FileObjectPointer       FilePntr;
  FileSystemObjectPointer FSOPntr = node;
  bool                    GotWriteLock = false;
  bool                    IsFakeSymbolicLink = false;
  status_t                ReturnCode = -1;
  uint32                  StorageAvailable;
  uint32                  StorageCopied;
  uint8                  *StoragePntr;
  const uint8            *UserPntr;

#if DPRINT_READ_AND_WRITE_OPERATIONS
  dprintf ("ARFS_write entered, NS:$%X, node:$%X, cookie:$%X, "
    "pos:%Ld, buf:$%X, *len:%u.\n",
    (int) ns, (int) node, (int) cookie,
    pos, (unsigned int) buf, (unsigned int) *len);
#endif

  if (INVALID_DISK_VOLUME (DiskVolumePntr))
  {
    dprintf ("ARFS_write: invalid input disk volume pointer $%X.\n",
      (unsigned int) DiskVolumePntr);
    goto ErrorExit;
  }

  if ((7 & (unsigned int) FSOPntr) == 1)
    IsFakeSymbolicLink = true; /* Don't care if target is valid or not. */
  else
  {
    if (INVALID_FILE_SYSTEM_OBJECT (FSOPntr))
    {
      dprintf ("ARFS_write: invalid input FSO pointer $%X.\n",
        (unsigned int) FSOPntr);
      goto ErrorExit;
    }
  }

  if (FileCookiePntr == NULL || FileCookiePntr->fsoPntr != FSOPntr)
  {
    dprintf ("ARFS_write: %s doesn't match cookie!\n",
      FSOAsString (FSOPntr));
    ReturnCode = EINVAL;
    goto ErrorExit;
  }

  /* Get the cookie lock here to make sure no other threads using the same file
  handle change the open flag. */

  ErrorCode = acquire_sem_etc (FileCookiePntr->mutexSemID, 1 /* Count */,
    B_CAN_INTERRUPT /* flags */, 0 /* unused timeout value */);
  if (ErrorCode != 0)
  {
    dprintf ("ARFS_write: failed to aquire cookie "
      "semaphore for %s, semaphoreID is %d, error %d.\n",
      FSOAsString (FSOPntr), (int) FileCookiePntr->mutexSemID,
      (int) ErrorCode);
    ReturnCode = ErrorCode;
    goto ErrorExit;
  }
  CookieLockDone = true;

  if (!FileCookiePntr->fileIsOpen)
  {
    dprintf ("ARFS_write: %s has been closed previously!\n",
      FSOAsString (FSOPntr));
    goto ErrorExit;
  }

  /* Also need to write lock the file itself, for obvious reasons. */

  if (!IsFakeSymbolicLink)
  {
    ErrorCode = WriteLock (FSOPntr);
    if (ErrorCode != 0)
    {
      dprintf ("ARFS_write: Unable to get write lock on %s!\n",
        FSOAsString (FSOPntr));
      ReturnCode = ErrorCode;
      goto ErrorExit;
    }
    GotWriteLock = true;
  }

  if (!FileCookiePntr->writingAllowed)
  {
    if (IsFakeSymbolicLink || FSOPntr->objectType != FSO_TYPE_FILE)
    {
      dprintf ("ARFS_write: %s is not a file, can't write to it.\n",
        FSOAsString (FSOPntr));
      ReturnCode = EISDIR;
    }
    else
    {
      dprintf ("ARFS_write: %s was not opened for writing.\n",
        FSOAsString (FSOPntr));
      ReturnCode = EACCES;
    }
    goto ErrorExit;
  }

  if (FSOPntr->objectType != FSO_TYPE_FILE)
  {
    dprintf ("ARFS_write: Bug!  %s is not a file.\n",
      FSOAsString (FSOPntr));
    goto ErrorExit;
  }

  /* Have a real file, now write the requested data. */

  ReturnCode = 0;
  FilePntr = (FileObjectPointer) FSOPntr;
  if (FileCookiePntr->appendMode)
    FileOffset = FilePntr->fileSize;
  else
  {
    if (pos < 0) pos = 0; /* Fix up negative positions. */
    if (pos > 0x70000000) pos = 0x70000000; /* Avoid number overflow. */
    FileOffset = pos;
  }
  BufferSizeRemaining = *len;
  UserPntr = buf;

  /* Reserve some file size space from the system total file size.  This limit
  prevents files from using up more memory than is available, and crashing the
  kernel. */

  EstimatedNewFileSize = FileOffset + BufferSizeRemaining;
  if (EstimatedNewFileSize > FilePntr->fileSize)
  {
    AmountReserved = ReserveTotalFileSize (DiskVolumePntr,
      EstimatedNewFileSize - FilePntr->fileSize);
    if (AmountReserved < EstimatedNewFileSize - FilePntr->fileSize)
    {
      BufferSizeRemaining -=
        EstimatedNewFileSize - FilePntr->fileSize - AmountReserved;
      dprintf ("ARFS_write: Had to reduce requested amount to write from %u "
        "to %lu because the maximum total file size limit would be exceeded "
        "(was only able to reserve %lu bytes, not the full %lu needed).  "
        "For writing at offset %lu in %s.\n",
        (unsigned int) *len, BufferSizeRemaining,
        AmountReserved, EstimatedNewFileSize - FilePntr->fileSize,
        FileOffset, FSOAsString (&FilePntr->fso));
      EstimatedNewFileSize = FileOffset + BufferSizeRemaining;
      if (BufferSizeRemaining <= 0)
      {
        ReturnCode = ENOSPC;
        goto ErrorExit;
      }
    }
  }
  else /* Not growing the file size. */
    AmountReserved = 0;

  /* Allocate more space in the file if there isn't any.  Add enough to get to
  the projected end of the file.  If the amount is less than 2K then add 2K (in
  case the user is writing to the file one byte at a time).  It may partially
  or fully fail and allocate less.  The surplus will be trimmed off when the
  file is closed and the VNode is flushed. */

  if (FileOffset + BufferSizeRemaining > FilePntr->totalAllocated)
  {
    if (FileOffset + BufferSizeRemaining < FilePntr->totalAllocated + 2048)
      EnlargeFileTotalAllocation (DiskVolumePntr, FilePntr,
      FilePntr->totalAllocated + 2048);
    else /* Just add on the minimum amount needed. */
      EnlargeFileTotalAllocation (DiskVolumePntr, FilePntr,
      FileOffset + BufferSizeRemaining);
  }

  /* Write the user's data into as much space as we can find. */

  while (BufferSizeRemaining > 0)
  {
    StoragePntr =
      GetFileBufferAtOffset (FilePntr, FileOffset, &StorageAvailable);
    if (StoragePntr == NULL)
    {
      dprintf ("ARFS_write: No buffer available to write data for %s, "
        "at offset %u (usually means out of memory).\n",
        FSOAsString (&FilePntr->fso), (unsigned int) FileOffset);
      if (AmountWritten == 0)
        ReturnCode = ENOSPC; /* Allow partial writes, error if nothing done. */
      break;
    }
    StorageCopied = StorageAvailable;
    if (StorageCopied > BufferSizeRemaining)
      StorageCopied = BufferSizeRemaining;

    memcpy (StoragePntr, UserPntr, StorageCopied);

    UserPntr += StorageCopied;
    FileOffset += StorageCopied;
    if (FileOffset > FilePntr->fileSize)
    {
      FilePntr->fileSize = FileOffset;

      /* File size is changing.  Notify the user once in a while. */

      CurrentTime = system_time ();
      if (CurrentTime >= FilePntr->timeOfLastSizeChangeNotification +
      FILE_SIZE_CHANGE_NOTIFICATION_TIME_PERIOD)
      {
        UpdateSizeIndexWithNewFileSize (DiskVolumePntr, FilePntr);
        ErrorCode = notify_listener (B_STAT_CHANGED,
          DiskVolumePntr->nameSpaceID, 0, 0, (unsigned int) FilePntr, NULL);
        if (ErrorCode != 0 || DPRINT_NOTIFICATIONS)
        {
          dprintf ("ARFS_write: B_STAT_CHANGED notification for %s, "
            "error code %d.\n",
            FSOAsString (FSOPntr), (int) ErrorCode);
        }
        FilePntr->timeOfLastSizeChangeNotification = CurrentTime;
        FilePntr->sizeChangedButNotificationNotSent = false;
      }
      else /* Not yet time for a notification. */
        FilePntr->sizeChangedButNotificationNotSent = true;
    }
    AmountWritten += StorageCopied;
    BufferSizeRemaining -= StorageCopied;
  }

  if (AmountReserved > 0) /* If the file size was growing. */
  {
    if (FilePntr->fileSize > EstimatedNewFileSize)
    {
      dprintf ("ARFS_write: Bug!  File size is %lu, expected it to only "
        "grow to %lu, for %s.\n", FilePntr->fileSize, EstimatedNewFileSize,
        FSOAsString (&FilePntr->fso));
      goto ErrorExit;
    }

    /* If we stopped early (probably due to memory allocation failing), return
    extra reserved space to the global total. */

    if (FilePntr->fileSize < EstimatedNewFileSize)
    {
      dprintf ("ARFS_write: Returning %lu bytes of oddly unused file size "
        "(had reserved %lu) to global total for %s.\n",
        EstimatedNewFileSize - FilePntr->fileSize, AmountReserved,
        FSOAsString (&FilePntr->fso));
      atomic_add (&DiskVolumePntr->totalFileSize,
        - (int32) (EstimatedNewFileSize - FilePntr->fileSize));
    }
  }

#if DPRINT_READ_AND_WRITE_OPERATIONS
  dprintf ("ARFS_write: Wrote %d bytes to %s, file size: %d, allocated: %d.\n",
    (int) AmountWritten, FSOAsString (&FilePntr->fso),
    (int) FilePntr->fileSize, (int) FilePntr->totalAllocated);
#endif

ErrorExit:
  *len = AmountWritten;
#if DPRINT_READ_AND_WRITE_OPERATIONS
  dprintf ("ARFS_write exited, return code %d.\n", (int) ReturnCode);
#endif
  if (GotWriteLock)
    WriteUnlock (FSOPntr);
  if (CookieLockDone)
    release_sem (FileCookiePntr->mutexSemID);
  return ReturnCode;
}



/******************************************************************************
 * Do special operations to the file.  Things like changing a file from
 * cachable to uncachable, or finding out internal information on the file
 * (like the list of blocks used).  It's up to you to make your own special
 * control codes and define what they do.  The device drivers have a standard
 * set of ioctl codes, so it's best to avoid them (use numbers over
 * B_DEVICE_OP_CODES_END/9999).  You may want to pass the ioctl operations down
 * to the hardware device you are using, perhaps only if the file is the root
 * node.
 */

int ARFS_ioctl (
  void *ns,
  void *node,
  void *cookie,
  int cmd,
  void *buf,
  size_t len)
{
  DiskVolumePointer       DiskVolumePntr = ns;
  FileCookiePointer       FileCookiePntr = cookie;
  FileSystemObjectPointer FSOPntr = node;
  ARFSLLinkIoctlPointer   LinkDataPntr;
  status_t                ReturnCode = -1;

#if DPRINT_NON_ERROR_MESSAGES
  dprintf ("ARFS_ioctl Entered, ns:$%X, node:$%X, cookie:$%X, "
    "cmd:%d, buf:$%X, len:%d.\n",
    (unsigned int) ns, (unsigned int) node, (unsigned int) cookie,
    cmd, (unsigned int) buf, (unsigned int) len);
#endif

  if (INVALID_DISK_VOLUME (DiskVolumePntr))
  {
    dprintf ("ARFS_ioctl: Invalid input disk volume pointer $%X.\n",
      (unsigned int) DiskVolumePntr);
    goto ErrorExit;
  }

  if (INVALID_FILE_SYSTEM_OBJECT (FSOPntr))
  {
    dprintf ("ARFS_ioctl: Invalid input FSO %s.\n",
      FSOAsString (FSOPntr));
    goto ErrorExit;
  }

  if (FileCookiePntr == NULL || FileCookiePntr->fsoPntr != FSOPntr)
  {
    dprintf ("ARFS_ioctl: %s doesn't match cookie!\n",
      FSOAsString (FSOPntr));
    ReturnCode = EINVAL;
    goto ErrorExit;
  }

  switch (cmd)
  {
    case ARFS_IOCTL_LLINK:
      LinkDataPntr = buf;
      if (LinkDataPntr == NULL || LinkDataPntr->newName == NULL ||
      LinkDataPntr->structureSize != sizeof (ARFSLLinkIoctlRecord))
      {
        dprintf ("ARFS_ioctl: llink command arguments aren't valid, for %s.\n",
          FSOAsString (FSOPntr));
        ReturnCode = EINVAL;
        goto ErrorExit;
      }
#if DPRINT_NON_ERROR_MESSAGES
      dprintf ("ARFS_ioctl Starting llink operation.\n");
#endif
      ReturnCode = ARFS_link (ns, FSOPntr, LinkDataPntr->newName,
        (FileSystemObjectPointer) (unsigned int) LinkDataPntr->targetVNodeID);
      break;

    case ARFS_IOCTL_GET_MAX_TOTAL_FILE_SIZE:
      * (int32 *) buf = DiskVolumePntr->maxTotalFileSize;
      ReturnCode = 0;
#if DPRINT_NON_ERROR_MESSAGES
      dprintf ("ARFS_ioctl Returning max total file size of %ld.\n",
        DiskVolumePntr->maxTotalFileSize);
#endif
      break;

    case ARFS_IOCTL_SET_MAX_TOTAL_FILE_SIZE:
      if (* (int32 *) buf <= 0)
      {
        dprintf ("ARFS_ioctl: ARFS_IOCTL_SET_MAX_TOTAL_FILE_SIZE "
          "value %ld isn't positive.\n", * (int32 *) buf);
        ReturnCode = EINVAL;
        goto ErrorExit;
      }
      DiskVolumePntr->maxTotalFileSize = * (int32 *) buf;
      ReturnCode = 0;
#if DPRINT_NON_ERROR_MESSAGES
      dprintf ("ARFS_ioctl Set max total file size to %ld.\n",
        DiskVolumePntr->maxTotalFileSize);
#endif
      break;

    default:
      dprintf ("ARFS_ioctl: Unrecognized control code %d ($%X) "
        "applied to %s.\n", cmd, cmd, FSOAsString (FSOPntr));
      ReturnCode = B_DEV_INVALID_IOCTL;
      break;
  }

ErrorExit:
#if DPRINT_NON_ERROR_MESSAGES
  dprintf ("ARFS_ioctl Exited, return code %d.\n", (int) ReturnCode);
#endif
  return ReturnCode;
}



/******************************************************************************
 * Reads file status information.  Things like owner, date modified, size, and
 * whether this is a file or directory.  See stat.h for details.  For POSIX
 * symbolic links (because there is no inode, just a directory entry), the
 * permissions are those of the directory it is in, though it seems that all
 * rwx flags are set when listed with ls under Linux and all rwx flags are
 * clear under BeOS's devfs and set under BeOS's BFS.  I don't know if the
 * links field includes open file references, but I expect it does not, and it
 * just counts the number of directories that the file is in.  Remember to set
 * the st_mode flags to also identify whether the item is a file or directory,
 * otherwise it won't work!
 */

int ARFS_rstat (void *ns, void *node, struct stat *stat)
{
  DiskVolumePointer       DiskVolumePntr = ns;
  status_t                ErrorCode;
  AGMSLinkObjectPointer  *ExtraLinkPntrPntr;
  FileObjectPointer       FilePntr;
  FileSystemObjectPointer FSOPntr = node;
  bool                    IsFakeSymbolicLink = false;
  bool                    ReadLockDone = false;
  status_t                ReturnCode = -1;
  bool                    WriteLockDone = false;

#if DPRINT_NON_ERROR_MESSAGES
  dprintf ("ARFS_rstat entered, NS:$%X, node:$%X.\n",
    (unsigned int) ns, (unsigned int) node);
#endif

  if (stat == NULL)
  {
    dprintf ("ARFS_rstat: NULL stat pointer provided by caller.\n");
    goto ErrorExit;
  }

  if (INVALID_DISK_VOLUME (DiskVolumePntr))
  {
    dprintf ("ARFS_rstat: invalid input disk volume pointer $%X.\n",
      (unsigned int) DiskVolumePntr);
    goto ErrorExit;
  }

  if ((7 & (unsigned int) FSOPntr) == 1)
  {
    IsFakeSymbolicLink = true;
    FSOPntr = (void *) (((unsigned int) FSOPntr) - 1);
  }

  if (INVALID_FILE_SYSTEM_OBJECT (FSOPntr))
  {
    dprintf ("ARFS_rstat: invalid input FSO pointer $%X.\n",
      (unsigned int) node);
    goto ErrorExit;
  }

  /* Can't read-lock the fake symbolic link or its target because we don't know
  that the target is currently loaded.  Would have to do a get_vnode etc. */

  if (!IsFakeSymbolicLink)
  {
    ErrorCode = ReadLock (FSOPntr, false);
    if (ErrorCode != 0)
    {
      dprintf ("ARFS_rstat unable to get read lock for %s.\n",
        FSOAsString (FSOPntr));
      ReturnCode = ErrorCode;
      goto ErrorExit;
    }
    ReadLockDone = true;

    /* Update the file size visible to the user to be the actual size, so odd
    things like doing a rstat after writing some data gives the expected
    result.  Also updates the "size" index to match and sends the notification.
    Use the old size if locking fails. */

    if (FSOPntr->objectType == FSO_TYPE_FILE)
    {
      FilePntr = (FileObjectPointer) FSOPntr;
      if (FilePntr->fileSizeInIndex != FilePntr->fileSize &&
      ConvertReadLockToWriteLock (FSOPntr) == 0)
      {
        ReadLockDone = false;
        WriteLockDone = true;
        UpdateSizeIndexWithNewFileSize (DiskVolumePntr, FilePntr);
        ErrorCode = notify_listener (B_STAT_CHANGED,
          DiskVolumePntr->nameSpaceID, 0, 0, (unsigned int) FSOPntr, NULL);
        if (ErrorCode != 0 || DPRINT_NOTIFICATIONS)
        {
          dprintf ("ARFS_rstat: B_STAT_CHANGED notification "
            "for %s, error code %d.\n",
            FSOAsString (FSOPntr), (int) ErrorCode);
        }
        FilePntr->timeOfLastSizeChangeNotification = system_time ();
        FilePntr->sizeChangedButNotificationNotSent = false;
      }
    }
  }

  /* Would check the permissions to see about getting info on this file here in
  the code, but that would normally be done in the parent directory, so we
  presume you are allowed to find out about the file if you got the vnode. */

  /* Start filling in fields, leave unused ones at zero. */

  memset (stat, 0, sizeof (struct stat));
  stat->st_dev = DiskVolumePntr->nameSpaceID;
  stat->st_ino = (unsigned int) node; /* Preserve lower 3 flag bits. */

  if (IsFakeSymbolicLink)
    stat->st_mode = S_IFLNK /* Symbolic link */ | S_IRWXU | S_IRWXG | S_IRWXO;
  else /* An ordinary file system item, don't have to make it up. */
    stat->st_mode = FSOPntr->modeBits;

  if (DiskVolumePntr->readOnly) /* Remove "w" bits from the file mode. */
    stat->st_mode &= ~(S_IWUSR | S_IWGRP | S_IWOTH);

  /* Number of hard links is equal to the number of parent directories, we
  won't include open file references (so you could have zero links if a file is
  deleted but not closed yet).  Note that the root directory normally doesn't
  have any parent directories (but the user could add some via hard links). */

  if (IsFakeSymbolicLink)
    stat->st_nlink = 0;
  else
  {
    if (FSOPntr->mainParentDirectory != NULL)
      stat->st_nlink++;
    if (FSOPntr->parentLinkArray != NULL)
    {
      for (ExtraLinkPntrPntr = FSOPntr->parentLinkArray;
      *ExtraLinkPntrPntr != NULL;
      ExtraLinkPntrPntr++)
        stat->st_nlink++;
    }
  }

  stat->st_uid = FSOPntr->uid;
  stat->st_gid = FSOPntr->gid;
  stat->st_size = GetUserFileSize (
    (FileSystemObjectPointer) node /* include fake symbolic links */);
  stat->st_blksize = 256 * 1024; /* Prefered block size for read/write ops. */
  stat->st_atime = FSOPntr->atime / 1000000;
  stat->st_mtime = FSOPntr->mtime / 1000000;
  stat->st_ctime = FSOPntr->ctime / 1000000;
  stat->st_crtime = FSOPntr->crtime / 1000000;

  ReturnCode = 0;

#if DPRINT_NON_ERROR_MESSAGES
  dprintf ("ARFS_rstat successful for %s.\n",
    FSOAsString ((FileSystemObjectPointer) node));
#endif

ErrorExit:
#if DPRINT_NON_ERROR_MESSAGES
  dprintf ("ARFS_rstat exited, return code %d.\n", (int) ReturnCode);
#endif
  if (ReadLockDone)
    ReadUnlock (FSOPntr);
  if (WriteLockDone)
    WriteUnlock (FSOPntr);
  return ReturnCode;
}



/******************************************************************************
 * Write the file status.  This is used to change some of the properties of the
 * file.  The mask specifies which fields to change, which can be a bitwise
 * combination of WSTAT_MODE, WSTAT_UID, WSTAT_GID, WSTAT_SIZE, WSTAT_ATIME,
 * WSTAT_MTIME, WSTAT_CRTIME).  That implies you can truncate a file by setting
 * the size smaller, or make it larger (contents may vary, it's up to the user
 * to overwrite it with zeroes if they don't like what's already in the disk
 * sectors, so if you are paranoid, overwrite your files with zeroes before
 * deleting them).  Of course, a few security checks before changing the file's
 * owner / group / mode bits could be appropriate here.
 */

int ARFS_wstat (
  void *ns,
  void *node,
  struct stat *statBuf,
  long mask)
{
  uint32                  AmountReserved;
  bool                    ChangesMade = false;
  DiskVolumePointer       DiskVolumePntr = ns;
  status_t                ErrorCode;
  FileObjectPointer       FilePntr = NULL;
  FileSystemObjectPointer FSOPntr = node;
  bool                    IsRootUser;
  uint32                  RequestedSize;
  status_t                ReturnCode = -1;
  uid_t                   UserID;
  bool                    WriteLockDone = false;

#if DPRINT_NON_ERROR_MESSAGES
  dprintf ("ARFS_wstat entered, NS:$%X, node:$%X, stat:$%X, mask:$%X.\n",
    (unsigned int) ns, (unsigned int) node,
    (unsigned int) statBuf, (unsigned int) mask);
#endif

  if (INVALID_DISK_VOLUME (DiskVolumePntr))
  {
    dprintf ("ARFS_wstat: invalid input disk volume pointer $%X.\n",
      (unsigned int) DiskVolumePntr);
    goto ErrorExit;
  }

  if ((7 & (unsigned int) FSOPntr) != 0)
  {
    dprintf ("ARFS_wstat: Fake objects like %s "
      "don't have settable stat info.\n", FSOAsString (FSOPntr));
    ReturnCode = EBADF;
    goto ErrorExit;
  }

  if (INVALID_FILE_SYSTEM_OBJECT (FSOPntr))
  {
    dprintf ("ARFS_wstat: invalid input FSO pointer $%X.\n",
      (unsigned int) FSOPntr);
    goto ErrorExit;
  }

  if (statBuf == NULL)
  {
    dprintf ("ARFS_wstat: NULL stat buffer pointer.\n");
    goto ErrorExit;
  }

  if (DiskVolumePntr->readOnly)
  {
    dprintf ("ARFS_wstat: can't change %s info on a read-only file system.\n",
      FSOAsString (FSOPntr));
    ReturnCode = EROFS;
    goto ErrorExit;
  }

  ErrorCode = WriteLock (FSOPntr);
  if (ErrorCode != 0)
  {
    dprintf ("ARFS_wstat unable to get write lock for %s.\n",
      FSOAsString (FSOPntr));
    ReturnCode = ErrorCode;
    goto ErrorExit;
  }
  WriteLockDone = true;

  UserID = geteuid ();
  IsRootUser = (UserID == 0);

  if (mask & WSTAT_MODE)
  {
    /* Have to be the owner or root to change mode bits. */

    if (!IsRootUser && UserID != FSOPntr->uid)
    {
      dprintf ("ARFS_wstat: Need to be owner or root to change mode of %s.\n",
        FSOAsString (FSOPntr));
      ReturnCode = EPERM;
      goto ErrorExit;
    }

    FSOPntr->modeBits &= ~S_IUMSK;
    FSOPntr->modeBits |= (statBuf->st_mode & S_IUMSK);
    ChangesMade = true;

#if DPRINT_NON_ERROR_MESSAGES
    dprintf ("ARFS_wstat: Changed mode bits of %s.\n", FSOAsString (FSOPntr));
#endif
  }

  if (mask & WSTAT_UID)
  {
    /* Have to be root to change the owner of a file. */

    if (!IsRootUser)
    {
      dprintf ("ARFS_wstat: Need to be root to change owner of %s.\n",
        FSOAsString (FSOPntr));
      ReturnCode = EPERM;
      goto ErrorExit;
    }

    FSOPntr->uid = statBuf->st_uid;
    ChangesMade = true;

#if DPRINT_NON_ERROR_MESSAGES
    dprintf ("ARFS_wstat: Changed owner of %s.\n", FSOAsString (FSOPntr));
#endif
  }

  if (mask & WSTAT_GID)
  {
    /* Have to be owner or root to change group, and at that change group to
    one which the owner is in.  Note: not supported under BeOS. */

    if (!IsRootUser && UserID != FSOPntr->uid)
    {
      dprintf ("ARFS_wstat: Need to be owner or root to change group of %s.\n",
        FSOAsString (FSOPntr));
      ReturnCode = EPERM;
      goto ErrorExit;
    }

    FSOPntr->gid = statBuf->st_gid;
    ChangesMade = true;

#if DPRINT_NON_ERROR_MESSAGES
    dprintf ("ARFS_wstat: Changed group of %s.\n", FSOAsString (FSOPntr));
#endif
  }

  if (mask & WSTAT_SIZE)
  {
    /* Have to have write permission to change file size. */

    ErrorCode = PermissionAllowed (DiskVolumePntr, FSOPntr, W_OK);
    if (ErrorCode != 0)
    {
      dprintf ("ARFS_wstat: Need to have write permission to change "
        "size of %s.\n", FSOAsString (FSOPntr));
      ReturnCode = ErrorCode;
      goto ErrorExit;
    }

    /* Also, it has to be a file! */

    if (FSOPntr->objectType != FSO_TYPE_FILE)
    {
      dprintf ("ARFS_wstat: Can only change size of files, got %s.\n",
        FSOAsString (FSOPntr));
      ReturnCode = EISDIR; /* Or a symbolic link. */
      goto ErrorExit;
    }

    FilePntr = (FileObjectPointer) FSOPntr;

    /* Convert from 64 bit sizes to 32 bit, and at that under BeOS, the most
    memory we can allocate is around 800MB, so cut it off at a large positive
    number that won't make malloc blow up. */

    if (statBuf->st_size <= 0)
      RequestedSize = 0;
    else if (statBuf->st_size < 0x70000000)
      RequestedSize = statBuf->st_size;
    else
      RequestedSize = 0x70000000; /* Avoid number overflow. */

    if (RequestedSize < FilePntr->fileSize)
    {
      /* Shrink the file.  May need to add a block to force the packer to work
      (it only runs if there is a blocksCollection). */

      atomic_add (&DiskVolumePntr->totalFileSize,
        RequestedSize - FilePntr->fileSize);
      FilePntr->fileSize = RequestedSize;
      if (FilePntr->blocksCollection == NULL)
        AddMemoryToFile (DiskVolumePntr, FilePntr, 100);
      PackFileMemory (DiskVolumePntr, FilePntr);
    }
    else if (RequestedSize > FilePntr->fileSize)
    {
      /* Enlarge the file. */

      AmountReserved = ReserveTotalFileSize (DiskVolumePntr,
        RequestedSize - FilePntr->fileSize);
      if (AmountReserved != RequestedSize - FilePntr->fileSize)
      {
        dprintf ("ARFS_wstat: Unable to reserve space to grow file by %lu "
          "bytes (only got %lu) when changing size from %lu to %lu.  "
          "For %s.\n", RequestedSize - FilePntr->fileSize,
          AmountReserved, FilePntr->fileSize, RequestedSize,
          FSOAsString (FSOPntr));
        atomic_add (&DiskVolumePntr->totalFileSize, -AmountReserved);
        ReturnCode = ENOSPC;
        goto ErrorExit;
      }

      if (!EnlargeFileTotalAllocation (DiskVolumePntr, FilePntr,
      RequestedSize))
      {
        dprintf ("ARFS_wstat: Unable to allocate enough space to change size "
          "of %s to %d.\n", FSOAsString (FSOPntr), (int) RequestedSize);
        PackFileMemory (DiskVolumePntr, FilePntr); /* Get rid of partials. */
        atomic_add (&DiskVolumePntr->totalFileSize, -AmountReserved);
        ReturnCode = ENOSPC;
        goto ErrorExit;
      }
      FilePntr->fileSize = RequestedSize;
    }
    ChangesMade = true;

#if DPRINT_NON_ERROR_MESSAGES
    dprintf ("ARFS_wstat: Changed size of %s to %d.\n",
      FSOAsString (FSOPntr), (int) FilePntr->fileSize);
#endif
  }

  if (mask & WSTAT_ATIME)
  {
    /* Have to be the owner or root to change time. */

    if (!IsRootUser && UserID != FSOPntr->uid)
    {
      dprintf ("ARFS_wstat: Need to be owner or root to change access "
        "time of %s.\n", FSOAsString (FSOPntr));
      ReturnCode = EPERM;
      goto ErrorExit;
    }

    ChangeFSODates (DiskVolumePntr, FSOPntr, DK_ACCESS_FLAG,
      statBuf->st_atime * 1000000LL);
    ChangesMade = true;

#if DPRINT_NON_ERROR_MESSAGES
    dprintf ("ARFS_wstat: Changed access time of %s.\n",
      FSOAsString (FSOPntr));
#endif
  }

  if (mask & WSTAT_MTIME)
  {
    /* Have to be the owner or root to change time. */

    if (!IsRootUser && UserID != FSOPntr->uid)
    {
      dprintf ("ARFS_wstat: Need to be owner or root to change modification "
        "time of %s.\n", FSOAsString (FSOPntr));
      ReturnCode = EPERM;
      goto ErrorExit;
    }

    ChangeFSODates (DiskVolumePntr, FSOPntr, DK_MODIFIED_FLAG,
      statBuf->st_mtime * 1000000LL);
    ChangesMade = true;

#if DPRINT_NON_ERROR_MESSAGES
    dprintf ("ARFS_wstat: Changed modification time of %s.\n",
      FSOAsString (FSOPntr));
#endif
  }

  if (mask & WSTAT_CRTIME)
  {
    /* Have to be the owner or root to change time. */

    if (!IsRootUser && UserID != FSOPntr->uid)
    {
      dprintf ("ARFS_wstat: Need to be owner or root to change creation "
        "time of %s.\n", FSOAsString (FSOPntr));
      ReturnCode = EPERM;
      goto ErrorExit;
    }

    ChangeFSODates (DiskVolumePntr, FSOPntr, DK_CREATED_FLAG,
      statBuf->st_crtime * 1000000LL);
    ChangesMade = true;

#if DPRINT_NON_ERROR_MESSAGES
    dprintf ("ARFS_wstat: Changed creation time of %s.\n",
      FSOAsString (FSOPntr));
#endif
  }

  ReturnCode = 0;

ErrorExit:
  if (ChangesMade)
  {
    ErrorCode = notify_listener (B_STAT_CHANGED, DiskVolumePntr->nameSpaceID,
    0, 0, (unsigned int) FSOPntr, NULL);
    if (ErrorCode != 0 || DPRINT_NOTIFICATIONS)
    {
      dprintf ("ARFS_wstat: B_STAT_CHANGED notification for %s, "
        "error code %d.\n",
        FSOAsString (FSOPntr), (int) ErrorCode);
    }

    /* For data files, sending the notification also counts as sending the
    size changed notification. */

    if (FSOPntr->objectType == FSO_TYPE_FILE)
    {
      FilePntr = (FileObjectPointer) FSOPntr;
      UpdateSizeIndexWithNewFileSize (DiskVolumePntr, FilePntr);
      FilePntr->timeOfLastSizeChangeNotification = system_time ();
      FilePntr->sizeChangedButNotificationNotSent = false;
    }
  }

#if DPRINT_NON_ERROR_MESSAGES
  dprintf ("ARFS_wstat exited, return code %d.\n", (int) ReturnCode);
#endif
  if (WriteLockDone)
    WriteUnlock (FSOPntr);
  return ReturnCode;
}



/******************************************************************************
 * Possibly used for formatting a disk with an empty file system.  I can't
 * find a user equivalent call to trigger this callback.  I even tried mount()
 * and cycled through all the flag bits.  Guess it's not implemented in BeOS.
 */

int ARFS_initialize (
  const char *devname,
  void *parms,
  size_t len)
{
  status_t ReturnCode = -1;

#if DPRINT_NON_ERROR_MESSAGES
  dprintf ("ARFS_initialize entered, devname:\"%s\", parms:$%X, len:%ld.\n",
    devname, (int) parms, len);
#endif

  kernel_debugger ("\nARFS_initialize: "
    "How did you get to this secret function?\n"
    "Please tell me how, send mail to agmsmith@rogers.com.\n"
    "Use the \"cont\" command to get out of the kernel debugger.\n");

#if DPRINT_NON_ERROR_MESSAGES
  dprintf ("ARFS_initialize exited, return code %d.\n", (int) ReturnCode);
#endif
  return ReturnCode;
}



/******************************************************************************
 * Convert problematic characters into escape codes (RemoveBackslashEscapeCodes
 * has a list of the codes).  The resulting text is easy to edit in vi, or a
 * UTF-8 editor.  Returns 0 if successful, else an error code (usually because
 * of NULL pointers, or a too small output buffer).
 *
 * Input and output have to be different strings, since the output text is
 * usually larger than the input text (output buffer must be at least 6 bytes
 * long to hold the worst case escape code).  You can also have a NULL output
 * string pointer if you just wish to count up the amount of space the output
 * would take.  NUL bytes in the input string get converted to escape codes in
 * the output, which will never have NUL bytes (you'll have to add one if you
 * need it).  On input, source and destination lengths describe the sizes of
 * the buffers.  On output, they describe the number of bytes actually used.
 * Typically you should repeatedly call this function until the returned
 * DestinationLength is zero.  Initialise your state variable to zero for the
 * first iteration.  If you are not doing multiple iterations, you can pass in
 * a NULL StatePntr instead.  Internally the state is used for counting the
 * number of characters output since the last line break - an invisible line
 * break escape sequence (backslash followed by line feed) is inserted after
 * around 78 characters are output.  If you use TRUE for Encode8BitCharacters
 * then all bytes with the high bit set are output as numeric escape codes
 * (useful for arbitrary binary data).  If it is FALSE then they are output as
 * themselves (useful for UTF-8 strings).
 */

static status_t AddBackslashEscapeCodes (
  const char *SourceString,
  int *SourceLengthPntr,
  char *DestinationString,
  int *DestinationLengthPntr,
  int32 *StatePntr,
  bool Encode8BitCharacters)
{
  int                  InputChar;
  const unsigned char *InputPntr;
  int                  InputRemaining;
  bool                 OutputAllowed;
  unsigned char       *OutputPntr;
  int                  OutputRemaining;
  int                  OutputStringLength;
  const unsigned char *OutputString;
  unsigned char        TempString [6];
  int                  WrapCount;

  if (SourceString == NULL ||
  SourceLengthPntr == NULL || DestinationLengthPntr == NULL ||
  *SourceLengthPntr < 0 || *DestinationLengthPntr < 6)
  {
    dprintf ("ARFS:AddBackslashEscapeCodes: Called with invalid inputs.\n");
    return EINVAL;
  }

  InputRemaining = *SourceLengthPntr;
  InputPntr = SourceString;
  OutputRemaining = *DestinationLengthPntr;
  OutputPntr = DestinationString;
  OutputAllowed = (DestinationString != NULL);
  WrapCount = 0;
  if (StatePntr != NULL)
    WrapCount = *StatePntr;

  while (InputRemaining > 0)
  {
    InputChar = *InputPntr;
    switch (InputChar)
    {
      case 0x07: OutputString = "\\a"; OutputStringLength = 2; break;
      case 0x08: OutputString = "\\b"; OutputStringLength = 2; break;
      case 0x09: OutputString = "\\t"; OutputStringLength = 2; break;
      case 0x0A: OutputString = "\\n"; OutputStringLength = 2; break;
      case 0x0B: OutputString = "\\v"; OutputStringLength = 2; break;
      case 0x0C: OutputString = "\\f"; OutputStringLength = 2; break;
      case 0x0D: OutputString = "\\r"; OutputStringLength = 2; break;
      case '\\': OutputString = "\\\\"; OutputStringLength = 2; break;
      default:
        if (InputChar < 32 || (InputChar >= 128 && Encode8BitCharacters))
        {
          /* Use an octal number for this unprintable character. */

          if (InputRemaining >= 2 &&
          (InputPntr [1] < '0' || InputPntr [1] > '7'))
          {
            /* Can remove leading zeroes - the next letter isn't an octal digit
            so that won't cause confusion.  No octal in kernel, so can't use:
            sprintf (TempString, "\\%o", InputChar); */

            TempString [0] = '\\';
            if (InputChar < 8)
            {
              TempString [1] = InputChar + '0';
              OutputStringLength = 2;
            }
            else if (InputChar < 8 * 8)
            {
              TempString [1] = (InputChar >> 3) + '0';
              TempString [2] = (InputChar & 7) + '0';
              OutputStringLength = 3;
            }
            else
            {
              TempString [1] = (InputChar >> 6) + '0';
              TempString [2] = ((InputChar >> 3) & 7) + '0';
              TempString [3] = (InputChar & 7) + '0';
              OutputStringLength = 4;
            }
          }
          else /* Need the full 3 digit number to avoid ambiguity. */
          {
            /* Equivalent of: sprintf (TempString, "\\%03o", InputChar); */
            TempString [0] = '\\';
            TempString [1] = (InputChar >> 6) + '0';
            TempString [2] = ((InputChar >> 3) & 7) + '0';
            TempString [3] = (InputChar & 7) + '0';
            OutputStringLength = 4;
          }
          OutputString = TempString;
        }
        else /* Character doesn't need encoding. */
        {
          TempString [0] = InputChar;
          OutputStringLength = 1;
          OutputString = TempString;
        }
        break;
    }

    /* See if the output escape codes fit in the remaining buffer space.  Need
    to account for line wrap too.  Just stop if it doesn't fit.  No need to
    fiddle with partial escape codes (assume the output buffer is at least 6
    bytes long - backslash, LF, backslash, three octal digits). */

    if (OutputRemaining < OutputStringLength +
    ((WrapCount + OutputStringLength > 78) ? 2 : 0))
      break;

    if (WrapCount + OutputStringLength > 78)
    {
      /* Time for a line wrap.  Insert the backslash linefeed escape code. */
      if (OutputAllowed)
      {
        OutputPntr[0] = '\\';
        OutputPntr[1] = '\n';
      }
      OutputPntr += 2;
      OutputRemaining -= 2;
      WrapCount = 0;
    }

    if (OutputAllowed)
    {
      if (OutputStringLength == 1)
        *OutputPntr = *OutputString;
      else
        memcpy (OutputPntr, OutputString, OutputStringLength);
    }
    OutputPntr += OutputStringLength;
    OutputRemaining -= OutputStringLength;
    WrapCount += OutputStringLength;

    InputPntr++;
    InputRemaining--;
  }

  /* Return the amount of data processed back to the user. */

  *SourceLengthPntr = InputPntr - (const unsigned char *) SourceString;
  *DestinationLengthPntr = OutputPntr - (unsigned char *) DestinationString;
  if (StatePntr != NULL)
    *StatePntr = WrapCount;

  return 0;
}



/******************************************************************************
 * Process the escape codes in the input string and write the result to the
 * output string.  Returns 0 if successful, else an error code (usually because
 * of NULL pointers).
 *
 * Input and output can be the same string, but only safely on the first pass.
 * You can also have a NULL output string pointer if you just wish to count up
 * the amount of space the output would take.  NUL bytes in the strings work
 * OK, they get treated like any other character, not as end of string markers
 * (that's why buffer size is specified by a number for the input).  If you
 * want the output to be NUL terminated, include the NUL in the input string
 * and its length count.  On input, source and destination lengths describe the
 * sizes of the buffers.  On output, they describe the number of bytes actually
 * used.  Typically you should repeatedly call this function until the returned
 * DestinationLength is zero (it can read all the input and still have some
 * more output to produce if you happen to hit the end of the output buffer at
 * just the wrong time).  Initialise your state variable to zero for the first
 * iteration.  If you are not doing multiple iterations, you can pass in a NULL
 * StatePntr instead.  Internally the state contains a really short string with
 * the characters not processed from the last iteration, usually with the
 * backslash character in the front.
 *
 * Backslash (\) is the first character in an escape sequence.  The
 * following ones are cribbed from printf:
 *
 * \a      Write a <bell> 0x07 character (Alert noise).
 * \b      Write a <backspace> 0x08 character.
 * \f      Write a <form-feed> 0x0C character.
 * \n      Write a <new-line> 0x0A character.
 * \r      Write a <carriage return> 0x0D character.
 * \t      Write a <tab> 0x09 character.
 * \v      Write a <vertical tab> 0x0B character.
 * \\      Write a backslash character.
 * \num    Write an 8-bit character whose ASCII value is the 1-, 2-,
 * \0num   or 3-digit octal number num.
 * \LF     Write nothing, used for making line breaks invisible.
 * \Other  Write the Other character.
 */

static status_t RemoveBackslashEscapeCodes (
  const char *SourceString,
  int *SourceLengthPntr,
  char *DestinationString,
  int *DestinationLengthPntr,
  int32 *StatePntr)
{
  unsigned int         i;
  int                  InputChar;
  const unsigned char *InputPntr;
  int                  InputRemaining;
  int                  LoopsWithoutAnInputAdvance;
  bool                 OutputAllowed;
  int                  OutputChar;
  unsigned char       *OutputPntr;
  int                  OutputRemaining;
  unsigned char        StateString [5];

  if (SourceString == NULL ||
  SourceLengthPntr == NULL || DestinationLengthPntr == NULL ||
  *SourceLengthPntr < 0 || *DestinationLengthPntr < 0)
  {
    dprintf ("ARFS:RemoveBackslashEscapeCodes: Called with invalid inputs.\n");
    return EINVAL;
  }

  InputRemaining = *SourceLengthPntr;
  InputPntr = SourceString;
  OutputRemaining = *DestinationLengthPntr;
  OutputPntr = DestinationString;
  OutputAllowed = (DestinationString != NULL);
  memset (StateString, 0, sizeof (StateString));
  if (StatePntr != NULL)
    memcpy (StateString, StatePntr, 4 /* string size 5 - last byte zeroed */);

  LoopsWithoutAnInputAdvance = 0;
  while (LoopsWithoutAnInputAdvance++ < 2)
  {
    /* Look ahead at the next character to see if the stuff accumulated in
    StateString should dump out a character (happens for a non-digit ending an
    octal number). */

    if (InputRemaining <= 0)
      InputChar = -1;
    else
      InputChar = *InputPntr;

    OutputChar = -1; /* Negative for no output character. */

    if (StateString [0] == 0)
    {
      /* No stored up characters from before.  NULs handled elsewhere. */
    }
    else if (StateString [0] != '\\')
      OutputChar = StateString [0]; /* Not an escape code, char is itself. */
    else /* Have an escape code building up in StateString. */
    {
      switch (StateString [1])
      {
        case 0: /* No second character available yet in the escape code. */
          break;

        case 'a': /* Write a <bell> 0x07 character (Alert noise). */
          OutputChar = 7;
          break;

        case 'b': /* Write a <backspace> 0x08 character. */
          OutputChar = 8;
          break;

        case 'f': /* Write a <form-feed> 0x0C character. */
          OutputChar = 12;
          break;

        case 'n': /* Write a <new-line> 0x0A character. */
          OutputChar = 10;
          break;

        case 'r': /* Write a <carriage return> 0x0D character. */
          OutputChar = 13;
          break;

        case 't': /* Write a <tab> 0x09 character. */
          OutputChar = 9;
          break;

        case 'v': /* Write a <vertical tab> 0x0B character. */
          OutputChar = 11;
          break;

        case '\\': /* Write a backslash character. */
          OutputChar = '\\';
          break;

        case '0':
        case '1':
        case '2':
        case '3':
        case '4':
        case '5':
        case '6':
        case '7': /* See if the end of the octal number has been reached. */
          if (StateString [2] != 0)
          {
            if (StateString [3] != 0)
            {
              /* Have a 3 digit octal number, no more digits allowed. */
              OutputChar =
                (StateString [1] - '0') * 8 * 8 +
                (StateString [2] - '0') * 8 +
                (StateString [3] - '0');
            }
            else if (InputChar >= 0 && (InputChar < '0' || InputChar > '7'))
            {
              /* Have a two digit number, followed by something else. */
              OutputChar =
                (StateString [1] - '0') * 8 +
                (StateString [2] - '0');
            }
          }
          else if (InputChar >= 0 && (InputChar < '0' || InputChar > '7'))
          {
            /* Have a one digit number, followed by something else. */
            OutputChar =
              (StateString [1] - '0');
          }
          break;

        case '\n': /* Write nothing, used for making line breaks invisible. */
          StateString [0] = 0; /* Poof!  The backslash linefeed vanishes. */
          break;

        default: /* Write the Other character. */
          OutputChar = StateString [1];
          break;
      }
    }

    /* Output the character derived from the current state contents. */

    if (OutputChar >= 0)
    {
      if (OutputRemaining <= 0)
        break; /* No more output space when we need it. */

      if (OutputChar > 255)
        OutputChar = 255; /* Peg large octal numbers at the max. */
      if (OutputAllowed)
        *OutputPntr++ = OutputChar;
      else
        OutputPntr++; /* Just counting, not a valid pointer. */
      OutputRemaining--;
      StateString [0] = 0;
    }

    if (InputChar == 0)
    {
      /* Special NUL case handling, since it can't be represented in the
      StateString, output it now. */

      if (OutputRemaining <= 0)
        break; /* No more output space when we need it. */

      if (OutputAllowed)
        *OutputPntr++ = 0;
      else
        OutputPntr++; /* Just counting, not a valid pointer. */
      OutputRemaining--;
      InputPntr++;
      InputRemaining--;
      LoopsWithoutAnInputAdvance = 0;
      /* Can clear StateString since the NUL forced it to dump any octal
      numbers, and if we trash a simple backslash in the StateString, that's OK
      too since it would output a NUL anyway. */
      StateString [0] = 0;
    }
    else if (InputChar > 0)
    {
      /* Append the character to the state string. */

      for (i = 0; i < sizeof (StateString); i++)
      {
        if (StateString [i] == 0)
        {
          StateString [i] = InputChar;
          StateString [i+1] = 0;
          break;
        }
      }
      InputPntr++;
      InputRemaining--;
      LoopsWithoutAnInputAdvance = 0;
    }
  }

  /* Return the amount of data processed back to the user. */

  *SourceLengthPntr = InputPntr - (const unsigned char *) SourceString;
  *DestinationLengthPntr = OutputPntr - (unsigned char *) DestinationString;
  if (StatePntr != NULL)
    strncpy ((char *) StatePntr, StateString, 4); /* Zeroes trailing bytes. */

  return 0;
}



/******************************************************************************
 * Data structures to support the dump file.
 */

typedef struct CompressionBuffersStruct
{
#define COMPRESSION_BUFFER_SIZE (1024 * 256)
  bool              compressionEnabled;
    /* TRUE for zlib deflate compression, FALSE for none. */
  bool              deflateInitDone;
  DiskVolumePointer diskVolumePntr;
  int               dumpFileHandle;
    /* Set to -1 if the file is closed. */
  int               inputBufferIndex;
    /* Index of the next free byte in the input buffer, starts at zero. */
  uint8            *inputBufferPntr;
    /* The input buffer, where data is copied to before being compressed and
    written to the output buffer.  NULL if the buffer hasn't been allocated.
    Will be allocated with kernel malloc, with size COMPRESSION_BUFFER_SIZE. */
  int               outputBufferIndex;
  uint8            *outputBufferPntr;
    /* Same usage as inputBufferPntr, same size too. */
  z_stream          zLibStream;
} CompressionBuffersRecord, *CompressionBuffersPointer;



/******************************************************************************
 * Utility function for the ZLib library to allocate memory.  We will just
 * use the kernel malloc, even though it does lock pages in memory etc.
 */

voidpf MyAllocForZLib (voidpf opaque, uInt items, uInt size)
{
  unsigned long Amount;
  void         *Pntr;

  Amount = items * (unsigned long) size;
  Pntr = malloc (Amount);
  if (Pntr == NULL)
  {
    dprintf ("ARFS:MyAllocForZLib: failed to allocate %lu bytes "
      "(%d items of %d size).\n", Amount, (int) items, (int) size);
    return Z_NULL;
  }

  return Pntr;
}



/******************************************************************************
 * Utility function for the ZLib library to deallocate memory.
 */

void MyFreeForZLib (voidpf opaque, voidpf address)
{
  free (address);
}



/******************************************************************************
 * An internal utility function to move data from the input buffer to the
 * output buffer, possibly compressing it, and finally to the dump file.  Only
 * moves data when the buffers are half full, unless you set FinalFlush to TRUE
 * (then it totally empties the buffers and closes off the compression stream
 * too).  Returns 0 if successful, else a negative error code.
 */

static status_t CompressSome (
  CompressionBuffersPointer CmpBufsPntr,
  bool FinalFlush)
{
  uint AmountCopied;
  uint AmountRead;
  uint AmountWritten;
  int  ErrorCode = -1;

  if (CmpBufsPntr == NULL ||
  CmpBufsPntr->inputBufferPntr == NULL ||
  CmpBufsPntr->outputBufferPntr == NULL ||
  CmpBufsPntr->dumpFileHandle < 0)
  {
    dprintf ("ARFS:CompressSome: Called with bad parameters.\n");
    return -1;
  }

  /* Do some compression, only if there is a large enough amount of data ready
  to be compressed, or if it's flushing time. */

  while ((FinalFlush && ErrorCode != Z_STREAM_END) ||
  CmpBufsPntr->inputBufferIndex > COMPRESSION_BUFFER_SIZE / 2)
  {
    CmpBufsPntr->zLibStream.next_in = CmpBufsPntr->inputBufferPntr;
    CmpBufsPntr->zLibStream.avail_in = CmpBufsPntr->inputBufferIndex;

    CmpBufsPntr->zLibStream.next_out =
      CmpBufsPntr->outputBufferPntr + CmpBufsPntr->outputBufferIndex;
    CmpBufsPntr->zLibStream.avail_out =
      COMPRESSION_BUFFER_SIZE - CmpBufsPntr->outputBufferIndex;

    AmountRead = 0;
    AmountWritten = 0;

    if (CmpBufsPntr->compressionEnabled)
    {
      /* Compress some of the input buffer into the tail end of the output
      buffer, if there's any space there. */

      if (FinalFlush ||
      CmpBufsPntr->outputBufferIndex < COMPRESSION_BUFFER_SIZE)
      {
        ErrorCode = deflate (&CmpBufsPntr->zLibStream,
          FinalFlush ? Z_FINISH : Z_NO_FLUSH);
        if (ErrorCode != Z_STREAM_END && ErrorCode != Z_OK)
        {
          dprintf ("ARFS:CompressSome: zlib error %d when compressing.\n",
            ErrorCode);
          return -1;
        }
        AmountRead =
          CmpBufsPntr->inputBufferIndex - CmpBufsPntr->zLibStream.avail_in;
        AmountWritten = COMPRESSION_BUFFER_SIZE -
          CmpBufsPntr->outputBufferIndex - CmpBufsPntr->zLibStream.avail_out;
      }
    }
    else /* No compression, just copy data. */
    {
      AmountCopied = CmpBufsPntr->zLibStream.avail_in;
      if (AmountCopied > CmpBufsPntr->zLibStream.avail_out)
        AmountCopied = CmpBufsPntr->zLibStream.avail_out;
      if (AmountCopied > 0)
        memcpy (CmpBufsPntr->zLibStream.next_out,
          CmpBufsPntr->zLibStream.next_in, AmountCopied);
      AmountRead = AmountCopied;
      AmountWritten = AmountCopied;
      ErrorCode = Z_STREAM_END; /* So FinalFlush will exit after one loop. */
    }

    /* Update our buffer settings for the amount of data which was removed from
    the input and added to the output. */

    CmpBufsPntr->outputBufferIndex += AmountWritten;
    if (AmountRead > 0)
    {
      /* Move any remaining data at the end of the input buffer up to the front
      of the buffer. */
      if (AmountRead < CmpBufsPntr->inputBufferIndex)
      {
        memmove (CmpBufsPntr->inputBufferPntr,
          CmpBufsPntr->inputBufferPntr + AmountRead,
          CmpBufsPntr->inputBufferIndex - AmountRead);
        CmpBufsPntr->inputBufferIndex -= AmountRead;
      }
      else
        CmpBufsPntr->inputBufferIndex = 0;
    }

    /* Write the output buffer to the file, if it is full enough. */

    if ((FinalFlush && CmpBufsPntr->outputBufferIndex > 0) ||
    CmpBufsPntr->outputBufferIndex > COMPRESSION_BUFFER_SIZE / 2)
    {
      AmountCopied = CmpBufsPntr->outputBufferIndex;
      AmountWritten = write (CmpBufsPntr->dumpFileHandle,
        CmpBufsPntr->outputBufferPntr, AmountCopied);
      if (AmountWritten < 0 || AmountWritten != AmountCopied)
      {
        ErrorCode = errno;
        dprintf ("ARFS:CompressSome: Unable to write to dump file, "
          "error code %d (%s).\n", ErrorCode, strerror (ErrorCode));
        return ErrorCode;
      }
      CmpBufsPntr->outputBufferIndex = 0;
    }
  }
  return 0;
}



/******************************************************************************
 * Dumps the file system to the dump file, if specified earlier in the mount
 * arguments.
 */

static status_t DumpFileSystem (
  DiskVolumePointer DiskVolumePntr)
{
  CompressionBuffersRecord CompressionBuffers;
  status_t                 ErrorCode;
  status_t                 ReturnCode = -1;

  if (DiskVolumePntr->dumpFilePath == NULL)
    return 0; /* No dump requested. */

  memset (&CompressionBuffers, 0, sizeof (CompressionBuffers));
  CompressionBuffers.diskVolumePntr = DiskVolumePntr;
  CompressionBuffers.dumpFileHandle = -1;

  CompressionBuffers.dumpFileHandle = open (DiskVolumePntr->dumpFilePath,
    O_CREAT | O_TRUNC | O_WRONLY | O_BINARY, 0666 /* file mode/permission */);
  if (CompressionBuffers.dumpFileHandle < 0)
  {
    ReturnCode = errno;
    dprintf ("ARFS:DumpFileSystem: Unable to create file \"%s\", "
      "error code %d (%s).\n",
      DiskVolumePntr->dumpFilePath, (int) ReturnCode, strerror (ReturnCode));
    goto ErrorExit;
  }

  CompressionBuffers.inputBufferPntr = malloc (COMPRESSION_BUFFER_SIZE);
  CompressionBuffers.outputBufferPntr = malloc (COMPRESSION_BUFFER_SIZE);
  if (CompressionBuffers.inputBufferPntr == NULL ||
  CompressionBuffers.outputBufferPntr == NULL)
  {
    dprintf ("ARFS:DumpFileSystem: Unable to allocate memory for "
      "compression buffers.\n");
    ReturnCode = ENOMEM;
    goto ErrorExit;
  }

  CompressionBuffers.compressionEnabled =
    (strcasecmp (DiskVolumePntr->dumpCompressionMethod, "zlibDeflate") == 0);
  if (CompressionBuffers.compressionEnabled)
  {
    CompressionBuffers.zLibStream.zalloc = MyAllocForZLib;
    CompressionBuffers.zLibStream.zfree = MyFreeForZLib;
    ErrorCode = deflateInit (&CompressionBuffers.zLibStream,
      DiskVolumePntr->dumpZlibCompressionLevel);
    if (ErrorCode != Z_OK)
    {
      dprintf ("ARFS:DumpFileSystem: deflateInit failed, code %d (%s).\n",
        (int) ErrorCode, CompressionBuffers.zLibStream.msg ?
        CompressionBuffers.zLibStream.msg : "?");
      goto ErrorExit;
    }
    CompressionBuffers.deflateInitDone = true;
  }

// bleeble
  strcpy (CompressionBuffers.inputBufferPntr, "This is a");
  CompressionBuffers.inputBufferIndex = strlen (CompressionBuffers.inputBufferPntr);
  CompressSome (&CompressionBuffers, false);

  strcpy (CompressionBuffers.inputBufferPntr, "test.");
  CompressionBuffers.inputBufferIndex = strlen (CompressionBuffers.inputBufferPntr);
  CompressSome (&CompressionBuffers, true);

ErrorExit:
  if (CompressionBuffers.deflateInitDone)
  {
    ErrorCode = deflateEnd (&CompressionBuffers.zLibStream);
    if (ErrorCode != Z_OK)
    {
      dprintf ("ARFS:DumpFileSystem: deflateEnd failed, code %d (%s).\n",
        (int) ErrorCode, CompressionBuffers.zLibStream.msg ?
        CompressionBuffers.zLibStream.msg : "?");
    }
  }
  if (CompressionBuffers.inputBufferPntr != NULL)
    free (CompressionBuffers.inputBufferPntr);
  if (CompressionBuffers.outputBufferPntr != NULL)
    free (CompressionBuffers.outputBufferPntr);
  if (CompressionBuffers.dumpFileHandle >= 0)
  {
    fsync (CompressionBuffers.dumpFileHandle);
    close (CompressionBuffers.dumpFileHandle);
  }
  return ReturnCode;
}



/******************************************************************************
 * Parse the startup parameters.  This can either just count up the number of
 * extra bytes of storage needed to store them, or actually store them in a
 * DiskVolumeRecord.  Counts if DiskVolumePntr is NULL, otherwise stores them
 * in the DiskVolumeRecord in the string storage space.  Returns the number of
 * string space bytes needed / stored, or a negative error code.
 *
 * The arguments are in the form NAME=VALUE with a tab character separating
 * them.  Spaces are treated like any other character, so don't have extra ones
 * around the equals sign.  Case is ignored for matching NAME items.  Escape
 * codes can be in the NAME and VALUE portions, and are interpreted after they
 * have been isolated from the rest of the command line (isolated by looking
 * for unescaped tabs and equals signs).  The last NAME occurence is the one
 * used, the earlier ones just waste a bit of space in the string storage
 * space.
 */

static int ParseMountCommandArguments (
  DiskVolumePointer DiskVolumePntr,
  const char *CommandLinePntr)
{
#define MAX_MOUNT_ARGUMENT_VALUE_LENGTH 65000
  int         DestinationLength;
  status_t    ErrorCode;
  int         i;
  char        KeywordCopy [30];
  char       *OutputPntr;
  int         SourceLength;
  int         SpaceUsed;
  const char *ValuePntr;

  SpaceUsed = 0;
  while (*CommandLinePntr != 0)
  {
    /* Extract the NAME keyword - everything up to the next equals sign. */

    KeywordCopy [0] = 0;
    OutputPntr = KeywordCopy;
    while (*CommandLinePntr != 0 && *CommandLinePntr != '=')
    {
      if (OutputPntr < KeywordCopy + (sizeof (KeywordCopy) - 1))
      {
        *OutputPntr++ = *CommandLinePntr;
        *OutputPntr = 0;
      }
      CommandLinePntr++;
    }

    SourceLength = strlen (KeywordCopy) + 1;
    DestinationLength = sizeof (KeywordCopy);
    RemoveBackslashEscapeCodes (KeywordCopy, &SourceLength,
      KeywordCopy, &DestinationLength, NULL);

    /* Skip over the equals sign. */

    if (*CommandLinePntr != '=')
    {
      dprintf ("ARFS:ParseMountCommandArguments: "
        "Keyword \"%s\" isn't followed by an equals sign.\n", KeywordCopy);
      return EINVAL;
    }
    CommandLinePntr++;

    /* Find the end of the value - at a tab character or end of string. */

    ValuePntr = CommandLinePntr;
    while (*CommandLinePntr != '\t' && *CommandLinePntr != 0)
      CommandLinePntr++;

    /* Process the escape codes in the value, finding out how long the
    resulting string is, and possibly storing it too. */

    SourceLength = CommandLinePntr - ValuePntr;
    if (DiskVolumePntr == NULL)
    {
      DestinationLength = MAX_MOUNT_ARGUMENT_VALUE_LENGTH;
      OutputPntr = NULL;
    }
    else
    {
      DestinationLength = DiskVolumePntr->stringStorageSize - SpaceUsed;
      OutputPntr = DiskVolumePntr->stringStorageSpace + SpaceUsed;
    }
    ErrorCode = RemoveBackslashEscapeCodes (ValuePntr, &SourceLength,
      OutputPntr, &DestinationLength, NULL);
    if (ErrorCode != 0)
    {
      dprintf ("ARFS:ParseMountCommandArguments: Unable to de-escape "
        "argument string!  Code %d, keyword: %s\n",
        (int) ErrorCode, KeywordCopy);
      return ErrorCode;
    }
    if (DestinationLength >= MAX_MOUNT_ARGUMENT_VALUE_LENGTH)
    {
      dprintf ("ARFS:ParseMountCommandArguments: Argument is %d bytes, "
        "over 65k, too long!  Keyword: %s\n",
        DestinationLength, KeywordCopy);
      return ENAMETOOLONG;
    }
    if (DiskVolumePntr != NULL)
      OutputPntr [DestinationLength] = 0;

    /* Process the known keywords.  Unknown ones get ignored. */

#if DPRINT_NON_ERROR_MESSAGES
    if (OutputPntr != NULL)
      dprintf ("ARFS:ParseMountCommandArguments: "
        "Keyword \"%s\", value \"%s\", length %d.\n", KeywordCopy,
        (DestinationLength > 1050) ? "Too long to display!" : OutputPntr,
        DestinationLength);
#endif

    if (strcasecmp (KeywordCopy, "MountDir") == 0)
    {
      SpaceUsed += DestinationLength + 1 /* for NUL at string end */;
      if (OutputPntr != NULL)
        DiskVolumePntr->pathToRootParent = OutputPntr;
    }
    else if (strcasecmp (KeywordCopy, "MaxVolumeSize") == 0)
    {
      if (OutputPntr != NULL)
      {
        i = atoi (OutputPntr);
        if (i > 0)
          DiskVolumePntr->maxTotalFileSize = i;
      }
    }
    else if (strcasecmp (KeywordCopy, "DumpFilePath") == 0)
    {
      SpaceUsed += DestinationLength + 1;
      if (OutputPntr != NULL)
        DiskVolumePntr->dumpFilePath = OutputPntr;
    }
    else if (strcasecmp (KeywordCopy, "DumpCompressionMethod") == 0)
    {
      SpaceUsed += DestinationLength + 1;
      if (OutputPntr != NULL)
        DiskVolumePntr->dumpCompressionMethod = OutputPntr;
    }
    else if (strcasecmp (KeywordCopy, "DumpZlibCompressionLevel") == 0)
    {
      if (OutputPntr != NULL)
      {
        i = atoi (OutputPntr);
        if (i >= 0 && i <= 9)
          DiskVolumePntr->dumpZlibCompressionLevel = i;
      }
    }
    else
      dprintf ("ARFS:ParseMountCommandArguments: "
        "Keyword \"%s\" isn't recognized, ignoring it.\n", KeywordCopy);

    /* Skip over the tab character between the commands. */

    while (*CommandLinePntr == '\t')
      CommandLinePntr++;
  }

  return SpaceUsed + 20 /* Include temp space for parsing numeric arguments */;
}



/******************************************************************************
 * Starts up a file system.  After some user program calls the Posix mount
 * (const char *filesystem, const char *where, const char *device, ulong flags,
 * void *parms, int len) function (often done when a disk is inserted, possibly
 * automatically by a daemon looking for new disks), the OS loads your file
 * system add-on if it isn't already there, and calls your mount function.  The
 * nsid is used by the OS to identify this particular file system instance, and
 * should be saved for later use.  Devname is the path name to the disk device
 * to use, specified by the user program.  The flags are also user specified
 * bits, B_MOUNT_READ_ONLY (1) being the only documented one.  The parms and
 * len (length of parameters) are also provided by the user program, and can be
 * anything you want.  Once you have finished setting things up, return a
 * pointer to your instance specific semi-global data in *data.  By semi-global
 * data, I mean a struct holding things which are specific to the file system
 * instance as a whole (like disk space remaining, free list, volume name,
 * etc).  Since there could be more than one instance of your file system
 * running (one instance for each disk volume/partition mounted using your kind
 * of file system), allocate a new semi-global data area for each instance to
 * keep them separate.  This semi-global pointer will be passed in future calls
 * in the "void *ns" field so you know which instance to use.  The final vnid
 * argument is used for returning the vnode_id of the root directory.  You need
 * to set up the root vnode in BeOS by using new_vnode and initialising it as
 * if read_vnode had processed it (no, it doesn't call read_vnode, but it will
 * call write_vnode when it unmounts), otherwise we get a kernel panic about
 * root being NULL.  You'd think it would do its own get_vnode when it needed
 * the root, but it does not.  You need to force feed it.  Return 0 if
 * everything went OK, otherwise return a negative number error code (see
 * support/Errors.h for a list).
 */

int ARFS_mount (
  nspace_id nsid,
  const char *devname,
  ulong flags,
  void *parms,
  size_t len,
  void **data,
  vnode_id *vnid)
{
  area_malloc_module_info *AreaMallocModuleInfoPntr = NULL;
  DiskVolumePointer        DiskVolumePntr = NULL;
  status_t                 ErrorCode;
  int                      ExtraSpaceNeeded;
  void                    *HoleAreaAddress = NULL;
  area_id                  HoleAreaID = -1;
  int                      i;
  IndexObjectPointer       IndexFSOPntr;
  const char              *IndexNamePntr;
  DirectoryObjectPointer   IndicesFSOPntr;
  const void              *MainPool = NULL;
  status_t                 ReturnCode = -1;
  int                      RootPathLength;
  char                    *SlashPntr;
  char                    *VolumeName;

  dprintf ("ARFS version: $Header: /CommonBe/agmsmith/Programming/AGMSRAMFileSystem/RCS/AGMSRAMFileSystem.c,v 1.156 2004/04/04 02:00:16 agmsmith Exp agmsmith $\n");
  dprintf ("ARFS compiled on " __DATE__ " at " __TIME__ ".\n");
  ErrorCode = load_driver_symbols ("ARFSv0");
  if (ErrorCode != 0)
    dprintf ("ARFS loading driver symbols returned %d.\n", (int) ErrorCode);
  dprintf ("ARFS AGMSRAMFileSystem is copyright © 2001 by Alexander G. M. Smith.\n");
  dprintf ("ARFS This software is distributed under the GNU Lesser General Public License.\n");

#if DPRINT_NON_ERROR_MESSAGES
  dprintf ("ARFS_mount entered, nsid:%d, devname:$%X \"%s\", flags:$%lX, "
    "parms:$%X, len:%ld.\n", (int) nsid, (int) devname,
    (devname == NULL) ? "NULL" : devname, flags, (int) parms, len);
#endif

  /* Validate the command line parameters.  Start by counting up the space
  needed to store them.  The parameters are a NUL terminated UTF-8 string
  containing NAME=VALUE pairs, separated by tabs.  Backslash escape codes are
  supported. */

  if (parms == NULL)
  {
    dprintf ("ARFS_mount: Parameter string is NULL, need some parameters!\n");
    ReturnCode = EINVAL;
    goto ErrorExit;
  }

  ExtraSpaceNeeded = strlen (parms);
  if (ExtraSpaceNeeded >= len)
  {
    dprintf ("ARFS_mount: Just a warning - parameter size of %d provided "
      "by the caller is to small to hold a string of length %d (plus one "
      "for the NUL string end byte).\n",
      (int) len, ExtraSpaceNeeded);
  }

  ExtraSpaceNeeded = ParseMountCommandArguments (NULL, parms);
  if (ExtraSpaceNeeded < 0)
  {
    dprintf ("ARFS_mount: Unable to parse arguments string, code %d.\n",
      ExtraSpaceNeeded);
    ReturnCode = ExtraSpaceNeeded;
    goto ErrorExit;
  }

  /* Open a memory pool for this disk volume to use for all its allocations. */

  ErrorCode = get_module (g_AreaMallocModuleName,
    (module_info **) &AreaMallocModuleInfoPntr);
  if (ErrorCode != 0 || AreaMallocModuleInfoPntr == NULL)
  {
    dprintf ("ARFS_mount: unable to get the area malloc module.\n");
    ReturnCode = ErrorCode;
    goto ErrorExit;
  }

#if DPRINT_NON_ERROR_MESSAGES
  dprintf ("ARFS_mount: got area malloc module named \"%s\".\n",
    AreaMallocModuleInfoPntr->minfo.name);
#endif

  /* Buggy BeOS kernel memory manager hack - allocate an area almost as big as
  the size used for our allocations, then free it a bit later, to leave a hole
  in memory which other kernel things can grow into, like the ptty device or
  the DOS file system or disk cache.  Don't care much if it fails. */

  HoleAreaID = create_area (
    "ARFS Memory Hole",
    &HoleAreaAddress,
    B_ANY_KERNEL_ADDRESS,
    15000 * 4096 /* Slightly less than the pool size increment */,
    g_UseVirtualMemory ? B_NO_LOCK : B_FULL_LOCK,
    B_READ_AREA | B_WRITE_AREA);

  MainPool = AreaMallocModuleInfoPntr->create_pool (
    B_ANY_KERNEL_ADDRESS,
    16384 * 4096 /* Size it grows by when more space is needed,
    note that 1MB is too small and fragments kernel memory space,
    making it fail to mount new disks and have other problems. */,
    g_UseVirtualMemory ? B_NO_LOCK : B_FULL_LOCK,
    B_READ_AREA | B_WRITE_AREA);

  if (MainPool == NULL)
  {
    dprintf ("ARFS_mount: unable to allocate main memory pool.\n");
    goto ErrorExit;
  }

#if DPRINT_NON_ERROR_MESSAGES
  dprintf ("ARFS_mount: main memory pool $%X\n", (unsigned int) MainPool);
#endif

  DiskVolumePntr = AreaMallocModuleInfoPntr->malloc (
    MainPool, sizeof (DiskVolumeRecord) + ExtraSpaceNeeded);
  if (DiskVolumePntr == NULL)
  {
    dprintf ("ARFS_mount: unable to allocate %d bytes for the disk volume "
      "structure.\n", (int) sizeof (DiskVolumeRecord) + ExtraSpaceNeeded);
    goto ErrorExit;
  }

#if DPRINT_NON_ERROR_MESSAGES
  dprintf ("ARFS_mount: DiskVolumeRecord at $%X, size %d (includes %d bytes "
    "of extra string space).\n",
    (unsigned int) DiskVolumePntr,
    (int) sizeof (DiskVolumeRecord) + ExtraSpaceNeeded, ExtraSpaceNeeded);
#endif

  memset (DiskVolumePntr, 0,
    sizeof (DiskVolumeRecord) + ExtraSpaceNeeded);
  DiskVolumePntr->nameSpaceID = nsid;
  DiskVolumePntr->magicCode =
    (DiskVolumePntr->nameSpaceID ^ DISK_VOLUME_MAGIC);
  DiskVolumePntr->readOnly = (flags & B_MOUNT_READ_ONLY) != 0;
  DiskVolumePntr->volumePoolID = MainPool;
  DiskVolumePntr->deletePoolPntr = AreaMallocModuleInfoPntr->delete_pool;
  DiskVolumePntr->mallocPntr = AreaMallocModuleInfoPntr->malloc;
  DiskVolumePntr->freePntr = AreaMallocModuleInfoPntr->free;
  DiskVolumePntr->liveQueryListSemaphoreID = -1;
  DiskVolumePntr->maxTotalFileSize = 50000000; /* Default value. */
  DiskVolumePntr->stringStorageSize = ExtraSpaceNeeded;
  DiskVolumePntr->dumpCompressionMethod = "None";
  DiskVolumePntr->dumpZlibCompressionLevel = 6;

  /* Copy the command line arguments into the DiskVolumeRecord. */

  ErrorCode = ParseMountCommandArguments (DiskVolumePntr, parms);
  if (ErrorCode < 0)
  {
    dprintf ("ARFS_mount: Unable to copy command line arguments into the "
      "internal data structures, code %d.\n", (int) ErrorCode);
    ReturnCode = ErrorCode;
    goto ErrorExit;
  }
  if (ErrorCode != ExtraSpaceNeeded)
  {
    dprintf ("ARFS_mount: Bug!  Second pass of parameter parsing returned a "
      "different extra space size, got %d, was expecting %d.\n",
      (int) ErrorCode, ExtraSpaceNeeded);
    goto ErrorExit;
  }

  /* The "MountDir" argument should be a string with the absolute path to the
  mount point directory.  Has a leading slash, no trailing slash and isn't too
  long.  The volume name will be taken from the last part of that path. */

  if (DiskVolumePntr->pathToRootParent == NULL)
  {
    dprintf ("ARFS_mount: Need the mount point directory, use "
      "MountDir=/blah in the mount() arguments, tab separated "
      "from other arguments in the parameter string.\n");
    ReturnCode = EINVAL;
    goto ErrorExit;
  }
  RootPathLength = strlen (DiskVolumePntr->pathToRootParent);
  if (RootPathLength >= PATH_MAX ||
  DiskVolumePntr->pathToRootParent[0] != '/' ||
  DiskVolumePntr->pathToRootParent[RootPathLength-1] == '/')
  {
    dprintf ("ARFS_mount: \"MountDir\" parameter should be a string "
      "with an absolute path (leading slash, no trailing slash).\n");
    ReturnCode = EINVAL;
    goto ErrorExit;
  }

  SlashPntr = strrchr (DiskVolumePntr->pathToRootParent, '/');
  *SlashPntr = 0; /* Break the string into two - path prefix, volume name. */
  DiskVolumePntr->stringLengthOfPathToRootParent =
    SlashPntr - DiskVolumePntr->pathToRootParent;
  VolumeName = SlashPntr + 1;

  /* Set up the semaphore which is used for guarding the live query list. */

  ErrorCode = create_sem (MAX_SIMULTANEOUS_READERS,
    "ARFS Live Query List");
  if (ErrorCode < 0)
  {
    dprintf ("ARFS_mount: Unable to create access semaphore for "
      "live query list, error code %d!\n", (int) ErrorCode);
    goto ErrorExit;
  }
  DiskVolumePntr->liveQueryListSemaphoreID = ErrorCode;

  /* Set the ownership to the kernel, so that the semaphore doesn't disappear
  when the caller dies. */

  ErrorCode =
    set_sem_owner (DiskVolumePntr->liveQueryListSemaphoreID, B_SYSTEM_TEAM);
  if (ErrorCode < 0)
  {
    dprintf ("ARFS_mount Unable to kernelize ownership of live query list "
      "semaphore, error code %d!\n", (int) ErrorCode);
    goto ErrorExit;
  }

  /* Create an empty root directory.  Note that we leave the parent directory
  set to NULL, because its parent is outside the file system.  Take the name
  from the mount point. */

  DiskVolumePntr->rootFSOPntr =
    AllocateDirectoryFSO (DiskVolumePntr, VolumeName);
  if (DiskVolumePntr->rootFSOPntr == NULL)
  {
    dprintf ("ARFS_mount: unable to create root directory.\n");
    goto ErrorExit;
  }
  DiskVolumePntr->rootVNodeID = (unsigned int) DiskVolumePntr->rootFSOPntr;

  /* Give the root directory rwxrwxrwx permissions. */

  DiskVolumePntr->rootFSOPntr->fso.modeBits |=
    (S_IRWXU | S_IRWXG | S_IRWXO
    /* S_IRGRP | S_IXGRP | S_IROTH | S_IXOTH */);

  /* Need to set up the root vnode in BeOS, otherwise we get a kernel panic
  about root being NULL.  You'd think it would do its own get_vnode when it
  needed the root, but it does not.  You need to force feed it with new_vnode
  and make it active as if it had gone through read_vnode. */

  ErrorCode = ActivateFSO (&DiskVolumePntr->rootFSOPntr->fso);
  if (ErrorCode < 0)
  {
    dprintf ("ARFS_mount: ActivateFSO failed for root directory, "
      "error code %d.\n", (int) ErrorCode);
    ReturnCode = ErrorCode;
    goto ErrorExit;
  }

  ErrorCode =
    new_vnode (nsid, DiskVolumePntr->rootVNodeID, DiskVolumePntr->rootFSOPntr);
  if (ErrorCode < 0)
  {
    dprintf ("ARFS_mount: new_vnode failed for root directory, "
      "the OS doesn't want to know about us.\n");
    ReturnCode = ErrorCode;
    goto ErrorExit;
  }

  /* Set up the magic indices directory.  It doesn't have an outstanding
  get_vnode like the root directory (I tried), because that would prevent you
  from unmounting the file system. */

  IndicesFSOPntr =
    AllocateDirectoryFSO (DiskVolumePntr, g_IndicesDirectoryName);
  if (IndicesFSOPntr == NULL)
  {
    dprintf ("ARFS_mount: Unable to create indices (%s) directory.\n",
      g_IndicesDirectoryName);
    goto ErrorExit;
  }

  /* The indices directory is readable by anyone, writeable by nobody. */

  IndicesFSOPntr->fso.modeBits |=
    (S_IRUSR | S_IXUSR | S_IRGRP | S_IXGRP | S_IROTH | S_IXOTH);

  ErrorCode = AddFreshFSOToDirectory (DiskVolumePntr,
    &IndicesFSOPntr->fso, DiskVolumePntr->rootFSOPntr);
  if (ErrorCode < 0)
  {
    dprintf ("ARFS_mount: Failed to add indices directory to "
      "root directory, code %d.\n", (int) ErrorCode);
    ReturnCode = ErrorCode;
    goto ErrorExit;
  }

  /* Make the standard system indices: name, size and last_modified. */

  for (i = 0; i < STDINUM_MAX; i++)
  {
    IndexNamePntr = g_NamesOfStandardIndices [i];
    IndexFSOPntr = AllocateIndexFSO (DiskVolumePntr, IndexNamePntr,
      g_TypesOfStandardIndices [i], false /* InterpretAsMultipleKeywords */);
    if (IndexFSOPntr == NULL)
    {
      dprintf ("ARFS_mount: Unable to create index \"%s\".\n", IndexNamePntr);
      ReturnCode = ENOMEM;
      goto ErrorExit;
    }
    IndexFSOPntr->fso.modeBits |=
      (S_IRUSR | S_IXUSR | S_IRGRP | S_IXGRP | S_IROTH | S_IXOTH);
    ErrorCode = AddFreshFSOToDirectory (DiskVolumePntr,
      &IndexFSOPntr->fso, IndicesFSOPntr);
    if (ErrorCode < 0)
    {
      dprintf ("ARFS_mount: Failed to add index \"%s\" to indices directory, "
        "code %d.\n", IndexNamePntr, (int) ErrorCode);
      ReturnCode = ErrorCode;
      goto ErrorExit;
    }
    DiskVolumePntr->standardIndicesVnodes[i] = (unsigned int) IndexFSOPntr;
  }

  /* It looks good.  Return success and don't undo the allocations we made. */

#if DPRINT_NON_ERROR_MESSAGES
  dprintf ("ARFS_mount successful, volume name: \"%s\", "
    "path to root's parent: \"%s\" (cached length %d), "
    "max total file size: %ld.\n",
    VolumeName, DiskVolumePntr->pathToRootParent,
    DiskVolumePntr->stringLengthOfPathToRootParent,
    DiskVolumePntr->maxTotalFileSize);
  if (DiskVolumePntr->dumpFilePath != NULL)
    dprintf ("ARFS_mount Dump file enabled, will save contents when "
      "unmounting to path: \"%s\", compression method: \"%s\", "
      "zlib compression level: %d.\n",
      DiskVolumePntr->dumpFilePath, DiskVolumePntr->dumpCompressionMethod,
      DiskVolumePntr->dumpZlibCompressionLevel);
#endif

  *data = DiskVolumePntr;
  *vnid = DiskVolumePntr->rootVNodeID;
  ReturnCode = 0;
  goto NormalExit;

ErrorExit:
  if (DiskVolumePntr != NULL && DiskVolumePntr->liveQueryListSemaphoreID >= 0)
    delete_sem (DiskVolumePntr->liveQueryListSemaphoreID);
  if (MainPool != NULL)
    AreaMallocModuleInfoPntr->delete_pool (MainPool);
  if (AreaMallocModuleInfoPntr != NULL)
    put_module (g_AreaMallocModuleName);

NormalExit:
  if (HoleAreaID >= 0)
    delete_area (HoleAreaID);
#if DPRINT_NON_ERROR_MESSAGES
  dprintf ("ARFS_mount exited, return code %d.\n", (int) ReturnCode);
#endif
  return ReturnCode;
}



/******************************************************************************
 * Unmount the given file system instance.  Only gets called when no open files
 * are using your file system.  Usually this means flushing any cached data to
 * the media, closing any device drivers used, then returning when it is
 * complete.  You should also deallocate your semi-global data here too.  BeOS
 * does a write_vnode for all outstanding vnodes (including the root directory)
 * before calling this function.
 */

int ARFS_unmount (void *ns)
{
  DiskVolumePointer  DiskVolumePntr = ns;
  status_t           ReturnCode = -1;

#if DPRINT_NON_ERROR_MESSAGES
  dprintf ("ARFS_unmount entered, namespace: $%X.\n", (int) ns);
#endif

  if (INVALID_DISK_VOLUME (DiskVolumePntr))
  {
    dprintf ("ARFS_unmount: invalid input disk volume pointer $%X, "
      "don't know what to unmount.\n", (unsigned int) DiskVolumePntr);
    goto ErrorExit;
  }

  DumpFileSystem (DiskVolumePntr);

  if (DiskVolumePntr->liveQueryListSemaphoreID >= 0)
  {
    delete_sem (DiskVolumePntr->liveQueryListSemaphoreID);
    DiskVolumePntr->liveQueryListSemaphoreID = -1;
  }

  DiskVolumePntr->magicCode += 1234567; /* Make the magic code invalid. */
  DiskVolumePntr->deletePoolPntr (DiskVolumePntr->volumePoolID);
  put_module (g_AreaMallocModuleName);
  ReturnCode = 0;

ErrorExit:
  dprintf ("ARFS_unmount exited, return code %d.  Goodbye!\n",
    (int) ReturnCode);
  return ReturnCode;
}



/******************************************************************************
 * Reads information about the file system instance in the fs_info structure.
 * That includes things like the amount of free space and volume name.  Return
 * an error code or 0.  This needs to be implemented or the OS will crash right
 * after the mount, as many BeOS daemons are busy using it.  Return an empty
 * string for the device name if the file system doesn't run on a physical
 * device.  You have 8 letters for the file system name (otherwise the "df"
 * output looks bad).  The BSD statfs documentation says that fields undefined
 * for a particular file system will be set to -1.
 */

int ARFS_rfsstat (
  void *ns,
  struct fs_info *fsinfo)
{
  DiskVolumePointer  DiskVolumePntr = ns;
  status_t           ErrorCode;
  bool               ReadLockDone = false;
  status_t           ReturnCode = -1;

#if DPRINT_NON_ERROR_MESSAGES
  dprintf ("ARFS_rfsstat entered, NS:$%08X, fsinfo:$%08X.\n",
    (int) ns, (int) fsinfo);
#endif

  if (INVALID_DISK_VOLUME (DiskVolumePntr))
  {
    dprintf ("ARFS_rfsstat: invalid input disk volume pointer $%X, "
      "don't know what to get status of.\n", (unsigned int) DiskVolumePntr);
    goto ErrorExit;
  }

  if (fsinfo == NULL)
  {
    dprintf ("ARFS_rfsstat: NULL fsinfo pointer from caller.\n");
    goto ErrorExit;
  }

  if (INVALID_FILE_SYSTEM_OBJECT (&DiskVolumePntr->rootFSOPntr->fso))
  {
    dprintf ("ARFS_rfsstat: Invalid root directory FSO $%X.\n",
      (unsigned int) DiskVolumePntr->rootFSOPntr);
    goto ErrorExit;
  }

  ErrorCode = ReadLock (&DiskVolumePntr->rootFSOPntr->fso, false);
  if (ErrorCode != 0)
  {
    dprintf ("ARFS_rfsstat: Unable to get read lock for %s.\n",
      FSOAsString (&DiskVolumePntr->rootFSOPntr->fso));
    ReturnCode = ErrorCode;
    goto ErrorExit;
  }
  ReadLockDone = true;

  memset (fsinfo, -1, sizeof (fs_info));
  fsinfo->dev = DiskVolumePntr->nameSpaceID;
  fsinfo->root = DiskVolumePntr->rootVNodeID;
  /* Possible flags: B_FS_IS_READONLY, B_FS_IS_REMOVABLE, B_FS_IS_PERSISTENT,
  B_FS_IS_SHARED, B_FS_HAS_MIME, B_FS_HAS_ATTR, B_FS_HAS_QUERY. */
  fsinfo->flags = B_FS_HAS_MIME | B_FS_HAS_ATTR | B_FS_HAS_QUERY |
    B_FS_IS_PERSISTENT /* Tracker only does queries on persistent volumes */;
  if (DiskVolumePntr->readOnly)
    fsinfo->flags |= B_FS_IS_READONLY;
  fsinfo->block_size = 1; /* Byte sized in effect. */
  fsinfo->io_size = 256 * 1024; /* Optimal IO size. */
  fsinfo->total_blocks = DiskVolumePntr->maxTotalFileSize;
  fsinfo->free_blocks =
    DiskVolumePntr->maxTotalFileSize - DiskVolumePntr->totalFileSize;
  /* fsinfo->total_nodes = -1; */
  /* fsinfo->free_nodes = -1; */
  strcpy (fsinfo->device_name, "");
  strcpy (fsinfo->volume_name, DiskVolumePntr->rootFSOPntr->fso.trueNamePntr);
  strcpy (fsinfo->fsh_name, "ARFSv0");

  ReturnCode = 0;

#if DPRINT_NON_ERROR_MESSAGES
  dprintf ("ARFS_rfsstat: "
    "dev:%ld, rootID:%016LX, flags:$%X, blksz:%Ld, iosz:%Ld, "
    "totblk:%Ld, freeblk:%Ld, totnod:%Ld, freenod:%Ld, "
    "devname:\"%s\", volname:\"%s\", fsname:\"%s\".\n",
    fsinfo->dev, fsinfo->root, (int) fsinfo->flags,
    fsinfo->block_size, fsinfo->io_size,
    fsinfo->total_blocks, fsinfo->free_blocks,
    fsinfo->total_nodes, fsinfo->free_nodes,
    fsinfo->device_name, fsinfo->volume_name, fsinfo->fsh_name);
#endif

ErrorExit:
#if DPRINT_NON_ERROR_MESSAGES
  dprintf ("ARFS_rfsstat exited, return code %d.\n", (int) ReturnCode);
#endif
  if (ReadLockDone)
    ReadUnlock (&DiskVolumePntr->rootFSOPntr->fso);
  return ReturnCode;
}



/******************************************************************************
 * Used for setting (W for writing) information about the file system.
 * Currently just the volume name, with WFSSTAT_NAME (1) for the mask.  Maybe
 * also the file system read-only flag too?  I'll step in and suggest defining
 * WFSSTAT_USEFLAG_READONLY (2) which will copy the read only bit
 * (B_FS_IS_READONLY) from fs_info.flags into the active setting used by the
 * file system.  If you change to read only mode, open files can still be
 * written to, but new opens are required to be read-only.
 */

int ARFS_wfsstat (
  void *ns,
  struct fs_info *fsinfo,
  long mask)
{
  DiskVolumePointer  DiskVolumePntr = ns;
  status_t           ErrorCode;
  char              *NewNamePntr;
  status_t           ReturnCode = -1;
  int                StringLength;
  bool               WriteLockDone = false;

#if DPRINT_NON_ERROR_MESSAGES
  dprintf ("ARFS_wfsstat entered, NS:$%08X, fsinfo:$%08X, mask:$%X.\n",
    (int) ns, (int) fsinfo, (int) mask);
#endif

  if (INVALID_DISK_VOLUME (DiskVolumePntr))
  {
    dprintf ("ARFS_wfsstat: invalid input disk volume pointer $%X, "
      "don't know what to get status of.\n", (unsigned int) DiskVolumePntr);
    goto ErrorExit;
  }

  if (fsinfo == NULL)
  {
    dprintf ("ARFS_wfsstat: NULL fsinfo pointer from caller.\n");
    goto ErrorExit;
  }

  if (INVALID_FILE_SYSTEM_OBJECT (&DiskVolumePntr->rootFSOPntr->fso))
  {
    dprintf ("ARFS_wfsstat: Invalid root directory FSO $%X.\n",
      (unsigned int) DiskVolumePntr->rootFSOPntr);
    goto ErrorExit;
  }

  ErrorCode = WriteLock (&DiskVolumePntr->rootFSOPntr->fso);
  if (ErrorCode != 0)
  {
    dprintf ("ARFS_wfsstat: Unable to get write lock for %s.\n",
      FSOAsString (&DiskVolumePntr->rootFSOPntr->fso));
    ReturnCode = ErrorCode;
    goto ErrorExit;
  }
  WriteLockDone = true;

  if (mask & WFSSTAT_NAME)
  {
    StringLength = strlen (fsinfo->volume_name);
    if (StringLength <= 0 || StringLength >= B_FILE_NAME_LENGTH)
    {
      dprintf ("ARFS_wfsstat: New volume name string has a bad length.\n");
      ReturnCode = ENAMETOOLONG;
      goto ErrorExit;
    }
    NewNamePntr = mallocDVP (DiskVolumePntr, StringLength + 1);
    if (NewNamePntr == NULL)
    {
      dprintf ("ARFS_wfsstat: Unable to allocate memory for new "
        "volume name.\n");
      ReturnCode = ENOMEM;
      goto ErrorExit;
    }
    strcpy (NewNamePntr, fsinfo->volume_name);

#if DPRINT_NON_ERROR_MESSAGES
    dprintf ("ARFS_wfsstat: Changing volume name from \"%s\" to \"%s\".\n",
      DiskVolumePntr->rootFSOPntr->fso.trueNamePntr, NewNamePntr);
#endif

    freeDVP (DiskVolumePntr, DiskVolumePntr->rootFSOPntr->fso.trueNamePntr);
    DiskVolumePntr->rootFSOPntr->fso.trueNamePntr = NewNamePntr;
  }

  if (mask & WFSSTAT_USEFLAG_READONLY)
  {
    DiskVolumePntr->readOnly = ((fsinfo->flags & B_FS_IS_READONLY) != 0);
    dprintf ("ARFS_wfsstat: Set read only flag for volume \"%s\" to %d.\n",
      DiskVolumePntr->rootFSOPntr->fso.trueNamePntr,
      (int) DiskVolumePntr->readOnly);
  }

  ReturnCode = 0;

ErrorExit:
#if DPRINT_NON_ERROR_MESSAGES
  dprintf ("ARFS_wfsstat exited, return code %d.\n", (int) ReturnCode);
#endif
  if (WriteLockDone)
    WriteUnlock (&DiskVolumePntr->rootFSOPntr->fso);
  return ReturnCode;
}



/******************************************************************************
 * Start reading the list of names of indices, much like opendir reads names of
 * things in a directory.
 */

int ARFS_open_indexdir (
  void *ns,
  void **cookie)
{
  DiskVolumePointer      DiskVolumePntr = ns;
  DirectoryObjectPointer IndicesDirPntr;
  status_t               ReturnCode = -1;

#if DPRINT_NON_ERROR_MESSAGES
  dprintf ("ARFS_open_indexdir entered, NS:$%X, cookie: $%X.\n",
    (unsigned int) ns, (unsigned int) cookie);
#endif

  if (INVALID_DISK_VOLUME (DiskVolumePntr))
  {
    dprintf ("ARFS_open_indexdir: invalid input disk volume pointer $%X.\n",
      (unsigned int) DiskVolumePntr);
    goto ErrorExit;
  }

  IndicesDirPntr = GetAndLockIndicesDirectory (DiskVolumePntr, NULL, NULL);
  if (IndicesDirPntr == NULL)
  {
    dprintf ("ARFS_open_indexdir: Can't find indices directory "
      "named \"%s\".\n", g_IndicesDirectoryName);
    ReturnCode = ENOENT;
    goto ErrorExit;
  }

  /* Can't use PutIndicesDirectory().  Only need to leave the get_vnode pending
  so that the indices directory doesn't get unloaded or even deleted while
  we're using it.  GenericFreeDirCookie will do the put_vnode for us. */

  ReadUnlock (&IndicesDirPntr->fso);

  ReturnCode = GenericOpenDir (ns, IndicesDirPntr, cookie, GDV_INDEX_CONTENTS,
    NULL, NULL);

ErrorExit:
#if DPRINT_NON_ERROR_MESSAGES
  dprintf ("ARFS_open_indexdir exited, return code %d.\n", (int) ReturnCode);
#endif
  return ReturnCode;
}



/******************************************************************************
 * Finished reading the list of indices, but don't free the cookie yet as other
 * threads may still be reading.
 */

int ARFS_close_indexdir (
  void *ns,
  void *cookie)
{
  GenericDirCookiePointer CookiePntr = cookie;

#if DPRINT_NON_ERROR_MESSAGES
  dprintf ("ARFS_close_indexdir entered, NS:$%X, cookie:$%X.\n",
    (unsigned int) ns, (unsigned int) cookie);
#endif

  return GenericCloseDir (ns, CookiePntr->fsoPntr, cookie);
}



/******************************************************************************
 * The last thread has finished reading the list of indices, you can now really
 * close it.  Deallocate the cookie and otherwise clean up here.  Note that
 * node is NULL (there is no VNodeID for the index directory under BFS) but the
 * argument is still present.
 */

int ARFS_free_indexdircookie (
  void *ns,
  void *node,
  void *cookie)
{
  GenericDirCookiePointer CookiePntr = cookie;

#if DPRINT_NON_ERROR_MESSAGES
  dprintf ("ARFS_free_indexdircookie entered, NS:$%X, node:$%X, cookie:$%X.\n",
    (unsigned int) ns, (unsigned int) node, (unsigned int) cookie);
#endif

  return GenericFreeDirCookie (ns, CookiePntr->fsoPntr, cookie);
}



/******************************************************************************
 * See rewinddir, except that this is for the names of the indices.
 */

int ARFS_rewind_indexdir (void *ns, void *cookie)
{
  GenericDirCookiePointer CookiePntr = cookie;

#if DPRINT_NON_ERROR_MESSAGES
  dprintf ("ARFS_rewind_indexdir entered, NS:$%X, cookie:$%X.\n",
    (unsigned int) ns, (unsigned int) cookie);
#endif

  return GenericRewindDir (ns, CookiePntr->fsoPntr, cookie);
}



/******************************************************************************
 * See readdir, except that this is for the names of the indices.
 */

int ARFS_read_indexdir (
  void *ns,
  void *cookie,
  long *num,
  struct dirent *buf,
  size_t bufsize)
{
  GenericDirCookiePointer CookiePntr = cookie;

#if DPRINT_NON_ERROR_MESSAGES
  dprintf ("ARFS_read_indexdir entered, "
    "NS:$%X, cookie:$%X, *num:%d, buf:$%X, bufsize:%d.\n",
    (unsigned int) ns, (unsigned int) cookie,
    (int) *num, (unsigned int) buf, (int) bufsize);
#endif

  return GenericReadDir (ns, CookiePntr->fsoPntr, cookie, num, buf, bufsize);
}



/******************************************************************************
 * Create a new index for attributes with the given name.  Fails if it already
 * exists.  Futhur changes to attributes with the same name will add/remove the
 * file's vnode_id and the attribute value from the index.  Unfortunately
 * attributes from pre-existing files aren't automatically added to the index
 * under BFS, though you could do that in your file system.  I've
 * (AGMS20020630) defined the flag B_CREATE_INDEX_WITH_REBUILD (0x00000001) to
 * enable filling the index with the files that should be there.  The type
 * specifies the type of the index for sorting purposes.  It doesn't exactly
 * have to match the attribute type used in files, so you could have MIME
 * string attributes (B_MIME_STRING_TYPE) using the generic string
 * (B_STRING_TYPE) for the index type (a convenient idea, otherwise the file
 * system would have to find out about new types somehow).  Note that the
 * PFSDWTBFS book says that they should match, but I don't think it's enforced.
 * BFS supports B_INT32_TYPE, B_INT64_TYPE, B_FLOAT_TYPE, B_DOUBLE_TYPE,
 * B_STRING_TYPE.  I've also added a related flag,
 * B_CREATE_INDEX_MULTIPLE_KEYWORDS (0x00000002), which will split up strings
 * into multiple keywords (separated by spaces or control characters) and treat
 * fixed size types like arrays of the base type.  Each of the multiple
 * keywords gets added to the index just like a single attribute would have
 * been.  This does mean that the query system users have to put up with
 * getting the same file several times as a query result.
 */

int ARFS_create_index (
  void *ns,
  const char *name,
  int type,
  int flags)
{
  DiskVolumePointer       DiskVolumePntr = ns;
  status_t                ErrorCode;
  bool                    GotWriteLock = false;
  int                     i;
  IndexObjectPointer      IndexPntr = NULL;
  DirectoryObjectPointer  IndicesDirPntr = NULL;
  AVLDupThingRecord       Key;
  status_t                ReturnCode = -1;
  unsigned int            TempUINT;

#if DPRINT_NON_ERROR_MESSAGES
  dprintf ("ARFS_create_index entered, NS:$%X, name:\"%s\", type:%s, "
    "flags:$%X.\n", (unsigned int) ns, name, TypeCodeAsString (type), flags);
#endif

  if (name == NULL)
  {
    dprintf ("ARFS_create_index: NULL arguments.\n");
    goto ErrorExit;
  }

  if (AVLDupGetThingTypeSize (type) < 0)
  {
    dprintf ("ARFS_create_index: Can't make index \"%s\" because type %s "
      "isn't a recognized one - no way to sort it so it can't be used for "
      "an index.\n", name, TypeCodeAsString (type));
    ReturnCode = ERANGE;
    goto ErrorExit;
  }

  if (INVALID_DISK_VOLUME (DiskVolumePntr))
  {
    dprintf ("ARFS_create_index: invalid input disk volume pointer $%X.\n",
      (unsigned int) DiskVolumePntr);
    goto ErrorExit;
  }

  if (INVALID_FILE_SYSTEM_OBJECT (&DiskVolumePntr->rootFSOPntr->fso))
  {
    dprintf ("ARFS_create_index: Invalid root directory FSO $%X.\n",
      (unsigned int) DiskVolumePntr->rootFSOPntr);
    goto ErrorExit;
  }

  IndicesDirPntr = GetAndLockIndicesDirectory (DiskVolumePntr, NULL, NULL);
  if (IndicesDirPntr == NULL)
  {
    dprintf ("ARFS_create_index: Couldn't find indices directory \"%s\" "
      "when trying to create index \"%s\".\n",
      g_IndicesDirectoryName, name);
    ReturnCode = ENOENT;
    goto ErrorExit;
  }

  /* See if the index is already there. */

  Key.longStringThing.stringPntr = (char *) name;
  Key.longStringThing.isLongString = true;
  TempUINT = AVLDupFind (IndicesDirPntr->directory, &Key);
  if (TempUINT != 0)
  {
    dprintf ("ARFS_create_index: %s is already in the index directory.\n",
      FSOAsString ((FileSystemObjectPointer) TempUINT));
    ReturnCode = EEXIST;
    goto ErrorExit;
  }

  /* Upgrade the indices directory lock to a write lock. */

  ErrorCode = ConvertReadLockToWriteLock (&IndicesDirPntr->fso);
  if (ErrorCode < 0)
  {
    dprintf ("ARFS_create_index: Unable to get a write lock on %s.\n",
      FSOAsString (&IndicesDirPntr->fso));
    ReturnCode = ErrorCode;
    goto ErrorExit;
  }
  GotWriteLock = true;

  /* Create the index. */

  IndexPntr = AllocateIndexFSO (DiskVolumePntr, name, type,
    (flags & B_CREATE_INDEX_MULTIPLE_KEYWORDS) != 0);

  if (IndexPntr == NULL)
  {
    dprintf ("ARFS_create_index: Unable to create the index \"%s\", "
      "out of memory.\n", name);
    ReturnCode = ENOMEM;
    goto ErrorExit;
  }

  /* Add the index to the index directory. */

  IndexPntr->fso.modeBits |=
    (S_IRUSR | S_IXUSR | S_IRGRP | S_IXGRP | S_IROTH | S_IXOTH);

  ErrorCode = AddFreshFSOToDirectory (DiskVolumePntr,
    &IndexPntr->fso, IndicesDirPntr);
  if (ErrorCode < 0)
  {
    DeallocateAnyFSO (DiskVolumePntr, &IndexPntr->fso);
    IndexPntr = NULL;

    dprintf ("ARFS_create_index: Failed to add index \"%s\" to indices "
      "%s, code %d.\n",
      name, FSOAsString (&IndicesDirPntr->fso), (int) ErrorCode);
    ReturnCode = ErrorCode;
    goto ErrorExit;
  }

  ChangeFSODates (DiskVolumePntr, &IndicesDirPntr->fso, DK_MODIFIED_FLAG, -1);

  /* See if it is one of the standard indices, which have their vnode_ids saved
  for quick access. */

  for (i = 0; i < STDINUM_MAX; i++)
  {
    if (strcmp (name, g_NamesOfStandardIndices [i]) == 0)
    {
#if DPRINT_NON_ERROR_MESSAGES
      dprintf ("ARFS_create_index: Replacing standard system index for "
        "\"%s\", was %s, now will be %s.\n",
        name, FSOAsString ((FileSystemObjectPointer) (unsigned int)
        DiskVolumePntr->standardIndicesVnodes[i]),
        FSOAsString (&IndexPntr->fso));
#endif
      DiskVolumePntr->standardIndicesVnodes[i] = (unsigned int) IndexPntr;
    }
  }

#if DPRINT_NON_ERROR_MESSAGES
  dprintf ("ARFS_create_index: Successfully made %s in %s, type %s.\n",
    FSOAsString (&IndexPntr->fso), FSOAsString (&IndicesDirPntr->fso),
    TypeCodeAsString (type));
#endif

  ErrorCode = notify_listener (B_ENTRY_CREATED,
    DiskVolumePntr->nameSpaceID,
    (unsigned int) IndicesDirPntr, 0,
    (unsigned int) IndexPntr, name);
  if (ErrorCode != 0 || DPRINT_NOTIFICATIONS)
  {
    dprintf ("ARFS_create_index: B_ENTRY_CREATED notification for "
      "creating %s in %s, error code %d.\n",
      FSOAsString (&IndexPntr->fso), FSOAsString (&IndicesDirPntr->fso),
      (int) ErrorCode);
  }

  /* Successful. */

  ReturnCode = 0;

ErrorExit:
#if DPRINT_NON_ERROR_MESSAGES
  dprintf ("ARFS_create_index exited, return code %d.\n", (int) ReturnCode);
#endif
  /* PutIndicesDirectory() can't be used due to write lock. */
  if (IndicesDirPntr != NULL)
  {
    if (GotWriteLock)
      WriteUnlock (&IndicesDirPntr->fso);
    else
      ReadUnlock (&IndicesDirPntr->fso);
    put_vnode (DiskVolumePntr->nameSpaceID, (unsigned int) IndicesDirPntr);
  }
  return ReturnCode;
}



/******************************************************************************
 * Immediately deletes the given index.  Zap!  Probably also blows away any
 * currently open queries using that index, or they switch to unindexed mode.
 */

int ARFS_remove_index (
  void *ns,
  const char *name)
{
  DiskVolumePointer       DiskVolumePntr = ns;
  int                     i;
  IndexObjectPointer      IndexPntr = NULL;
  bool                    IndicesDirGetVNodeDone = false;
  DirectoryObjectPointer  IndicesDirPntr = NULL;
  bool                    IndicesDirReadLockDone = false;
  AVLDupThingRecord       Key;
  status_t                ReturnCode = -1;
  RDWorkStateRecord       State;
  unsigned int            TempUINT;

#if DPRINT_NON_ERROR_MESSAGES
  dprintf ("ARFS_remove_index entered, NS:$%X, name:\"%s\".\n",
    (unsigned int) ns, name);
#endif

  if (name == NULL)
  {
    dprintf ("ARFS_remove_index: invalid input NULL arguments.\n");
    goto ErrorExit;
  }

  if (INVALID_DISK_VOLUME (DiskVolumePntr))
  {
    dprintf ("ARFS_remove_index: invalid input disk volume pointer $%X.\n",
      (unsigned int) DiskVolumePntr);
    goto ErrorExit;
  }

  IndicesDirPntr = GetAndLockIndicesDirectory (DiskVolumePntr, NULL, NULL);
  if (IndicesDirPntr == NULL)
  {
    dprintf ("ARFS_remove_index: Can't find indices directory "
      "named \"%s\".\n", g_IndicesDirectoryName);
    ReturnCode = ENOENT;
    goto ErrorExit;
  }
  IndicesDirGetVNodeDone = true;
  IndicesDirReadLockDone = true;

  /* See if the index is there. */

  Key.longStringThing.stringPntr = (char *) name;
  Key.longStringThing.isLongString = true;
  TempUINT = AVLDupFind (IndicesDirPntr->directory, &Key);
  IndexPntr = (IndexObjectPointer) TempUINT;

  if (IndexPntr == NULL)
  {
    dprintf ("ARFS_remove_index: Can't find \"%s\" in the "
      "indices directory.\n", name);
    ReturnCode = ENOENT;
    goto ErrorExit;
  }

  /* Is it really an index?  Since the indices directory is locked, it can't be
  deleted while we're looking at it (though fields could change since the index
  isn't read locked, but we only are looking at the object type which never
  changes). */

  if (IndexPntr->fso.objectType != FSO_TYPE_INDEX)
  {
    dprintf ("ARFS_remove_index: %s isn't an index object!\n",
      FSOAsString (&IndexPntr->fso));
    ReturnCode = ENOTDIR;
    goto ErrorExit;
  }

  /* Deactivate this index if it is a standard index.  Done before the delete,
  in case you are deleting the date last modified index (which would get
  updated by the change in the indice's directory date).  Of course, if the
  delete fails then the standard index pointer will be wrong, but that should
  be rare and unimportant. */

  for (i = 0; i < STDINUM_MAX; i++)
  {
    if (DiskVolumePntr->standardIndicesVnodes[i] == (unsigned int) IndexPntr)
    {
#if DPRINT_NON_ERROR_MESSAGES
      dprintf ("ARFS_remove_index: Forgetting standard system index for "
        "\"%s\", was %s.\n", name, FSOAsString (&IndexPntr->fso));
#endif
      DiskVolumePntr->standardIndicesVnodes[i] = 0;
    }
  }

  ReadUnlock (&IndicesDirPntr->fso);
  IndicesDirReadLockDone = false;

  /* Set up things to delete the index, or something with a similar name. */

  memset (&State, 0, sizeof (State));
  State.diskVolumePntr = DiskVolumePntr;
  State.destinationDirectoryPntr = IndicesDirPntr;
  State.newNamePntr = name;
  State.rmdirNotUnlink = true; /* An index is a bit like a directory. */

  ReturnCode = RenameOrDeleteWithRetry (&State);

ErrorExit:
#if DPRINT_NON_ERROR_MESSAGES
  dprintf ("ARFS_remove_index exited, return code %d.\n", (int) ReturnCode);
#endif
  if (IndicesDirReadLockDone)
    ReadUnlock (&IndicesDirPntr->fso);
  if (IndicesDirGetVNodeDone)
    put_vnode (DiskVolumePntr->nameSpaceID, (unsigned int) IndicesDirPntr);
  return ReturnCode;
}



/******************************************************************************
 * Find size and type of an index, also gives ownership and a few datestamps.
 * Returns an error if it doesn't exist (useful for quickly checking if a
 * particular index exists).
 */

int ARFS_stat_index (
  void *ns,
  const char *name,
  struct index_info *buf)
{
  DiskVolumePointer       DiskVolumePntr = ns;
  status_t                ErrorCode;
  bool                    GetIndexVNodeDone = false;
  IndexObjectPointer      IndexPntr = NULL;
  DirectoryObjectPointer  IndicesDirPntr = NULL;
  AVLDupThingRecord       Key;
  bool                    ReadLockIndexDone = false;
  status_t                ReturnCode = -1;
  unsigned int            TempUINT;
  void                   *VoidPntr;

#if DPRINT_NON_ERROR_MESSAGES
  dprintf ("ARFS_stat_index entered, NS:$%X, name:\"%s\", buf:$%X.\n",
    (unsigned int) ns, name, (unsigned int) buf);
#endif

  if (name == NULL || buf == NULL)
  {
    dprintf ("ARFS_stat_index: invalid input NULL arguments.\n");
    goto ErrorExit;
  }

  if (INVALID_DISK_VOLUME (DiskVolumePntr))
  {
    dprintf ("ARFS_stat_index: invalid input disk volume pointer $%X.\n",
      (unsigned int) DiskVolumePntr);
    goto ErrorExit;
  }

  IndicesDirPntr = GetAndLockIndicesDirectory (DiskVolumePntr, NULL, NULL);
  if (IndicesDirPntr == NULL)
  {
    dprintf ("ARFS_stat_index: Can't find indices directory "
      "named \"%s\".\n", g_IndicesDirectoryName);
    ReturnCode = ENOENT;
    goto ErrorExit;
  }

  /* See if the index is there. */

  Key.longStringThing.stringPntr = (char *) name;
  Key.longStringThing.isLongString = true;
  TempUINT = AVLDupFind (IndicesDirPntr->directory, &Key);
  if (TempUINT == 0)
  {
    dprintf ("ARFS_stat_index: Can't find \"%s\" in the "
      "indices directory.\n", name);
    ReturnCode = ENOENT;
    goto ErrorExit;
  }
  IndexPntr = (IndexObjectPointer) TempUINT;

  /* Load the index file system object, since the OS hasn't done it for us. */

  ErrorCode = get_vnode (DiskVolumePntr->nameSpaceID,
    (unsigned int) IndexPntr, &VoidPntr);
  if (ErrorCode != 0)
  {
    dprintf ("ARFS_stat_index: get_vnode failed for %s, code %d!\n",
      FSOAsString (&IndexPntr->fso), (int) ErrorCode);
    ReturnCode = ErrorCode;
    goto ErrorExit;
  }
  GetIndexVNodeDone = true;

  if (VoidPntr != IndexPntr ||
  INVALID_FILE_SYSTEM_OBJECT (&IndexPntr->fso))
  {
    dprintf ("ARFS_stat_index: Corrupt index for \"%s\": %s.\n",
      name, FSOAsString (&IndexPntr->fso));
    goto ErrorExit;
  }

  if (IndexPntr->fso.objectType != FSO_TYPE_INDEX)
  {
    dprintf ("ARFS_stat_index: %s isn't an index object!\n",
      FSOAsString (&IndexPntr->fso));
    ReturnCode = ENOTDIR;
    goto ErrorExit;
  }

  ErrorCode = ReadLock (&IndexPntr->fso, false);
  if (ErrorCode != 0)
  {
    dprintf ("ARFS_stat_index: Read lock failed for %s, code %d!\n",
      FSOAsString (&IndexPntr->fso), (int) ErrorCode);
    ReturnCode = ErrorCode;
    goto ErrorExit;
  }
  ReadLockIndexDone = true;

  /* Output the stat info about the index. */

  memset (buf, 0, sizeof (index_info));
  buf->type = IndexPntr->keyType;
  buf->size = AVLDupGetTreeCount (IndexPntr->treePntr);
  buf->modification_time = IndexPntr->fso.mtime;
  buf->creation_time = IndexPntr->fso.crtime;
  buf->uid = IndexPntr->fso.uid;
  buf->gid = IndexPntr->fso.gid;

  ReturnCode = 0;

ErrorExit:
#if DPRINT_NON_ERROR_MESSAGES
  dprintf ("ARFS_stat_index exited, return code %d.\n", (int) ReturnCode);
#endif
  if (ReadLockIndexDone)
    ReadUnlock (&IndexPntr->fso);
  if (GetIndexVNodeDone)
    put_vnode (DiskVolumePntr->nameSpaceID, (unsigned int) IndexPntr);
  if (IndicesDirPntr != NULL)
    PutIndicesDirectory (DiskVolumePntr, IndicesDirPntr, NULL, NULL);
  return ReturnCode;
}



/******************************************************************************
 * Starts examining the list of attributes associated with a file (or
 * directory, or even a symbolic link if your implementation allows that).
 * Otherwise similar to open_dir in operation.  Most of these operations are
 * similar to directories, except that instead of file names in a directory you
 * are dealing with attribute names of a file.
 */

int ARFS_open_attrdir (
  void *ns,
  void *node,
  void **cookie)
{
#if DPRINT_NON_ERROR_MESSAGES
  dprintf ("ARFS_open_attrdir entered, NS:$%X, node:$%X, cookie: $%X.\n",
    (unsigned int) ns, (unsigned int) node, (unsigned int) cookie);
#endif

  return GenericOpenDir (ns, node, cookie, GDV_ATTRIBUTE_LISTING, NULL, NULL);
}



/******************************************************************************
 * Stop accessing the attribute list, but don't free the cookie just yet (other
 * threads could be reading the list).
 */

int ARFS_close_attrdir (
  void *ns,
  void *node,
  void *cookie)
{
#if DPRINT_NON_ERROR_MESSAGES
  dprintf ("ARFS_close_attrdir entered, NS:$%X, node:$%X, cookie:$%X.\n",
    (unsigned int) ns, (unsigned int) node, (unsigned int) cookie);
#endif

  return GenericCloseDir (ns, node, cookie);
}



/******************************************************************************
 * The last thread has finished reading the list of attributes, you can now
 * really close it.  Deallocate the cookie and otherwise clean up here.
 */

int ARFS_free_attrdircookie (
  void *ns,
  void *node,
  void *cookie)
{
#if DPRINT_NON_ERROR_MESSAGES
  dprintf ("ARFS_free_attrdircookie entered, NS:$%X, node:$%X, cookie:$%X.\n",
    (unsigned int) ns, (unsigned int) node, (unsigned int) cookie);
#endif

  return GenericFreeDirCookie (ns, node, cookie);
}



/******************************************************************************
 * Go back to the first attribute in the list.
 */

int ARFS_rewind_attrdir (
  void *ns,
  void *node,
  void *cookie)
{
#if DPRINT_NON_ERROR_MESSAGES
  dprintf ("ARFS_rewind_attrdir entered, NS:$%X, node:$%X, cookie:$%X.\n",
    (unsigned int) ns, (unsigned int) node, (unsigned int) cookie);
#endif

  return GenericRewindDir (ns, node, cookie);
}



/******************************************************************************
 * Get the names of a bunch of attributes.  Much like readdir() does.  Remember
 * to use dirent.d_reclen correctly!
 */

int ARFS_read_attrdir (
  void *ns,
  void *node,
  void *cookie,
  long *num,
  struct dirent *buf,
  size_t bufsize)
{
#if DPRINT_NON_ERROR_MESSAGES
  dprintf ("ARFS_read_attrdir entered, "
    "NS:$%X, node:$%X, cookie:$%X, *num:%d, buf:$%X, bufsize:%d.\n",
    (unsigned int) ns, (unsigned int) node, (unsigned int) cookie,
    (int) *num, (unsigned int) buf, (int) bufsize);
#endif

  return GenericReadDir (ns, node, cookie, num, buf, bufsize);
}



/******************************************************************************
 * Write some bytes to the value of the named (by name) attribute of the given
 * file (by shadow vnode).  If the attribute doesn't exist, it is created.  The
 * type is a 4 byte code, usually as 4 ASCII characters, that identifies how
 * the user interprets the attribute (see TypeConstants.h for definitions of
 * things like B_FLOAT_TYPE ('FLOT'), B_TIME_TYPE ('TIME'), B_RECT_TYPE
 * ('RECT') and so on).  For safety, if the new attribute type is zero, the old
 * attribute type should be kept if it exists.  If the attribute is indexed,
 * the index with the same name has to be updated to include the file's
 * vnode_id and attribute value after the value is changed.  Because attributes
 * are stateless, there is no file handle, and all writes provide the position
 * to write at, as well as the usual data buffer and length.  You may wish to
 * deny writes to magic attributes like "size", "name", "last_modified".
 */

int ARFS_write_attr (
  void *ns,
  void *node,
  const char *name,
  int type,
  const void *buf,
  size_t *len,
  off_t pos)
{
  AttributeDataPointer    AttributePntr;
  DiskVolumePointer       DiskVolumePntr = ns;
  status_t                ErrorCode;
  FileSystemObjectPointer FSOPntr = node;
  bool                    GotWriteLock = false;
  AVLDupThingRecord       Key;
  int                     i;
  DirectoryObjectPointer  IndicesDirPntr = NULL;
  AttributeDataPointer    NewAttributePntr = NULL;
  size_t                  NewSize;
  status_t                ReturnCode = -1;
  unsigned int            TempUINT;

#if DPRINT_READ_AND_WRITE_OPERATIONS
  dprintf ("ARFS_write_attr entered, NS:$%X, node:$%X, name:\"%s\", "
    "type:%s, buf:$%X, *len:%u, pos:%Ld.\n",
    (int) ns, (int) node, name, TypeCodeAsString (type),
    (unsigned int) buf, (unsigned int) *len, pos);
#endif

  if (name == NULL || name [0] == 0 || len == NULL ||
  (*len > 0 && buf == NULL))
  {
    dprintf ("ARFS_write_attr: NULL or otherwise bad arguments when writing "
      "to %s.\n", FSOAsString (FSOPntr));
    goto ErrorExit;
  }

  for (i = 0; i < STDINUM_MAX; i++)
  {
    if (strcmp (name, g_NamesOfStandardIndices [i]) == 0)
    {
      dprintf ("ARFS_write_attr: Can't change reserved attribute \"%s\" in "
        "%s,\n", name, FSOAsString (FSOPntr));
      ReturnCode = EPERM;
      goto ErrorExit;
    }
  }

  if (INVALID_DISK_VOLUME (DiskVolumePntr))
  {
    dprintf ("ARFS_write_attr: invalid input disk volume pointer $%X.\n",
      (unsigned int) DiskVolumePntr);
    goto ErrorExit;
  }

  if ((7 & (unsigned int) FSOPntr) != 0)
  {
    dprintf ("ARFS_write_attr: Can't write to fake objects like %s.\n",
      FSOAsString (FSOPntr));
    ReturnCode = EBADF;
    goto ErrorExit;
  }

  if (INVALID_FILE_SYSTEM_OBJECT (FSOPntr))
  {
    dprintf ("ARFS_write_attr: invalid input FSO pointer $%X.\n",
      (unsigned int) FSOPntr);
    goto ErrorExit;
  }

  ErrorCode = WriteLock (FSOPntr);
  if (ErrorCode != 0)
  {
    dprintf ("ARFS_write_attr: Unable to get write lock on %s!\n",
      FSOAsString (FSOPntr));
    ReturnCode = ErrorCode;
    goto ErrorExit;
  }
  GotWriteLock = true;

  ErrorCode = PermissionAllowed (DiskVolumePntr, FSOPntr, W_OK);
  if (ErrorCode != 0)
  {
    dprintf ("ARFS_write_attr: %s doesn't have write access permission "
      "(need the w flag).\n", FSOAsString (FSOPntr));
    ReturnCode = ErrorCode;
    goto ErrorExit;
  }

  NewSize = pos + *len;
  if (pos > 0x70000000 || *len > 0x70000000 || NewSize > 0x70000000)
  {
    dprintf ("ARFS_write_attr: total size %ld (%Ld + %ld) is too big "
      "for %s.\n", NewSize, pos, *len, FSOAsString (FSOPntr));
    ReturnCode = EFBIG;
    goto ErrorExit;
  }

  /* Initialise the attribute storage tree if it doesn't exist yet.  Keys will
  be strings (the attribute names), values will be pointers to the
  AttributeDataRecord. */

  if (FSOPntr->attributesTree == NULL)
  {
    FSOPntr->attributesTree = AVLDupAllocTree (B_STRING_TYPE,
      DiskVolumePntr->mallocPntr, DiskVolumePntr->freePntr,
      DiskVolumePntr->volumePoolID);
    if (FSOPntr->attributesTree == NULL)
    {
      dprintf ("ARFS_write_attr: Unable to allocate attribute tree "
        "for %s.\n", FSOAsString (FSOPntr));
      ReturnCode = ENOMEM;
      goto ErrorExit;
    }
  }

  /* Find the old attribute record, if there is a tree and attribute exists. */

  Key.longStringThing.stringPntr = (char *) name;
  Key.longStringThing.isLongString = true;
  TempUINT = AVLDupFind (FSOPntr->attributesTree, &Key);
  AttributePntr = (AttributeDataPointer) TempUINT;

  if (AttributePntr != NULL && type != AttributePntr->type)
  {
    if (type == 0)
      dprintf ("ARFS_write_attr: Warning - NOT replacing \"%s\" attribute "
        "type %s with new type %s, for %s, because the new type is zero.\n",
        name, TypeCodeAsString (AttributePntr->type), TypeCodeAsString (type),
        FSOAsString (FSOPntr));
    else
      dprintf ("ARFS_write_attr: Warning - replacing \"%s\" attribute "
        "type %s with new type %s, for %s.\n",
        name, TypeCodeAsString (AttributePntr->type), TypeCodeAsString (type),
        FSOAsString (FSOPntr));
  }

  /* Set up the new attribute record.  Try to avoid changing the existing one
  for as long as possible so that errors leave it unchanged.  If possible,
  reuse the old one.  If the new one will be bigger (it never shrinks, the
  caller has to remove and recreate the attribute to do that), allocate space
  for the new one and copy as much of the old data as fits.  Initialise new
  space to zero. */

  if (AttributePntr == NULL || AttributePntr->size < NewSize)
  {
    NewAttributePntr =
      mallocDVP (DiskVolumePntr, NewSize + sizeof (AttributeDataRecord));
    if (NewAttributePntr == NULL)
    {
      dprintf ("ARFS_write_attr: Unable to allocate %ld bytes for %s.\n",
        NewSize, FSOAsString (FSOPntr));
      ReturnCode = ENOMEM;
      goto ErrorExit;
    }
    memset (NewAttributePntr, 0, NewSize + sizeof (AttributeDataRecord));
    NewAttributePntr->size = NewSize;

    if (AttributePntr != NULL)
    {
      /* Include the old attribute data in the new larger one. */

      NewAttributePntr->type = AttributePntr->type;

      if (AttributePntr->size > 0)
        memcpy (NewAttributePntr->value, AttributePntr->value,
        AttributePntr->size);

      if (!AVLDupChangeValueInPlace (FSOPntr->attributesTree, &Key,
      (unsigned int) AttributePntr, (unsigned int) NewAttributePntr))
      {
        dprintf ("ARFS_write_attr: Bug!  Failed to replace entry "
          "of \"%s\" with new replacement value in tree for %s.\n",
          name, FSOAsString (FSOPntr));
        goto ErrorExit;
      }
    }
    else /* No old attributes, just add the new one to the list. */
    {
      ErrorCode = AVLDupAdd (FSOPntr->attributesTree, &Key,
        (unsigned int) NewAttributePntr);
      if (ErrorCode != AVLADD_ADDED_A_NODE)
      {
        dprintf ("ARFS_write_attr: Failed to properly add a tree node, "
          "code %d, for attribute \"%s\" in %s.\n",
          (int) ErrorCode, name, FSOAsString (FSOPntr));
        ReturnCode = ENOMEM;
        goto ErrorExit;
      }
    }
  }
  else /* Will be reusing the existing attribute. */
    NewAttributePntr = NULL;

  /* Remove the old attributes from the index and live queries.  Don't stop for
  errors since the attribute tree is pointing to the new value but we haven't
  gotten rid of the old one yet.  First find and maybe also lock (if it is
  different from FSOPntr) the indices directory for use in updating the index
  corresponding to the attribute.  Can return NULL if there isn't an index
  directory, but that's OK too (only live queries will be processed in that
  case). */

  IndicesDirPntr = GetAndLockIndicesDirectory (DiskVolumePntr, FSOPntr, NULL);

  if (AttributePntr != NULL)
  {
    ErrorCode = AddOrRemoveAttributeInIndex (DiskVolumePntr,
      false /* false for remove operation */, IndicesDirPntr, name,
      AttributePntr, FSOPntr);
    if (ErrorCode < 0)
    {
      dprintf ("ARFS_write_attr: Warning - failed to remove old attributes "
        "from index \"%s\" while changing %s, error code %d.\n",
        name, FSOAsString (FSOPntr), (int) ErrorCode);
      /* No error should be reported to caller. */
    }
  }

  /* If not reusing the old attribute record, switch to the new one. */

  if (NewAttributePntr != NULL)
  {
    if (AttributePntr != NULL)
      freeDVP (DiskVolumePntr, AttributePntr);
    AttributePntr = NewAttributePntr;
    NewAttributePntr = NULL; /* It's been used up. */
  }

  /* Copy the new data into the attribute. */

  if (*len > 0)
    memcpy (AttributePntr->value + pos, buf, *len);

  if (type != 0)
    AttributePntr->type = type;

  /* Update the index and live queries for the new attribute value.  This can
  fail if it runs out of memory, leaving the index inconsistent with what's
  actually in the RAM disk. */

  ErrorCode = AddOrRemoveAttributeInIndex (DiskVolumePntr, true /* add it */,
    IndicesDirPntr, name, AttributePntr, FSOPntr);
  if (ErrorCode < 0)
  {
    dprintf ("ARFS_write_attr: Failed to add values to index for "
      "attribute \"%s\", error code %d.  Note that the actual attribute "
      "attached to %s is correct, just the index is now incorrect and "
      "notifications have been missed.\n",
      name, (int) ErrorCode, FSOAsString (FSOPntr));
    /* The caller doesn't see this error. */
  }

#if DPRINT_READ_AND_WRITE_OPERATIONS
  dprintf ("ARFS_write_attr: Wrote %d bytes at position %Ld into "
    "attribute \"%s\", type %s, total size %ld, in %s, "
    "attribute tree $%X now has %d items.\n",
    (int) *len, pos, name, TypeCodeAsString (type),
    AttributePntr->size, FSOAsString (FSOPntr),
    (int) FSOPntr->attributesTree,
    AVLDupGetTreeCount (FSOPntr->attributesTree));
#endif

  ErrorCode = notify_listener (B_ATTR_CHANGED,
    DiskVolumePntr->nameSpaceID, 0, 0, (unsigned int) FSOPntr, name);
  if (ErrorCode != 0 || DPRINT_NOTIFICATIONS)
  {
    dprintf ("ARFS_write_attr: B_ATTR_CHANGED notification for "
      "writing to attribute \"%s\", error code %d, %s.\n",
      name, (int) ErrorCode, FSOAsString (FSOPntr));
  }

  ReturnCode = 0;

ErrorExit:
  if (ReturnCode != 0 && len != NULL)
    *len = 0;
  if (NewAttributePntr != NULL)
    freeDVP (DiskVolumePntr, NewAttributePntr);
#if DPRINT_READ_AND_WRITE_OPERATIONS
  dprintf ("ARFS_write_attr exited, return code %d.\n", (int) ReturnCode);
#endif
  PutIndicesDirectory (DiskVolumePntr, IndicesDirPntr, FSOPntr, NULL);
  if (GotWriteLock)
    WriteUnlock (FSOPntr);
  return ReturnCode;
}



/******************************************************************************
 * Reads data from the attribute, starting at the given byte offset, up to the
 * specified length in bytes, into the buffer.  When successful, sets the
 * length to the amount actually read.  Probably best to set the length to zero
 * on failure.  If the attribute doesn't exist returns B_ENTRY_NOT_FOUND.  If
 * the attribute type is different from the requested type and neither are zero
 * (don't fail for zero types so that bad programs can continue working with
 * bad data) then return B_BAD_VALUE.  Other access errors are as usual.
 * Actually, the BeBook says it ignores the type and offset.  But it shouldn't!
 */

int ARFS_read_attr (
  void *ns,
  void *node,
  const char *name,
  int type,
  void *buf,
  size_t *len,
  off_t pos)
{
  size_t                  AmountToRead;
  AttributeDataPointer    AttributePntr;
  DiskVolumePointer       DiskVolumePntr = ns;
  status_t                ErrorCode;
  FileSystemObjectPointer FSOPntr = node;
  bool                    GotReadLock = false;
  AVLDupThingRecord       Key;
  status_t                ReturnCode = -1;
  unsigned int            TempUINT;
  bool                    TypesAreCompatible;

#if DPRINT_READ_AND_WRITE_OPERATIONS
  dprintf ("ARFS_read_attr entered, NS:$%X, node:$%X, name:\"%s\", "
    "type:%s, buf:$%X, *len:%u, pos:%Ld.\n",
    (int) ns, (int) node, name, TypeCodeAsString (type),
    (unsigned int) buf, (unsigned int) *len, pos);
#endif

  if (name == NULL || name [0] == 0 || len == NULL ||
  (*len > 0 && buf == NULL))
  {
    dprintf ("ARFS_read_attr: NULL or otherwise bad arguments when writing "
      "to %s.\n", FSOAsString (FSOPntr));
    goto ErrorExit;
  }

  if (INVALID_DISK_VOLUME (DiskVolumePntr))
  {
    dprintf ("ARFS_read_attr: invalid input disk volume pointer $%X.\n",
      (unsigned int) DiskVolumePntr);
    goto ErrorExit;
  }

  if ((7 & (unsigned int) FSOPntr) != 0)
  {
#if DPRINT_NON_ERROR_MESSAGES
    dprintf ("ARFS_read_attr: Can't read fake objects like %s.\n",
      FSOAsString (FSOPntr));
#endif
    ReturnCode = EBADF;
    goto ErrorExit;
  }

  if (INVALID_FILE_SYSTEM_OBJECT (FSOPntr))
  {
    dprintf ("ARFS_read_attr: invalid input FSO pointer $%X.\n",
      (unsigned int) FSOPntr);
    goto ErrorExit;
  }

  ErrorCode = ReadLock (FSOPntr, false);
  if (ErrorCode != 0)
  {
    dprintf ("ARFS_read_attr: Unable to get read lock on %s!\n",
      FSOAsString (FSOPntr));
    ReturnCode = ErrorCode;
    goto ErrorExit;
  }
  GotReadLock = true;

  ErrorCode = PermissionAllowed (DiskVolumePntr, FSOPntr, R_OK);
  if (ErrorCode != 0)
  {
    dprintf ("ARFS_read_attr: %s doesn't have read access permission "
      "(need the r flag).\n", FSOAsString (FSOPntr));
    ReturnCode = ErrorCode;
    goto ErrorExit;
  }

  AmountToRead = *len;
  if (pos > 0x70000000 || AmountToRead > 0x70000000 ||
  pos + AmountToRead > 0x70000000)
  {
    dprintf ("ARFS_read_attr: End offset (%Ld + %ld) is too big "
      "for %s.\n", pos, AmountToRead, FSOAsString (FSOPntr));
    ReturnCode = EFBIG;
    goto ErrorExit;
  }

  /* Find the attribute record. */

  Key.longStringThing.stringPntr = (char *) name;
  Key.longStringThing.isLongString = true;
  TempUINT = AVLDupFind (FSOPntr->attributesTree, &Key);
  AttributePntr = (AttributeDataPointer) TempUINT;

  if (AttributePntr == NULL)
  {
#if DPRINT_NON_ERROR_MESSAGES
    dprintf ("ARFS_read_attr: Attribute \"%s\" doesn't exist for %s.\n",
      name, FSOAsString (FSOPntr));
#endif
    ReturnCode = ENOENT;
    goto ErrorExit;
  }

  /* Check that the attribute type matches the requested type. */

  TypesAreCompatible = (AttributePntr->type == type);

  if ((type == B_MIME_STRING_TYPE || type == B_STRING_TYPE) &&
  (AttributePntr->type == B_MIME_STRING_TYPE ||
  AttributePntr->type == B_STRING_TYPE))
    TypesAreCompatible = true; /* Two string types, with different names. */

  if (AttributePntr->type == 0 || type == 0)
    TypesAreCompatible = true; /* Zero types are backwards compatible wild. */

  if (!TypesAreCompatible ||
  (AttributePntr->type != type && DPRINT_NON_ERROR_MESSAGES))
  {
    dprintf ("ARFS_read_attr: Attribute \"%s\" type %s doesn't match "
      "requested type %s, for %s.  %s.\n", name,
      TypeCodeAsString (AttributePntr->type), TypeCodeAsString (type),
      FSOAsString (FSOPntr),
      TypesAreCompatible
        ? "Will ignore this problem for backwards compatibility reasons"
        : "This is a problem, will return an EINVAL error");
  }

  if (!TypesAreCompatible)
  {
    ReturnCode = EINVAL;
    goto ErrorExit;
  }

  /* Copy the data from the attribute. */

  if (pos >= AttributePntr->size)
    AmountToRead = 0;
  else if (AmountToRead > AttributePntr->size - pos)
    AmountToRead = AttributePntr->size - pos;
  if (AmountToRead > 0)
    memcpy (buf, AttributePntr->value + pos, AmountToRead);

#if DPRINT_READ_AND_WRITE_OPERATIONS
  dprintf ("ARFS_read_attr: Read %d bytes at position %Ld into "
    "attribute \"%s\", type %s, total size %ld, in %s.\n",
    (int) AmountToRead, pos, name, TypeCodeAsString (type),
    AttributePntr->size, FSOAsString (FSOPntr));
#endif

  *len = AmountToRead;
  ReturnCode = 0;

ErrorExit:
  if (ReturnCode != 0 && len != NULL)
    *len = 0;
#if DPRINT_READ_AND_WRITE_OPERATIONS
  dprintf ("ARFS_read_attr exited, return code %d.\n", (int) ReturnCode);
#endif
  if (GotReadLock)
    ReadUnlock (FSOPntr);
  return ReturnCode;
}




/******************************************************************************
 * Deletes the attribute immediately, no waiting for open files to close, as
 * there are no file handles to attributes.  Also removes the attribute value
 * from the index with the same name.
 */

int ARFS_remove_attr (
  void *ns,
  void *node,
  const char *name)
{
  DiskVolumePointer       DiskVolumePntr = ns;
  status_t                ErrorCode;
  FileSystemObjectPointer FSOPntr = node;
  bool                    GotWriteLock = false;
  DirectoryObjectPointer  IndicesDirPntr;
  AVLDupThingRecord       Key;
  int                     NumberOfAttributes;
  status_t                ReturnCode = -1;
  unsigned int            TempUINT;

#if DPRINT_NON_ERROR_MESSAGES
  dprintf ("ARFS_remove_attr entered, NS:$%X, node:$%X, name:\"%s\".\n",
    (int) ns, (int) node, name);
#endif

  if (name == NULL || name [0] == 0)
  {
    dprintf ("ARFS_remove_attr: NULL or otherwise bad arguments "
      "when removing attributes from %s.\n", FSOAsString (FSOPntr));
    goto ErrorExit;
  }

  if (INVALID_DISK_VOLUME (DiskVolumePntr))
  {
    dprintf ("ARFS_remove_attr: invalid input disk volume pointer $%X.\n",
      (unsigned int) DiskVolumePntr);
    goto ErrorExit;
  }

  if ((7 & (unsigned int) FSOPntr) != 0)
  {
    dprintf ("ARFS_remove_attr: Fake objects like %s don't have attributes.\n",
      FSOAsString (FSOPntr));
    ReturnCode = EBADF;
    goto ErrorExit;
  }

  if (INVALID_FILE_SYSTEM_OBJECT (FSOPntr))
  {
    dprintf ("ARFS_remove_attr: invalid input FSO pointer $%X.\n",
      (unsigned int) FSOPntr);
    goto ErrorExit;
  }

  ErrorCode = WriteLock (FSOPntr);
  if (ErrorCode != 0)
  {
    dprintf ("ARFS_remove_attr: Unable to get write lock on %s!\n",
      FSOAsString (FSOPntr));
    ReturnCode = ErrorCode;
    goto ErrorExit;
  }
  GotWriteLock = true;

  ErrorCode = PermissionAllowed (DiskVolumePntr, FSOPntr, W_OK);
  if (ErrorCode != 0)
  {
    dprintf ("ARFS_remove_attr: %s doesn't have write access permission "
      "(need the w flag).\n", FSOAsString (FSOPntr));
    ReturnCode = ErrorCode;
    goto ErrorExit;
  }

  /* Look for the attribute in the attribute tree for the file.  OK if the tree
  is NULL. */

  Key.longStringThing.stringPntr = (char *) name;
  Key.longStringThing.isLongString = true;
  TempUINT = AVLDupFind (FSOPntr->attributesTree, &Key);

  if ((void *) TempUINT == NULL)
  {
#if DPRINT_NON_ERROR_MESSAGES
    dprintf ("ARFS_remove_attr: %s doesn't have an attribute named "
      "\"%s\".\n", FSOAsString (FSOPntr), name);
#endif
    ReturnCode = ENOENT;
    goto ErrorExit;
  }

  /* Remove the old attribute from the index and live queries. */

  IndicesDirPntr = GetAndLockIndicesDirectory (DiskVolumePntr, FSOPntr, NULL);

  ErrorCode = AddOrRemoveAttributeInIndex (DiskVolumePntr,
    false /* false for remove operation */, IndicesDirPntr, name,
    (AttributeDataPointer) TempUINT, FSOPntr);

  PutIndicesDirectory (DiskVolumePntr, IndicesDirPntr, FSOPntr, NULL);

  if (ErrorCode < 0)
  {
    dprintf ("ARFS_remove_attr: Warning - failed to remove old attribute "
      "from index \"%s\" while changing %s, error code %d.\n",
      name, FSOAsString (FSOPntr), (int) ErrorCode);
    /* No error should be reported to caller. */
  }

  /* Deallocate the attribute's data and remove it from the tree. */

  DeallocateOneAttribute (&Key, &TempUINT, DiskVolumePntr);

  if (!AVLDupDelete (FSOPntr->attributesTree, &Key, TempUINT))
  {
    dprintf ("ARFS_remove_attr: Bug!  Failed to delete \"%s\" from %s "
      "attribute tree.\n", name, FSOAsString (FSOPntr));
    goto ErrorExit;
  }

  /* Free the tree if it is now empty. */

  NumberOfAttributes = AVLDupGetTreeCount (FSOPntr->attributesTree);
  if (NumberOfAttributes <= 0)
  {
    AVLDupFreeTree (FSOPntr->attributesTree);
    FSOPntr->attributesTree = NULL;
  }

#if DPRINT_NON_ERROR_MESSAGES
  dprintf ("ARFS_remove_attr: Successfully removed \"%s\" from %s, "
    "leaving %d attributes in tree at $%X.\n",
    name, FSOAsString (FSOPntr), NumberOfAttributes,
    (int) FSOPntr->attributesTree);
#endif

  ErrorCode = notify_listener (B_ATTR_CHANGED,
    DiskVolumePntr->nameSpaceID, 0, 0, (unsigned int) FSOPntr, name);
  if (ErrorCode != 0 || DPRINT_NOTIFICATIONS)
  {
    dprintf ("ARFS_remove_attr: B_ATTR_CHANGED notification for "
      "removing attribute \"%s\", error code %d, %s.\n",
      name, (int) ErrorCode, FSOAsString (FSOPntr));
  }

  ReturnCode = 0;

ErrorExit:
#if DPRINT_NON_ERROR_MESSAGES
  dprintf ("ARFS_remove_attr exited, return code %d.\n", (int) ReturnCode);
#endif
  if (GotWriteLock)
    WriteUnlock (FSOPntr);
  return ReturnCode;
}



/******************************************************************************
 * Changes the name of the attribute.  Not implemented by BFS.  If you
 * implement it, also remember to update the indices (remove the old name/value
 * and add the new name/value).  Watch out for renaming magic attributes.
 */

int ARFS_rename_attr (
  void *ns,
  void *node,
  const char *oldname,
  const char *newname)
{
  DiskVolumePointer       DiskVolumePntr = ns;
  status_t                ErrorCode;
  FileSystemObjectPointer FSOPntr = node;
  bool                    GotWriteLock = false;
  int                     i;
  DirectoryObjectPointer  IndicesDirPntr = NULL;
  AVLDupThingRecord       Key;
  AttributeDataPointer    OldAttributePntr;
  AttributeDataPointer    OverwrittenAttributePntr;
  status_t                ReturnCode = -1;
  unsigned int            TempUINT;

#if DPRINT_NON_ERROR_MESSAGES
  dprintf ("ARFS_rename_attr entered, NS:$%X, node:$%X, "
    "oldname:\"%s\", newname:\"%s\".\n",
    (int) ns, (int) node, oldname, newname);
#endif

  if (oldname == NULL || oldname [0] == 0 ||
  newname == NULL || newname [0] == 0)
  {
    dprintf ("ARFS_rename_attr: NULL or otherwise bad arguments "
      "when renaming attributes of %s.\n", FSOAsString (FSOPntr));
    goto ErrorExit;
  }

  for (i = 0; i < STDINUM_MAX; i++)
  {
    if (strcmp (newname, g_NamesOfStandardIndices [i]) == 0)
    {
      dprintf ("ARFS_rename_attr: Can't rename to a reserved attribute "
        "\"%s\" in %s,\n", newname, FSOAsString (FSOPntr));
      ReturnCode = EPERM;
      goto ErrorExit;
    }
  }

  if (INVALID_DISK_VOLUME (DiskVolumePntr))
  {
    dprintf ("ARFS_rename_attr: invalid input disk volume pointer $%X.\n",
      (unsigned int) DiskVolumePntr);
    goto ErrorExit;
  }

  if ((7 & (unsigned int) FSOPntr) != 0)
  {
    dprintf ("ARFS_rename_attr: Fake objects like %s don't have attributes.\n",
      FSOAsString (FSOPntr));
    ReturnCode = EBADF;
    goto ErrorExit;
  }

  if (INVALID_FILE_SYSTEM_OBJECT (FSOPntr))
  {
    dprintf ("ARFS_rename_attr: invalid input FSO pointer $%X.\n",
      (unsigned int) FSOPntr);
    goto ErrorExit;
  }

  ErrorCode = WriteLock (FSOPntr);
  if (ErrorCode != 0)
  {
    dprintf ("ARFS_rename_attr: Unable to get write lock on %s!\n",
      FSOAsString (FSOPntr));
    ReturnCode = ErrorCode;
    goto ErrorExit;
  }
  GotWriteLock = true;

  ErrorCode = PermissionAllowed (DiskVolumePntr, FSOPntr, W_OK);
  if (ErrorCode != 0)
  {
    dprintf ("ARFS_rename_attr: %s doesn't have write access permission "
      "(need the w flag).\n", FSOAsString (FSOPntr));
    ReturnCode = ErrorCode;
    goto ErrorExit;
  }

  if (FSOPntr->attributesTree == NULL)
  {
    dprintf ("ARFS_rename_attr: %s doesn't have any attributes, was trying "
      "to rename \"%s\".\n", FSOAsString (FSOPntr), oldname);
    ReturnCode = ENOENT;
    goto ErrorExit;
  }

  /* Look for the old attribute in the attribute tree for the file, nothing to
  do if it isn't there. */

  Key.longStringThing.stringPntr = (char *) oldname;
  Key.longStringThing.isLongString = true;
  TempUINT = AVLDupFind (FSOPntr->attributesTree, &Key);
  OldAttributePntr = (AttributeDataPointer) TempUINT;

  if (OldAttributePntr == NULL)
  {
    dprintf ("ARFS_rename_attr: %s doesn't have an attribute named "
      "\"%s\".\n", FSOAsString (FSOPntr), oldname);
    ReturnCode = ENOENT;
    goto ErrorExit;
  }

  /* Avoid problems if the source and destination are the same. */

  if (strcmp (oldname, newname) == 0)
  {
    ReturnCode = 0; /* Nothing to do if the names are the same. */
    goto ErrorExit;
  }

  IndicesDirPntr = GetAndLockIndicesDirectory (DiskVolumePntr, FSOPntr, NULL);

  /* Look for the overwritten attribute in the attribute tree for the file.
  The plan is to set up the tree node for the new name before removing the old
  one, so if it runs out of memory, it won't change the file's attributes. */

  Key.longStringThing.stringPntr = (char *) newname;
  Key.longStringThing.isLongString = true;
  TempUINT = AVLDupFind (FSOPntr->attributesTree, &Key);
  OverwrittenAttributePntr = (AttributeDataPointer) TempUINT;

  if (OverwrittenAttributePntr != NULL)
  {
    /* Remove the overwritten attribute from the index. */

    ErrorCode = AddOrRemoveAttributeInIndex (DiskVolumePntr,
      false /* false for remove operation */, IndicesDirPntr, newname,
      OverwrittenAttributePntr, FSOPntr);
    if (ErrorCode < 0)
    {
      dprintf ("ARFS_rename_attr: Warning - failed to remove overwritten "
        "attribute from index \"%s\" while changing %s, error code %d.\n",
        newname, FSOAsString (FSOPntr), (int) ErrorCode);
      /* No error should be reported to caller. */
    }

    /* Change the overwritten tree node to point to the new data.  This doesn't
    allocate memory so it shouldn't fail. */

    if (!AVLDupChangeValueInPlace (FSOPntr->attributesTree, &Key,
    (unsigned int) OverwrittenAttributePntr, (unsigned int) OldAttributePntr))
    {
      dprintf ("ARFS_rename_attr: Bug!  Failed to replace entry "
        "of \"%s\" with new replacement value in tree for %s.\n",
        newname, FSOAsString (FSOPntr));
      goto ErrorExit;
    }

    /* Deallocate the overwritten attribute's data.  Also fail-safe. */

    TempUINT = (unsigned int) OverwrittenAttributePntr;
    DeallocateOneAttribute (&Key, &TempUINT, DiskVolumePntr);
  }
  else /* No existing new name node in the tree, add one.  May fail. */
  {
    ErrorCode = AVLDupAdd (FSOPntr->attributesTree, &Key,
      (unsigned int) OldAttributePntr);
    if (ErrorCode != AVLADD_ADDED_A_NODE)
    {
      dprintf ("ARFS_rename_attr: Failed to properly add a tree node, "
        "code %d, for attribute \"%s\" in %s.\n",
        (int) ErrorCode, newname, FSOAsString (FSOPntr));
      ReturnCode = ENOMEM; /* AVLDupAdd error codes are specific to itself. */
      goto ErrorExit;
    }
  }

  /* Remove the renamed attribute from the old name index. */

  ErrorCode = AddOrRemoveAttributeInIndex (DiskVolumePntr,
    false /* false for remove operation */, IndicesDirPntr, oldname,
    OldAttributePntr, FSOPntr);
  if (ErrorCode < 0)
  {
    dprintf ("ARFS_rename_attr: Warning - failed to remove old "
      "attribute from index \"%s\" while changing %s, error code %d.\n",
      oldname, FSOAsString (FSOPntr), (int) ErrorCode);
    /* No error should be reported to caller. */
  }

  /* Delete the old name node.  Fail-safe. */

  Key.longStringThing.stringPntr = (char *) oldname;
  Key.longStringThing.isLongString = true;
  if (!AVLDupDelete (FSOPntr->attributesTree, &Key,
  (unsigned int) OldAttributePntr))
  {
    dprintf ("ARFS_rename_attr: Bug!  Failed to delete \"%s\" from %s "
      "attribute tree.\n", oldname, FSOAsString (FSOPntr));
    goto ErrorExit;
  }

  /* Add the renamed attribute to the index with the new name. */

  ErrorCode = AddOrRemoveAttributeInIndex (DiskVolumePntr, true /* add it */,
    IndicesDirPntr, newname, OldAttributePntr, FSOPntr);
  if (ErrorCode < 0)
  {
    dprintf ("ARFS_rename_attr: Failed to add values to index for new "
      "attribute \"%s\", error code %d.  Note that the actual attribute "
      "attached to %s is correct, just the index is now incorrect and "
      "notifications have been missed.\n",
      newname, (int) ErrorCode, FSOAsString (FSOPntr));
    /* No error reported to the caller. */
  }

#if DPRINT_NON_ERROR_MESSAGES
  dprintf ("ARFS_rename_attr: Successfully renamed \"%s\" to \"%s\" "
    "from %s, leaving %d attributes in tree at $%X.\n",
    oldname, newname, FSOAsString (FSOPntr),
    AVLDupGetTreeCount (FSOPntr->attributesTree),
    (int) FSOPntr->attributesTree);
#endif

  ErrorCode = notify_listener (B_ATTR_CHANGED /* Really need B_ATTR_MOVED */,
    DiskVolumePntr->nameSpaceID, 0, 0, (unsigned int) FSOPntr, newname);
  if (ErrorCode != 0 || DPRINT_NOTIFICATIONS)
  {
    dprintf ("ARFS_rename_attr: B_ATTR_CHANGED notification for "
      "renaming attribute to \"%s\" (was \"%s\"), error code %d, %s.\n",
      newname, oldname, (int) ErrorCode, FSOAsString (FSOPntr));
  }

  ReturnCode = 0;

ErrorExit:
#if DPRINT_NON_ERROR_MESSAGES
  dprintf ("ARFS_rename_attr exited, return code %d.\n", (int) ReturnCode);
#endif
  PutIndicesDirectory (DiskVolumePntr, IndicesDirPntr, FSOPntr, NULL);
  if (GotWriteLock)
    WriteUnlock (FSOPntr);
  return ReturnCode;
}



/******************************************************************************
 * Returns information about the attribute, just size and type.
 */

int ARFS_stat_attr (
  void *ns,
  void *node,
  const char *name,
  struct attr_info *buf)
{
  AttributeDataPointer    AttributePntr;
  DiskVolumePointer       DiskVolumePntr = ns;
  status_t                ErrorCode;
  FileSystemObjectPointer FSOPntr = node;
  bool                    GotReadLock = false;
  AVLDupThingRecord       Key;
  status_t                ReturnCode = -1;
  unsigned int            TempUINT;

#if DPRINT_NON_ERROR_MESSAGES
  dprintf ("ARFS_stat_attr entered, NS:$%X, node:$%X, name:\"%s\", buf:$%X.\n",
    (int) ns, (int) node, name, (unsigned int) buf);
#endif

  if (name == NULL || name [0] == 0 || buf == NULL)
  {
    dprintf ("ARFS_stat_attr: NULL or otherwise bad arguments when "
      "examining %s.\n", FSOAsString (FSOPntr));
    goto ErrorExit;
  }

  if (INVALID_DISK_VOLUME (DiskVolumePntr))
  {
    dprintf ("ARFS_stat_attr: invalid input disk volume pointer $%X.\n",
      (unsigned int) DiskVolumePntr);
    goto ErrorExit;
  }

  if ((7 & (unsigned int) FSOPntr) != 0)
  {
    dprintf ("ARFS_stat_attr: Fake objects like %s don't have attributes.\n",
      FSOAsString (FSOPntr));
    ReturnCode = EBADF;
    goto ErrorExit;
  }

  if (INVALID_FILE_SYSTEM_OBJECT (FSOPntr))
  {
    dprintf ("ARFS_stat_attr: invalid input FSO pointer $%X.\n",
      (unsigned int) FSOPntr);
    goto ErrorExit;
  }

  ErrorCode = ReadLock (FSOPntr, false);
  if (ErrorCode != 0)
  {
    dprintf ("ARFS_stat_attr: Unable to get read lock on %s!\n",
      FSOAsString (FSOPntr));
    ReturnCode = ErrorCode;
    goto ErrorExit;
  }
  GotReadLock = true;

  ErrorCode = PermissionAllowed (DiskVolumePntr, FSOPntr, R_OK);
  if (ErrorCode != 0)
  {
    dprintf ("ARFS_stat_attr: %s doesn't have read access permission "
      "(need the r flag).\n", FSOAsString (FSOPntr));
    ReturnCode = ErrorCode;
    goto ErrorExit;
  }

  /* Find the attribute record. */

  Key.longStringThing.stringPntr = (char *) name;
  Key.longStringThing.isLongString = true;
  TempUINT = AVLDupFind (FSOPntr->attributesTree, &Key);
  AttributePntr = (AttributeDataPointer) TempUINT;

  if (AttributePntr == NULL)
  {
#if DPRINT_NON_ERROR_MESSAGES
    dprintf ("ARFS_stat_attr: Attribute \"%s\" doesn't exist for %s.\n",
      name, FSOAsString (FSOPntr));
#endif
    ReturnCode = ENOENT;
    goto ErrorExit;
  }

  buf->type = AttributePntr->type;
  buf->size = AttributePntr->size;

#if DPRINT_NON_ERROR_MESSAGES
  dprintf ("ARFS_stat_attr: Attribute \"%s\" has type %s, "
    "total size %ld, in %s.\n", name, TypeCodeAsString (AttributePntr->type),
    AttributePntr->size, FSOAsString (FSOPntr));
#endif

  ReturnCode = 0;

ErrorExit:
#if DPRINT_NON_ERROR_MESSAGES
  dprintf ("ARFS_stat_attr exited, return code %d.\n", (int) ReturnCode);
#endif
  if (GotReadLock)
    ReadUnlock (FSOPntr);
  return ReturnCode;
}



/******************************************************************************
 * Use this to start reading the list of files which match a query string.
 * "query" is the string which describes the files you are looking for using
 * the query language (BFS constructs a parse tree from your string, stores it
 * in the cookie, and uses it to match or reject files).
 *
 * If it is a live query, the port is specified and notifications are sent
 * there (your file system calls send_notification to do that) when new
 * matching files appear and old files disappear.  The token is provided by the
 * end user, and is returned with the notification message sent to the port.
 * Under the hood, your file system monitors all attribute changes mentioned in
 * the query string, and when any of them change, re-evaluates the query on
 * that new/deleted file to see if a notification needs to be sent.  So, having
 * lots of live queries slows things down a bit.  Files already in existance
 * get sent to the user via read_query, not notifications.  Files that change
 * the indices after open_query get sent as notification messages (if it would
 * change the value reported by read_query, even if the caller hasn't gotten
 * that far in read_query), and maybe also show up in read_query if the
 * underlying iteration of read_query hasn't reached that file yet.  For
 * example, a new file is created which matches the query, the end user may get
 * informed twice: in read_query (if the iteration hasn't gotten that far in
 * alphabetical or whatever order it is using) and definitely once as a
 * notification.
 *
 * The query language consists of comparison expressions joined by "&&"
 * (logical and) and "||" (logical or) operations, with round parenthesis "()"
 * to control grouping.  && has grouping precedence over || if you don't
 * override it with parenthesis (in general, the same operator precidence is
 * used as in the C programming language).  You can also put a "!" (logical
 * negation) before the expression to reverse its meaning.  The comparison
 * expression has three parts:
 *
 * AttributeName Relation Value
 *
 * The AttributeName is just the name of an attribute, which is case sensitive,
 * and for speed should match an index.  The Relation is one of "==" (equal),
 * "=" (equal), "!=" (not equal), "<" (less than), ">" (greater than), ">="
 * (greater than or equal), "<=" (less than or equal) which have the usual
 * meanings.  The value is a string which gets converted to a number if the
 * attribute is numeric (dates are big numbers: number of seconds since January
 * 1, 1970), or gets converted to a regular expression if the attribute is a
 * string.  Note that hex values (0xabcd1234) can be specified even for
 * floating point numbers.  The same attribute on different files can have a
 * different data type, making comparisons more interesting.  If the attribute
 * isn't present, the file won't be listed.  Regular expressions use "*" to
 * stand for any number (including zero) of any character, "\" to escape the
 * next character, "?" to match any single character, [ad-f] to match a single
 * character specified by the actual characters or a range, [!...] to match a
 * single character not of the specified ones.  The pattern matching works on
 * Unicode characters, so that if you store the file name as UTF-8 multibyte
 * characters, you need to be careful about what counts as a single character
 * for "?" and [a-z] patterns.  For example a "?" will match "事", which is some
 * Japanese character (shows up as the 3 bytes E4, BA, 8B in UTF-8 format).
 *
 * UnaryOperator AttributeName
 *
 * As a new (January 10 2003) feature in AGMSRAMFileSystem, I'm adding a couple
 * of unary operators <> and >< to find all files that have or don't have an
 * attribute.  Use "<>AttributeName" for files with the attribute and
 * "><AttributeName" to find files that don't have it.  Note how >< looks like
 * an X in shape, kind of implying crossing out of something, so I use it to
 * find files that do not have the attribute.  Rather than using new symbols, <
 * and > are reused so that the list of reserved symbols doesn't change.
 *
 * After a bit of experimentation (finding out you can put quotes around
 * attribute names etc), I've come up with this description of the language.
 * The order of the expansions is important, if the first alternative fits
 * (such as QuotedString vs UnquotedString in the definition of String), use it
 * before trying the remaining alternatives.
 *
 * <String> ::= <WhiteSpace> ( <QuotedString> | <UnquotedString> )
 *
 * <WhiteSpace> ::= { <Space-like characters> }
 * Meaning zero or more space or other space-like characters (tabs, newlines,
 * and others - see the standard isspace() function).  I'm not going to insert
 * <WhiteSpace> all over the place in this language description, but it is
 * there in places you'd expect.  In practice, the query will be converted to
 * tokens before being parsed, at that point white space is removed and the
 * parser doesn't have to worry about it.
 *
 * <QuotedString> ::= "\"" { <NonQuote> | "\\\"" } ( "\"" | "\0" )
 * Meaning a quote character followed by zero or more of (non-quote character
 * or backslash quote (\") which puts a quote character in the string) followed
 * by a quote or end of string.  The string will contain the text between the
 * quotes (including white space), with the backslash quote combination for
 * entering one quote into the string.  Other backslash combinations aren't
 * special; they appear in the string as the backslash and the following
 * character.  You can leave off the trailing quote if the QuotedString is the
 * last thing in the query.
 *
 * <UnquotedString> ::= [ <NonSpecial character> ]
 * Meaning one or more of any character except (, ), !, |, &, <, >, =, white
 * space.  The string will be all characters up to the next character that
 * might have special meaning.  Yes, quotes just are ordinary characters; you
 * can use name==*"* to find all files with a quote mark in their name.  No,
 * it's not smart enough to look ahead, so a "=" will end the string even if
 * the next character doesn't make a "==" out of it.  You may want to ignore
 * that BFS simplification, but it doesn't matter since most queries use quoted
 * strings.
 *
 * <Expression> ::= <SimplerExpression> | <SimplerExpression> "||" <Expression>
 * Meaning that the || operator has lowest precedence.  Also implies that a
 * sequence of || operations will be evaluated left to right, though query
 * optimization may change that order.
 *
 * <SimplerExpression> ::= <Term> | <Term> "&&" <SimplerExpression>
 * Meaning that the && operator has the second lowest precedence and associates
 * left to right.
 *
 * <Term> ::= <Comparison> | "(" <Expression> ( ")" | "\0" ) | "!" <Term>
 * Meaning that !  (the not operator) binds tightly, so !a<b&&c<d is equivalent
 * to (!(a<b))&&(c<d).  To save on complexity, the parser can remove !
 * operations and reverse the logic of the items being notted, so that
 * statement would be equivalent to (a>=b)&&(c<d).  Use De Morgan's Law to
 * apply not to logic operations.  Parenthesis let you group subexpressions.
 * You can leave off the trailing ")" for expressions that end at the end of
 * the query string.
 *
 * <Comparison> ::=
 *   <String> <RelationalOperator> <String> | <UnaryOperator> <String>
 * Meaning that the basic comparison operation is done between a named
 * attribute (the first string) and some sort of value (the second string).
 * The second string will be converted to the appropriate data type to match
 * the attribute (if it is an integer attribute, the value string will be
 * converted to an integer, same for floating point attributes, date
 * attributes, etc).  If the conversion fails, it will use some default as the
 * value (usually zero).  In the unary version, the string is the name of an
 * attribute.
 *
 * <RelationalOperator> ::= "<" | "<=" | "==" | "=" | "!=" | ">=" | ">"
 * As a special case for string comparisons, if the value is a pattern
 * (containing * [ ?  or \) and the comparison is == or = or != then pattern
 * matching will be done rather than the usual relational comparison.  BFS
 * overdoes this and gives no results if you use less or greater than with a
 * pattern.  If you are using multiple keyword indices, then the test succeeds
 * if any one of the keywords succeeds.
 *
 * <UnaryOperator> ::= "<>" | "><"
 * These are used for seeing if an attribute exists or if it does not exist.
 */

int ARFS_open_query (
  void *ns,
  const char *query,
  ulong flags,
  port_id port,
  long token,
  void **cookie)
{
  DiskVolumePointer       DiskVolumePntr = ns;
  status_t                ErrorCode;
  DirectoryObjectPointer  IndexDirectoryFSOPntr = NULL;
  IndexObjectPointer      IndexFSOPntr = NULL;
  AVLIterationRangeRecord IterationRange;
  int                     NextTokenIndex;
  ParseNodePointer        ParseTreePntr = NULL;
  char                    PrintableQueryString [500];
  QueryStatePointer       QueryPntr = NULL;
  status_t                ReturnCode = -1;
  TokenizedQueryPointer   TokenizedQueryPntr = NULL;

#if DPRINT_NON_ERROR_MESSAGES
  dprintf ("ARFS_open_query entered, NS:$%X, query:$%X, flags:$%lX, "
    "port:$%lX, token:$%lX, cookie:$%X.\n",
    (int) ns, (unsigned int) query, flags, port, token, (unsigned int) cookie);
#endif

  if (INVALID_DISK_VOLUME (DiskVolumePntr))
  {
    dprintf ("ARFS_open_query: invalid input disk volume pointer $%X.\n",
      (unsigned int) DiskVolumePntr);
    goto ErrorExit;
  }

  if (query == NULL)
  {
    dprintf ("ARFS_open_query: NULL query string.\n");
    goto ErrorExit;
  }

  /* Make a short printable query string for use in error messages, since
  dprintf crashes if it prints out more than about 1 kilobyte of text. */

  if (strlen (query) > sizeof (PrintableQueryString) - 1)
  {
    strncpy (PrintableQueryString, query, sizeof (PrintableQueryString) - 35);
    PrintableQueryString [sizeof (PrintableQueryString) - 35] = 0;
    sprintf (PrintableQueryString + strlen (PrintableQueryString),
      "... and %ld more bytes",
      strlen (query) - (sizeof (PrintableQueryString) - 35));
  }
  else
    strcpy (PrintableQueryString, query);

  /* Tokenize the query string into easier to handle token objects. */

  TokenizedQueryPntr =
    StringToTokenizedQueryRecord (DiskVolumePntr, query);
  if (TokenizedQueryPntr == NULL)
  {
    dprintf ("ARFS_open_query: Unable to tokenize this query: %s\n",
      PrintableQueryString);
    ReturnCode = ENOMEM;
    goto ErrorExit;
  }

  /* Make a parse tree from the tokens to show the order of operations. */

  IndexDirectoryFSOPntr =
    GetAndLockIndicesDirectory (DiskVolumePntr, NULL, NULL);
  if (IndexDirectoryFSOPntr == NULL)
  {
    dprintf ("ARFS_open_query: Unable to find indices directory.\n");
    ReturnCode = ENOTDIR;
    goto ErrorExit;
  }

  NextTokenIndex = 0;
  ErrorCode = TokensToParseTree (DiskVolumePntr, IndexDirectoryFSOPntr,
    TokenizedQueryPntr, &NextTokenIndex, &ParseTreePntr);
  if (ErrorCode != 0)
  {
    dprintf ("ARFS_open_query: Unable to parse this query: %s\n",
      PrintableQueryString);
    ReturnCode = ErrorCode;
    goto ErrorExit;
  }

  /* Check that it parsed the whole query.  Sometimes it just ignores gibberish
  near the end, which may or may not be a good thing. */

  if (NextTokenIndex != TokenizedQueryPntr->numberOfTokens - 1)
  {
    dprintf ("ARFS_open_query: There were leftover tokens, stopped at index "
      "%d when it should have stopped at %d.  While parsing this query: "
      "%s\n", NextTokenIndex, TokenizedQueryPntr->numberOfTokens - 1,
      PrintableQueryString);
    ReturnCode = EINVAL;
    goto ErrorExit;
  }

#if DPRINT_NON_ERROR_MESSAGES
  dprintf ("ARFS_open_query: Parse tree for query {%s}:\n",
    PrintableQueryString);
  {
    char PrintString [500];
    if (!PrintParseTree (ParseTreePntr, PrintString, sizeof (PrintString) - 4))
      strcat (PrintString, "...");
    dprintf ("ARFS_open_query: %s\n", PrintString);
  }
#endif

  /* Find the best index to use and the range of values to iterate over. */

  ErrorCode = FindOptimalIndexAndRange (DiskVolumePntr, IndexDirectoryFSOPntr,
    ParseTreePntr, &IndexFSOPntr, &IterationRange);
  if (ErrorCode != 0)
  {
    dprintf ("ARFS_open_query: Failed to find optimal index, code %d.  "
      "While parsing this query: %s\n", (int) ErrorCode, PrintableQueryString);
    ReturnCode = ErrorCode;
    goto ErrorExit;
  }

  /* Let other threads have access to the indices directory since we've
  finished with it and there's still a bit more processing to do. */

  PutIndicesDirectory (DiskVolumePntr, IndexDirectoryFSOPntr, NULL, NULL);
  IndexDirectoryFSOPntr = NULL;

  /* Don't need the other indices now that we know the best one. */

  PutVnodeParseTree (DiskVolumePntr, ParseTreePntr);

  /* Prepare the query state data which will be persistant so long as the query
  is active.  It holds the parse tree and a few other things so that we can
  test files to see if they match the query. */

  QueryPntr = mallocDVP (DiskVolumePntr, sizeof (QueryStateRecord));
  if (QueryPntr == NULL)
  {
    dprintf ("ARFS_open_query: Unable to allocate memory for the "
      "query state record.  While parsing this query: %s\n",
      PrintableQueryString);
    ReturnCode = ENOMEM;
    goto ErrorExit;
  }
  memset (QueryPntr, 0, sizeof (QueryStateRecord));

  QueryPntr->tokenizedQueryPntr = TokenizedQueryPntr;
  QueryPntr->parseTreePntr = ParseTreePntr;
  if (port <= 0)
    /* Got a static query, does not have a port.  Zero is often used by the
    caller for static queries, even though that is a valid port number. */
    QueryPntr->liveQueryPortID = -1;
  else
    QueryPntr->liveQueryPortID = port;
  QueryPntr->callbackToken = token;

  ErrorCode = GenericOpenDir (ns, IndexFSOPntr, cookie, GDV_QUERY,
    QueryPntr, &IterationRange);
  if (ErrorCode != 0)
  {
    dprintf ("ARFS_open_query: Failed to set up the cookie, code %d.  "
      "While parsing this query: %s\n", (int) ErrorCode, PrintableQueryString);
    ReturnCode = ErrorCode;
    goto ErrorExit;
  }

  /* Query cookie is all set up now.  Make it live if it is live. */

  if (QueryPntr->liveQueryPortID >= 0 &&
  WriteLockLiveQueryList (DiskVolumePntr) == 0)
  {
    QueryPntr->nextLiveQueryPntr = DiskVolumePntr->liveQueryListHead;
    DiskVolumePntr->liveQueryListHead = QueryPntr;
    WriteUnlockLiveQueryList (DiskVolumePntr);
  }

  ReturnCode = 0;
  goto NormalExit;

ErrorExit:
  if (IndexFSOPntr != NULL)
  {
    AVLDupFreeThingArray (&IterationRange.lowerKey,
      IndexFSOPntr->keyType, 1 /* how many */,
      DiskVolumePntr->freePntr, DiskVolumePntr->volumePoolID);
    AVLDupFreeThingArray (&IterationRange.upperKey,
      IndexFSOPntr->keyType, 1 /* how many */,
      DiskVolumePntr->freePntr, DiskVolumePntr->volumePoolID);
  }
  if (QueryPntr != NULL)
    freeDVP (DiskVolumePntr, QueryPntr);
  if (ParseTreePntr != NULL)
    FreeParseTree (DiskVolumePntr, ParseTreePntr);
  if (IndexDirectoryFSOPntr != NULL)
    PutIndicesDirectory (DiskVolumePntr, IndexDirectoryFSOPntr, NULL, NULL);
  if (TokenizedQueryPntr != NULL)
    freeDVP (DiskVolumePntr, TokenizedQueryPntr);

NormalExit:
#if DPRINT_NON_ERROR_MESSAGES
  dprintf ("ARFS_open_query exited, return code %d.\n", (int) ReturnCode);
#endif
  return ReturnCode;
}



/******************************************************************************
 * Finished reading the list of files matching the query, but don't free the
 * cookie yet as other threads may still be reading.  Other threads blocked
 * while reading queries should be unblocked and return a suitable error code.
 */

int ARFS_close_query (
  void *ns,
  void *cookie)
{
  GenericDirCookiePointer CookiePntr = cookie;
  QueryStatePointer       CurrentQueryPntr;
  DiskVolumePointer       DiskVolumePntr = ns;
  QueryStatePointer       NextQueryPntr;
  QueryStatePointer       PreviousQueryPntr;
  QueryStatePointer       QueryToBeRemovedPntr;

#if DPRINT_NON_ERROR_MESSAGES
  dprintf ("ARFS_close_query entered, NS:$%X, cookie:$%X.\n",
    (unsigned int) ns, (unsigned int) cookie);
#endif

  if (INVALID_DISK_VOLUME (DiskVolumePntr))
  {
    dprintf ("ARFS_close_query: invalid input disk volume pointer $%X.\n",
      (unsigned int) DiskVolumePntr);
    return -1;
  }

  if (CookiePntr == NULL)
  {
    dprintf ("ARFS_close_query: NULL cookie pointer!\n");
    return -1;
  }

  /* Remove the query from the global live query list first, before closing it
  and otherwise wiping out the query data.  Works even if close called multiple
  times, by different threads (unlikely, but just in case…). */

  QueryToBeRemovedPntr = CookiePntr->queryPntr;
  if (QueryToBeRemovedPntr != NULL &&
  QueryToBeRemovedPntr->liveQueryPortID >= 0 &&
  WriteLockLiveQueryList (DiskVolumePntr) == 0)
  {
    PreviousQueryPntr = NULL;
    CurrentQueryPntr = DiskVolumePntr->liveQueryListHead;
    while (CurrentQueryPntr != NULL)
    {
      NextQueryPntr = CurrentQueryPntr->nextLiveQueryPntr;
      if (CurrentQueryPntr == QueryToBeRemovedPntr)
      {
        if (PreviousQueryPntr == NULL)
          DiskVolumePntr->liveQueryListHead = NextQueryPntr;
        else
          PreviousQueryPntr->nextLiveQueryPntr = NextQueryPntr;
        break;
      }
      CurrentQueryPntr = NextQueryPntr;
    }
    WriteUnlockLiveQueryList (DiskVolumePntr);
  }

  return GenericCloseDir (ns, CookiePntr->fsoPntr, cookie);
}



/******************************************************************************
 * The last thread has finished reading the query results, you can now really
 * close it.  Deallocate the cookie and otherwise clean up here.  Watch out for
 * the meaningless node parameter.
 */

int ARFS_free_querycookie (
  void *ns,
  void *node,
  void *cookie)
{
  GenericDirCookiePointer  CookiePntr = cookie;

#if DPRINT_NON_ERROR_MESSAGES
  dprintf ("ARFS_free_querycookie entered, NS:$%X, useless node:$%X, "
    "cookie:$%X.\n",
    (unsigned int) ns, (unsigned int) node, (unsigned int) cookie);
#endif

  if (CookiePntr == NULL)
    return -1;

  return GenericFreeDirCookie (ns, CookiePntr->fsoPntr, cookie);
}



/******************************************************************************
 * Gets a list of files which match the query.  Returns the number of files
 * being described in *num, and the information about them as variable sized
 * dirent structures in the buffer you provided (which is bufsize bytes long).
 * To get all the files matching the query, keep on calling read_query until no
 * more are returned (should just return zero in *num, not an error code).
 * Note that BeOS will crash if you don't provide an implementation of this
 * function - there doesn't seem to be a stub function that just returns an
 * error code like it does for all other API calls.
 */

int ARFS_read_query (
  void *ns,
  void *cookie,
  long *num,
  struct dirent *buf,
  size_t bufsize)
{
  GenericDirCookiePointer  CookiePntr = cookie;
  status_t                 ErrorCode;
  long                     OriginalNum;
  int                      RetryCount;
  unsigned int             SavedRandomState = 0;

#if DPRINT_NON_ERROR_MESSAGES
  dprintf ("ARFS_read_query entered, "
    "NS:$%X, cookie:$%X, *num:%d, buf:$%X, bufsize:%d.\n",
    (unsigned int) ns, (unsigned int) cookie,
    (int) *num, (unsigned int) buf, (int) bufsize);
#endif

  if (CookiePntr == NULL)
    return -1;
  OriginalNum = *num;

  for (RetryCount = 0; RetryCount < 5; RetryCount++)
  {
    ErrorCode =
      GenericReadDir (ns, CookiePntr->fsoPntr, cookie, num, buf, bufsize);

    if (ErrorCode != EDEADLK)
      return ErrorCode; /* Success or ordinary error. */

#if DPRINT_NON_ERROR_MESSAGES
    dprintf ("ARFS_read_query: Deadlock detected.  Got %ld results already, "
      "for cookie:$%X.  Retry count %d.\n",
      *num, (unsigned int) cookie, RetryCount);
#endif

    /* The query got stuck.  The usual situation is that the query readlocks
    the index it is travering and then readlocks the files it is examining.  If
    some other thread is modifying a file (write locks it) then the query soon
    afterwards runs into it and waits, meanwhile the second thread finishes
    modifying the file and then tries to update the unavailable index,
    resulting in a deadlock.  To get around that, the query uses a tentative
    readlock for things that it examines, so it will stop with an EDEADLK when
    it tries examining a busy file, coincidentally unlocking the index too.  If
    it has retrieved at least one query result, we can return that right now.
    Otherwise, wait a while and try again. */

    if (*num > 0)
      return 0; /* Success, got at least one result. */

    *num = OriginalNum;
    RandomDelay (&SavedRandomState);
  }
  return ErrorCode;
}



/******************************************************************************
 * Some global variables which the file system loader looks for (by name) and
 * uses to access the functions which our file system exports.
 */

/* This is a magic global variable which the add-on loader will look for to
decide how to use this file system.  The value comes from the fsproto.h file,
so that the OS knows which variety of vnode_ops structure we implemented. */

int32 api_version = B_CUR_FS_API_VERSION;

#if B_CUR_FS_API_VERSION != 2
  #error This file system was designed for version 2 of the API, please update it for the new version before compiling and using it.
#endif

/* Table of our callback functions that the kernal can use.  NULL for
unimplemented ones, which the kernel will usually do a default action for. */

vnode_ops fs_entry =
{
  ARFS_read_vnode /* read_vnode */,
  ARFS_write_vnode /* write_vnode */,
  ARFS_remove_vnode /* remove_vnode */,
  NULL /* secure_vnode */,
  ARFS_walk /* walk */,
  ARFS_access /* access */,
  ARFS_create /* create */,
  ARFS_mkdir /* mkdir */,
  ARFS_symlink /* symlink */,
  ARFS_link /* link */,
  ARFS_rename /* rename */,
  ARFS_unlink /* unlink */,
  ARFS_rmdir /* rmdir */,
  ARFS_readlink /* readlink */,
  ARFS_opendir /* opendir */,
  ARFS_closedir /* closedir */,
  ARFS_free_dircookie /* free_dircookie */,
  ARFS_rewinddir /* rewinddir */,
  ARFS_readdir /* readdir */,
  ARFS_open /* open */,
  ARFS_close /* close */,
  ARFS_free_cookie /* free_cookie */,
  ARFS_read /* read */,
  ARFS_write /* write */,
  NULL /* readv */,
  NULL /* writev */,
  ARFS_ioctl /* ioctl */,
  NULL /* setflags */,
  ARFS_rstat /* rstat */,
  ARFS_wstat /* wstat */,
  NULL /* fsync */,
  ARFS_initialize /* initialize */,
  ARFS_mount /* mount */,
  ARFS_unmount /* unmount */,
  NULL /* sync */,
  ARFS_rfsstat /* rfsstat */,
  ARFS_wfsstat /* wfsstat */,
  NULL /* select */,
  NULL /* deselect */,
  ARFS_open_indexdir /* open_indexdir */,
  ARFS_close_indexdir /* close_indexdir */,
  ARFS_free_indexdircookie /* free_indexdircookie */,
  ARFS_rewind_indexdir /* rewind_indexdir */,
  ARFS_read_indexdir /* read_indexdir */,
  ARFS_create_index /* create_index */,
  ARFS_remove_index /* remove_index */,
  NULL /* rename_index */,
  ARFS_stat_index /* stat_index */,
  ARFS_open_attrdir /* open_attrdir */,
  ARFS_close_attrdir /* close_attrdir */,
  ARFS_free_attrdircookie /* free_attrdircookie */,
  ARFS_rewind_attrdir /* rewind_attrdir */,
  ARFS_read_attrdir /* read_attrdir */,
  ARFS_write_attr /* write_attr */,
  ARFS_read_attr /* read_attr */,
  ARFS_remove_attr /* remove_attr */,
  ARFS_rename_attr /* rename_attr */,
  ARFS_stat_attr /* stat_attr */,
  ARFS_open_query /* open_query */,
  ARFS_close_query /* close_query */,
  ARFS_free_querycookie /* free_querycookie */,
  ARFS_read_query /* read_query */
};
